import unittest
import logging

from pyswagger import SwaggerApp
from pyswagger.contrib.client.requests import Client
from pyswagger.utils import jp_compose
from pyswagger.primitives import SwaggerPrimitive
from pyswagger.primitives._int import validate_int, create_int

factory = SwaggerPrimitive()
factory.register('integer', '', create_int, validate_int)
factory.register('integer', None, create_int, validate_int)


class TestDNS(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        app = SwaggerApp.load('./api.json', prim=factory)
        app.prepare()
        self.client = Client(send_opt={"verify": False})
        self.DNS = app.resolve(jp_compose('/network/dns', base='#/paths'))
        self.ETHERNET = app.resolve(
            jp_compose('/network/ethernets', base='#/paths'))

    @classmethod
    def tearDownClass(self):
        pass

    def test_get_dns(self):
        """Get current DNS"""
        self.client.request(self.DNS.get()).data

    def test_put_dns(self):
        """Change DNS to fixed, it should override all dns settings"""
        self.client.request(self.DNS.put(body={
            "enableFixed": True,
            "fixedDNS": ["1.1.1.1", "2.2.2.2"]
        }))

        with open("/etc/resolv.conf") as file:
            lines = file.read()
            self.assertTrue("nameserver 1.1.1.1" in lines)
            self.assertTrue("nameserver 2.2.2.2" in lines)

    def test_put_dns_2(self):
        """Change DNS to non-fixed, it should use wan's dns settings"""
        self.client.request(self.DNS.put(body={
            "enableFixed": False
        }))

        eths = self.client.request(self.ETHERNET.get()).data
        dns_list = []
        for eth in eths:
            if eth["wan"] is True:
                dns_list = eth["dns"]

        with open("/etc/resolv.conf") as file:
            lines = file.read()
            for dns in dns_list:
                if len(dns) > 0:
                    self.assertIn("nameserver {}".format(dns), lines)

if __name__ == "__main__":
    FORMAT = '%(asctime)s - %(levelname)s - %(lineno)s - %(message)s'
    logging.basicConfig(level=30, format=FORMAT)
    logger = logging.getLogger('SanjiKeeper')
    unittest.main()
