import unittest
import logging
import sh

from datetime import datetime
from pyswagger import SwaggerApp
from pyswagger.contrib.client.requests import Client
from pyswagger.utils import jp_compose
from pyswagger.primitives import SwaggerPrimitive
from pyswagger.primitives._int import validate_int, create_int

factory = SwaggerPrimitive()
factory.register('integer', '', create_int, validate_int)
factory.register('integer', None, create_int, validate_int)


class TestTime(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        app = SwaggerApp.load('./api.json', prim=factory)
        app.prepare()
        self.client = Client(send_opt={"verify": False})
        self.TIME = app.resolve(jp_compose('/system/time', base='#/paths'))

    @classmethod
    def tearDownClass(self):
        pass

    def test_case_1_get_time(self):
        """Get current TIME"""
        resp = self.client.request(self.TIME.get()).data
        dt = datetime.strptime(
            resp["time"], "%Y-%m-%dT%H:%M:%S.%fZ")

        # date_string = sh.date()
        self.assertLessEqual((datetime.utcnow() - dt).seconds, 1)

    def test_case_2_manually(self):
        """Disable NTP Server, Update Time"""
        self.client.request(self.TIME.put(body={
            "time": "2000-01-01T00:00:00.000000Z",
            "timezone": "+01:00,0",
            "ntp": {
                "enable": False,
                "interval": 7200,
                "server": "192.168.31.254"
            }
        }))
        # Check dateime (within 3 seconds)
        dt = datetime.strptime(
            "2000-01-01T00:00:00.000000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        self.assertLessEqual((datetime.utcnow() - dt).seconds, 3)
        # Check time zone
        date_string = sh.date()
        hwclock_string = sh.hwclock()
        self.assertIn("CAT", date_string)
        self.assertIn("CAT", hwclock_string)

    def test_case_3_ntp(self):
        """Enable NTP Server"""
        self.client.request(self.TIME.put(body={
            "timezone": "+08:00,0",
            "ntp": {
                "enable": True,
                "interval": 7200,
                "server": "192.168.50.33"
            }
        }))

        # Check time zone
        date_string = sh.date()
        hwclock_string = sh.hwclock()
        self.assertIn("CST", date_string)
        self.assertIn("CST", hwclock_string)
        # Check Year
        self.assertIn("2016", date_string)
        self.assertIn("2016", hwclock_string)

if __name__ == "__main__":
    FORMAT = '%(asctime)s - %(levelname)s - %(lineno)s - %(message)s'
    logging.basicConfig(level=30, format=FORMAT)
    logger = logging.getLogger('test')
    unittest.main()
