# Using C API

C libraries comes with C-header files and shared-objects.

`libmxidaf` provides *pkg-config* support, to compile your source code:

```
$ gcc -c example.c -o example.o `pkg-config --cflags`
```

To link your code:

```
$ gcc -o example example.o `pkg-config --libs`
```

## Library Initialization

`libmxidaf` should be initialized before use.
To initialize the library, simply get the instance of `Service`.

To get `Service` instance in C:

```c
mxidaf_Service *service = mxidaf_Service_instance();
```

## Import Tags

Tags are imported by `TagImporter`.

```c
mxidaf_Time *at = mxidaf_Time_new_system_time();

mxidaf_Value *volt = mxidaf_Value_new_float(
    5.010);

mxidaf_Value *curr = mxidaf_Value_new_float(
    0.344);

mxidaf_Tag *volt_tag = mxidaf_Tag_new(
    volt,
    at,
    "v");

mxidaf_Tag *curr_tag = mxidaf_Tag_new(
    curr,
    at,
    "a");

/* get TagImporter object */
mxidaf_TagImporter *tag_importer = mxidaf_Service_get_tag_importer(
    service);

mxidaf_TagImporter_import(
    tag_importer,
    "electrical-probe",     /* equipment name */
    "voltage",              /* tag name */
    &volt_tag,
    1);

mxidaf_TagImporter_import(
    tag_importer,
    "electrical-probe",
    "current",
    &curr_tag,
    1);
```

Note: If the `equipment name` or `tag name` is named start with `-`, the whole tag will be ignored.

### Sample

Sample application can be found at

  `/usr/share/mxdaf/libmxidaf/sample/import.c`

This sample imports a `voltage` Tag and a `current` Tag under `electrical-probe` Equipment.

To run sample:

```
$ ./import
```

You can use `mxtaghistory` to see the result:

```
$ mxtaghistory
"equipment","tag","at","value","unit"
"electrical-probe","current","2016-03-16T16:14:19.206Z","0.344","a"
"electrical-probe","voltage","2016-03-16T16:14:19.206Z","5.01","v"
```

## Query History Tags

To query history Tags, first get a `Cache` object.
Use `Cache` object to get the `Equipment` object of interest.
Use the `Equipment` object to get the `Stream` object you want.
After you obtain the `Stream` object, you can query the `Tags` as needed.

```c
mxidaf_Cache *cache = mxidaf_Service_get_cache(
    service);

mxidaf_Equipment *equipment = mxidaf_Cache_get_equipment(
    cache,
    "electrical-probe");

mxidaf_Stream *stream = mxidaf_Equipment_get_stream(
    equipment,
    "voltage");

mxidaf_Tag *tag_list[3] = {0};
const int len = mxidaf_Stream_get_last_tag_list(
    stream,
    tag_list,
    3);

/* ...do something with tag_list... */

/* cleanup */
int ax;
for(ax = 0; ax < len; ++ax)
    mxidaf_Tag_delete(tag_list[ax]);

mxidaf_Stream_delete(stream);
mxidaf_Equipment_delete(equipment);
mxidaf_Cache_delete(cache);
```

Note: Please DO NOT query history tags when there is no tag imported.

### Sample

Sample application can be found at

  `/usr/share/mxdaf/libmxidaf/sample/query.c`

This sample queries `voltage` Tag of `electrical-probe` Equipment and dumps Tag list to stdout.

To run sample:

```
$ ./query
queried 2 tags
2016-03-16T16:14:51.994Z:5.010000:v
2016-03-16T16:14:19.206Z:5.010000:v
```

## Arithmetic of Tag value

With the ability to *Query History Tags* and *Import Tags*,
you can now inject your data processing logic into *Data Acquisition Framework*.

For example, we want to produce a new value by multiplying two input values:

```c
double
    last_value_of_stream(
        mxidaf_Stream *stream)
{
    mxidaf_Tag *tag = NULL;
    mxidaf_Stream_get_last_tag_list(
        stream,
        &tag,
        1);

    mxidaf_Value *value = NULL;
    mxidaf_Tag_get(
        tag,
        &value,
        NULL,
        NULL);

    return mxidaf_Value_as_float(
        value);
}

/* ... */
double voltage_v = last_value_of_stream(
    voltage_stream);
double current_a = last_vlaue_of_stream(
    current_stream);

mxidaf_Value *watt = mxidaf_Value_new_float(
    voltage_v * current_a);
mxidaf_Tag *tag = mxidaf_Tag_new_system_time(
    value,
    "w");
mxidaf_TagImporter_import(
    "electrical-probe",
    "power",
    tag);
/* ... */

```

### Sample

Sample application can be found at

  `/usr/share/mxdaf/libmxidaf/sample/calculate.c`

This sample imports a `power` Tag and its value, which is the latest value of `voltage` multiply the latest value of `current`.

To run sample:

```
$ ./calculate
```

After running the sample, you can see that `power` Tag is imported:

```
$ mxtaghistory --equipment electrical-probe --tag power
"equipment","tag","at","value","unit"
"electrical-probe","power","2016-03-16T17:03:31.508Z","1.72344","w"
```

## Error Handling

Error codes and messages can be retrieved through `mxidaf_Service_get_last_error()` and `mxidaf_Service_get_last_error_msg()`,
for example:

```c
const int len = mxidaf_Stream_get_last_tag_list(
    stream,
    &tag_list,
    10);

if (len == -1)
{
    const mxidaf_Error err = mxidaf_Stream_get_last_error(service);
    const char *errmsg = mxidaf_Stream_get_last_error_msg(service);

    /* ...handle error... */
}
```
