#include <mxhtag/Database.hpp>

#include <boost/date_time/gregorian/gregorian.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/tuple/tuple.hpp>

#include <mxutil/log.hpp>

#include <sqlite3.h>

#include <mxsqlite3/Sqlite3.hpp>
#include <mxsqlite3/Transaction.hpp>

using boost::filesystem::path;
using boost::get_error_info;
using boost::gregorian::date;
using boost::lexical_cast;
using boost::make_shared;
using boost::make_tuple;
using boost::optional;
using boost::posix_time::hours;
using boost::posix_time::microsec_clock;
using boost::posix_time::milliseconds;
using boost::posix_time::minutes;
using boost::posix_time::ptime;
using boost::posix_time::seconds;
using boost::posix_time::time_duration;
using boost::shared_ptr;
using boost::tuple;

using mxsqlite3::Connection;
using mxsqlite3::Sqlite3;
using mxsqlite3::Sqlite3ErrCodeInfo;
using mxsqlite3::Statement;
using mxsqlite3::Transaction;
using mxsqlite3::Value;

using mxtag::Tag;

using std::string;
using std::vector;

namespace
{

const string    VERSION = "0.2.1";
const ptime     PTIME_OFFSET = ptime(date(1970, 1, 1), seconds(0));

int_least64_t
    millisec_from_ptime(const ptime &pt)
{
    const time_duration td = pt - PTIME_OFFSET;

    // Since time_duration::total_milliseconds() returns long, which might be
    // only 4 bytes long on some platforms, which causes its value truncated,
    // we calculate its value our selves.
    int_least64_t ms = td.hours();
    ms *= 60;
    ms += td.minutes();
    ms *= 60;
    ms += td.seconds();
    ms *= 1000;

    long frag = td.fractional_seconds();
    for(size_t ax = 3; ax < time_duration::num_fractional_digits(); ++ax)
        frag /= 10;

    ms += frag;

    return ms;
}

const ptime
    ptime_from_millisec(int_least64_t ms)
{
    const long ms_part = ms % 1000;
    const long sec = (ms / 1000) % 60;
    const long minute = (ms / 1000 / 60) % 60;
    const long hour = ms / 1000 / 60 / 60;

    return PTIME_OFFSET + hours(hour) + minutes(minute) + seconds(sec) + milliseconds(ms_part);
}

typedef
    tuple<string, string>
    EquipmentTagPair;

typedef
    tuple<string, string>
    TypeValuePair;

const TypeValuePair
    db_from_value(const mxtag::Value &v)
{
    const string type =
        v.is_int() ? "i" :
        v.is_uint() ? "u" :
        v.is_float() ? "f" :
        v.is_string() ? "s" :
        "e";

    BOOST_VERIFY(type != "e");

    return TypeValuePair(type, lexical_cast<string>(v));
}

const mxtag::Value
    value_from_db(
        const TypeValuePair &tvp)
{
    const string t = tvp.get<0>();
    const string v = tvp.get<1>();

    return
        t == "i" ? mxtag::Value(lexical_cast<int_least64_t>(v)) :
        t == "u" ? mxtag::Value(lexical_cast<uint_least64_t>(v)) :
        t == "f" ? mxtag::Value(lexical_cast<double>(v)) :
        mxtag::Value(v);
}

}

namespace mxhtag
{

void
    Database::create_tables(
        const path &dbpath)
{
    try
    {
        shared_ptr<Connection> conn = Sqlite3::instance().connect(
            dbpath,
            10 * 1000,
            10,
            true);

        *conn
        <<
            "CREATE TABLE readings("
                " id            INTEGER PRIMARY KEY AUTOINCREMENT,"
                " equipment     VARCHAR(256) NOT NULL,"
                " tag           VARCHAR(256) NOT NULL,"
                " at_ms         INTEGER NOT NULL,"
                " value         VARCHAR(256) NOT NULL,"
                " value_type    CHARACTER(1) NOT NULL,"    // i, u, f, s
                " unit          VARCHAR(256) NOT NULL,"
                " created_at_ms INTEGER NOT NULL"
            " );"
        <<
            "CREATE UNIQUE INDEX readings_equipment_tag_at"
            " ON readings(equipment, tag, at_ms);"
        <<
            "CREATE TABLE configs("
                " key   VARCHAR(64) NOT NULL,"
                " value VARCHAR(64) NOT NULL"
            " );"
        ;

        shared_ptr<Statement> insert = conn->create_statement(
            "INSERT INTO configs VALUES(?, ?);"
        );

        insert->bind(1, "version");
        insert->bind(2, VERSION);
        insert->execute();
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(Error() << EInfoMsg(diagnostic_information(e)));
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(Error());
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(Error());
    }
}

struct Database::Impl
{
    Impl(
        shared_ptr<Connection> connection_)
        : connection(connection_)
        , delete_by_id_stmt(                        connection->create_statement(
            "DELETE FROM readings"
            " WHERE id <= ?"))
        , insert_reading_list_stmt(                 connection->create_statement(
            "INSERT INTO readings (equipment, tag, at_ms, value_type, value, unit, created_at_ms)"
            " VALUES (?, ?, ?, ?, ?, ?, ?)"))
        , query_equipment_name_list_stmt(           connection->create_statement(
            "SELECT DISTINCT equipment FROM readings"))
        , query_is_equipment_exist_stmt(            connection->create_statement(
            "SELECT 1 FROM readings WHERE equipment = ?"))
        , query_is_equipment_tag_exist_stmt(        connection->create_statement(
            "SELECT 1 FROM readings WHERE equipment = ? AND tag = ?"))
        , query_last_tag_list_by_equipment_tag_stmt(connection->create_statement(
            "SELECT at_ms, value, value_type, unit FROM readings"
            " WHERE equipment = ?"
            " AND tag = ?"
            " ORDER BY id DESC"
            " LIMIT ?"))
        , query_reading_count_stmt(                 connection->create_statement(
            "SELECT COUNT(*) FROM readings"))
        , query_reading_id_by_offset_stmt(          connection->create_statement(
            "SELECT id FROM readings ORDER BY id ASC LIMIT 1 OFFSET ?"))
        , query_tag_list_by_equipment_tag_stmt(     connection->create_statement(
            "SELECT at_ms, value, value_type, unit FROM readings"
            " WHERE equipment = ?"
            " AND tag = ?"
            " ORDER BY id DESC"))
        , query_tag_name_list_by_equipment_stmt(    connection->create_statement(
            "SELECT DISTINCT tag FROM readings WHERE equipment = ?"))
    {
    }

    const shared_ptr<Connection> connection;

    size_t
        query_reading_count();

    int_least64_t
        query_reading_id_by_offset(
            size_t offset);

    const shared_ptr<Statement> delete_by_id_stmt;
    const shared_ptr<Statement> insert_reading_list_stmt;
    const shared_ptr<Statement> query_equipment_name_list_stmt;
    const shared_ptr<Statement> query_is_equipment_exist_stmt;
    const shared_ptr<Statement> query_is_equipment_tag_exist_stmt;
    const shared_ptr<Statement> query_last_tag_list_by_equipment_tag_stmt;
    const shared_ptr<Statement> query_reading_count_stmt;
    const shared_ptr<Statement> query_reading_id_by_offset_stmt;
    const shared_ptr<Statement> query_tag_list_by_equipment_tag_stmt;
    const shared_ptr<Statement> query_tag_name_list_by_equipment_stmt;
};

Database::Database(
    const path &dbpath,
    int retry_timeout_ms,
    int retry_wait_ms)
{
    BOOST_VERIFY(retry_timeout_ms >= 0);
    BOOST_VERIFY(retry_wait_ms >= 0);

    try
    {
        const shared_ptr<Connection> connection = Sqlite3::instance().connect(
            dbpath,
            retry_timeout_ms,
            retry_wait_ms);

        *connection
        <<
            "PRAGMA page_size = 1024"
        <<
            "PRAGMA max_page_count = 10240"
        <<
            "PRAGMA synchronous = OFF"
        ;

        const shared_ptr<Statement> select = connection->create_statement(
            "SELECT value FROM configs WHERE key = ?;"
        );

        select->bind(1, "version");
        optional<vector<Value> > row = select->execute();
        if (!row ||
            row->size() != 1 ||
            row->at(0).as_string() != VERSION)
        {
            BOOST_THROW_EXCEPTION(VersionError());
        }

        pimpl.reset(new Impl(connection));
    }
    catch(mxsqlite3::Error &e)
    {
        const int *errcode = get_error_info<mxsqlite3::Sqlite3ErrCodeInfo>(e);
        if (errcode &&
            *errcode == SQLITE_CANTOPEN)
        {
            BOOST_THROW_EXCEPTION(
                OpenError() << EInfoMsg(diagnostic_information(e))
            );
        }

        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception")
        );
    }
}

Database::~Database()
{
}

void
    Database::insert_reading(
        const Tag &tag)
{
    this->insert_reading_list(vector<Tag>(1, tag));
}

void
    Database::insert_reading_list(const vector<Tag> &tag_list)
{
    try
    {
        const int_least64_t created_at_ms = millisec_from_ptime(microsec_clock::universal_time());

        const shared_ptr<Statement> insert = pimpl->insert_reading_list_stmt;

        Transaction transaction(*pimpl->connection);

        BOOST_FOREACH(const Tag &tag, tag_list)
        {
            insert->reset();

            const TypeValuePair tvp = db_from_value(tag.value());

            insert->bind(1, tag.equipment());
            insert->bind(2, tag.tag());
            insert->bind(3, millisec_from_ptime(tag.at()));
            insert->bind(4, tvp.get<0>());
            insert->bind(5, tvp.get<1>());
            insert->bind(6, tag.unit());
            insert->bind(7, created_at_ms);

            insert->execute();
        }

        transaction.commit();
    }
    catch(mxsqlite3::Error &e)
    {
        const int * ec = get_error_info<Sqlite3ErrCodeInfo>(e);
        if (ec &&
            *ec == SQLITE_CONSTRAINT)
        {
            // reading exist
            MXLOG_DBUG("insert_reading_list: Tag already exist");
            if (tag_list.size() > 1)
            {
                // insert tags one-by-one
                BOOST_FOREACH(const Tag &tag, tag_list)
                {
                    this->insert_reading(tag);
                }
            }
        }
        else if(ec &&
                *ec == SQLITE_FULL)
        {
            // disk full
            BOOST_THROW_EXCEPTION(
                DiskFullError() << EInfoMsg(diagnostic_information(e))
            );
        }
        else
        {
            BOOST_THROW_EXCEPTION(
                Error() << EInfoMsg(diagnostic_information(e))
            );
        }
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

void
    Database::delete_oldest_readings(
        float rate)
{
    BOOST_VERIFY(rate <= 1.0);

    try
    {
        Transaction transaction(*pimpl->connection);

        const size_t reading_count = pimpl->query_reading_count();
        size_t remove_count = ceil(reading_count * rate);

        if (0 == remove_count)
            return;

        const int_least64_t id = pimpl->query_reading_id_by_offset(
            remove_count - 1);

        const shared_ptr<Statement> stmt = pimpl->delete_by_id_stmt;
        stmt->reset();

        stmt->bind(1, id);

        stmt->execute();

        transaction.commit();
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

const vector<string>
    Database::get_equipment_name_list() const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_equipment_name_list_stmt;
        select->reset();

        vector<string> equipment_list;
        optional<vector<Value> > row;
        while((row = select->execute()))
        {
            equipment_list.push_back(row->at(0).as_string());
        }

        return equipment_list;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

bool
    Database::is_equipment_exist(
        const string &equipment) const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_is_equipment_exist_stmt;
        select->reset();

        select->bind(1, equipment);

        optional<vector<Value> > row;
        row = select->execute();

        return row;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

const vector<string>
    Database::get_tag_name_list_by_equipment(
        const string &equipment) const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_tag_name_list_by_equipment_stmt;
        select->reset();

        select->bind(1, equipment);

        vector<string> tag_list;
        optional<vector<Value> > row;
        while((row = select->execute()))
        {
            tag_list.push_back(row->at(0).as_string());
        }

        return tag_list;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

bool
    Database::is_equipment_tag_exist(
        const string &equipment,
        const string &tag) const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_is_equipment_tag_exist_stmt;
        select->reset();

        select->bind(1, equipment);
        select->bind(2, tag);

        optional<vector<Value> > row;
        row = select->execute();

        return row;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

const vector<Tag>
    Database::get_tag_list_by_equipment_tag(
        const string &equipment,
        const string &tag) const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_tag_list_by_equipment_tag_stmt;
        select->reset();

        select->bind(1, equipment);
        select->bind(2, tag);

        vector<Tag> tag_list;
        optional<vector<Value> > row;
        while((row = select->execute()))
        {
            tag_list.push_back(Tag(
                equipment,
                tag,
                ptime_from_millisec(row->at(0).as_int()),
                value_from_db(TypeValuePair(row->at(2).as_string(), row->at(1).as_string())),
                row->at(3).as_string()
            ));
        }

        return tag_list;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

const vector<Tag>
    Database::get_last_tag_list_by_equipment_tag(
        const string &equipment,
        const string &tag,
        size_t count) const
{
    try
    {
        const shared_ptr<Statement> select = pimpl->query_last_tag_list_by_equipment_tag_stmt;
        select->reset();

        select->bind(1, equipment);
        select->bind(2, tag);
        select->bind(3, static_cast<int64_t>(count));

        vector<Tag> tag_list;
        optional<vector<Value> > row;
        while((row = select->execute()))
        {
            tag_list.push_back(Tag(
                equipment,
                tag,
                ptime_from_millisec(row->at(0).as_int()),
                value_from_db(TypeValuePair(row->at(2).as_string(), row->at(1).as_string())),
                row->at(3).as_string()
            ));
        }

        return tag_list;
    }
    catch(mxsqlite3::Error &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(diagnostic_information(e))
        );
    }
    catch(std::exception &e)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg(e.what())
        );
    }
    catch(...)
    {
        BOOST_THROW_EXCEPTION(
            Error() << EInfoMsg("unexpected exception caught")
        );
    }
}

// Database::Impl member functions ////////////////////////////////////////////
size_t
    Database::Impl::query_reading_count()
{
    const shared_ptr<Statement> stmt = this->query_reading_count_stmt;
    stmt->reset();

    optional<vector<Value> > row = stmt->execute();
    BOOST_VERIFY(row);

    return row->at(0).as_int();
}

int_least64_t
    Database::Impl::query_reading_id_by_offset(
        size_t offset)
{
    const shared_ptr<Statement> stmt = this->query_reading_id_by_offset_stmt;
    stmt->reset();

    stmt->bind(1, static_cast<int64_t>(offset));
    optional<vector<Value> > row = stmt->execute();
    BOOST_VERIFY(row);

    return row->at(0).as_int();
}

}
