#include <mxsqlite3/ConnectionImpl.hpp>

#include <cstring>

#include <boost/chrono/chrono.hpp>
#include <boost/lexical_cast.hpp>

#include <mxutil/log.hpp>

#include <mxsqlite3/StatementImpl.hpp>

using boost::chrono::nanoseconds;
using boost::chrono::steady_clock;
using boost::lexical_cast;
using boost::shared_ptr;

using std::string;

namespace mxsqlite3
{

struct ConnectionImpl::Impl
{
    Impl(
        sqlite3 *handle_,
        int retry_timeout_ms_,
        int retry_wait_ms_)
        : handle(handle_)
        , retry_timeout_ms(retry_timeout_ms_)
        , retry_wait_ms(retry_wait_ms_)
    {
    }

    sqlite3 *   handle;
    const int   retry_timeout_ms;
    const int   retry_wait_ms;
};

ConnectionImpl::ConnectionImpl(
    sqlite3 *handle,
    int retry_timeout_ms,
    int retry_wait_ms)
    : pimpl(new Impl(
        handle,
        retry_timeout_ms,
        retry_wait_ms))
{
    BOOST_VERIFY(pimpl->handle);
    BOOST_VERIFY(pimpl->retry_timeout_ms >= 0);
    BOOST_VERIFY(pimpl->retry_wait_ms >= 0);
}

ConnectionImpl::~ConnectionImpl()
{
    if (!pimpl->handle)
        return;

    if (SQLITE_OK != sqlite3_close(pimpl->handle))
    {
        // do nothing
    }

    pimpl->handle = 0;
}

shared_ptr<Statement>
    ConnectionImpl::do_create_statement(
        const string &sql)
{
    sqlite3_stmt *stmt = 0;
    const char *tail = 0;

    steady_clock::time_point expired = steady_clock::now() + nanoseconds(pimpl->retry_timeout_ms * 1000000ll);

    do
    {
        const int result = sqlite3_prepare_v2(
            pimpl->handle,
            sql.c_str(),
            sql.size() + 1,
            &stmt,
            &tail);

        if (result == SQLITE_BUSY)
        {
            usleep(pimpl->retry_wait_ms * 1000);
            continue;
        }
        else if (result != SQLITE_OK)
        {
            BOOST_THROW_EXCEPTION(
                Error()
                    << Sqlite3ErrCodeInfo(result)
                    << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
                    << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
            );
        }

        BOOST_ASSERT(stmt);

        if (tail != 0 &&
            strlen(tail) != 0)
        {
            if (stmt)
                sqlite3_finalize(stmt);

            BOOST_THROW_EXCEPTION(Error() << ErrMsgInfo("not all SQL statement consumed"));
        }

        return shared_ptr<Statement>(new StatementImpl(
            stmt,
            pimpl->handle,
            pimpl->retry_timeout_ms,
            pimpl->retry_wait_ms));
    }
    while(steady_clock::now() <= expired);

    BOOST_THROW_EXCEPTION(
        Error()
            << ErrMsgInfo("retry timeout: retry_timeout_ms = " + lexical_cast<string>(pimpl->retry_timeout_ms))
    );
}

}
