#include <mxsqlite3/StatementImpl.hpp>

#include <boost/chrono/chrono.hpp>
#include <boost/lexical_cast.hpp>

#include <mxutil/log.hpp>

using boost::chrono::nanoseconds;
using boost::chrono::steady_clock;
using boost::lexical_cast;
using boost::optional;

using std::string;
using std::vector;

namespace mxsqlite3
{

struct StatementImpl::Impl
{
    Impl(
        sqlite3_stmt *stmt_handle_,
        sqlite3 *handle_,
        int retry_timeout_ms_,
        int retry_wait_ms_)
        : stmt_handle(stmt_handle_)
        , handle(handle_)
        , retry_timeout_ms(retry_timeout_ms_)
        , retry_wait_ms(retry_wait_ms_)
    {
    }

    sqlite3_stmt *  stmt_handle;
    sqlite3 *       handle;
    const int       retry_timeout_ms;
    const int       retry_wait_ms;
};

StatementImpl::StatementImpl(
    sqlite3_stmt *stmt_handle,
    sqlite3 *handle,
    int retry_timeout_ms,
    int retry_wait_ms)
    : pimpl(new Impl(
        stmt_handle,
        handle,
        retry_timeout_ms,
        retry_wait_ms))
{
    BOOST_VERIFY(pimpl->stmt_handle && pimpl->handle);
    BOOST_VERIFY(pimpl->retry_timeout_ms >= 0);
    BOOST_VERIFY(pimpl->retry_wait_ms >= 0);
}

StatementImpl::~StatementImpl()
{
    sqlite3_finalize(pimpl->stmt_handle);
    pimpl->stmt_handle = NULL;
}

void
    StatementImpl::do_reset()
{
    int result;
    if (SQLITE_OK != (result = sqlite3_reset(pimpl->stmt_handle)))
        BOOST_THROW_EXCEPTION(
            Error()
                << Sqlite3ErrCodeInfo(result)
                << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
        );
}

void
    StatementImpl::do_bind(size_t i)
{
    int result;
    if (SQLITE_OK != (result = sqlite3_bind_null(pimpl->stmt_handle, i)))
        BOOST_THROW_EXCEPTION(
            Error()
                << Sqlite3ErrCodeInfo(result)
                << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
        );
}

void
    StatementImpl::do_bind(
        size_t i,
        int64_t v)
{
    int result;
    if (SQLITE_OK != (result = sqlite3_bind_int64(pimpl->stmt_handle, i, v)))
        BOOST_THROW_EXCEPTION(
            Error()
                << Sqlite3ErrCodeInfo(result)
                << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
        );
}

void
    StatementImpl::do_bind(
        size_t i,
        double v)
{
    int result;
    if (SQLITE_OK != (result = sqlite3_bind_double(pimpl->stmt_handle, i, v)))
        BOOST_THROW_EXCEPTION(
            Error()
                << Sqlite3ErrCodeInfo(result)
                << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
        );
}

void
    StatementImpl::do_bind(
        size_t i,
        const string &v)
{
    int result;
    if (SQLITE_OK != (result = sqlite3_bind_text(pimpl->stmt_handle, i, v.c_str(), v.size(), SQLITE_TRANSIENT)))
        BOOST_THROW_EXCEPTION(
            Error()
                << Sqlite3ErrCodeInfo(result)
                << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
        );
}

optional<vector<Value> >
    StatementImpl::do_execute()
{
    steady_clock::time_point expired = steady_clock::now() + nanoseconds(pimpl->retry_timeout_ms * 1000000ll);

    do
    {
        const int result = sqlite3_step(pimpl->stmt_handle);
        if (result == SQLITE_ROW)
        {
            const int column_count = sqlite3_column_count(pimpl->stmt_handle);

            vector<Value> value_list;
            for(int ax = 0; ax < column_count; ++ax)
            {
                const int column_type = sqlite3_column_type(pimpl->stmt_handle, ax);
                switch(column_type)
                {
                case SQLITE_INTEGER:
                    value_list.push_back(Value(
                        static_cast<int64_t>(sqlite3_column_int64(pimpl->stmt_handle, ax))
                    ));
                    break;

                case SQLITE_FLOAT:
                    value_list.push_back(Value(sqlite3_column_double(pimpl->stmt_handle, ax)));
                    break;

                case SQLITE_TEXT:
                    value_list.push_back(Value(string(
                        reinterpret_cast<const char *>(sqlite3_column_text(pimpl->stmt_handle, ax))
                    )));
                    break;

                case SQLITE_NULL:
                    value_list.push_back(Value());
                    break;

                default:
                    BOOST_THROW_EXCEPTION(
                        Error() << ErrMsgInfo("unexpected column type: " + lexical_cast<string>(column_type))
                    );
                }
            }

            return value_list;
        }
        else if (result == SQLITE_DONE)
        {
            return boost::none;
        }
        else if (result == SQLITE_BUSY)
        {
            usleep(pimpl->retry_wait_ms * 1000);

            continue;
        }
        else
            BOOST_THROW_EXCEPTION(
                Error()
                    << Sqlite3ErrCodeInfo(result)
                    << Sqlite3ErrMsgInfo(sqlite3_errmsg(pimpl->handle))
                    << Sqlite3ExtErrCodeInfo(sqlite3_extended_errcode(pimpl->handle))
                    << ErrMsgInfo("unexpected sqlite3_step() return code")
            );
    }
    while(steady_clock::now() <= expired);

    BOOST_THROW_EXCEPTION(
        Error()
            << ErrMsgInfo("retry timeout: retry_timeout_ms = " + lexical_cast<string>(pimpl->retry_timeout_ms))
    );
}

}
