#include <boost/test/unit_test.hpp>

#include <mxhtag/Database.hpp>

#include <boost/foreach.hpp>

#include <mxsqlite3/Sqlite3.hpp>

using boost::optional;
using boost::posix_time::time_from_string;
using boost::shared_ptr;

using mxhtag::Database;

using mxsqlite3::Connection;
using mxsqlite3::Sqlite3;
using mxsqlite3::Statement;
using mxsqlite3::Value;

using mxtag::Tag;

using std::string;
using std::vector;

namespace
{

struct TestFixture
{
    TestFixture()
        : dbpath("/dev/shm/test.sqlite3")
    {
        if (0 != unlink(this->dbpath.c_str()))
        {
            // do nothing
        }
    }

    virtual ~TestFixture()
    {
        if (0 != unlink(this->dbpath.c_str()))
        {
            // do nothing
        }
    }

    const string    dbpath;
};

struct CreateTablesTestFixture
    : public TestFixture
{
    using TestFixture::dbpath;

    CreateTablesTestFixture()
    {
        Database::create_tables(dbpath);
    }
};

}

namespace mxhtag
{

BOOST_FIXTURE_TEST_CASE(test_Database_create_tables_ShouldPass, TestFixture)
{
    // arrange

    // act
    Database::create_tables(dbpath);

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_Database_ctor_ShouldPass, CreateTablesTestFixture)
{
    // arrange

    // act
    Database db(dbpath);

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_Database_ctor_WhenTableNotCreated_ShouldThrowError, TestFixture)
{
    // arrange

    // act & assert
    BOOST_CHECK_THROW(Database d(dbpath), Error);
}

BOOST_FIXTURE_TEST_CASE(test_Database_insert_reading_ShouldPass, CreateTablesTestFixture)
{
    // arrange
    Database db(dbpath);

    // act
    db.insert_reading(Tag(
        "equipment-name",
        "channel-name",
        time_from_string("2010-01-01 00:00:00"),
        mxtag::Value("this-is-a-string-value"),
        "unit-name"
    ));

    // assert
    const shared_ptr<Connection> conn = Sqlite3::instance().connect(dbpath);
    const shared_ptr<Statement> select = conn->create_statement(
        "SELECT COUNT(*) FROM readings"
    );

    const optional<vector<Value> > row = select->execute();

    BOOST_REQUIRE(row && row->size() == 1);
    BOOST_CHECK_EQUAL(1, row->at(0).as_int());
}

BOOST_FIXTURE_TEST_CASE(test_Database_insert_reading_WhenReadingExist_ShouldRollback, CreateTablesTestFixture)
{
    // arrange
    Database db(dbpath);
    const Tag t(
        "equipment-name",
        "channel-name",
        time_from_string("2010-01-01 00:00:00"),
        mxtag::Value("this-is-a-string-value"),
        "unit-name"
    );

    db.insert_reading(t);

    // act
    db.insert_reading(t);

    // assert
    const vector<Tag> tag_list = db.get_tag_list_by_equipment_tag(
        "equipment-name",
        "channel-name");
    BOOST_CHECK_EQUAL(1, tag_list.size());
}

BOOST_FIXTURE_TEST_CASE(
    test_Database_insert_reading_list_WhenSomeReadingExist_ShouldInsertNotExistReadings,
    CreateTablesTestFixture)
{
    // arrange
    Database db(dbpath);
    const Tag t(
        "equipment-name",
        "channel-name",
        time_from_string("2010-01-01 00:00:00"),
        mxtag::Value("this-is-a-string-value"),
        "unit-name"
    );

    db.insert_reading(t);

    const Tag t2(
        "equipment-name",
        "channel-name",
        time_from_string("2010-01-01 00:00:01"),
        mxtag::Value("this-is-a-string-value"),
        "unit-name"
    );
    vector<Tag> tag_list;
    {
        tag_list.push_back(t);
        tag_list.push_back(t2);
    }

    // act
    db.insert_reading_list(tag_list);

    // assert
    const vector<Tag> queried_tag_list = db.get_tag_list_by_equipment_tag(
        "equipment-name",
        "channel-name");
    BOOST_CHECK_EQUAL(2, queried_tag_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_Database_delete_oldest_readings_ShouldPass, CreateTablesTestFixture)
{
    // arrange
    Database db(dbpath);
    const Tag reading_list[] = {
        {"equipment-name", "channel-name", time_from_string("2010-01-01 00:00:00"), mxtag::Value("value"), "unit"},
        {"equipment-name", "channel-name", time_from_string("2010-01-02 00:00:00"), mxtag::Value("value"), "unit"},
        {"equipment-name", "channel-name", time_from_string("2010-01-03 00:00:00"), mxtag::Value("value"), "unit"}
    };

    BOOST_FOREACH(const Tag &reading, reading_list)
    {
        db.insert_reading(reading);
    }

    // act
    db.delete_oldest_readings(0.3);

    // assert
    const shared_ptr<Connection> conn = Sqlite3::instance().connect(dbpath);
    const shared_ptr<Statement> select = conn->create_statement(
        "SELECT COUNT(*) FROM readings"
    );

    const optional<vector<Value> > row = select->execute();

    BOOST_REQUIRE(row && row->size() == 1);
    BOOST_CHECK_EQUAL(2, row->at(0).as_int());
}

}
