#include <boost/test/unit_test.hpp>

#include <mxhtag/Database.hpp>

#include <boost/foreach.hpp>

#include <mxsqlite3/Sqlite3.hpp>

using boost::optional;
using boost::lexical_cast;
using boost::posix_time::ptime;
using boost::posix_time::seconds;
using boost::posix_time::time_from_string;
using boost::shared_ptr;

using mxhtag::Database;

using mxsqlite3::Connection;
using mxsqlite3::Sqlite3;
using mxsqlite3::Statement;
using mxsqlite3::Value;

using mxtag::Tag;

using std::string;
using std::vector;

namespace
{

struct Fixture
{
    Fixture()
        : dbpath("/dev/shm/test.sqlite3")
    {
        if (0 != unlink(this->dbpath.c_str()))
        {
            // do nothing
        }
    }

    virtual ~Fixture()
    {
        if (0 != unlink(this->dbpath.c_str()))
        {
            // do nothing
        }
    }

    const string    dbpath;
};

struct CreateTablesFixture
    : public Fixture
{
    using Fixture::dbpath;

    CreateTablesFixture()
    {
        Database::create_tables(dbpath);
    }
};

struct TestDataFixture
    : public CreateTablesFixture
{
    using CreateTablesFixture::dbpath;

    TestDataFixture()
        : database(dbpath)
    {
        const unsigned int readings_per_tag = 10;
        const unsigned int equ_count = 3;
        const unsigned int tags_per_equ = 3;

        const ptime base_time = time_from_string("2010-01-01 00:00:00");

        for(unsigned int reading = 0; reading < readings_per_tag; ++reading)
        {
            const ptime reading_time = base_time + seconds(reading);

            vector<Tag> reading_list;
            for(unsigned int equ = 0; equ < equ_count; ++equ)
            {
                for(unsigned int tag = 0; tag < tags_per_equ; ++tag)
                {
                    reading_list.push_back(Tag(
                        "equipment" + lexical_cast<string>(equ),
                        "e" + lexical_cast<string>(equ) + "tag" + lexical_cast<string>(tag),
                        reading_time,
                        mxtag::Value("test-value"),
                        "test-unit"));
                }
            }

            database.insert_reading_list(reading_list);
        }
    }

    Database    database;
};

}

namespace mxhtag
{

BOOST_FIXTURE_TEST_CASE(test_Database_get_equipment_name_list_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<string> equ_list = database.get_equipment_name_list();

    // assert
    BOOST_CHECK_EQUAL(3, equ_list.size());
    BOOST_CHECK(equ_list.end() != find(equ_list.begin(), equ_list.end(), "equipment0"));
    BOOST_CHECK(equ_list.end() != find(equ_list.begin(), equ_list.end(), "equipment1"));
    BOOST_CHECK(equ_list.end() != find(equ_list.begin(), equ_list.end(), "equipment2"));
}

BOOST_FIXTURE_TEST_CASE(test_Database_is_equipment_exist_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const bool exist = database.is_equipment_exist("equipment0");

    // assert
    BOOST_CHECK(exist);
}

BOOST_FIXTURE_TEST_CASE(test_Database_is_equipment_exist_WithNotExistEquipment_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const bool not_exist = database.is_equipment_exist("equipment5");

    // assert
    BOOST_CHECK(!not_exist);
}

BOOST_FIXTURE_TEST_CASE(test_Database_get_tag_name_list_by_equipment_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<string> tag_list = database.get_tag_name_list_by_equipment("equipment2");

    // assert
    BOOST_CHECK_EQUAL(3, tag_list.size());
    BOOST_CHECK(tag_list.end() != find(tag_list.begin(), tag_list.end(), "e2tag0"));
    BOOST_CHECK(tag_list.end() != find(tag_list.begin(), tag_list.end(), "e2tag1"));
    BOOST_CHECK(tag_list.end() != find(tag_list.begin(), tag_list.end(), "e2tag2"));
}

BOOST_FIXTURE_TEST_CASE(test_Database_is_equipment_tag_exist_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const bool exist = database.is_equipment_tag_exist("equipment2", "e2tag0");

    // assert
    BOOST_CHECK(exist);
}

BOOST_FIXTURE_TEST_CASE(test_Database_is_equipment_tag_exist_WithNotExistEquipmentTag_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const bool not_exist = database.is_equipment_tag_exist("equipment2", "tag0");

    // assert
    BOOST_CHECK(!not_exist);
}

BOOST_FIXTURE_TEST_CASE(test_Database_get_tag_list_by_equipment_tag_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<mxtag::Tag> tag_list = database.get_tag_list_by_equipment_tag(
        "equipment1",
        "e1tag2");

    // assert
    BOOST_CHECK_EQUAL(10, tag_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_Database_get_tag_list_by_equipment_tag_WithNotExistTag_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<mxtag::Tag> tag_list = database.get_tag_list_by_equipment_tag(
        "equipment1",
        "e1tag100");

    // assert
    BOOST_CHECK_EQUAL(0, tag_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_Database_get_last_tag_list_by_equipment_tag_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<mxtag::Tag> tag_list = database.get_last_tag_list_by_equipment_tag(
        "equipment1",
        "e1tag2",
        2);

    // assert
    BOOST_CHECK_EQUAL(2, tag_list.size());
    BOOST_CHECK(Tag("equipment1", "e1tag2", time_from_string("2010-01-01 00:00:09"), mxtag::Value("test-value"), "test-unit") == tag_list[0]);
    BOOST_CHECK(Tag("equipment1", "e1tag2", time_from_string("2010-01-01 00:00:08"), mxtag::Value("test-value"), "test-unit") == tag_list[1]);
}

BOOST_FIXTURE_TEST_CASE(test_Database_get_last_tag_list_by_equipment_tag_WithZeroCount_ShouldPass, TestDataFixture)
{
    // arrange

    // act
    const vector<mxtag::Tag> tag_list = database.get_last_tag_list_by_equipment_tag(
        "equipment1",
        "e1tag2",
        0);

    // assert
    BOOST_CHECK_EQUAL(0, tag_list.size());
}

}
