#include <boost/test/unit_test.hpp>

#include <mxsqlite3/Statement.hpp>

#include <mxsqlite3/Sqlite3.hpp>

namespace mxsqlite3
{

using boost::optional;
using boost::shared_ptr;

using std::string;
using std::vector;

namespace
{

struct TestFixture
{
    TestFixture()
        : path("/dev/shm/test.sqlite3")
    {
        if (0 != unlink(path.c_str()))
        {
            // do nothing
        }

        connection = Sqlite3::instance().connect(
            path,
            10 * 1000,
            10,
            true);

        *connection <<
            "CREATE TABLE records("
            " int_value BIGINT,"
            " double_value FLOAT,"
            " str_value TEXT"
            " );";
    }

    virtual ~TestFixture()
    {
        if (0 != unlink(path.c_str()))
        {
            // do nothing
        }
    }

    const string            path;
    shared_ptr<Connection>  connection;
};

}

BOOST_FIXTURE_TEST_CASE(test_Statement_bind_null_ShouldPass, TestFixture)
{
    // arrange
    shared_ptr<Statement> stmt = connection->create_statement(
        "INSERT INTO records (int_value) VALUES (?);"
    );

    // act
    stmt->bind(1);

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_Statement_bind_int_ShouldPass, TestFixture)
{
    // arrange
    shared_ptr<Statement> stmt = connection->create_statement(
        "INSERT INTO records (int_value) VALUES (?);"
    );

    // act
    stmt->bind(1, static_cast<int64_t>(345));

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_Statement_bind_double_ShouldPass, TestFixture)
{
    // arrange
    shared_ptr<Statement> stmt = connection->create_statement(
        "INSERT INTO records (double_value) VALUES (?);"
    );

    // act
    stmt->bind(1, 123.456);

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_Statement_bind_string_ShouldPass, TestFixture)
{
    // arrange
    shared_ptr<Statement> stmt = connection->create_statement(
        "INSERT INTO records (str_value) VALUES (?);"
    );

    // act
    stmt->bind(1, "123.456");

    // assert
}

namespace
{

struct RecordInsertedTestFixture
    : TestFixture
{
    RecordInsertedTestFixture()
    {
        const shared_ptr<Statement> insert = connection->create_statement(
            "INSERT INTO records VALUES (?, ?, ?);"
        );

        insert->bind(1, static_cast<int64_t>(10));
        insert->bind(2, 3.1415);
        insert->bind(3, "3.1415");

        insert->execute();
    }
};

}

BOOST_FIXTURE_TEST_CASE(test_Statement_execute_WithSelect_ShouldPass, RecordInsertedTestFixture)
{
    // arrange
    const shared_ptr<Statement> select = connection->create_statement(
        "SELECT * FROM records;"
    );

    // act
    const optional<vector<Value> > row = select->execute();

    // assert
    BOOST_REQUIRE(row);
    BOOST_REQUIRE_EQUAL(3, row->size());
    BOOST_CHECK_EQUAL(10, row->at(0).as_int());
    BOOST_CHECK_CLOSE(3.1415, row->at(1).as_double(), 0.0001);
    BOOST_CHECK_EQUAL("3.1415", row->at(2).as_string());
}

}
