/**
 * @file
 * @author tc.huang@moxa.com
 * @brief Service object
 *
 */

#ifndef _MXIDAF_SERVICE_H_
#define _MXIDAF_SERVICE_H_

#include <mxidaf/Cache.h>
#include <mxidaf/TagImporter.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @defgroup mxidaf-service Service object
 *
 * @ingroup mxidaf
 */

/**
 * @brief Service instance
 *
 * @ingroup mxidaf-service
 */
typedef struct mxidaf_Service_ mxidaf_Service;

/**
 * @brief Get Service instance.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Service instance.\n
 *      Return NULL on failure.
 */
mxidaf_Service *
    mxidaf_Service_instance();

/**
 * @brief Create Cache.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Cache instance.\n
 *      Return NULL on failure.
 *
 * @see
 *      mxidaf_Cache_delete()
 */
mxidaf_Cache *
    mxidaf_Service_get_cache(
        const mxidaf_Service *self  /**< Service instance. */
    );

/**
 * @brief Create Cache.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Cache instance.\n
 *      Return NULL on failure.
 *
 * @see
 *      mxidaf_Cache_delete()
 */
mxidaf_Cache *
    mxidaf_Service_get_cache_2(
        const mxidaf_Service *self, /**< Service instance. */
        const char *dbpath          /**< Path to db-file. */
    );

/**
 * @brief Create TagImporter.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Return mxidaf_TagImporter instance.\n
 *      Return NULL on failure.
 *
 * @see
 *      mxidaf_TagImporter_delete()
 */
mxidaf_TagImporter *
    mxidaf_Service_get_tag_importer(
        const mxidaf_Service *self  /**< [in] Service instance. */
    );

/**
 * @brief Create TagImporter.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Return mxidaf_TagImporter instance.\n
 *      Return NULL on failure.
 *
 * @see
 *      mxidaf_TagImporter_delete()
 */
mxidaf_TagImporter *
    mxidaf_Service_get_tag_importer_2(
        const mxidaf_Service *self, /**< [in] Service instance. */
        const char *host,           /**< [in] Broker hostname. */
        int port                    /**< [in] Broker port. */
    );

/**
 * @brief Error codes.
 *
 * @ingroup mxidaf-service
 */
enum mxidaf_Error
{
    MXIDAF_ERROR_NONE,              /**< No error */
    MXIDAF_ERROR_ILLEGAL_ARGUMENT,  /**< Illegal argument */
    MXIDAF_ERROR_ILLEGAL_SERVICE,   /**< Illegal service instance */
    MXIDAF_ERROR_NOT_FOUND,         /**< Resource requested not found */
    MXIDAF_ERROR_VERSION,           /**< Version mismatch */
    MXIDAF_ERROR_INTERNAL,          /**< Internal error */
};

/**
 * @brief Get error code.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Error code.
 */
enum mxidaf_Error
    mxidaf_Service_get_last_error(
        const mxidaf_Service *self  /**< [in] Service instance. */
    );

/**
 * @brief Get error message.
 *
 * @ingroup mxidaf-service
 *
 * @return
 *      Error message.\n
 *      Return NULL if no error occurred.
 */
const char *
    mxidaf_Service_get_last_error_msg(
        const mxidaf_Service *self  /**< [in] Service instance. */
    );

#ifdef __cplusplus
}
#endif

#endif
