/**
 * @file
 * @author tc.huang@moxa.com
 * @brief Time object
 *
 */

#ifndef _MXIDAF_TIME_H_
#define _MXIDAF_TIME_H_

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @defgroup mxidaf-time Time object
 *
 * @ingroup mxidaf
 */

/**
 * @brief Time instance.
 *
 * Represent time in UTC.\n
 * Precision: milliseconds.
 *
 * @ingroup mxidaf-time
 */
typedef struct mxidaf_Time_ mxidaf_Time;

/**
 * @brief Value helper.
 *
 * @ingroup mxidaf-time
 */
typedef struct mxidaf_Time_Value_
{
    int year;       /**< Year */
    int month;      /**< Month */
    int day;        /**< Day */
    int hour;       /**< Hour */
    int minute;     /**< Minute */
    int second;     /**< Second */
    int millisec;   /**< Millisecond */
} mxidaf_Time_Value;

/**
 * @brief Create Time object with given time point.
 *
 * @ingroup mxidaf-time
 *
 * @return
 *      Return Time instance.\n
 *      Return NULL on failure.
 */
mxidaf_Time *
    mxidaf_Time_new(
        const mxidaf_Time_Value *tv     /**< [in] Initial value */
    );

/**
 * @brief Create Time object using system time.
 *
 * @ingroup mxidaf-time
 *
 * @return
 *      Return Time instance.\n
 *      Return NULL on failure.
 */
mxidaf_Time *
    mxidaf_Time_new_system_time();

/**
 * @brief Delete Time object.
 *
 * @ingroup mxidaf-time
 */
void
    mxidaf_Time_delete(
        mxidaf_Time *self   /**< [in] Time instance to be deleted */
    );

/**
 * @brief Get time values.
 *
 * @ingroup mxidaf-time
 *
 * @return
 *      Return TimeValue instance passed by tv.\n
 *      Return NULL on failure.
 */
mxidaf_Time_Value *
    mxidaf_Time_get(
        const mxidaf_Time *self,    /**< [in] Time instance */
        mxidaf_Time_Value *tv       /**< [out] Value to fill */
    );

/**
 * @brief Create a copy of existing Time object.
 *
 * @ingroup mxidaf-time
 *
 * @return
 *      Return a new Time object.\n
 *      Return NULL on failure.
 */
mxidaf_Time *
    mxidaf_Time_clone(
        const mxidaf_Time *self     /**< [in] Time instance to be copied */
    );

/**
 * @brief Compare two Time instances.
 *
 * @ingroup mxidaf-time
 *
 * @return
 *      -1: lhs earlier than rhs.\n
 *      0: lhs == rhs.\n
 *      1: lhs later than rhs.
 */
int
    mxidaf_Time_compare(
        const mxidaf_Time *lhs,     /**< [in] Time instance */
        const mxidaf_Time *rhs      /**< [in] Time instance */
    );

#ifdef __cplusplus
}
#endif

#endif
