/**
 * @file
 * @author tc.huang@moxa.com
 * @brief Value object
 */

#ifndef _MXIDAF_VALUE_H_
#define _MXIDAF_VALUE_H_

#include <stdbool.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @defgroup mxidaf-value Value object
 *
 * @ingroup mxidaf
 */

/**
 * @brief Value instance.
 *
 * @ingroup mxidaf-value
 */
typedef struct mxidaf_Value_ mxidaf_Value;

/**
 * @brief Create an integer Value instance.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return a Value object.\n
 *      Return NULL on failure.
 */
mxidaf_Value *
    mxidaf_Value_new_int(
        int64_t i                   /**< [in] Initial value */
    );

/**
 * @brief Create an unsigned integer Value instance.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return a Value object.\n
 *      Return NULL on failure.
 */
mxidaf_Value *
    mxidaf_Value_new_uint(
        uint64_t u                  /**< [in] Initial value */
    );

/**
 * @brief Create a float Value instance.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return a Value object.\n
 *      Return NULL on failure.
 */
mxidaf_Value *
    mxidaf_Value_new_float(
        double f                    /**< [in] Initial value */
    );

/**
 * @brief Create a string Value instance.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return a Value object.\n
 *      Return NULL on failure.
 */
mxidaf_Value *
    mxidaf_Value_new_string(
        const char *s               /**< [in] Initial value */
    );


/**
 * @brief Delete a Value instance.
 *
 * @ingroup mxidaf-value
 */
void
    mxidaf_Value_delete(
        mxidaf_Value *self      /**< [in] Value instance to be deleted */
    );

/**
 * @brief Check Value type.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return TRUE if Value is int, otherwise FALSE.
 */
int
    mxidaf_Value_is_int(
        const mxidaf_Value *self    /**< [in] Value instance */
    );

/**
 * @brief Check Value type.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return TRUE if Value is uint, otherwise FALSE.
 */
int
    mxidaf_Value_is_uint(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Check Value type.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return TRUE if Value is float, otherwise FALSE.
 */
int
    mxidaf_Value_is_float(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Check Value type.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Return TRUE if Value is string, otherwise FALSE.
 */
int
    mxidaf_Value_is_string(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Get integer value.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      The value.
 *
 * @note
 *      Can only be called when mxidaf_Value_is_int(self) is TRUE.
 */
int64_t
    mxidaf_Value_as_int(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Get unsigned integer value.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      The value.
 *
 * @note
 *      Can only be called when mxidaf_Value_is_uint(self) is TRUE.
 */
uint64_t
    mxidaf_Value_as_uint(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Get float value.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      The value.
 *
 * @note
 *      Can only be called when mxidaf_Value_is_float(self) is TRUE.
 */
double
    mxidaf_Value_as_float(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Get string value.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      The value.\n
 *      The returned pointer is maintained by Value object.
 *      free() is not permitted.
 *
 * @note
 *      Can only be called when mxidaf_Value_is_string(self) is TRUE.
 */
const char *
    mxidaf_Value_as_string(
        const mxidaf_Value *self   /**< [in] Value instance */
    );

/**
 * @brief Create a copy of an existing Value object.
 *
 * @ingroup mxidaf-value
 *
 * @return
 *      Value object.\n
 *      Return NULL on failure.
 */
mxidaf_Value *
    mxidaf_Value_clone(
        const mxidaf_Value *from   /**< [in] Value to be copied */
    );

#ifdef __cplusplus
}
#endif

#endif
