#include <boost/test/unit_test.hpp>

#include <mxidaf/Equipment.h>

#include <boost/range.hpp>
#include <boost/scope_exit.hpp>

#include <mxidaf/Service.h>

#include "DatabaseFixture.hpp"

using boost::size;

using std::string;

namespace
{

struct TestFixture
    : public DatabaseFixture
{
    TestFixture()
    {
        mxidaf_Service *service = mxidaf_Service_instance();
        cache = mxidaf_Service_get_cache(
            service);

        equipment = mxidaf_Cache_get_equipment(
            cache,
            "equipment1");
    }

    ~TestFixture()
    {
        mxidaf_Equipment_delete(equipment);
        mxidaf_Cache_delete(cache);
    }

    mxidaf_Cache *cache;
    mxidaf_Equipment *equipment;
};

}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_destruct_ShouldPass, TestFixture)
{
    // arrange
    mxidaf_Equipment *inst = mxidaf_Cache_get_equipment(
        cache,
        "equipment2");

    // act
    mxidaf_Equipment_delete(
        inst);

    // assert
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_get_name_ShouldPass, TestFixture)
{
    // arrange

    // act
    const char *name = mxidaf_Equipment_get_name(
        equipment);

    // assert
    BOOST_CHECK_EQUAL(string("equipment1"), name);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_get_stream_list_ShouldPass, TestFixture)
{
    // arrange
    mxidaf_Stream *stream_list[10];

    // act
    const int len = mxidaf_Equipment_get_stream_list(
        equipment,
        &stream_list[0],
        size(stream_list));

    BOOST_SCOPE_EXIT((&len)(&stream_list))
    {
        for (int ax = 0; ax < len; ++ax)
        {
            mxidaf_Stream_delete(stream_list[ax]);
            stream_list[ax] = NULL;
        }
    }
    BOOST_SCOPE_EXIT_END

    // assert
    BOOST_CHECK_EQUAL(3, len);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_get_stream_list_WithZeroLen_ShouldPass, TestFixture)
{
    // arrange

    // act
    const int len = mxidaf_Equipment_get_stream_list(
        equipment,
        0,
        0);

    // assert
    BOOST_CHECK_EQUAL(3, len);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_get_stream_ShouldPass, TestFixture)
{
    // arrange

    // act
    mxidaf_Stream *stream = mxidaf_Equipment_get_stream(
        equipment,
        "e1tag0");

    BOOST_SCOPE_EXIT((&stream))
    {
        mxidaf_Stream_delete(stream);
    }
    BOOST_SCOPE_EXIT_END

    // assert
    BOOST_CHECK(NULL != stream);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Equipment_get_stream_WithNotExitTag_ShouldFail, TestFixture)
{
    // arrange

    // act
    mxidaf_Stream *stream = mxidaf_Equipment_get_stream(
        equipment,
        "not-exist-tag");

    const int ec = mxidaf_Service_get_last_error(
        mxidaf_Service_instance());

    // assert
    BOOST_CHECK(NULL == stream);
    BOOST_CHECK_EQUAL(MXIDAF_ERROR_NOT_FOUND, ec);
}
