#include <boost/test/unit_test.hpp>

#include <mxidaf/Tag.h>

#include <string>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/scope_exit.hpp>

using boost::posix_time::microsec_clock;
using boost::posix_time::ptime;
using boost::posix_time::time_duration;

using std::string;

namespace
{

struct TestFixture
{
    TestFixture()
        : value(mxidaf_Value_new_string("hello world"))
        , at(NULL)
    {
        const mxidaf_Time_Value ctv = {
            year: 2016,
            month: 2,
            day: 19,
            hour: 9,
            minute: 9,
            second: 20,
            millisec: 392};

        at = mxidaf_Time_new(
            &ctv);
    }

    ~TestFixture()
    {
        mxidaf_Value_delete(value);
        mxidaf_Time_delete(at);
    }

    mxidaf_Value *value;
    mxidaf_Time *at;
};

}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Tag_new_ShouldPass, TestFixture)
{
    // arrange

    // act
    mxidaf_Tag *inst = mxidaf_Tag_new(
        value,
        at,
        "test-unit");

    BOOST_SCOPE_EXIT((&inst))
    {
        mxidaf_Tag_delete(inst);
    }
    BOOST_SCOPE_EXIT_END

    // assert
    BOOST_CHECK(inst);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Tag_new_system_time_ShouldPass, TestFixture)
{
    // arrange
    const ptime now = microsec_clock::universal_time();

    // act
    mxidaf_Tag *inst = mxidaf_Tag_new_system_time(
        value,
        "test-unit");

    BOOST_SCOPE_EXIT((&inst))
    {
        mxidaf_Tag_delete(inst);
    }
    BOOST_SCOPE_EXIT_END

    // assert
    BOOST_REQUIRE(inst);

    mxidaf_Time *at_ = NULL;
    mxidaf_Tag_get(
        inst,
        NULL,
        &at_,
        NULL);

    BOOST_SCOPE_EXIT((&at_))
    {
        mxidaf_Time_delete(at_);
    }
    BOOST_SCOPE_EXIT_END

    mxidaf_Time_Value tv;
    mxidaf_Time_get(
        at_,
        &tv);

    const ptime ref = ptime(
        boost::gregorian::date(tv.year, tv.month, tv.day),
        time_duration(tv.hour, tv.minute, tv.second)
            + boost::posix_time::milliseconds(tv.millisec));

    const time_duration diff = ref - now;
    BOOST_CHECK(diff.total_milliseconds() < 1000);
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Tag_get_ShouldPass, TestFixture)
{
    // arrange
    mxidaf_Tag *inst = mxidaf_Tag_new(
        value,
        at,
        "test-unit");

    BOOST_SCOPE_EXIT((&inst))
    {
        mxidaf_Tag_delete(inst);
    }
    BOOST_SCOPE_EXIT_END

    // act
    mxidaf_Value *value_ = NULL;
    mxidaf_Time *at_ = NULL;
    mxidaf_String *unit = NULL;

    const int rc = mxidaf_Tag_get(
        inst,
        &value_,
        &at_,
        &unit);

    BOOST_SCOPE_EXIT((&value_)(&at_)(&unit))
    {
        mxidaf_Value_delete(value_);
        mxidaf_Time_delete(at_);
        mxidaf_String_delete(unit);
    }
    BOOST_SCOPE_EXIT_END

    // assert
    BOOST_REQUIRE_EQUAL(0, rc);

    BOOST_REQUIRE(value_);
    BOOST_REQUIRE(mxidaf_Value_is_string(value_));
    BOOST_CHECK_EQUAL(string("hello world"), mxidaf_Value_as_string(value_));

    BOOST_REQUIRE(at_);
    mxidaf_Time_Value tv;
    mxidaf_Time_get(
        at_,
        &tv);
    BOOST_CHECK_EQUAL(2016, tv.year);
    BOOST_CHECK_EQUAL(2, tv.month);
    BOOST_CHECK_EQUAL(19, tv.day);
    BOOST_CHECK_EQUAL(9, tv.hour);
    BOOST_CHECK_EQUAL(9, tv.minute);
    BOOST_CHECK_EQUAL(20, tv.second);
    BOOST_CHECK_EQUAL(392, tv.millisec);

    BOOST_REQUIRE(unit);
    BOOST_CHECK_EQUAL(string("test-unit"), mxidaf_String_str(unit));
}

BOOST_FIXTURE_TEST_CASE(test_mxidaf_Tag_copy_ShouldPass, TestFixture)
{
    // arrange
    mxidaf_Tag *from = mxidaf_Tag_new(
        value,
        at,
        "test-unit");

    BOOST_SCOPE_EXIT((&from))
    {
        mxidaf_Tag_delete(from);
    }
    BOOST_SCOPE_EXIT_END

    // act
    mxidaf_Tag *to = mxidaf_Tag_clone(
        from);

    BOOST_SCOPE_EXIT((&to))
    {
        mxidaf_Tag_delete(to);
    }
    BOOST_SCOPE_EXIT_END

    // assert
    mxidaf_Value *value_ = NULL;
    mxidaf_Time *at_ = NULL;
    mxidaf_String *unit = NULL;

    mxidaf_Tag_get(
        to,
        &value_,
        &at_,
        &unit);

    BOOST_SCOPE_EXIT((&value_)(&at_)(&unit))
    {
        mxidaf_Value_delete(value_);
        mxidaf_Time_delete(at_);
        mxidaf_String_delete(unit);
    }
    BOOST_SCOPE_EXIT_END

    BOOST_REQUIRE(mxidaf_Value_is_string(value_));
    BOOST_CHECK_EQUAL(string("hello world"), mxidaf_Value_as_string(value_));

    mxidaf_Time_Value tv;
    mxidaf_Time_get(
        at_,
        &tv);
    BOOST_CHECK_EQUAL(2016, tv.year);
    BOOST_CHECK_EQUAL(2, tv.month);
    BOOST_CHECK_EQUAL(19, tv.day);
    BOOST_CHECK_EQUAL(9, tv.hour);
    BOOST_CHECK_EQUAL(9, tv.minute);
    BOOST_CHECK_EQUAL(20, tv.second);
    BOOST_CHECK_EQUAL(392, tv.millisec);

    BOOST_CHECK_EQUAL(string("test-unit"), mxidaf_String_str(unit));
}
