#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/python.hpp>
#include <boost/python/suite/indexing/vector_indexing_suite.hpp>

#include <mxidafpp/Service.hpp>

using boost::make_shared;
using boost::lexical_cast;
using boost::shared_ptr;

using mxidafpp::Cache;
using mxidafpp::Error;
using mxidafpp::Equipment;
using mxidafpp::Stream;
using mxidafpp::Tag;
using mxidafpp::TagImporter;
using mxidafpp::Time;
using mxidafpp::Value;

using std::string;
using std::vector;

using namespace boost::python;

namespace
{

shared_ptr<Time>
    make_Time(
        int y,
        int mon,
        int d,
        int h,
        int m,
        int s,
        int ms)
{
    return make_shared<Time>(
        y, mon, d, h, m, s, ms
    );
}

shared_ptr<Tag>
    make_Tag(
        const Value &value,
        const Time &at,
        const string &unit)
{
    return make_shared<Tag>(
        value,
        at,
        unit);
}

}

namespace
{

const string
    Value_str(const Value *v)
{
    BOOST_VERIFY(v);

    if (v->is_int())
        return lexical_cast<string>(v->as_int());
    else if (v->is_uint())
        return lexical_cast<string>(v->as_uint());
    else if (v->is_float())
        return lexical_cast<string>(v->as_float());
    else
        return lexical_cast<string>(v->as_string());
}

}

namespace
{

boost::python::list
    Stream_get_last_tag_list(
        const Stream *self,
        size_t s)
{
    BOOST_VERIFY(self);

    const vector<Tag> tag_list = self->get_last_tag_list(s);

    boost::python::list rtn;
    BOOST_FOREACH(const Tag &t, tag_list)
    {
        rtn.append(t);
    }

    return rtn;
}

}

namespace
{

PyObject *error_ptr = NULL;

void
    translate_Error(
        const Error &e)
{
    BOOST_VERIFY(error_ptr);

    boost::python::object exception_inst(e);
    PyErr_SetObject(error_ptr, exception_inst.ptr());
}

const string
    Error_str(
        const Error *p)
{
    return diagnostic_information(*p);
}

}

namespace
{

shared_ptr<Value>
    make_Value(const boost::python::object &o)
{
    if (PyInt_Check(o.ptr()))
    {
        return make_shared<Value>(extract<int64_t>(o));
    }
    else if (PyFloat_Check(o.ptr()))
    {
        return make_shared<Value>(extract<double>(o));
    }
    else if (PyString_Check(o.ptr()))
    {
        return make_shared<Value>(extract<string>(o));
    }
    else
        return shared_ptr<Value>();
}

}

namespace
{

void
    TagImporter_import(
        TagImporter *self,
        const string &equipment_name,
        const string &tag_name,
        boost::python::list &l)
{
    vector<Tag> tag_list;
    for(int ax = 0; ax < len(l); ++ax)
    {
        tag_list.push_back(extract<Tag>(l[ax]));
    }

    self->import(
        equipment_name,
        tag_name,
        tag_list);
}

}

BOOST_PYTHON_MODULE(libmxidaf_py)
{
    docstring_options docs_opts(true, true, false);

    const TagImporter
        (mxidafpp::Service::*get_tag_importer)() = &mxidafpp::Service::get_tag_importer;
    const Cache
        (mxidafpp::Service::*get_cache)() = &mxidafpp::Service::get_cache;

    class_<mxidafpp::Service, boost::noncopyable>(
            "Service",
            "Library entry.\n"
            "Use it to get helper objects.",
            no_init)
        .def(
            "instance",
            &mxidafpp::Service::instance,
            "Get Service instance.",
            return_value_policy<reference_existing_object>())
            .staticmethod("instance")
        .def(
            "get_tag_importer",
            get_tag_importer,
            "Create a TagImporter object.")
        .def(
            "get_cache",
            get_cache,
            "Create a Cache object.")
        ;

    class_<TagImporter>(
            "TagImporter",
            "Helper object to import Tags.",
            no_init)
        .def(
            "import_tags",
            &TagImporter_import,
            args("self", "equipment_name", "tag_name", "tag_list"),
            "Import a list of Tags associated to given equipment_name/tag_name.")
        ;

    class_<Tag>(
            "Tag",
            "Tag object.\n"
            "Models a reading.",
            no_init)
        .def(
            "__init__",
            make_constructor(
                make_Tag,
                default_call_policies(),
                args("value", "at", "unit")),
            "Constructor.")
        .def(
            "value",
            &Tag::value,
            "Get Tag value.",
            return_internal_reference<>())
        .def(
            "at",
            &Tag::at,
            "Get Tag timestamp.",
            return_internal_reference<>())
        .def(
            "unit",
            &Tag::unit,
            "Get value unit.",
            return_value_policy<copy_const_reference>())
        ;

    class_<Value>(
            "Value",
            "Represents a reading value.\n"
            "Can either be an integer, unsigned integer, float or string.\n"
            "Use is_*() functions to determine its type.\n"
            "Use as_*() functions to retrieve the actual value.",
            no_init)
        .def(
            "__init__",
            make_constructor(
                make_Value,
                default_call_policies(),
                args("value")),
            "Constructor.")
        .def("__str__", &Value_str)
        .def("is_int", &Value::is_int)
        .def("as_int", &Value::as_int)
        .def("is_uint", &Value::is_uint)
        .def("as_uint", &Value::as_uint)
        .def("is_float", &Value::is_float)
        .def("as_float", &Value::as_float)
        .def("is_string", &Value::is_string)
        .def("as_string", &Value::as_string,
            return_value_policy<copy_const_reference>())
        ;

    class_<Time>(
            "Time",
            "Represents a time point in UTC.\n"
            "Precision to milliseconds.",
            no_init)
        .def(
            "__init__",
            make_constructor(
                make_Time,
                default_call_policies(),
                args("year", "month", "day", "hour", "minute", "second", "millisec")),
            "Constructor.")
        .def("__str__", &Time::to_string)
        .def("now", &Time::now)
            .staticmethod("now")
        .def("year", &Time::year)
        .def("month", &Time::month)
        .def("day", &Time::day)
        .def("hour", &Time::hour)
        .def("minute", &Time::minute)
        .def("second", &Time::second)
        .def("millisec", &Time::millisec)
        ;

    class_<Cache>(
            "Cache",
            "Helper object to retrieve cached Tags.",
            no_init)
        .def(
            "get_equipment_list",
            &Cache::get_equipment_list,
            "Get all Equipment objects in cache.")
        .def(
            "get_equipment",
            &Cache::get_equipment,
            args("self", "equipment_name"),
            "Get specific Equipment object.")
        ;

    class_<Equipment>(
            "Equipment",
            "Helper object to retrieve cached Tags.",
            no_init)
        .def(
            "name",
            &Equipment::name,
            "Equipment name associated to this object",
            return_value_policy<copy_const_reference>())
        .def(
            "get_stream_list",
            &Equipment::get_stream_list,
            "Get all Stream objects that in cache.")
        .def(
            "get_stream",
            &Equipment::get_stream,
            args("self", "tag_name"),
            "Get specific Stream object.")
        ;

    class_<std::vector<Equipment> >("EquipmentList")
        .def(vector_indexing_suite<std::vector<Equipment> >())
        ;

    class_<Stream>(
            "Stream",
            "Helper object to retrieve cached Tags.",
            no_init)
        .def(
            "equipment_name",
            &Stream::equipment_name,
            "Equipment name associated to the object.",
            return_value_policy<copy_const_reference>())
        .def(
            "tag_name",
            &Stream::tag_name,
            "Tag name associated to the object.",
            return_value_policy<copy_const_reference>())
        .def(
            "get_last_tag_list",
            &Stream_get_last_tag_list,
            args("self", "count"),
            "Query the last few Tags in cache.")
        ;

    class_<std::vector<Stream> >("StreamList")
        .def(vector_indexing_suite<std::vector<Stream> >())
        ;

    class_<Error> error_inst(
            "Error",
            "Thrown whenever an error is occurred.",
            no_init);
    error_inst
        .def("__str__", &Error_str)
        ;
    error_ptr = error_inst.ptr();

    register_exception_translator<Error>(
        &translate_Error);
}
