#include <mxmodbus/ModbusTcpClient.hpp>

#include <boost/assert.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>

#include <modbus/modbus.h>

#include <mxmodbus/priv/ModbusClientImpl.hpp>

using boost::lexical_cast;
using boost::make_shared;
using boost::shared_ptr;

using mxmodbus::priv::ModbusClientImpl;

using std::string;
using std::vector;

namespace mxmodbus
{

struct ModbusTcpClient::Impl
{
    Impl()
    {
    }

    shared_ptr<ModbusClientImpl>    impl;
};

ModbusTcpClient::ModbusTcpClient(
    const string &host,
    int service,
    int defaultResponseTimeoutMS)
    : pimpl(new Impl())
{
    modbus_t *modbus_inst = modbus_new_tcp(host.c_str(), service);
    BOOST_ASSERT(modbus_inst);

    // uncomment to enabled libmodbus debug message
    //modbus_set_debug(pimpl->modbus_inst);

    // disable error recovery
    if (0 != modbus_set_error_recovery(modbus_inst, MODBUS_ERROR_RECOVERY_NONE))
    {
        modbus_free(modbus_inst);
        BOOST_THROW_EXCEPTION(Error());
    }

    struct timeval response_timeout;
    response_timeout.tv_sec = defaultResponseTimeoutMS / 1000;
    response_timeout.tv_usec = (defaultResponseTimeoutMS % 1000) * 1000;
    modbus_set_response_timeout(modbus_inst, &response_timeout);

    // turn-off byte timeout
    struct timeval byte_timeout;
    byte_timeout.tv_sec = -1;
    modbus_set_byte_timeout(modbus_inst, &byte_timeout);

    if (0 != modbus_connect(modbus_inst))
    {
        modbus_free(modbus_inst);
        BOOST_THROW_EXCEPTION(Error() << EInfoMsg(
            "host = " + host
            + ", service = " + lexical_cast<string>(service)));
    }

    pimpl->impl = make_shared<ModbusClientImpl>(modbus_inst);
}

ModbusTcpClient::~ModbusTcpClient()
{
}

const vector<bool>
    ModbusTcpClient::do_read_coils(
        device_t device,
        address_t addr,
        size_t quantity)
{
    return pimpl->impl->read_coils(device, addr, quantity);
}

const vector<bool>
    ModbusTcpClient::do_read_discrete_inputs(
        device_t device,
        address_t addr,
        size_t quantity)
{
    return pimpl->impl->read_discrete_inputs(device, addr, quantity);
}

const vector<uint16_t>
    ModbusTcpClient::do_read_holding_registers(
        device_t device,
        address_t addr,
        size_t quantity)
{
    return pimpl->impl->read_holding_registers(device, addr, quantity);
}

const vector<uint16_t>
    ModbusTcpClient::do_read_input_registers(
        device_t device,
        address_t addr,
        size_t quantity)
{
    return pimpl->impl->read_input_registers(device, addr, quantity);
}

void
    ModbusTcpClient::do_write_single_coil(
        device_t device,
        address_t addr,
        bool value)
{
    return pimpl->impl->write_single_coil(device, addr, value);
}

void
    ModbusTcpClient::do_write_single_register(
        device_t device,
        address_t addr,
        uint16_t value)
{
    return pimpl->impl->write_single_register(device, addr, value);
}

void
    ModbusTcpClient::do_write_multiple_coils(
        device_t device,
        address_t addr,
        const vector<bool> &value)
{
    return pimpl->impl->write_multiple_coils(device, addr, value);
}

void
    ModbusTcpClient::do_write_multiple_registers(
        device_t device,
        address_t addr,
        const vector<uint16_t> &value)
{
    return pimpl->impl->write_multiple_registers(device, addr, value);
}

}
