#include <mxmodbus/priv/ModbusClientImpl.hpp>

#include <mxmodbus/priv/utility.hpp>

using mxmodbus::priv::bool_vector_from_uint8_vector;
using mxmodbus::priv::uint8_vector_from_bool_vector;
using mxmodbus::priv::append_exception_info;

using std::vector;

namespace mxmodbus
{
namespace priv
{

struct ModbusClientImpl::Impl
{
    Impl(modbus_t *modbus_inst_)
        : modbus_inst(modbus_inst_)
    {
    }

    modbus_t * modbus_inst;

    void
        set_slave(device_t device)
    {
        if (0 != modbus_set_slave(this->modbus_inst, device))
            BOOST_THROW_EXCEPTION(Error());
    }
};

ModbusClientImpl::ModbusClientImpl(modbus_t *inst)
    : pimpl(new Impl(inst))
{
    BOOST_ASSERT(pimpl->modbus_inst);
}

ModbusClientImpl::~ModbusClientImpl()
{
    if (pimpl->modbus_inst)
    {
        modbus_close(pimpl->modbus_inst);
        modbus_free(pimpl->modbus_inst);

        pimpl->modbus_inst = 0;
    }
}

const vector<bool>
    ModbusClientImpl::read_coils(
        device_t device,
        address_t addr,
        size_t quantity)
{
    pimpl->set_slave(device);

    vector<uint8_t> buf(quantity);
    if (-1 == modbus_read_bits(pimpl->modbus_inst, addr, quantity, &buf.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(quantity);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }

    return bool_vector_from_uint8_vector(buf);
}

const vector<bool>
    ModbusClientImpl::read_discrete_inputs(
        device_t device,
        address_t addr,
        size_t quantity)
{
    pimpl->set_slave(device);

    vector<uint8_t> buf(quantity);
    if (-1 == modbus_read_input_bits(pimpl->modbus_inst, addr, quantity, &buf.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(quantity);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }

    return bool_vector_from_uint8_vector(buf);
}

const vector<uint16_t>
    ModbusClientImpl::read_holding_registers(
        device_t device,
        address_t addr,
        size_t quantity)
{
    pimpl->set_slave(device);

    vector<uint16_t> buf(quantity);
    if (-1 == modbus_read_registers(pimpl->modbus_inst, addr, quantity, &buf.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(quantity);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }

    return buf;
}

const vector<uint16_t>
    ModbusClientImpl::read_input_registers(
        device_t device,
        address_t addr,
        size_t quantity)
{
    pimpl->set_slave(device);

    vector<uint16_t> buf(quantity);
    if (-1 == modbus_read_input_registers(pimpl->modbus_inst, addr, quantity, &buf.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(quantity);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }

    return buf;
}

void
    ModbusClientImpl::write_single_coil(
        device_t device,
        address_t addr,
        bool value)
{
    pimpl->set_slave(device);

    if (-1 == modbus_write_bit(pimpl->modbus_inst, addr, (value ? TRUE : FALSE)))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }
}

void
    ModbusClientImpl::write_single_register(
        device_t device,
        address_t addr,
        uint16_t value)
{
    pimpl->set_slave(device);

    if (-1 == modbus_write_register(pimpl->modbus_inst, addr, value))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr);

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }
}

void
    ModbusClientImpl::write_multiple_coils(
        device_t device,
        address_t addr,
        const vector<bool> &value)
{
    pimpl->set_slave(device);

    const vector<uint8_t> buf = uint8_vector_from_bool_vector(value);
    if (-1 == modbus_write_bits(pimpl->modbus_inst, addr, buf.size(), &buf.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(value.size());

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }
}

void
    ModbusClientImpl::write_multiple_registers(
        device_t device,
        address_t addr,
        const vector<uint16_t> &value)
{
    pimpl->set_slave(device);

    if (-1 == modbus_write_registers(pimpl->modbus_inst, addr, value.size(), &value.front()))
    {
        ProtocolError e;
        e << EInfoDevice(device)
            << EInfoAddress(addr)
            << EInfoQuantity(value.size());

        append_exception_info(e);

        BOOST_THROW_EXCEPTION(e);
    }
}

}
}
