#include <mxmodbus/tcp/modbus_tcp_protocol.hpp>

#include <boost/foreach.hpp>
#include <boost/range.hpp>

using boost::size;

using mxmodbus::address_t;
using mxmodbus::device_t;
using mxmodbus::exception_t;

using std::vector;

namespace
{

typedef uint8_t function_code_t;

const function_code_t READ_COILS = 0x01;
const function_code_t READ_DISCRETE_INPUTS = 0x02;
const function_code_t READ_HOLDING_REGISTERS = 0x03;
const function_code_t READ_INPUT_REGISTERS = 0x04;
const function_code_t WRITE_SINGLE_COIL = 0x05;
const function_code_t WRITE_SINGLE_REGISTER = 0x06;
const function_code_t WRITE_MULTIPLE_COILS = 0x0F;
const function_code_t WRITE_MULTIPLE_REGISTERS = 0x10;

int
    parse_read_request(
        const vector<uint8_t> &packet,
        function_code_t function_code,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity)
{
    // Insufficient packet length
    if (packet.size() < (7 + 5))
        return 0;

    size_t index = 0;

    uint16_t id_ = packet[index++];
    id_ <<= 8;
    id_ += packet[index++];

    uint16_t protocol_id = packet[index++];
    protocol_id <<= 8;
    protocol_id += packet[index++];
    if (0x0000 != protocol_id)
        return -1;

    uint16_t length = packet[index++];
    length <<= 8;
    length += packet[index++];
    if (6 != length)
        return -1;

    const uint8_t device_ = packet[index++];
    const uint8_t function_code_ = packet[index++];
    if (function_code != function_code_)
        return -1;

    uint16_t address_ = packet[index++];
    address_ <<= 8;
    address_ += packet[index++];

    uint16_t quantity_ = packet[index++];
    quantity_ <<= 8;
    quantity_ += packet[index++];

    BOOST_ASSERT(index == 12);

    id = id_;
    device = device_;
    address = address_;
    quantity = quantity_;

    return 12;
}

int
    parse_write_single_request(
        const vector<uint8_t> &packet,
        function_code_t function_code,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &value)
{
    // Insufficient packet length
    if (packet.size() < (7 + 5))
        return 0;

    size_t index = 0;

    uint16_t id_ = packet[index++];
    id_ <<= 8;
    id_ += packet[index++];

    uint16_t protocol_id = packet[index++];
    protocol_id <<= 8;
    protocol_id += packet[index++];
    if (0x0000 != protocol_id)
        return -1;

    uint16_t length = packet[index++];
    length <<= 8;
    length += packet[index++];
    if (6 != length)
        return -1;

    const uint8_t device_ = packet[index++];
    const uint8_t function_code_ = packet[index++];
    if (function_code != function_code_)
        return -1;

    uint16_t address_ = packet[index++];
    address_ <<= 8;
    address_ += packet[index++];

    uint16_t value_ = packet[index++];
    value_ <<= 8;
    value_ += packet[index++];

    BOOST_ASSERT(index == 12);

    id = id_;
    device = device_;
    address = address_;
    value = value_;

    return 12;
}

int
    parse_write_multiple_request_header(
        const vector<uint8_t> &packet,
        function_code_t function_code,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity,
        uint8_t &byte_count)
{
    // Insufficient packet length of header
    if (packet.size() < (7 + 6))
        return 0;

    size_t index = 0;

    uint16_t id_ = packet[index++];
    id_ <<= 8;
    id_ += packet[index++];

    uint16_t protocol_id = packet[index++];
    protocol_id <<= 8;
    protocol_id += packet[index++];
    if (0x0000 != protocol_id)
        return -1;

    uint16_t length = packet[index++];
    length <<= 8;
    length += packet[index++];

    const uint8_t device_ = packet[index++];
    const uint8_t function_code_ = packet[index++];
    if (function_code != function_code_)
        return -1;

    uint16_t address_ = packet[index++];
    address_ <<= 8;
    address_ += packet[index++];

    uint16_t quantity_ = packet[index++];
    quantity_ <<= 8;
    quantity_ += packet[index++];

    const uint8_t byte_count_ = packet[index++];
    if (packet.size() < static_cast<size_t>((7 + 6 + byte_count_)))
        return 0;
    if (length != (7 + byte_count_))
        return -1;

    id = id_;
    device = device_;
    address = address_;
    quantity = quantity_;
    byte_count = byte_count_;

    return 13;
}

const vector<uint8_t>
    form_read_status_response(
        function_code_t function_code,
        uint16_t id,
        device_t device,
        const vector<bool> &status)
{
    const size_t count = (status.size() + 7) / 8;
    const uint16_t length = 1 + 2 + count;

    const uint8_t HEADER[] = {
        static_cast<uint8_t>((id >> 8) & 0xFF),     // Transaction Identifier
        static_cast<uint8_t>((id >> 0) & 0xFF),
        0x00,                                       // Protocol Identifier
        0x00,
        static_cast<uint8_t>((length >> 8) & 0xFF), // Length
        static_cast<uint8_t>((length >> 0) & 0xFF),
        device,                                     // Unit Identifier
        function_code,                              // Function Code
        static_cast<uint8_t>(count)                 // Byte Count
    };

    vector<uint8_t> rtn(HEADER, HEADER + size(HEADER));

    // Status
    uint8_t temp = 0;
    for(size_t i = 0; i < status.size(); i++)
    {
        temp |= (status[i] ? 1 : 0) << (i % 8);

        if ((i % 8) == 7)
        {
            rtn.push_back(temp);
            temp = 0;
        }
    }
    if (status.size() % 8)
        rtn.push_back(temp);

    return rtn;
}

const vector<uint8_t>
    form_read_registers_response(
        function_code_t function_code,
        uint16_t id,
        device_t device,
        const vector<uint16_t> &registers)
{
    const size_t count = registers.size() * 2;
    const uint16_t length = 1 + 2 + count;

    const uint8_t HEADER[] = {
        static_cast<uint8_t>((id >> 8) & 0xFF),     // Transaction Identifier
        static_cast<uint8_t>((id >> 0) & 0xFF),
        0x00,                                       // Protocol Identifier
        0x00,
        static_cast<uint8_t>((length >> 8) & 0xFF), // Length
        static_cast<uint8_t>((length >> 0) & 0xFF),
        device,                                     // Unit Identifier
        function_code,                              // Function Code
        static_cast<uint8_t>(count)                 // Byte Count
    };

    vector<uint8_t> rtn(HEADER, HEADER + size(HEADER));

    // Registers
    BOOST_FOREACH(uint16_t reg, registers)
    {
        rtn.push_back((reg >> 8) & 0xFF);
        rtn.push_back((reg >> 0) & 0xFF);
    }

    return rtn;
}

const vector<uint8_t>
    form_write_response(
        function_code_t function_code,
        uint16_t id,
        device_t device,
        address_t address,
        uint16_t data)
{
    const uint16_t length = 1 + 5;

    const uint8_t RESPONSE[] = {
        static_cast<uint8_t>((id >> 8) & 0xFF),         // Transaction Identifier
        static_cast<uint8_t>((id >> 0) & 0xFF),
        0x00,                                           // Protocol Identifier
        0x00,
        static_cast<uint8_t>((length >> 8) & 0xFF),     // Length
        static_cast<uint8_t>((length >> 0) & 0xFF),
        device,                                         // Unit Identifier
        function_code,                                  // Function Code
        static_cast<uint8_t>((address >> 8) & 0xFF),    // Starting Address
        static_cast<uint8_t>((address >> 0) & 0xFF),
        static_cast<uint8_t>((data >> 8) & 0xFF),       // Output Value for write_single commands,
        static_cast<uint8_t>((data >> 0) & 0xFF)        //   or Quantity of Outputs for write_multiple commands.
    };

    return vector<uint8_t>(RESPONSE, RESPONSE + size(RESPONSE));
}

const vector<uint8_t>
    form_exception(
        function_code_t function_code,
        uint16_t id,
        device_t device,
        exception_t exception)
{
    const uint16_t length = 3;

    const uint8_t PACKET[] = {
        static_cast<uint8_t>((id >> 8) & 0xFF),
        static_cast<uint8_t>((id >> 0) & 0xFF),
        0x00,
        0x00,
        static_cast<uint8_t>((length >> 8) & 0xFF),
        static_cast<uint8_t>((length >> 0) & 0xFF),
        device,
        static_cast<uint8_t>(function_code | 0x80),
        exception
    };

    return vector<uint8_t>(PACKET, PACKET + size(PACKET));
}

}

namespace mxmodbus
{
namespace tcp
{

int
    parse_read_coils_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity)
{
    return parse_read_request(
        packet,
        READ_COILS,
        id,
        device,
        address,
        quantity);
}

int
    parse_read_discrete_inputs_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity)
{
    return parse_read_request(
        packet,
        READ_DISCRETE_INPUTS,
        id,
        device,
        address,
        quantity);
}

int
    parse_read_holding_registers_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity)
{
    return parse_read_request(
        packet,
        READ_HOLDING_REGISTERS,
        id,
        device,
        address,
        quantity);
}

int
    parse_read_input_registers_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &quantity)
{
    return parse_read_request(
        packet,
        READ_INPUT_REGISTERS,
        id,
        device,
        address,
        quantity);
}

int
    parse_write_single_coil_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        bool &value)
{
    uint16_t id_;
    device_t device_;
    address_t address_;
    uint16_t value_;

    const int rtn = parse_write_single_request(
        packet,
        WRITE_SINGLE_COIL,
        id_,
        device_,
        address_,
        value_);

    if (rtn <= 0)
        return rtn;

    bool rtn_value;
    if (value_ == 0xFF00)
        rtn_value = true;
    else if (value_ == 0x0000)
        rtn_value = false;
    else
        return -1;

    id = id_;
    device = device_;
    address = address_;
    value = rtn_value;

    return rtn;
}

int
    parse_write_single_register_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        uint16_t &value)
{
    return parse_write_single_request(
        packet,
        WRITE_SINGLE_REGISTER,
        id,
        device,
        address,
        value);
}

int
    parse_write_multiple_coils_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        vector<bool> &value)
{
    uint16_t id_;
    device_t device_;
    address_t address_;
    uint16_t quantity;
    uint8_t byte_count;

    const int header_size = parse_write_multiple_request_header(
        packet,
        WRITE_MULTIPLE_COILS,
        id_,
        device_,
        address_,
        quantity,
        byte_count);

    if (header_size <= 0)
        return header_size;

    if (byte_count != (quantity + 7) / 8)
        return -1;

    uint8_t index = 0;
    vector<bool> value_;
    for(size_t i = 0; i < quantity; ++i)
    {
        value_.push_back((packet[header_size + index] >> (i % 8)) & 0x01);
        if (i % 8 == 7)
            index++;
    }

    id = id_;
    device = device_;
    address = address_;
    value = value_;

    return header_size + byte_count;
}

int
    parse_write_multiple_registers_request(
        const vector<uint8_t> &packet,
        uint16_t &id,
        device_t &device,
        address_t &address,
        vector<uint16_t> &value)
{
    uint16_t id_;
    device_t device_;
    address_t address_;
    uint16_t quantity;
    uint8_t byte_count;

    const int header_size = parse_write_multiple_request_header(
        packet,
        WRITE_MULTIPLE_REGISTERS,
        id_,
        device_,
        address_,
        quantity,
        byte_count);

    if (header_size <= 0)
        return header_size;

    if (byte_count != quantity * 2)
        return -1;

    vector<uint16_t> value_;
    for(size_t i = 0; i < quantity; ++i)
    {
        value_.push_back(
            (packet[header_size + 2 * i] << 8) + packet[header_size + 2 * i + 1]
        );
    }

    id = id_;
    device = device_;
    address = address_;
    value = value_;

    return header_size + byte_count;
}

const vector<uint8_t>
    form_read_coils_response(
        uint16_t id,
        device_t device,
        const vector<bool> &status)
{
    return form_read_status_response(
        READ_COILS,
        id,
        device,
        status);
}

const vector<uint8_t>
    form_read_coils_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        READ_COILS,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_read_discrete_inputs_response(
        uint16_t id,
        device_t device,
        const vector<bool> &status)
{
    return form_read_status_response(
        READ_DISCRETE_INPUTS,
        id,
        device,
        status);
}

const vector<uint8_t>
    form_read_discrete_inputs_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        READ_DISCRETE_INPUTS,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_read_holding_registers_response(
        uint16_t id,
        device_t device,
        const vector<uint16_t> &value)
{
    return form_read_registers_response(
        READ_HOLDING_REGISTERS,
        id,
        device,
        value);
}

const vector<uint8_t>
    form_read_holding_registers_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        READ_HOLDING_REGISTERS,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_read_input_registers_response(
        uint16_t id,
        device_t device,
        const vector<uint16_t> &registers)
{
    return form_read_registers_response(
        READ_INPUT_REGISTERS,
        id,
        device,
        registers);
}

const vector<uint8_t>
    form_read_input_registers_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        READ_INPUT_REGISTERS,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_write_single_coil_response(
        uint16_t id,
        device_t device,
        address_t address,
        bool value)
{
    uint16_t data;
    if (value)
        data = 0xFF00;
    else
        data = 0x0000;

    return form_write_response(
        WRITE_SINGLE_COIL,
        id,
        device,
        address,
        data);
}

const vector<uint8_t>
    form_write_single_coil_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        WRITE_SINGLE_COIL,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_write_single_register_response(
        uint16_t id,
        device_t device,
        address_t address,
        uint16_t value)
{
    return form_write_response(
        WRITE_SINGLE_REGISTER,
        id,
        device,
        address,
        value);
}

const vector<uint8_t>
    form_write_single_register_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        WRITE_SINGLE_REGISTER,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_write_multiple_coils_response(
        uint16_t id,
        device_t device,
        address_t address,
        uint16_t quantity)
{
    return form_write_response(
        WRITE_MULTIPLE_COILS,
        id,
        device,
        address,
        quantity);
}

const vector<uint8_t>
    form_write_multiple_coils_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        WRITE_MULTIPLE_COILS,
        id,
        device,
        exception);
}

const vector<uint8_t>
    form_write_multiple_registers_response(
        uint16_t id,
        device_t device,
        address_t address,
        uint16_t quantity)
{
    return form_write_response(
        WRITE_MULTIPLE_REGISTERS,
        id,
        device,
        address,
        quantity);
}

const vector<uint8_t>
    form_write_multiple_registers_exception(
        uint16_t id,
        device_t device,
        exception_t exception)
{
    return form_exception(
        WRITE_MULTIPLE_REGISTERS,
        id,
        device,
        exception);
}

}
}
