#include <mxmqtt/Client.hpp>

#include <boost/bind.hpp>
#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>

#include <mxmqtt/init.hpp>
#include <mxmqtt/MosquittoClient.hpp>

using boost::bind;
using boost::make_shared;
using boost::shared_ptr;

using mxmqtt::Message;

using std::string;
using std::vector;

namespace
{

void
    dummy(const Message &)
{
    // do nothing
}

}

namespace mxmqtt
{

struct Client::Impl
{
    Impl()
    {
        initialize_external_library();
    }

    shared_ptr<MosquittoClient> mosquitto_client;
};

Client::Client(
    const string &host,
    int port,
    const string &client_id,
    bool clean_session)
    : pimpl(new Impl())
{
    pimpl->mosquitto_client = make_shared<MosquittoClient>(
        host,
        port,
        client_id,
        &dummy,
        vector<string>(),
        clean_session);

    try
    {
        pimpl->mosquitto_client->start();
    }
    catch(MosquittoError &e)
    {
        MXLOG_WARN(diagnostic_information(e));
        pimpl->mosquitto_client.reset();
    }
}

Client::Client(
    const string &host,
    int port,
    const string &client_id,
    const OnMessageCallback &omc,
    const vector<string> &subscribe_topic_list,
    bool clean_session)
    : pimpl(new Impl())
{
    pimpl->mosquitto_client = make_shared<MosquittoClient>(
        host,
        port,
        client_id,
        omc,
        subscribe_topic_list,
        clean_session);

    try
    {
        pimpl->mosquitto_client->start();
    }
    catch(MosquittoError &e)
    {
        MXLOG_WARN(diagnostic_information(e));
        pimpl->mosquitto_client.reset();
    }
}

Client::~Client()
{
}

void
    Client::loop(
        int timeout_ms)
{
    if (!pimpl->mosquitto_client)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    try
    {
        pimpl->mosquitto_client->run_once(timeout_ms);
    }
    catch(IllegalStateError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        throw;
    }
}

bool
    Client::is_valid() const
{
    if (!pimpl->mosquitto_client)
        return false;

    return pimpl->mosquitto_client->is_valid();
}

bool
    Client::is_connected() const
{
    if (!pimpl->mosquitto_client)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    return pimpl->mosquitto_client->is_connected();
}

message_id_t
    Client::publish(
        const Message &msg)
{
    if (!pimpl->mosquitto_client)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    try
    {
        return pimpl->mosquitto_client->publish_message(msg);
    }
    catch(IllegalStateError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        throw;
    }
}

bool
    Client::is_publish_complete(
        message_id_t mid) const
{
    if (!pimpl->mosquitto_client)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    try
    {
        return pimpl->mosquitto_client->is_publish_complete(mid);
    }
    catch(IllegalStateError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        throw;
    }
}

}
