#include <mxmqtt/MosquittoClient.hpp>

#include <set>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/scope_exit.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <mxutil/log.hpp>

using boost::lexical_cast;
using boost::recursive_mutex;

using std::set;
using std::string;
using std::vector;

namespace mxmqtt
{

struct MosquittoClient::Impl
{
    Impl(
        const string &host_,
        int port_,
        const OnMessageCallback &omc_,
        const vector<string> &subscribe_topic_list_)
        : host(host_)
        , port(port_)
        , omc(omc_)
        , subscribe_topic_list(subscribe_topic_list_)
        , is_valid(true)
        , is_connected(false)
    {
    }

    const string            host;
    const int               port;
    const OnMessageCallback omc;
    const vector<string>    subscribe_topic_list;

    vector<int>             subscribe_mid_list;

    set<int>                publish_mid_list;

    bool    is_valid;
    bool    is_connected;

    recursive_mutex     lock;
};

MosquittoClient::MosquittoClient(
    const string &host,
    int port,
    const string &client_id,
    const OnMessageCallback &omc,
    const vector<string> &subscribe_topic_list,
    bool clean_session)
    : mosquittopp(client_id.c_str(), clean_session)
    , pimpl(new Impl(
        host,
        port,
        omc,
        subscribe_topic_list))
{
}

MosquittoClient::~MosquittoClient()
{
}

void
    MosquittoClient::start()
{
    if (!pimpl->is_valid)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    recursive_mutex::scoped_lock lock(pimpl->lock);

    const int rc = this->connect_async(
        pimpl->host.c_str(),
        pimpl->port,
        5);

    if (rc != MOSQ_ERR_SUCCESS)
    {
        pimpl->is_valid = false;

        BOOST_THROW_EXCEPTION(
            MosquittoError()
                << MosquittoErrCode(rc)
                << ErrMsg("connect_async")
        );
    }
}

void
    MosquittoClient::run_once(
        int timeout_ms)
{
    if (!pimpl->is_valid)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    recursive_mutex::scoped_lock lock(pimpl->lock);

    const int rc = this->loop(timeout_ms, 1);
    if (rc != MOSQ_ERR_SUCCESS)
    {
        pimpl->is_valid = false;

        BOOST_THROW_EXCEPTION(
            MosquittoError()
                << MosquittoErrCode(rc)
                << ErrMsg("connect_async")
        );
    }
}

bool
    MosquittoClient::is_valid() const
{
    recursive_mutex::scoped_lock lock(pimpl->lock);

    return pimpl->is_valid;
}

bool
    MosquittoClient::is_connected() const
{
    recursive_mutex::scoped_lock lock(pimpl->lock);

    return pimpl->is_connected;
}

void
    MosquittoClient::on_connect(
        int rc)
{
    MXLOG_DBUG("on_connect");

    if (0 != rc)
    {
        MXLOG_WARN("connect failed: rc = " + lexical_cast<string>(rc));

        pimpl->is_valid = false;
        return;
    }

    pimpl->is_connected = true;

    for(size_t ax = 0; ax < pimpl->subscribe_topic_list.size(); ++ax)
    {
        const int result = this->subscribe(
            &pimpl->subscribe_mid_list[ax],
            pimpl->subscribe_topic_list[ax].c_str(),
            2);

        if (result != MOSQ_ERR_SUCCESS)
        {
            MXLOG_WARN("subscribe failed");

            pimpl->is_valid = false;
            break;
        }
    }
}

void
    MosquittoClient::on_disconnect(
        int rc)
{
    MXLOG_DBUG("on_disconnect");

    if (0 == rc) // this->disconnect() is called
    {
        return;
    }

    MXLOG_WARN("disconnected: rc = " + lexical_cast<string>(rc));

    pimpl->is_valid = false;
}

void
    MosquittoClient::on_subscribe(
        int mid,
        int,
        const int *)
{
    MXLOG_DBUG("on_subscribe");

    vector<int>::iterator iter = find(
        pimpl->subscribe_mid_list.begin(),
        pimpl->subscribe_mid_list.end(),
        mid);

    if (iter == pimpl->subscribe_mid_list.end())
    {
        MXLOG_WARN("unexpected mid: " + lexical_cast<string>(mid));
    }
    else
    {
        *iter = -1;
    }
}

void
    MosquittoClient::on_message(
        const struct mosquitto_message *msg)
{
    MXLOG_DBUG("on_message");

    const vector<uint8_t> payload(
        reinterpret_cast<const uint8_t *>(msg->payload),
        reinterpret_cast<const uint8_t *>(msg->payload) + msg->payloadlen);

    pimpl->omc(Message(
        msg->topic,
        payload,
        msg->qos,
        msg->retain));
}

void
    MosquittoClient::on_publish(
        int mid)
{
    set<int>::iterator iter = pimpl->publish_mid_list.find(mid);
    if (iter == pimpl->publish_mid_list.end())
    {
        MXLOG_DBUG("unexpected mid = " + lexical_cast<string>(mid));

        return;
    }

    pimpl->publish_mid_list.erase(iter);
}

message_id_t
    MosquittoClient::publish_message(
        const Message &msg)
{
    if (!pimpl->is_valid)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    recursive_mutex::scoped_lock lock(pimpl->lock);

    if (!pimpl->is_connected)
        BOOST_THROW_EXCEPTION(
            NotConnectedError());

    int mid;
    const int rc = this->publish(
        &mid,
        msg.topic().c_str(),
        msg.payload().size(),
        &msg.payload()[0],
        msg.qos(),
        msg.retain());

    if (rc != MOSQ_ERR_SUCCESS)
    {
        pimpl->is_valid = false;

        BOOST_THROW_EXCEPTION(
            MosquittoError()
                << MosquittoErrCode(rc)
                << ErrMsg("publish")
        );
    }

    const set<int>::const_iterator iter = pimpl->publish_mid_list.find(mid);
    if (iter != pimpl->publish_mid_list.end())
    {
        pimpl->is_valid = false;
        MXLOG_WARN("duplicated mid " + lexical_cast<string>(mid));

        BOOST_THROW_EXCEPTION(
            IllegalStateError()
                << ErrMsg("duplicated mid " + lexical_cast<string>(mid))
        );
    }

    if (msg.qos() >= 1)
        pimpl->publish_mid_list.insert(mid);

    return mid;
}

bool
    MosquittoClient::is_publish_complete(
        int mid) const
{
    if (!pimpl->is_valid)
        BOOST_THROW_EXCEPTION(
            IllegalStateError());

    recursive_mutex::scoped_lock lock(pimpl->lock);

    const set<int>::const_iterator iter = pimpl->publish_mid_list.find(mid);
    return iter == pimpl->publish_mid_list.end();
}

}
