#include <mxmqtt/Worker.hpp>

#include <boost/bind.hpp>
#include <boost/chrono/chrono.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/weak_ptr.hpp>

#include <mxutil/log.hpp>

#include <mxmqtt/Client.hpp>

using boost::bind;
using boost::condition_variable;
using boost::chrono::milliseconds;
using boost::chrono::steady_clock;
using boost::function;
using boost::get_system_time;
using boost::lexical_cast;
using boost::make_shared;
using boost::mutex;
using boost::posix_time::ptime;
using boost::shared_ptr;
using boost::this_thread::interruption_point;
using boost::thread;
using boost::thread_interrupted;
using boost::weak_ptr;

using mxmqtt::Client;
using mxmqtt::Message;
using mxmqtt::message_id_t;
using mxmqtt::Worker;

using std::max;
using std::string;
using std::vector;

namespace
{

typedef
    function<
        shared_ptr<Client>
            (void)
    >
    CreateClient;

shared_ptr<Client>
    create_client(
        const string &host,
        int port,
        const string &client_id,
        bool clean_session)
{
    return make_shared<Client>(
        host,
        port,
        client_id,
        clean_session);
}

shared_ptr<Client>
    create_client(
        const string &host,
        int port,
        const string &client_id,
        const Worker::OnMessageCallback &omc,
        const vector<string> &subscribe_topic_list,
        bool clean_session)
{
    return make_shared<Client>(
        host,
        port,
        client_id,
        omc,
        subscribe_topic_list,
        clean_session);
}

class PublishCommand
    : private boost::noncopyable
{
public:
    enum status_t
    {
        STATUS_WAITING,
        STATUS_RUNNING,
        STATUS_SUCCESS,
        STATUS_FAILURE
    };

public:
    explicit PublishCommand(
        const Message &msg)
        : msg_(msg)
        , status_(STATUS_WAITING)
    {
    }

    void
        run(Client &client)
    {
        try
        {
            status_ = STATUS_RUNNING;
            cond_.notify_all();

            const message_id_t mid = client.publish(msg_);

            while(!client.is_publish_complete(mid))
            {
                interruption_point();
                client.loop(0);
            }

            status_ = STATUS_SUCCESS;
            cond_.notify_all();
        }
        catch(...)
        {
            status_ = STATUS_FAILURE;
            cond_.notify_all();

            throw;
        }
    }

    status_t
        status() const
    {
        return status_;
    }

    void
        wait_complete(int timeout_ms)
    {
        const ptime expired = get_system_time() + boost::posix_time::milliseconds(timeout_ms);
        mutex::scoped_lock lock(mutex_);

        while(status_ <= STATUS_RUNNING)
        {
            cond_.timed_wait(lock, expired);

            if (get_system_time() >= expired)
                break;
        }
    }

private:
    const Message   msg_;
    status_t        status_;

    mutex               mutex_;
    condition_variable  cond_;
};

}

namespace mxmqtt
{

struct Worker::Impl
{
    Impl()
    {
    }

    CreateClient    create_client_callback;

    shared_ptr<Client>  client;

    shared_ptr<thread>  thread_object;

    vector<shared_ptr<PublishCommand> > command_queue;
    mutex                               command_queue_mutex;
    condition_variable                  command_queue_cond;

    void
        worker_thread()
    {
        try
        {
            while(true)
            {
                interruption_point();

                try
                {
                    this->client = this->create_client_callback();

                    while(true)
                    {
                        interruption_point();

                        this->client->loop(0);

                        if (!this->client->is_valid())
                        {
                            usleep(500 * 1000);
                            break;
                        }

                        if (!this->client->is_connected())
                            continue;

                        shared_ptr<PublishCommand> ptr = this->pop_command();
                        if (NULL == ptr)
                            continue;

                        ptr->run(*this->client);
                    }

                    this->client.reset();
                }
                catch(mxmqtt::Error &e)
                {
                    MXLOG_WARN(diagnostic_information(e));
                    this->client.reset();

                    usleep(500 * 1000);
                }
            }
        }
        catch(thread_interrupted &e)
        {
            MXLOG_INFO("terminating");
        }
    }

    void
        add_command(shared_ptr<PublishCommand> cmd)
    {
        BOOST_ASSERT(cmd != NULL);

        mutex::scoped_lock lock(this->command_queue_mutex);

        this->command_queue.push_back(cmd);

        this->command_queue_cond.notify_one();
    }

    shared_ptr<PublishCommand>
        pop_command()
    {
        mutex::scoped_lock lock(this->command_queue_mutex);

        if (this->command_queue.empty())
        {
            command_queue_cond.timed_wait(lock, boost::posix_time::seconds(3));

            if (this->command_queue.empty())
                return shared_ptr<PublishCommand>();
        }

        shared_ptr<PublishCommand> ptr = this->command_queue.front();
        this->command_queue.erase(this->command_queue.begin());

        return ptr;
    }

    void
        remove_command(shared_ptr<PublishCommand> cmd)
    {
        mutex::scoped_lock lock(this->command_queue_mutex);

        const vector<shared_ptr<PublishCommand> >::iterator iter = find(
            this->command_queue.begin(),
            this->command_queue.end(),
            cmd);

        if (iter == this->command_queue.end())
            return;

        this->command_queue.erase(iter);
    }
};

Worker::Worker(
    const string &host,
    int port,
    const string &client_id,
    bool clean_session)
    : pimpl(new Impl())
{
    pimpl->create_client_callback = bind(
        &create_client,
        host,
        port,
        client_id,
        clean_session);

    pimpl->thread_object.reset(new thread(bind(
        &Impl::worker_thread,
        &*pimpl
    )));
}

Worker::Worker(
    const string &host,
    int port,
    const string &client_id,
    const OnMessageCallback &omc,
    const vector<string> &subscribe_topic_list,
    bool clean_session)
    : pimpl(new Impl())
{
    pimpl->create_client_callback = bind(
        &create_client,
        host,
        port,
        client_id,
        omc,
        subscribe_topic_list,
        clean_session);

    pimpl->thread_object.reset(new thread(bind(
        &Impl::worker_thread,
        &*pimpl
    )));
}

Worker::~Worker()
{
    pimpl->thread_object->interrupt();
    pimpl->thread_object->join();
}

void
    Worker::publish(
        const Message &msg,
        int timeout_ms)
{
    BOOST_ASSERT(timeout_ms > 0);

    const shared_ptr<PublishCommand> cmd = make_shared<PublishCommand>(msg);

    pimpl->add_command(cmd);

    cmd->wait_complete(timeout_ms);

    if (cmd->status() != PublishCommand::STATUS_SUCCESS)
    {
        pimpl->remove_command(cmd);

        BOOST_THROW_EXCEPTION(
            Error() << ErrMsg("publish failed")
        );
    }
}

}
