#include <mxtag/log/JsonSerializer.hpp>

#include <sstream>

#include <boost/algorithm/string/join.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <rapidjson/document.h>

#include <mxtag/tag_form.hpp>
#include <mxtag/tag_parse.hpp>

using boost::algorithm::join;
using boost::lexical_cast;
using boost::posix_time::ptime;

using mxtag::form_iso8601_time;
using mxtag::parse_iso8601_time;

using rapidjson::Document;

using std::map;
using std::set;
using std::string;
using std::stringstream;
using std::vector;

namespace mxtag
{
namespace log
{

JsonSerializer::JsonSerializer()
{
}

JsonSerializer::~JsonSerializer()
{
}

const vector<uint8_t>
    JsonSerializer::do_from_tag(const Tag &t) const
{
    stringstream ss;
    ss << '{'
        << "\"equ\":\"" << t.equipment() << "\","
        << "\"at\":\"" << form_iso8601_time(t.at()) << "\","
        << "\"tag\":\"" << t.tag() << "\",";

    if (!t.unit().empty())
        ss << "\"unit\":\"" << t.unit() << "\",";

    ss << "\"value\":";
    if (t.value().is_string())
        ss << '\"' << t.value().as_string() << '\"';
    else if (t.value().is_int())
        ss << lexical_cast<string>(t.value().as_int());
    else if (t.value().is_float())
        ss << lexical_cast<string>(t.value().as_float());
    ss << '}';

    const string ss_str(ss.str());

    return vector<uint8_t>(ss_str.begin(), ss_str.end());
}

const mxtag::Tag
    JsonSerializer::do_to_tag(const vector<uint8_t> &v) const
{
    string content(v.begin(), v.end());
    Document doc;
    doc.Parse(content.c_str());
    if (!doc.IsObject())
        BOOST_THROW_EXCEPTION(ParseError());

    if (!doc.HasMember("equ") ||
        !doc.HasMember("tag") ||
        !doc.HasMember("at") ||
        !doc.HasMember("value"))
        BOOST_THROW_EXCEPTION(ParseError());

    if (!doc["equ"].IsString() ||
        !doc["tag"].IsString() ||
        !doc["at"].IsString())
        BOOST_THROW_EXCEPTION(ParseError());

    rapidjson::Value &at_ = doc["at"];
    ptime at;
    if (!parse_iso8601_time(string(at_.GetString(), at_.GetString() + at_.GetStringLength()), at))
        BOOST_THROW_EXCEPTION(ParseError());

    Value value("");
    rapidjson::Value &value_ = doc["value"];
    if (value_.IsString())
        value = Value(value_.GetString());
    else if (value_.IsInt64())
        value = Value(value_.GetInt64());
    else if (value_.IsDouble())
        value = Value(value_.GetDouble());
    else
        BOOST_THROW_EXCEPTION(ParseError());

    string unit;
    if (doc.HasMember("unit"))
        if (doc["unit"].IsString())
            unit = doc["unit"].GetString();

    rapidjson::Value &equ_ = doc["equ"];
    rapidjson::Value &tag_ = doc["tag"];
    return Tag(
        string(equ_.GetString(), equ_.GetString() + equ_.GetStringLength()),
        string(tag_.GetString(), tag_.GetString() + tag_.GetStringLength()),
        at,
        value,
        unit);
}

const vector<uint8_t>
    JsonSerializer::do_from_tag_list(const map<string, set<string> > &tag_list) const
{
    vector<string> equ_list;
    typedef map<string, set<string> >::value_type vtype;
    BOOST_FOREACH(const vtype &e, tag_list)
    {
        const string &equipment = e.first;
        const set<string> &tag_list_ = e.second;

        stringstream equ;
        equ << "{\"equ\": \"" << equipment << '\"'
            << ", \"tagList\":[";

        vector<string> quoted_tag_list;
        BOOST_FOREACH(const string &t, tag_list_)
        {
            quoted_tag_list.push_back('\"' + t + '\"');
        }

        equ << join(quoted_tag_list, ",")
            << "]}";

        equ_list.push_back(equ.str());
    }

    stringstream lh;
    lh << "{\"equList\":[" << join(equ_list, ",") << "]}\n";
    const string lh_str(lh.str());

    return vector<uint8_t>(lh_str.begin(), lh_str.end());
}

const map<string, set<string> >
    JsonSerializer::do_to_tag_list(const vector<uint8_t> &v) const
{
    string content(v.begin(), v.end());
    Document doc;
    doc.Parse(content.c_str());
    if (!doc.IsObject())
        BOOST_THROW_EXCEPTION(ParseError());

    if (!doc.HasMember("equList"))
        BOOST_THROW_EXCEPTION(ParseError());

    const rapidjson::Value &equ_list = doc["equList"];
    if (!equ_list.IsArray())
        BOOST_THROW_EXCEPTION(ParseError());

    map<string, set<string> > rtn;
    for(rapidjson::Value::ConstValueIterator equ_iter = equ_list.Begin();
        equ_iter != equ_list.End();
        ++equ_iter)
    {
        if (!equ_iter->HasMember("equ"))
            BOOST_THROW_EXCEPTION(ParseError());

        const rapidjson::Value &equ_ = (*equ_iter)["equ"];
        if (!equ_.IsString())
            BOOST_THROW_EXCEPTION(ParseError());

        if (!equ_iter->HasMember("tagList"))
            BOOST_THROW_EXCEPTION(ParseError());

        const rapidjson::Value &tag_list_ = (*equ_iter)["tagList"];
        if (!tag_list_.IsArray())
            BOOST_THROW_EXCEPTION(ParseError());

        set<string> tag_list;
        for (rapidjson::Value::ConstValueIterator tag_iter = tag_list_.Begin();
            tag_iter != tag_list_.End();
            ++tag_iter)
        {
            const rapidjson::Value &tag = *tag_iter;
            if (!tag.IsString())
                BOOST_THROW_EXCEPTION(ParseError());

            tag_list.insert(tag.GetString());
        }

        rtn[equ_.GetString()] = tag_list;
    }

    return rtn;
}

}
}
