#include <mxtag/mqtt/JsonSerializer.hpp>

#include <sstream>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>

#include <mxtag/tag_form.hpp>
#include <mxtag/mqtt/tag_parse.hpp>

#include <mxtag/Value.hpp>

using boost::lexical_cast;
using boost::posix_time::ptime;
using boost::regex;

using mxtag::form_iso8601_time;
using mxtag::mqtt::parse_topic;
using mxtag::mqtt::parse_payload;

using mxtag::ParseError;
using mxtag::Value;

using std::make_pair;
using std::pair;
using std::string;
using std::stringstream;
using std::vector;

namespace
{
const Value
    value_from_string(const string &str)
{
    static regex e_int(
            "^[-]*[0-9]*$"
            );
    static regex e_float(
            "^[0-9]*\\.[0-9]*$"
            );
    static regex e_string(
            "^\".*\"$"
            );

    if (regex_match(str, e_int))
        return Value(lexical_cast<int64_t>(str));
    else if (regex_match(str, e_float))
        return Value(lexical_cast<double>(str));
    else if (regex_match(str, e_string))
        return Value(str);

    BOOST_THROW_EXCEPTION(ParseError());
}
}

namespace mxtag
{
namespace mqtt
{

JsonSerializer::JsonSerializer()
{
}

JsonSerializer::~JsonSerializer()
{
}

const pair<string, vector<uint8_t> >
    JsonSerializer::do_from_tag(const Tag &t) const
{
    stringstream ss;
    ss << '{'
        << "\"at\":\"" << form_iso8601_time(t.at()) << "\",";

    if (!t.unit().empty())
        ss << "\"unit\":\"" << t.unit() << "\",";

    ss << "\"value\":";
    if (t.value().is_string())
        ss << '\"' << t.value().as_string() << '\"';
    else if (t.value().is_int())
        ss << lexical_cast<string>(t.value().as_int());
    else if (t.value().is_float())
        ss << lexical_cast<string>(t.value().as_float());
    ss << '}';

    const string ss_str(ss.str());

    return make_pair<string, vector<uint8_t> >(
        "/equs/" + t.equipment() + "/tags/" + t.tag(),
        vector<uint8_t>(ss_str.begin(), ss_str.end()));
}

const Tag
    JsonSerializer::do_to_tag(
        const string &topic,
        const vector<uint8_t> &payload) const
{
    string equipment;
    string tag;
    if (!parse_topic(topic, equipment, tag))
        BOOST_THROW_EXCEPTION(ParseError());

    ptime at;
    string value;
    string unit;
    if (!parse_payload(string(payload.begin(), payload.end()), at, value, unit))
        BOOST_THROW_EXCEPTION(ParseError());

    const Value value_ = value_from_string(value);

    return Tag(
        equipment,
        tag,
        at,
        value_,
        unit);
}

}
}
