#include <mxtag/mqtt/TagSubscriber.hpp>

#include <boost/algorithm/string/join.hpp>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

#include <mxtag/mqtt/ProtobufSerializer.hpp>

#include <mxmqtt/Client.hpp>

using boost::algorithm::join;
using boost::bind;
using boost::make_shared;
using boost::scoped_ptr;
using boost::shared_ptr;

using mxmqtt::Client;
using mxmqtt::Message;

using mxtag::mqtt::ProtobufSerializer;
using mxtag::mqtt::Serializer;

using std::map;
using std::set;
using std::string;
using std::vector;

namespace
{

const vector<string>
    topic_list_from_eq_tag_list(
        const map<string, set<string> > &eq_tag_list)
{
    vector<string> topic_list;

    typedef map<string, set<string> >::value_type vtype;
    BOOST_FOREACH(const vtype & p, eq_tag_list)
    {
        const string &equipment = p.first;

        BOOST_FOREACH(const string &tag, p.second)
        {
            topic_list.push_back(
                "/equs/" + equipment + "/tags/" + tag
            );
        }
    }

    return topic_list;
}

}

namespace mxtag
{
namespace mqtt
{

struct TagSubscriber::Impl
{
    Impl(
        const string &host_,
        int port_,
        const string &client_id_,
        bool clean_session_,
        const map<string, set<string> > &eq_tag_list_,
        const ProcessTagCallback &process_tag_)
        : host(host_)
        , port(port_)
        , client_id(client_id_)
        , clean_session(clean_session_)
        , eq_tag_list(eq_tag_list_)
        , process_tag(process_tag_)
        , serializer(new ProtobufSerializer())
    {
    }

    const string                    host;
    const int                       port;
    const string                    client_id;
    const bool                      clean_session;
    const map<string, set<string> > eq_tag_list;
    const ProcessTagCallback        process_tag;

    scoped_ptr<Serializer>          serializer;

    shared_ptr<Client>              client;

    void
        process_message(const Message &msg)
    {
        MXLOG_DBUG("topic = " + msg.topic() + ", payload = " + string(msg.payload().begin(), msg.payload().end()));

        try
        {
            const mxtag::Tag t = this->serializer->to_tag(
                msg.topic(),
                msg.payload());

            try
            {
                this->process_tag(t);
            }
            catch(boost::exception &e)
            {
                MXLOG_EROR(diagnostic_information(e));
                throw;
            }
            catch(std::exception &e)
            {
                MXLOG_EROR(string("std::exception: ") + e.what());
                throw;
            }
            catch(...)
            {
                MXLOG_EROR("unexpected exception caught");
                throw;
            }
        }
        catch(mxtag::ParseError &e)
        {
            MXLOG_WARN(diagnostic_information(e));
        }
    }
};

TagSubscriber::TagSubscriber(
    const string &host,
    int port,
    const string &client_id,
    bool clean_session,
    const map<string, set<string> > &eq_tag_list,
    const ProcessTagCallback &process_tag)
    : pimpl(new Impl(
        host,
        port,
        client_id,
        clean_session,
        eq_tag_list,
        process_tag
    ))
{
    MXLOG_DBUG("host = " + host);
    MXLOG_DBUG("topics = " + join(topic_list_from_eq_tag_list(pimpl->eq_tag_list), ", "));
}

TagSubscriber::~TagSubscriber()
{
}

void
    TagSubscriber::loop(int timeoutMS)
{
    try
    {
        if (!pimpl->client)
        {
            pimpl->client = make_shared<Client>(
                pimpl->host,
                pimpl->port,
                pimpl->client_id,
                bind(
                    &Impl::process_message,
                    &*pimpl,
                    _1),
                topic_list_from_eq_tag_list(pimpl->eq_tag_list),
                pimpl->clean_session
            );
        }
        else
        {
            pimpl->client->loop(
                timeoutMS);
        }
    }
    catch(mxmqtt::Error &e)
    {
        MXLOG_WARN(diagnostic_information(e));
        pimpl->client.reset();

        usleep(timeoutMS * 1000 / 2);
    }
}

}
}
