#include <algorithm>
#include <fstream>
#include <iostream>
#include <iterator>
#include <limits>
#include <string>
#include <vector>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/random.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <mxtag/Tag.hpp>
#include <mxtag/Value.hpp>
#include <mxtag/log/EnvelopedOStream.hpp>
#include <mxtag/log/JsonSerializer.hpp>
#include <mxtag/log/ProtobufSerializer.hpp>
#include <mxtag/log/Serializer.hpp>

using boost::lexical_cast;
using boost::mt19937;
using boost::posix_time::ptime;
using boost::posix_time::seconds;
using boost::posix_time::time_from_string;
using boost::size;
using boost::shared_ptr;
using boost::uniform_int;
using boost::uniform_real;
using boost::variate_generator;

using mxtag::Tag;
using mxtag::Value;
using mxtag::log::EnvelopedOStream;
using mxtag::log::JsonSerializer;
using mxtag::log::ProtobufSerializer;
using mxtag::log::Serializer;

using std::cerr;
using std::copy;
using std::cout;
using std::endl;
using std::map;
using std::numeric_limits;
using std::ostream_iterator;
using std::ofstream;
using std::set;
using std::string;
using std::vector;

namespace
{

const map<string, set<string> >
    enumerate_tag_list(
        const vector<string> &equ_base_name,
        int max_equ_index,
        const vector<string> &tag_base_name,
        int max_tag_index)
{
    map<string, set<string> > tag_list;
    BOOST_FOREACH(const string &equ_base, equ_base_name)
    {
        for(int equ_index = 0; equ_index < max_equ_index; ++equ_index)
        {
            const string equ = equ_base + "-" + lexical_cast<string>(equ_index);

            BOOST_FOREACH(const string &tag_base, tag_base_name)
            {
                for(int tag_index = 0; tag_index < max_tag_index; ++tag_index)
                {
                    const string tag = tag_base + "-" + lexical_cast<string>(tag_index);
                    tag_list[equ].insert(tag);
                }
            }
        }
    }

    return tag_list;
}

}

int
    main(
        int argc,
        const char *argv[])
{
    if (argc != 4)
    {
        cerr << "usage: <loop-count> <seed-number> <json|protobuf>" << endl;
        return 1;
    }

    JsonSerializer json_serializer;
    ProtobufSerializer protobuf_serializer;

    const int loop_count = lexical_cast<int>(argv[1]);
    const int seed_number = lexical_cast<int>(argv[2]);
    const string format = argv[3];

    const shared_ptr<Serializer> serializer =
        format == "json" ? shared_ptr<Serializer>(new JsonSerializer) :
        format == "protobuf" ? shared_ptr<Serializer>(new ProtobufSerializer) :
        shared_ptr<Serializer>();

    if (!serializer)
    {
        cerr << "unexpected format: " << format << endl;
        return 1;
    }

    const string equ_base_name[] = {
        "UC-8100", "DA-662A", "W325A", "IA-241", "V2403A",
        "MD-124", "MPC-2240", "EXPC-1519"};
    const string tag_base_name[] = {
        "di", "do", "ai", "ao", "ri",
        "rtdi", "tci", "tag"};
    const string unit_name[] = {"Volt", "Amp", "Walt", "C", "F"};

    // create tag-list
    const map<string, set<string> > tag_list = enumerate_tag_list(
        vector<string>(equ_base_name, equ_base_name + size(equ_base_name)),
        8,
        vector<string>(tag_base_name, tag_base_name + size(tag_base_name)),
        8);

    // Generate random function
    variate_generator<mt19937, uniform_int<> >
        rand_in_0_8(mt19937(seed_number), uniform_int<>(0, 7));
    variate_generator<mt19937, uniform_int<> >
        rand_in_0_5(mt19937(seed_number), uniform_int<>(0, 4));
    variate_generator<mt19937, uniform_int<> >
        rand_in_int(
            mt19937(seed_number),
            uniform_int<>(numeric_limits<int>::min(), numeric_limits<int>::max()));
    variate_generator<mt19937, uniform_real<> >
        rand_in_double(
            mt19937(seed_number),
            uniform_real<>(numeric_limits<double>::min(), numeric_limits<double>::max()));

    EnvelopedOStream out(cout);
    out.write(serializer->from_tag_list(tag_list));

    for(int ax = 0; ax < loop_count; ++ax)
    {
        const string equ_name = string(equ_base_name[rand_in_0_8()]) + "-" + lexical_cast<string>(rand_in_0_8());
        const string tag_name = string(tag_base_name[rand_in_0_8()]) + "-" + lexical_cast<string>(rand_in_0_8());
        const string unit = string(unit_name[rand_in_0_5()]);
        const ptime at = time_from_string("1970-01-01 12:34:56.789") + seconds(ax);

        shared_ptr<Value> value;
        switch(rand_in_0_8() % 3)
        {
            case 0:
                // integer
                value.reset(new Value(lexical_cast<int64_t>(rand_in_int())));
                break;

            case 1:
                // double
                value.reset(new Value(lexical_cast<double>(rand_in_double())));
                break;

            case 2:
                // string
                value.reset(new Value(equ_base_name[rand_in_0_8()]));
                break;

            default:
                value.reset(new Value(""));
                break;
        }

        const Tag t(
            equ_name,
            tag_name,
            at,
            *value,
            unit);

        out.write(serializer->from_tag(t));
    }

    return 0;
}
