#include <boost/test/unit_test.hpp>

#include <mxtag/log/ProtobufSerializer.hpp>

using boost::posix_time::time_from_string;

using std::map;
using std::set;
using std::string;
using std::vector;

namespace
{

struct TestFixture
{
    mxtag::log::ProtobufSerializer  serializer;
};

}

namespace mxtag
{
namespace log
{

BOOST_FIXTURE_TEST_CASE(test_ProtobufSerializer_tag_Conversion_WithIntValue_ShouldPass, TestFixture)
{
    // arrange
    const Tag t(
        "Test_Equipment",
        "Test_Tag_0",
        time_from_string("1970-01-01 12:34:56.789"),
        Value(static_cast<int64_t>(-123456)),
        "fake-unit");

    // act
    const vector<uint8_t> pb_value = serializer.from_tag(t);
    const Tag r = serializer.to_tag(pb_value);

    // assert
    BOOST_CHECK_EQUAL("Test_Equipment", r.equipment());
    BOOST_CHECK_EQUAL("Test_Tag_0", r.tag());
    BOOST_CHECK_EQUAL(time_from_string("1970-01-01 12:34:56.789"), r.at());
    BOOST_CHECK(Value(static_cast<int64_t>(-123456)) == r.value());
    BOOST_CHECK_EQUAL("fake-unit", r.unit());
}

BOOST_FIXTURE_TEST_CASE(test_ProtobufSerializer_tag_Conversion_WithFloatValue_ShouldPass, TestFixture)
{
    // arrange
    const Tag t(
        "Test_Equipment",
        "Test_Tag_0",
        time_from_string("1970-01-01 12:34:56.789"),
        Value(123.456),
        "fake-unit");

    // act
    const vector<uint8_t> pb_value = serializer.from_tag(t);
    const Tag r = serializer.to_tag(pb_value);

    // assert
    BOOST_REQUIRE(r.value().is_float());
    BOOST_CHECK_CLOSE(123.456, r.value().as_float(), 0.0001);
}

BOOST_FIXTURE_TEST_CASE(test_ProtobufSerializer_tag_Conversion_WithStrValue_ShouldPass, TestFixture)
{
    // arrange
    const Tag t(
        "Test_Equipment",
        "Test_Tag_0",
        time_from_string("1970-01-01 12:34:56.789"),
        Value("test-string"),
        "fake-unit");

    // act
    const vector<uint8_t> pb_value = serializer.from_tag(t);
    const Tag r = serializer.to_tag(pb_value);

    // assert
    BOOST_REQUIRE(r.value().is_string());
    BOOST_CHECK_EQUAL("test-string", r.value().as_string());
}

BOOST_FIXTURE_TEST_CASE(test_ProtobufSerializer_tag_list_Conversion_ShouldPass, TestFixture)
{
    // arrange
    map<string, set<string> > tag_list;
    {
        tag_list["Test_Equipment"].insert("Test_Tag_0");
    }

    // act
    const vector<uint8_t> v = serializer.from_tag_list(tag_list);
    const map<string, set<string> > r = serializer.to_tag_list(v);

    // assert
    BOOST_CHECK(tag_list == r);
}

}
}
