#include <boost/test/unit_test.hpp>

#include <mxtag/mqtt/tag_parse.hpp>

#include <boost/lexical_cast.hpp>

using boost::lexical_cast;
using boost::posix_time::ptime;
using boost::posix_time::time_from_string;

using std::string;

namespace
{

class TestFixture
{
};

}

namespace mxtag
{
namespace mqtt
{

BOOST_FIXTURE_TEST_CASE(test_parse_topic_ShouldPass, TestFixture)
{
    // arrange
    const string SUT = "/equs/the equipment id/tags/123-tag-id";

    // act
    string equipment;
    string tag;
    const bool match = parse_topic(
        SUT,
        equipment,
        tag);

    // assert
    BOOST_REQUIRE(match);
    BOOST_CHECK_EQUAL("the equipment id", equipment);
    BOOST_CHECK_EQUAL("123-tag-id", tag);
}

BOOST_FIXTURE_TEST_CASE(test_parse_topic_WithShortEquipmentTagShouldPass, TestFixture)
{
    // arrange
    const string SUT = "/equs/e/tags/1";

    // act
    string equipment;
    string tag;
    const bool match = parse_topic(
        SUT,
        equipment,
        tag);

    // assert
    BOOST_REQUIRE(match);
    BOOST_CHECK_EQUAL("e", equipment);
    BOOST_CHECK_EQUAL("1", tag);
}

BOOST_FIXTURE_TEST_CASE(test_parse_topic_WithFaultyTopic_ShouldReturnFalse, TestFixture)
{
    // arrange
    const string SUT = "/equs/the equipment id/status";

    // act
    string equipment;
    string tag;
    const bool match = parse_topic(
        SUT,
        equipment,
        tag);

    // assert
    BOOST_CHECK(!match);
}

BOOST_FIXTURE_TEST_CASE(test_parse_topic_WithFaultyEquipment_ShouldReturnFalse, TestFixture)
{
    // arrange
    const string SUT = "/equs/the equipment id /tags/123-tag-id";

    // act
    string equipment;
    string tag;
    const bool match = parse_topic(
        SUT,
        equipment,
        tag);

    // assert
    BOOST_CHECK(!match);
}

BOOST_FIXTURE_TEST_CASE(test_parse_topic_WithFaultyTag_ShouldReturnFalse, TestFixture)
{
    // arrange
    const string SUT = "/equs/the equipment id/tags/ 123-tag-id";

    // act
    string equipment;
    string tag;
    const bool match = parse_topic(
        SUT,
        equipment,
        tag);

    // assert
    BOOST_CHECK(!match);
}

BOOST_FIXTURE_TEST_CASE(test_parse_payload_ShouldPass, TestFixture)
{
    // arrange
    const string JSON =
        "{\"value\": 123, \"at\": \"2015-08-04T16:33:19.553Z\", \"unit\": \"pcs.\"}";

    // act
    ptime at;
    string value;
    string unit;
    const bool match = parse_payload(
        JSON,
        at,
        value,
        unit
    );

    // assert
    BOOST_REQUIRE(match);
    BOOST_CHECK_EQUAL(time_from_string("2015-08-04 16:33:19.553"), at);
    BOOST_CHECK_EQUAL("123", value);
    BOOST_CHECK_EQUAL("pcs.", unit);
}

BOOST_FIXTURE_TEST_CASE(test_parse_payload_WithStringValue_ShouldPass, TestFixture)
{
    // arrange
    const string JSON =
        "{\"value\": \"123\", \"at\": \"2015-08-04T16:33:19.553Z\", \"unit\": \"pcs.\"}";

    // act
    ptime at;
    string value;
    string unit;
    const bool match = parse_payload(
        JSON,
        at,
        value,
        unit
    );

    // assert
    BOOST_REQUIRE(match);
    BOOST_CHECK_EQUAL("\"123\"", value);
}

BOOST_FIXTURE_TEST_CASE(test_parse_payload_WithDoubleValue_ShouldPass, TestFixture)
{
    // arrange
    const string JSON =
        "{\"value\": 123.456, \"at\": \"2015-08-04T16:33:19.553Z\", \"unit\": \"pcs.\"}";

    // act
    ptime at;
    string value;
    string unit;
    const bool match = parse_payload(
        JSON,
        at,
        value,
        unit
    );

    // assert
    BOOST_REQUIRE(match);
    BOOST_CHECK_CLOSE(lexical_cast<double>(value), 123.456, 0.0001);
}

}
}
