#include <mxutil/sched/DailyScheduler.hpp>

#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

using boost::make_shared;
using boost::posix_time::ptime;
using boost::posix_time::milliseconds;
using boost::posix_time::time_duration;
using boost::shared_ptr;

namespace
{

const ptime
    get_next_fire_ptime(
        const ptime &now,
        const time_duration &period)
{
    const long period_ms = period.total_milliseconds();
    const time_duration tod = now.time_of_day();
    return ptime(
        now.date(),
        milliseconds((tod.total_milliseconds() + period_ms) / period_ms * period_ms)
    );
}

}

namespace mxutil
{
namespace sched
{

struct DailyScheduler::Impl
{
    Impl(const time_duration &period_)
        : period(period_)
    {
    }

    const time_duration period;

    shared_ptr<ptime>   next_fire;
};

DailyScheduler::DailyScheduler(
    const time_duration &period,
    const FireCallback &fire_callback)
    : Scheduler(fire_callback)
    , pimpl(new Impl(period))
{
    BOOST_ASSERT(pimpl->period.total_milliseconds() > 0 && pimpl->period.total_milliseconds() < 86400000);
}

DailyScheduler::~DailyScheduler()
{
}

bool
    DailyScheduler::is_timeout(const ptime &now) const
{
    if (!pimpl->next_fire)
        return false;

    return now >= *pimpl->next_fire;
}

void
    DailyScheduler::reset_timeout(const ptime &now)
{
    pimpl->next_fire = make_shared<ptime>(
        get_next_fire_ptime(
            now,
            pimpl->period
        )
    );
}

}
}
