#include <iostream>
#include <string>

#include <boost/exception/all.hpp>
#include <boost/make_shared.hpp>

#include <mxsl/mxsl.hpp>

#include <mxutil/error.hpp>
#include <mxutil/log.hpp>
#include <mxutil/PidFile.hpp>
#include <mxutil/Signal.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StderrLogger.hpp>
#include <mxutil/log/SyslogLogger.hpp>

#include <mxlogarchiver/App.hpp>
#include <mxlogarchiver/Builder.hpp>
#include <mxlogarchiver/Option.hpp>

using boost::logic::tribool;
using boost::make_shared;
using boost::shared_ptr;

using mxlogarchiver::App;
using mxlogarchiver::Builder;
using mxlogarchiver::Option;

using mxutil::PidFile;
using mxutil::Signal;
using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StderrLogger;
using mxutil::log::SyslogLogger;

using std::cerr;
using std::endl;
using std::string;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

shared_ptr<App>
    get_app(
        const string &mqtt_broker_host,
        int mqtt_broker_port);

volatile bool stop = false;
void
    on_term(int)
{
    stop = true;
}

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StderrLogger());
        logger->set_level(Logger::DBUG);
        logger->enable_line_of_file();
        if (!opt.foreground)
        {
            const int result = daemon(0, 0);
            if (0 != result)
            {
                cerr << strerror(errno) << endl;
                return 1;
            }

            logger.reset(new SyslogLogger("mxlogarchiver"));
            logger->set_level(Logger::WARN);
        }
        Log::get().set_logger(logger);

        const shared_ptr<App> app = get_app(
            opt.mqtt_broker_host,
            opt.mqtt_broker_port
        );

        Signal::instance().sigterm(&on_term);

        shared_ptr<PidFile> pidfile;
        if(!opt.foreground)
            pidfile = PidFile::get("/var/run/mxlogarchiver.pid");

        mxsl::validate();

        app->start();

        while(!stop)
        {
            app->loop();
        }

        return 0;
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

shared_ptr<App>
    get_app(
        const string &mqtt_broker_host,
        int mqtt_broker_port)
{
    Builder builder;
    builder.set_configuration("etc/mxlogarchiver/configuration");
    builder.set_mqtt_broker(
        mqtt_broker_host,
        mqtt_broker_port
    );
    builder.set_profile_directory("etc/mxlogarchiver/conf.d");

    builder.clear_working_directory();

    return builder.build();
}

}
