#include <mxlogarchiver/ArchiveDirectory.hpp>

#include <vector>

#include <sys/stat.h>
#include <sys/types.h>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

using boost::filesystem::directory_iterator;
using boost::filesystem::path;
using boost::lexical_cast;
using boost::make_shared;
using boost::shared_ptr;

using std::string;
using std::vector;

namespace
{

uint64_t
    block_size(const path &p);

void
    remove_old_archives_by_count(
        const path &directory,
        unsigned int archives_to_keep);

void
    remove_old_archives_by_directory_size_limit(
        const path &directory,
        uint64_t limit_byte);

uint64_t
    required_size(
        const path &p,
        uint64_t bsize = 0);

}

namespace mxlogarchiver
{

struct ArchiveDirectory::Impl
{
    Impl(const path &directory_)
        : directory(directory_)
    {
    }

    const path  directory;

    shared_ptr<unsigned int>    file_count_limit;
    shared_ptr<uint64_t>        directory_size_limit_byte;
};

ArchiveDirectory::ArchiveDirectory(const path &directory)
    : pimpl(new Impl(directory))
{
    if (!is_directory(pimpl->directory))
        create_directories(pimpl->directory);
}

ArchiveDirectory::~ArchiveDirectory()
{
}

void
    ArchiveDirectory::set_file_count_limit(unsigned int count)
{
    if (count < 1)
        BOOST_THROW_EXCEPTION(ArgumentError());

    pimpl->file_count_limit = make_shared<unsigned int>(count);
}

void
    ArchiveDirectory::set_directory_size_limit_byte(uint64_t limit_byte)
{
    if (limit_byte < 1 * 1024 * 1024)
        BOOST_THROW_EXCEPTION(ArgumentError());

    pimpl->directory_size_limit_byte = make_shared<uint64_t>(limit_byte);
}

void
    ArchiveDirectory::add(
        const path &from,
        const string &name)
{
    if (pimpl->file_count_limit)
    {
        remove_old_archives_by_count(
            pimpl->directory,
            *pimpl->file_count_limit - 1);
    }

    if (pimpl->directory_size_limit_byte)
    {
        remove_old_archives_by_directory_size_limit(
            pimpl->directory,
            *pimpl->directory_size_limit_byte -
                required_size(
                    from,
                    block_size(pimpl->directory)
                )
        );
    }

    const path to = pimpl->directory / (name + ".log.gz");

    if (exists(to))
        remove(to);

    copy(from, to);
}

}

namespace
{

uint64_t
    block_size(const path &p)
{
    struct stat fi;
    stat(p.string().c_str(), &fi);

    return fi.st_blksize;
}

const vector<path>
    listdir(const path &dir)
{
    vector<path> v;
    copy(directory_iterator(dir), directory_iterator(), back_inserter(v));

    vector<path>rtn;
    BOOST_FOREACH(const path &p, v)
    {
        if(is_regular_file(p))
            rtn.push_back(p);
    }

    return rtn;
}

void
    remove_old_archives_by_count(
        const path &directory,
        unsigned int archives_to_keep)
{
    vector<path> archive_list = listdir(directory);
    if (archive_list.size() > archives_to_keep)
    {
        sort(archive_list.begin(), archive_list.end());

        const size_t remove_count = archive_list.size() - archives_to_keep;
        const vector<path> remove_list(archive_list.begin(), archive_list.begin() + remove_count);
        BOOST_FOREACH(const path &p, remove_list)
        {
            remove(p);
        }
    }
}

void
    remove_old_archives_by_directory_size_limit(
        const path &directory,
        uint64_t limit_byte)
{
    vector<path> archive_list = listdir(directory);
    sort(archive_list.begin(), archive_list.end());

    uint64_t total_size = 0;
    BOOST_REVERSE_FOREACH(const path &p, archive_list)
    {
        total_size += required_size(p);
        MXLOG_DBUG("required_size(p) = " + lexical_cast<string>(required_size(p))
            + ", total_size = " + lexical_cast<string>(total_size));

        if (total_size >= limit_byte)
            remove(p);
    }
}

uint64_t
    required_size(
        const path &p,
        uint64_t bsize)
{
    if (bsize == 0)
        bsize = block_size(p);

    return (file_size(p) + bsize - 1) / bsize * bsize;
}

}
