#include <mxlogarchiver/ArchiveRemoveWorker.hpp>

#include <deque>
#include <fstream>

#include <boost/bind.hpp>
#include <boost/exception/all.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/condition_variable.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#include <boost/tuple/tuple.hpp>

#include <mxutil/log.hpp>

using boost::bind;
using boost::condition_variable;
using boost::filesystem::path;
using boost::iostreams::filtering_ostream;
using boost::iostreams::gzip_compressor;
using boost::make_shared;
using boost::mutex;
using boost::shared_ptr;
using boost::thread;
using boost::tuple;

using mxlogarchiver::ArchiveDirectory;
using mxlogarchiver::SystemError;

using std::deque;
using std::ifstream;
using std::ios_base;
using std::ofstream;

namespace
{

typedef
    tuple<
        path,
        shared_ptr<ArchiveDirectory>
    > ArchiveParameter;

void
    gzip(
        const path &from,
        const path &to);

void
    archive_and_remove(
        const path &from,
        ArchiveDirectory &directory);

}

namespace mxlogarchiver
{

struct ArchiveRemoveWorker::Impl
{
    Impl()
        : thread_loop(true)
    {
    }

    deque<shared_ptr<ArchiveParameter> >     work_queue;
    mutex                   work_queue_mutex;
    condition_variable      work_queue_cond;

    shared_ptr<thread>      thread_object;
    bool                    thread_loop;


    const shared_ptr<ArchiveParameter>
        get_next_parameter()
    {
        mutex::scoped_lock lock(this->work_queue_mutex);
        while(this->work_queue.empty())
        {
            this->work_queue_cond.wait(lock);

            if (!this->thread_loop)
                return shared_ptr<ArchiveParameter>();
        }

        shared_ptr<ArchiveParameter> rtn = this->work_queue.front();
        this->work_queue.pop_front();

        return rtn;
    }

    bool
        is_not_empty()
    {
        mutex::scoped_lock lock(this->work_queue_mutex);

        return !this->work_queue.empty();
    }

    void
        worker_thread()
    {
        try
        {
            while(this->thread_loop || this->is_not_empty())
            {
                shared_ptr<ArchiveParameter> archive_parameter = this->get_next_parameter();

                if (!archive_parameter)
                    continue;

                archive_and_remove(
                    archive_parameter->get<0>(),
                    *archive_parameter->get<1>());
            }
            MXLOG_INFO("terminating");
        }
        catch(boost::exception &e)
        {
            MXLOG_EROR(diagnostic_information(e));

            exit(1);
        }
        catch(...)
        {
            MXLOG_EROR("unexpected exception caught");

            exit(1);
        }
    }
};

ArchiveRemoveWorker::ArchiveRemoveWorker()
    : pimpl(new Impl())
{
    pimpl->thread_object.reset(new thread(
        bind(
            &Impl::worker_thread,
            &*pimpl)
    ));
}

ArchiveRemoveWorker::~ArchiveRemoveWorker()
{
    pimpl->thread_loop = false;
    pimpl->work_queue_cond.notify_all();
    pimpl->thread_object->join();
}

void
    ArchiveRemoveWorker::add_path_and_directory(
        const path &p,
        shared_ptr<ArchiveDirectory> directory)
{
    BOOST_ASSERT(directory);

    mutex::scoped_lock lock(pimpl->work_queue_mutex);

    pimpl->work_queue.push_back(make_shared<ArchiveParameter>(p, directory));

    pimpl->work_queue_cond.notify_one();
}

}

namespace
{

void
    gzip(
        const path &from,
        const path &to)
{
    ofstream outfile(to.string().c_str(), ios_base::binary);
    if (!outfile)
        BOOST_THROW_EXCEPTION(SystemError());

    filtering_ostream out;
    out.push(gzip_compressor());
    out.push(outfile);

    ifstream infile(from.string().c_str(), ios_base::binary);
    if (!infile)
        BOOST_THROW_EXCEPTION(SystemError());

    boost::iostreams::copy(infile, out);
}

void
    archive_and_remove(
        const path &from,
        ArchiveDirectory &directory)
{
    path raw = from.stem();
    while(!raw.extension().empty())
        raw = raw.stem();

    const path to = from.parent_path() / (raw.string() + ".gz");
    gzip(from, to);

    directory.add(to, raw.string());

    remove(to);

    remove(from);
}

}
