#include <mxlogarchiver/Builder.hpp>

#include <fstream>
#include <map>
#include <set>
#include <utility>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>
#include <mxutil/sched/DailyScheduler.hpp>

#include <mxtag/mqtt/TagSubscriber.hpp>

#include <mxlogarchiver/App.hpp>
#include <mxlogarchiver/ArchiveDirectory.hpp>
#include <mxlogarchiver/ArchiveRemoveWorker.hpp>
#include <mxlogarchiver/Configuration.hpp>
#include <mxlogarchiver/LogArchiver.hpp>
#include <mxlogarchiver/LogStreamFactory.hpp>
#include <mxlogarchiver/ProfileConfiguration.hpp>

using boost::bind;
using boost::filesystem::directory_iterator;
using boost::filesystem::path;
using boost::make_shared;
using boost::posix_time::seconds;
using boost::shared_ptr;

using mxlogarchiver::App;
using mxlogarchiver::ArchiveRemoveWorker;
using mxlogarchiver::ProfileConfiguration;
using mxlogarchiver::LogArchiver;

using mxtag::mqtt::TagSubscriber;
using mxtag::Tag;

using mxutil::sched::Scheduler;
using mxutil::sched::DailyScheduler;

using std::ifstream;
using std::ios_base;
using std::make_pair;
using std::map;
using std::pair;
using std::set;
using std::string;
using std::vector;

namespace
{

const vector<path>
    listdir(const path &dir)
{
    vector<path> rtn;
    for(directory_iterator iter = directory_iterator(dir);
        iter != directory_iterator();
        ++iter)
    {
        const path &p = *iter;

        if (is_regular_file(p) ||
            is_symlink(p))
            rtn.push_back(p);
    }

    return rtn;
}

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if(!infile)
    {
        MXLOG_WARN("open file failed: " + p.string());

        return string();
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_WARN("read file failed: " + p.string());

        return string();
    }

    return string(buf.begin(), buf.end());
}

void
    process_message(
        const vector<shared_ptr<LogArchiver> > &archiver_list,
        const Tag &t)
{
    BOOST_FOREACH(const shared_ptr<LogArchiver> &a, archiver_list)
    {
        a->write(t);
    }
}

}

namespace mxlogarchiver
{

struct Builder::Impl
{
    Impl()
        : mqtt_broker_host("127.0.0.1")
        , mqtt_broker_service(1883)
    {
    }

    string                                      mqtt_broker_host;
    int                                         mqtt_broker_service;

    shared_ptr<Configuration>                       configuration;
    shared_ptr<map<string, ProfileConfiguration> >  profile_list;

    shared_ptr<LogArchiver>
        build_log_archiver(
            const string &profile_name,
            const ProfileConfiguration &profile,
            shared_ptr<ArchiveRemoveWorker> worker)
    {
        const path working_directory = path(this->configuration->working_directory) / profile_name;

        // create and configure LogStreamFactory
        const shared_ptr<LogStreamFactory> factory = make_shared<LogStreamFactory>(
            working_directory);

        BOOST_FOREACH(const ProfileConfiguration::Equipment &e, profile.equipment_list)
        {
            BOOST_FOREACH(const ProfileConfiguration::Equipment::Tag &t, e.tag_list)
            {
                factory->add_tag(e.equipment, t.tag, t.log_on_change);
            }
        }

        // create and configure ArchiveDirectory
        const shared_ptr<ArchiveDirectory> directory = make_shared<ArchiveDirectory>(
            profile.archive.directory);

        if (profile.archive.archives_to_keep)
            directory->set_file_count_limit(
                *profile.archive.archives_to_keep);

        if (profile.archive.directory_size_limit_byte)
            directory->set_directory_size_limit_byte(
                *profile.archive.directory_size_limit_byte);


        return make_shared<LogArchiver>(
            factory,
            directory,
            profile.archive.threshold_byte_per_file,
            worker);
    }
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_configuration(const path &p)
{
    pimpl->configuration = make_shared<Configuration>(Configuration::parse(readfile(p)));
}

void
    Builder::set_profile_directory(const path &dir)
{
    // build mxlogarchiver::ProfileConfiguration list
    BOOST_VERIFY(is_directory(dir));

    shared_ptr<map<string, ProfileConfiguration> > cfg_list = make_shared<map<string, ProfileConfiguration> >();
    const vector<path> filelist = listdir(dir);
    BOOST_FOREACH(const path &p, filelist)
    {
        try
        {
            if (p.extension() == ".conf")
            {
                MXLOG_DBUG("loading " + p.string());

                cfg_list->insert(pair<string, ProfileConfiguration>(
                    p.stem().string(),
                    ProfileConfiguration::parse(readfile(p))));
            }
        }
        catch(ParseError &e)
        {
            MXLOG_WARN(diagnostic_information(e));
        }
    }

    pimpl->profile_list = cfg_list;
}

void
    Builder::set_mqtt_broker(
        const string &host,
        int service)
{
    if (service > 65535 ||
        service <= 0)
        BOOST_THROW_EXCEPTION(ArgumentError());

    pimpl->mqtt_broker_host = host;
    pimpl->mqtt_broker_service = service;
}

void
    Builder::clear_working_directory()
{
    BOOST_ASSERT(pimpl->configuration && pimpl->profile_list);

    shared_ptr<ArchiveRemoveWorker> worker(new ArchiveRemoveWorker());

    typedef map<string, ProfileConfiguration> conf_t;
    BOOST_FOREACH(const conf_t::value_type &c, *pimpl->profile_list)
    {
        const string &profile_name = c.first;
        const ProfileConfiguration &profile = c.second;

        // Archiver's ctor would search for existing log-file and archive it
        //   when one is found.
        const shared_ptr<LogArchiver> archiver = pimpl->build_log_archiver(
            profile_name,
            profile,
            worker);
    }

    // wait for archive existing log-file
    worker.reset();

    if (!exists(path(pimpl->configuration->working_directory)))
        return;

    vector<path> rlist;
    copy(
        directory_iterator(pimpl->configuration->working_directory),
        directory_iterator(),
        back_inserter(rlist));
    BOOST_FOREACH(const path &p, rlist)
    {
        MXLOG_DBUG("removing: " + p.string());

        remove_all(p);
    }
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(pimpl->configuration && pimpl->profile_list);

    shared_ptr<ArchiveRemoveWorker> worker(new ArchiveRemoveWorker());

    // build mxlogarchiver::Filter list, eq_tag_list

    vector<shared_ptr<Scheduler> > scheduler_list;

    vector<shared_ptr<LogArchiver> > archiver_list;
    typedef map<string, ProfileConfiguration> conf_t;
    BOOST_FOREACH(const conf_t::value_type &c, *pimpl->profile_list)
    {
        const string &profile_name = c.first;
        const ProfileConfiguration &profile = c.second;

        const shared_ptr<LogArchiver> archiver = pimpl->build_log_archiver(
            profile_name,
            profile,
            worker);

        // archive log periodically
        if (profile.archive.period_sec)
        {
            scheduler_list.push_back(make_shared<DailyScheduler>(
                seconds(*profile.archive.period_sec),
                bind(
                    &LogArchiver::archive,
                    archiver
                )
            ));
        }

        archiver_list.push_back(archiver);
    }

    map<string, set<string> > eq_tag_list;
    eq_tag_list["+"].insert("+");

    shared_ptr<TagSubscriber> tag_subscriber(new TagSubscriber(
        pimpl->mqtt_broker_host,
        pimpl->mqtt_broker_service,
        "log-archiver",
        false,
        eq_tag_list,
        bind(
            process_message,
            archiver_list,
            _1
        )
    ));

    return make_shared<App>(
        tag_subscriber,
        scheduler_list);
}

}
