#include <mxlogarchiver/LogArchiver.hpp>

#include <vector>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/optional.hpp>
#include <boost/regex.hpp>
#include <boost/thread/thread.hpp>

#include <mxlogarchiver/ArchiveRemoveWorker.hpp>

using boost::filesystem::directory_iterator;
using boost::filesystem::path;
using boost::optional;
using boost::none;
using boost::regex;
using boost::shared_ptr;

using mxlogarchiver::ArchiveDirectory;
using mxlogarchiver::ArchiveRemoveWorker;

using mxtag::Tag;

using std::string;
using std::vector;

namespace
{

const vector<path>
    search_logfile(const path &temporary_directory);

const path
    make_stage_copy(const path &p);

void
    archive_stage_directory(
        const path &stage_dir,
        shared_ptr<ArchiveDirectory> directory,
        ArchiveRemoveWorker &worker);

}

namespace mxlogarchiver
{

struct LogArchiver::Impl
{
    Impl(
        shared_ptr<LogStreamFactory> factory_,
        shared_ptr<ArchiveDirectory> directory_,
        unsigned int log_threshold_byte_,
        shared_ptr<ArchiveRemoveWorker> worker_)
        : factory(factory_)
        , directory(directory_)
        , log_threshold_byte(log_threshold_byte_)
        , worker(worker_)
    {
    }

    const shared_ptr<LogStreamFactory>  factory;
    const shared_ptr<ArchiveDirectory>  directory;
    const unsigned int                  log_threshold_byte;

    shared_ptr<LogStream>                       stream;
    const shared_ptr<ArchiveRemoveWorker>       worker;
};

LogArchiver::LogArchiver(
    shared_ptr<LogStreamFactory> factory,
    shared_ptr<ArchiveDirectory> directory,
    unsigned int log_threshold_byte,
    shared_ptr<ArchiveRemoveWorker> worker)
    : pimpl(new Impl(
        factory,
        directory,
        log_threshold_byte,
        worker))
{
    BOOST_ASSERT(pimpl->factory && pimpl->directory && pimpl->worker);

    const path stage_dir = pimpl->factory->directory() / "stage";
    if (is_directory(stage_dir))
    {
        archive_stage_directory(
            stage_dir,
            pimpl->directory,
            *pimpl->worker);
    }
    else
        create_directories(stage_dir);

    // archive existing logfile
    const vector<path> existing_log_list = search_logfile(pimpl->factory->directory());
    if (existing_log_list.empty())
        return;

    if (file_size(existing_log_list.back()) != 0)
    {
        const path stage_file_path = make_stage_copy(existing_log_list.back());
        pimpl->worker->add_path_and_directory(stage_file_path, pimpl->directory);
    }

    remove(existing_log_list.back());
}

LogArchiver::~LogArchiver()
{
    this->archive();
}

void
    LogArchiver::write(const Tag &t)
{
    if (!pimpl->stream)
        pimpl->stream = pimpl->factory->create();

    pimpl->stream->write(t); // throws mxlogarchiver::SystemError

    if (pimpl->stream->bytes_written() >= pimpl->log_threshold_byte)
        this->archive();
}

void
    LogArchiver::archive()
{
    if (!pimpl->stream ||
        pimpl->stream->bytes_written() == 0)
        return;

    const path from = pimpl->stream->path();
    pimpl->stream.reset();

    const path temp_file_path = make_stage_copy(from);
    pimpl->worker->add_path_and_directory(temp_file_path, pimpl->directory);

    remove(from);
}

}

namespace
{

const vector<path>
    search_logfile(const path &temporary_directory)
{
    static regex e(
        "^[0-9]{8}-[0-9]{6}\\.log$"
    );

    if (!exists(temporary_directory))
        return vector<path>();

    vector<path> rtn_list;
    for(directory_iterator iter = directory_iterator(temporary_directory);
        iter != directory_iterator();
        ++iter)
    {
        if (is_regular_file(*iter) &&
            regex_match(iter->path().filename().string(), e))
        {
            rtn_list.push_back(*iter);
        }
    }

    sort(rtn_list.begin(), rtn_list.end());
    return rtn_list;
}

const path
    make_stage_copy(const path &p)
{
    const path stage_dir = p.parent_path() / "stage";

    const path stage_file_path = stage_dir / p.filename();
    boost::filesystem::copy_file(p, stage_file_path);

    return stage_file_path;
}

void
    archive_stage_directory(
        const path &stage_dir,
        shared_ptr<ArchiveDirectory> directory,
        ArchiveRemoveWorker &worker)
{
    const vector<path> stage_log_list = search_logfile(stage_dir);
    if (stage_log_list.empty())
        return;

    BOOST_FOREACH(const path &p, stage_log_list)
    {
        worker.add_path_and_directory(p, directory);
    }
}

}
