#include <mxlogarchiver/LogStreamFactory.hpp>

#include <iomanip>
#include <sstream>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/make_shared.hpp>

#include <mxlogarchiver/LogOnChangeDecorator.hpp>
#include <mxlogarchiver/LogStreamImpl.hpp>
#include <mxlogarchiver/TagFilteringDecorator.hpp>

using boost::filesystem::path;
using boost::make_shared;
using boost::posix_time::ptime;
using boost::posix_time::second_clock;
using boost::shared_ptr;

using std::map;
using std::set;
using std::setw;
using std::setfill;
using std::string;
using std::stringstream;

namespace
{

const string
    log_filename_from_ptime(const ptime &time)
{
    stringstream log_name;
    log_name << setw(4) << time.date().year()
        << setw(2) << setfill('0') << static_cast<int>(time.date().month())
        << setw(2) << setfill('0') << time.date().day() << '-'
        << setw(2) << setfill('0') << time.time_of_day().hours()
        << setw(2) << setfill('0') << time.time_of_day().minutes()
        << setw(2) << setfill('0') << time.time_of_day().seconds()
        << ".log";

    return log_name.str();
}

}

namespace mxlogarchiver
{

struct LogStreamFactory::Impl
{
    Impl(const path &directory_)
        : directory(directory_)
    {
    }

    const path  directory;

    map<string, set<string> >   tag_list;
    map<string, set<string> >   log_on_change_tag_list;
};

LogStreamFactory::LogStreamFactory(const path &directory)
    : pimpl(new Impl(directory))
{
}

LogStreamFactory::~LogStreamFactory()
{
}

const path &
    LogStreamFactory::directory() const
{
    return pimpl->directory;
}

void
    LogStreamFactory::add_tag(
        const string &e,
        const string &t,
        bool log_on_change)
{
    pimpl->tag_list[e].insert(t);

    if (log_on_change)
        pimpl->log_on_change_tag_list[e].insert(t);
}

shared_ptr<LogStream>
    LogStreamFactory::create()
{
    shared_ptr<LogStream> ls = make_shared<LogStreamImpl>(
        pimpl->directory / log_filename_from_ptime(second_clock::universal_time()),
        pimpl->tag_list
    );

    ls = make_shared<TagFilteringDecorator>(
        pimpl->tag_list,
        ls);

    if (!pimpl->log_on_change_tag_list.empty())
        ls = make_shared<LogOnChangeDecorator>(
            pimpl->log_on_change_tag_list,
            ls);

    return ls;
}

}
