#include <mxlogarchiver/ProfileConfiguration.hpp>

#include <boost/make_shared.hpp>
#include <boost/regex.hpp>

#include <libconfig.h++>

using boost::make_shared;
using boost::regex;
using boost::shared_ptr;

using libconfig::Config;
using libconfig::ParseException;
using libconfig::Setting;
using libconfig::SettingNotFoundException;

using mxlogarchiver::ParseError;

using std::string;
using std::vector;

namespace
{

void
    validate_equipment(const string &);

void
    validate_tag(const string &);

}

namespace mxlogarchiver
{

const ProfileConfiguration
    ProfileConfiguration::parse(const string &content)
{
    Config config;

    try
    {
        config.readString(content.c_str());

        const Setting &archive = config.lookup("archive");
        const string directory = archive["directory"];

        shared_ptr<uint64_t> directory_size_limit_byte;
        if (archive.exists("directory-size-limit-byte"))
        {
            uint64_t v;

            unsigned int uint;
            long long uint64;
            if (archive.lookupValue("directory-size-limit-byte", uint))
            {
                v = uint;
            }
            else if (archive.lookupValue("directory-size-limit-byte", uint64))
            {
                v = uint64;
            }
            else
                BOOST_THROW_EXCEPTION(ParseError());

            if (v < 1024 * 1024) // 1 MB
                BOOST_THROW_EXCEPTION(ParseError());

            directory_size_limit_byte = make_shared<uint64_t>(v);
        }

        const unsigned int threshold_byte_per_file = archive["threshold-byte-per-file"];
        if (threshold_byte_per_file < 10240)
            BOOST_THROW_EXCEPTION(ParseError());

        shared_ptr<unsigned int> period_sec;
        if (archive.exists("period-sec"))
        {
            const unsigned int v = archive["period-sec"];
            if (v < 60 || v >= 86400)
                BOOST_THROW_EXCEPTION(ParseError());

            period_sec = make_shared<unsigned int>(v);
        }

        shared_ptr<unsigned int> archives_to_keep;
        if (archive.exists("archives-to-keep"))
        {
            const unsigned int v = archive["archives-to-keep"];
            if (v == 0)
                BOOST_THROW_EXCEPTION(ParseError());

            archives_to_keep = make_shared<unsigned int>(v);
        }

        const Setting &equipment_list = config.lookup("equipment-list");
        vector<ProfileConfiguration::Equipment> cfg_equipment_list;
        for(int i = 0; i < equipment_list.getLength(); i++)
        {
            const Setting &equipment = equipment_list[i];

            const string equipment_name = equipment["equipment"];
            validate_equipment(equipment_name);

            const Setting &tag_list = equipment["tag-list"];
            vector<ProfileConfiguration::Equipment::Tag> cfg_tag_list;
            for(int j = 0; j < tag_list.getLength(); j++)
            {
                const string tag = tag_list[j]["tag"];
                validate_tag(tag);

                bool log_on_change = false;
                if (tag_list[j].exists("log-on-change"))
                {
                    log_on_change = tag_list[j]["log-on-change"];
                }

                cfg_tag_list.push_back(ProfileConfiguration::Equipment::Tag(tag, log_on_change));
            }

            cfg_equipment_list.push_back(ProfileConfiguration::Equipment(
                equipment_name,
                cfg_tag_list
            ));
        }

        return ProfileConfiguration(
            ProfileConfiguration::Archive(
                directory,
                directory_size_limit_byte,
                threshold_byte_per_file,
                period_sec,
                archives_to_keep
            ),
            cfg_equipment_list
        );
    }
    catch(ParseException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
    catch(SettingNotFoundException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
}

}

namespace
{

void
    validate_equipment(const string &equipment)
{
    static const regex e(
        "^[A-Za-z0-9_-]{3,255}$"
    );

    if (!regex_match(equipment.begin(), equipment.end(), e))
        BOOST_THROW_EXCEPTION(ParseError()); // TODO error message
}

void
    validate_tag(const string &tag)
{
    static const regex e(
        "^[A-Za-z0-9_-]{2,255}$"
    );

    if (!regex_match(tag.begin(), tag.end(), e))
        BOOST_THROW_EXCEPTION(ParseError()); // TODO error message
}

}
