#include <boost/test/unit_test.hpp>
#include <boost/range.hpp>

#include <mxlogarchiver/ProfileConfiguration.hpp>

using boost::size;

using std::string;

namespace mxlogarchiver
{

class TestFixture
{
};

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_ShouldPass, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "directory = \"/var/mxc/upload\"\n"
            "threshold-byte-per-file = 5242800\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act
    const ProfileConfiguration result = ProfileConfiguration::parse(SUT);

    // assert
    const ProfileConfiguration::Archive archive = result.archive;
    BOOST_CHECK_EQUAL("/var/mxc/upload", archive.directory);
    BOOST_CHECK(!archive.directory_size_limit_byte);
    BOOST_CHECK_EQUAL(5242800, archive.threshold_byte_per_file);
    BOOST_CHECK(!archive.archives_to_keep);

    BOOST_REQUIRE_EQUAL(2, result.equipment_list.size());

    {
        const ProfileConfiguration::Equipment equipment = result.equipment_list[0];
        BOOST_CHECK_EQUAL("My_ioLogik-E2242", equipment.equipment);
        BOOST_REQUIRE_EQUAL(3, equipment.tag_list.size());
        BOOST_CHECK_EQUAL("di0", equipment.tag_list[0].tag);
        BOOST_CHECK_EQUAL(true, equipment.tag_list[0].log_on_change);
        BOOST_CHECK_EQUAL("di1", equipment.tag_list[1].tag);
        BOOST_CHECK_EQUAL(true, equipment.tag_list[1].log_on_change);
        BOOST_CHECK_EQUAL("ai0", equipment.tag_list[2].tag);
        BOOST_CHECK_EQUAL(false, equipment.tag_list[2].log_on_change);
    }

    {
        const ProfileConfiguration::Equipment equipment = result.equipment_list[1];
        BOOST_CHECK_EQUAL("My_other_ioLogik-E2242", equipment.equipment);
        BOOST_REQUIRE_EQUAL(3, equipment.tag_list.size());
        BOOST_CHECK_EQUAL("di0", equipment.tag_list[0].tag);
        BOOST_CHECK_EQUAL(false, equipment.tag_list[0].log_on_change);
        BOOST_CHECK_EQUAL("di1", equipment.tag_list[1].tag);
        BOOST_CHECK_EQUAL(false, equipment.tag_list[1].log_on_change);
        BOOST_CHECK_EQUAL("ai0", equipment.tag_list[2].tag);
        BOOST_CHECK_EQUAL(false, equipment.tag_list[2].log_on_change);
    }
}

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_WithArchivesToKeep_ShouldPass, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "directory = \"/var/mxc/upload\"\n"
            "threshold-byte-per-file = 5242800\n"
            "archives-to-keep = 24\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act
    const ProfileConfiguration result = ProfileConfiguration::parse(SUT);

    // assert
    BOOST_REQUIRE(result.archive.archives_to_keep);
    BOOST_CHECK_EQUAL(24, *result.archive.archives_to_keep);
}

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_WithPeriodSec_ShouldPass, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "directory = \"/var/mxc/upload\"\n"
            "threshold-byte-per-file = 5242800\n"
            "period-sec = 3600\n"
            "archives-to-keep = 24\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act
    const ProfileConfiguration result = ProfileConfiguration::parse(SUT);

    // assert
    const ProfileConfiguration::Archive archive = result.archive;
    BOOST_REQUIRE(result.archive.period_sec);
    BOOST_CHECK_EQUAL(3600, *result.archive.period_sec);
}

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_WithDirectorySizeLimitByte_ShouldPass, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "directory = \"/var/mxc/upload\"\n"
            "directory-size-limit-byte = 1234567890\n"
            "threshold-byte-per-file = 5242800\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act
    const ProfileConfiguration result = ProfileConfiguration::parse(SUT);

    // assert
    BOOST_REQUIRE(result.archive.directory_size_limit_byte);
    BOOST_CHECK_EQUAL(1234567890ull, *result.archive.directory_size_limit_byte);
}

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_WithFaultyConfig_ShouldReturnParseError, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "directory = \"/var/mxc/upload\"\n"
            "threshold-byte-per-file = 5242800\n"
            "archives-to-keep = 24\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        "}\n"; // oops

    // act & assert
    BOOST_CHECK_THROW(ProfileConfiguration::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE(test_ProfileConfiguration_parse_WithInsufficientConfig_ShouldReturnParseError, TestFixture)
{
    // arrange
    const string SUT =
        "archive = {\n"
            "#directory = \"/var/mxc/upload\"\n" // oops
            "threshold-byte-per-file = 5242800\n"
            "archives-to-keep = 24\n"
        "}\n"
        "equipment-list = (\n"
            "{\n"
                "equipment = \"My_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                        "log-on-change = true\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "equipment = \"My_other_ioLogik-E2242\"\n"
                "tag-list = (\n"
                    "{\n"
                        "tag = \"di0\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"di1\"\n"
                    "},\n"
                    "{\n"
                        "tag = \"ai0\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act & assert
    BOOST_CHECK_THROW(ProfileConfiguration::parse(SUT), ParseError);
}

}
