#include <iostream>
#include <string>

#include <boost/exception/all.hpp>
#include <boost/make_shared.hpp>
#include <boost/ref.hpp>
#include <boost/shared_ptr.hpp>

#include <mxsl/mxsl.hpp>

#include <mxutil/error.hpp>
#include <mxutil/log.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StderrLogger.hpp>

#include <mxlogconverter/App.hpp>
#include <mxlogconverter/Option.hpp>
#include <mxlogconverter/RawPrinter.hpp>
#include <mxlogconverter/CsvFormatter.hpp>
#include <mxlogconverter/JsonFormatter.hpp>
#include <mxlogconverter/XmlFormatter.hpp>

using boost::logic::tribool;
using boost::make_shared;
using boost::ref;
using boost::shared_ptr;
using boost::size;

using mxlogconverter::App;
using mxlogconverter::Extractor;
using mxlogconverter::Option;
using mxlogconverter::RawPrinter;
using mxlogconverter::CsvFormatter;
using mxlogconverter::JsonFormatter;
using mxlogconverter::XmlFormatter;

using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StderrLogger;

using std::cin;
using std::cout;
using std::string;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StderrLogger());
        logger->set_level(Logger::WARN);
        Log::get().set_logger(logger);

        shared_ptr<Extractor::Handler> handler;
        if (opt.dump)
            handler = make_shared<RawPrinter>(ref(cout));

        else if ("xml" == opt.format)
            handler = make_shared<XmlFormatter>(ref(cout), opt.window_size);

        else if ("csv" == opt.format)
            handler = make_shared<CsvFormatter>(ref(cout), opt.window_size);

        else if ("json" == opt.format)
            handler = make_shared<JsonFormatter>(ref(cout), opt.window_size);

        else
        {
            MXLOG_EROR("unexpected format: " + opt.format);
            return 1;
        }

        const shared_ptr<App> app(new App(
            cin,
            handler
        ));

        mxsl::validate();

        app->run();
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
        exit(1);
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
        exit(1);
    }

    return 0;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

}
