#include <mxlogconverter/CsvFormatter.hpp>

#include <algorithm>
#include <sstream>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <mxutil/log.hpp>

#include <mxtag/tag_form.hpp>

#include <mxlogconverter/error.hpp>
#include <mxlogconverter/Formatter.hpp>

using boost::bind;
using boost::lexical_cast;
using boost::posix_time::ptime;

using mxlogconverter::ArgumentError;
using mxlogconverter::Formatter;

using mxtag::form_iso8601_time;
using mxtag::Value;

using std::endl;
using std::find_if;
using std::map;
using std::ostream;
using std::set;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

struct EquTag
{
    string  equ;
    string  tag;

    EquTag(
        const string &equ_,
        const string &tag_)
        : equ(equ_)
        , tag(tag_)
    {
    }
};

bool
    tag_eq(
        const string &tag_name,
        const Formatter::Tag &tag)
{
    return tag_name == tag.tag;
}

const string
    string_from_value(const Value &v)
{
    if (v.is_string())
        return v.as_string();
    else if (v.is_int())
        return lexical_cast<string>(v.as_int());
    else if (v.is_uint())
        return lexical_cast<string>(v.as_uint());
    else if (v.is_float())
        return lexical_cast<string>(v.as_float());
    else
        BOOST_THROW_EXCEPTION(ArgumentError());
}

}

namespace mxlogconverter
{

struct CsvFormatter::Impl
{
    Impl()
    {
    }

    vector<EquTag>  equ_tag_list;
};

CsvFormatter::CsvFormatter(
    ostream &os,
    int window_size)
    : Formatter(os, window_size)
    , pimpl(new Impl())
{
}

CsvFormatter::~CsvFormatter()
{
}

void
    CsvFormatter::print_header(ostream &output_stream)
{
    BOOST_VERIFY(pimpl->equ_tag_list.empty());

    stringstream ss;
    ss << "at";

    typedef map<string, set<string> > equ_tag_t;
    BOOST_FOREACH(const equ_tag_t::value_type &e, this->equ_tag_list())
    {
        const string &equipment = e.first;
        const set<string> &tag_list = e.second;
        BOOST_FOREACH(const string &t, tag_list)
        {
            ss << ',' << '\"';
            ss << t;
            ss << '@' << equipment;
            ss << '\"';

            pimpl->equ_tag_list.push_back(EquTag(equipment, t));
        }
    }

    ss << endl;

    output_stream << ss.str();
}

void
    CsvFormatter::print_record(
        ostream &output_stream,
        const ptime &at,
        const map<string, vector<Formatter::Tag> > &record_list)
{
    stringstream ss;
    ss << '"' << form_iso8601_time(at) << '"';

    BOOST_FOREACH(const EquTag &et, pimpl->equ_tag_list)
    {
        ss << ',';

        const map<string, vector<Formatter::Tag> >::const_iterator iter = record_list.find(et.equ);
        if (iter == record_list.end())
            continue;

        const vector<Formatter::Tag> &tag_list = iter->second;
        const vector<Formatter::Tag>::const_iterator tag_iter =
            find_if(tag_list.begin(), tag_list.end(), bind(&tag_eq, et.tag, _1));
        if (tag_iter == tag_list.end())
            continue;

        const Formatter::Tag &tag = *tag_iter;
        try
        {
            ss << '"' << string_from_value(tag.value);

            if (!tag.unit.empty())
                ss << '(' << tag.unit << ')';

            ss << '"';
        }
        catch(ArgumentError &e)
        {
            MXLOG_EROR(diagnostic_information(e));
            continue;
        }
    }

    ss << endl;

    output_stream << ss.str();
}

void
    CsvFormatter::print_footer(ostream &)
{
}

}
