#include <mxlogconverter/Formatter.hpp>

#include <algorithm>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

#include <mxtag/Tag.hpp>

#include <mxlogconverter/Extractor.hpp>

using boost::bind;
using boost::function;
using boost::posix_time::ptime;
using boost::shared_ptr;

using mxlogconverter::Extractor;
using mxlogconverter::Formatter;

using std::map;
using std::ostream;
using std::pair;
using std::set;
using std::sort;
using std::string;
using std::vector;

namespace
{

typedef enum
{
    STATE_INIT,
    STATE_AFTER_EQU_TAG,
    STATE_EOF
} state_t;

struct sort_by_tag_at_lt
{
    inline bool operator() (
        const mxtag::Tag &lhs,
        const mxtag::Tag &rhs)
    {
        return (lhs.at() < rhs.at());
    }
};

bool tag_cmp_at_ne(
    const mxtag::Tag &lhs,
    const mxtag::Tag &rhs)
{
    return lhs.at() != rhs.at();
}

}

namespace mxlogconverter
{

struct Formatter::Impl
{
    Impl(
        ostream &output_stream_,
        int window_size_)
        : output_stream(output_stream_)
        , window_size(window_size_)
        , state(STATE_INIT)
    {
    }

    ostream &                               output_stream;
    const int                               window_size;
    state_t                                 state;

    shared_ptr<map<string, set<string> > >  equ_tag_list;

    vector<mxtag::Tag>                      tag_list;

    bool
        is_valid(
            const mxtag::Tag &tag)
    {
        const map<string, set<string> >::const_iterator eq_iter = (*this->equ_tag_list).find(tag.equipment());
        if (eq_iter == (*this->equ_tag_list).end())
            return false;

        const set<string> &tag_list_ = eq_iter->second;
        if (tag_list_.find(tag.tag()) == tag_list_.end())
            return false;

        return true;
    }

    void
        export_tag_list(Formatter &formatter)
    {
        vector<mxtag::Tag>::iterator print_end = find_if(
            this->tag_list.begin(),
            this->tag_list.end(),
            bind(
                &tag_cmp_at_ne,
                this->tag_list.front(),
                _1
            )
        );

        ptime at = this->tag_list.front().at();
        map<string, vector<Tag> > tag_list;
        for(vector<mxtag::Tag>::const_iterator iter = this->tag_list.begin();
            iter != print_end;
            ++iter)
        {
            tag_list[iter->equipment()].push_back(Tag(
                iter->tag(),
                iter->value(),
                iter->unit()));
        }

        this->tag_list.erase(this->tag_list.begin(), print_end);

        formatter.print_record(this->output_stream, at, tag_list);
    }
};

Formatter::Formatter(
    ostream &os,
    int window_size)
    : pimpl(new Impl(os, window_size))
{
}

Formatter::~Formatter()
{
}

const map<string, set<string> > &
    Formatter::equ_tag_list() const
{
    BOOST_ASSERT(pimpl->state != STATE_INIT && pimpl->equ_tag_list);

    return *pimpl->equ_tag_list;
}

void
    Formatter::on_equ_tag_list(const map<string, set<string> > &tag_list)
{
    BOOST_ASSERT(pimpl->state == STATE_INIT && !pimpl->equ_tag_list);
    pimpl->equ_tag_list.reset(new map<string, set<string> >(tag_list));

    pimpl->state = STATE_AFTER_EQU_TAG;

    print_header(pimpl->output_stream);
}

void
    Formatter::on_tag(const mxtag::Tag &tag)
{
    BOOST_ASSERT(pimpl->state == STATE_AFTER_EQU_TAG);

    if (!pimpl->is_valid(tag))
        return;

    const vector<mxtag::Tag>::iterator insert = upper_bound(
        pimpl->tag_list.begin(),
        pimpl->tag_list.end(),
        tag,
        sort_by_tag_at_lt());

    pimpl->tag_list.insert(insert, tag);
    const bool same_time_as_front = tag.at() == pimpl->tag_list.front().at();

    if (pimpl->tag_list.size() < static_cast<size_t>(pimpl->window_size))
        return;

    if (same_time_as_front)
    {
        // Skew detected, don't export.
        // Wait for consequent tags with same "at".
        MXLOG_WARN("tag.at() skew detected");
        return;
    }

    pimpl->export_tag_list(*this);
}

void
    Formatter::on_eof()
{
    BOOST_ASSERT(pimpl->state == STATE_AFTER_EQU_TAG);

    // flush record_list
    while(!pimpl->tag_list.empty())
    {
        pimpl->export_tag_list(*this);
    }

    print_footer(pimpl->output_stream);

    pimpl->state = STATE_EOF;
}

}
