#include <mxlogconverter/XmlFormatter.hpp>

#include <algorithm>
#include <exception>
#include <map>
#include <set>
#include <sstream>

#include <boost/exception/all.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

#include <mxtag/tag_form.hpp>
#include <mxtag/Value.hpp>

#include <mxlogconverter/error.hpp>
#include <mxlogconverter/Formatter.hpp>

using boost::lexical_cast;
using boost::posix_time::ptime;
using boost::regex;
using boost::shared_ptr;

using mxlogconverter::ArgumentError;
using mxlogconverter::Formatter;

using mxtag::form_iso8601_time;
using mxtag::Value;

using std::make_pair;
using std::map;
using std::ostream;
using std::pair;
using std::set;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

const pair<string, string>
    string_from_value(const Value &v)
{
    if (v.is_string())
        return make_pair<string, string>(v.as_string(), "string");
    else if (v.is_int())
        return make_pair<string, string>(lexical_cast<string>(v.as_int()), "int");
    else if (v.is_uint())
        return make_pair<string, string>(lexical_cast<string>(v.as_uint()), "uint");
    else if (v.is_float())
        return make_pair<string, string>(lexical_cast<string>(v.as_float()), "float");
    else
        BOOST_THROW_EXCEPTION(ArgumentError());
}

}

namespace mxlogconverter
{

XmlFormatter::XmlFormatter(
    ostream &os,
    int window_size)
    : Formatter(os, window_size)
{
}

XmlFormatter::~XmlFormatter()
{
}

void
    XmlFormatter::print_header(ostream &output_stream)
{
    output_stream << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
    output_stream << "<record_list>";
}

void
    XmlFormatter::print_record(
        ostream &output_stream,
        const ptime &at,
        const map<string, vector<Formatter::Tag> > &record_list)
{
    typedef map<string, vector<Formatter::Tag> > record_t;
    BOOST_FOREACH(const record_t::value_type &r, record_list)
    {
        const string &equipment = r.first;
        const vector<Formatter::Tag> &tag_list = r.second;

        stringstream ss;
        ss << "<record>";
        ss << "<equ>" << equipment << "</equ>";
        ss << "<at>" << form_iso8601_time(at) << "</at>";
        ss << "<tag_list>";

        int output_count = 0;
        BOOST_FOREACH(const Formatter::Tag &t, tag_list)
        {
            pair<string, string> value_and_type;
            try
            {
                value_and_type = string_from_value(t.value);
            }
            catch(ArgumentError &e)
            {
                MXLOG_EROR(diagnostic_information(e));
                continue;
            }

            const string value = value_and_type.first;
            const string value_type = value_and_type.second;

            ss << "<tag>";
            ss << "<name>" << t.tag << "</name>";
            ss << "<value type=\"" << value_type << "\"";
            if (!t.unit.empty())
                ss << " unit=\"" << t.unit << "\"";
            ss << ">" << value << "</value>";
            ss << "</tag>";
            output_count++;
        }

        ss << "</tag_list>";
        ss << "</record>";

        if (output_count)
            output_stream << ss.str();
    }
}

void
    XmlFormatter::print_footer(ostream &output_stream)
{
    output_stream << "</record_list>";
}

}
