#include <boost/test/unit_test.hpp>

#include <map>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxlogconverter/Extractor.hpp>

#include <mxtag/Tag.hpp>
#include <mxtag/Value.hpp>
#include <mxtag/log/EnvelopedOStream.hpp>
#include <mxtag/log/internal_log.hpp>
#include <mxtag/log/ProtobufSerializer.hpp>

using boost::make_shared;
using boost::posix_time::time_from_string;
using boost::shared_ptr;

using mxlogconverter::Extractor;

using mxtag::Tag;
using mxtag::Value;
using mxtag::log::EnvelopedOStream;
using mxtag::log::ProtobufSerializer;

using std::map;
using std::set;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

class TestFixture
{
};

class TestHandler
    : public Extractor::Handler
{
public:
    TestHandler()
        : eof(false)
    {
    }

public:
    virtual void
        on_equ_tag_list(const map<string, set<string> > &equ_tag_list)
    {
        this->equ_tag_list = make_shared<map<string, set<string> > >(equ_tag_list);
    }

    virtual void
        on_tag(const Tag &tag)
    {
        this->tag = make_shared<Tag>(tag);
    }

    virtual void
        on_eof()
    {
        eof = true;
    }

public:
    shared_ptr<map<string, set<string> > >      equ_tag_list;
    shared_ptr<Tag>                             tag;
    bool                                        eof;
};

}

namespace mxlogconverter
{

BOOST_FIXTURE_TEST_CASE(test_parse_ShouldReturnCorrectValue, TestFixture)
{
    // arrange
    TestHandler handler;
    stringstream ss;
    EnvelopedOStream eos(ss);
    map<string, set<string> > equ_tag_list;
    equ_tag_list["My_ioLogik-E2242"].insert("ai0");
    equ_tag_list["My_ioLogik-E2242"].insert("di0");
    equ_tag_list["My_other_ioLogik-E2242"].insert("DOOR_STATUS");
    equ_tag_list["My_other_ioLogik-E2242"].insert("WINDOW_STATUS");

    ProtobufSerializer serializer;
    eos << serializer.from_tag_list(equ_tag_list);
    const Tag tag(
        "My_ioLogik-E2242",
        "ai0",
        time_from_string("2015-07-06 14:51:31"),
        Value(1.23),
        "V");
    eos << serializer.from_tag(tag);

    // act
    Extractor ext;
    ext.parse(ss, handler);

    // assert
    BOOST_REQUIRE(handler.equ_tag_list);

    BOOST_CHECK(equ_tag_list == *handler.equ_tag_list);

    BOOST_CHECK_EQUAL("My_ioLogik-E2242", handler.tag->equipment());
    BOOST_CHECK_EQUAL("ai0", handler.tag->tag());
    BOOST_CHECK_CLOSE(1.23, handler.tag->value().as_float(), 0.0001);
    BOOST_CHECK_EQUAL(time_from_string("2015-07-06 14:51:31.000"), handler.tag->at());
    BOOST_CHECK_EQUAL("V", handler.tag->unit());

    BOOST_CHECK(handler.eof);
}

}
