#include <boost/test/unit_test.hpp>

#include <map>
#include <set>
#include <sstream>

#include <boost/range.hpp>

#include <mxlogconverter/XmlFormatter.hpp>

#include <mxtag/Tag.hpp>
#include <mxtag/Value.hpp>

using boost::posix_time::time_from_string;
using boost::size;

using mxlogconverter::Extractor;
using mxlogconverter::XmlFormatter;

using mxtag::Tag;
using mxtag::Value;

using std::map;
using std::set;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

class TestFixture
{
};

}

namespace mxlogconverter
{

BOOST_FIXTURE_TEST_CASE(test_XmlFormatter_ShouldPass, TestFixture)
{
    // arrange
    const string tags[] = {"ai0", "di0"};
    map<string, set<string> > equ_tag_list;
    equ_tag_list["My_ioLogik-E2242"] = set<string>(tags, tags + size(tags));

    stringstream ss;
    XmlFormatter formatter(ss, 1);

    // act
    formatter.on_equ_tag_list(equ_tag_list);
    formatter.on_tag(Tag(
        "My_ioLogik-E2242",
        "ai0",
        time_from_string("2015-08-31 12:01:05.123"),
        Value(1.23),
        "V"));
    formatter.on_eof();

    // assert
    BOOST_CHECK_EQUAL(
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        "<record_list>"
            "<record>"
                "<equ>My_ioLogik-E2242</equ>"
                "<at>2015-08-31T12:01:05.123Z</at>"
                "<tag_list>"
                    "<tag>"
                        "<name>ai0</name>"
                        "<value type=\"float\" unit=\"V\">1.23</value>"
                    "</tag>"
                "</tag_list>"
            "</record>"
        "</record_list>",
        ss.str());
}

BOOST_FIXTURE_TEST_CASE(test_XmlFormatter_on_tag_withFaultyEquipmentRecord_ShouldNotStoreInFormatter, TestFixture)
{
    // arrange
    const string tags[] = {"ai0", "di0"};
    map<string, set<string> > equ_tag_list;
    equ_tag_list["My_ioLogik-E2242"] = set<string>(tags, tags + size(tags));

    stringstream ss;
    XmlFormatter formatter(ss, 1);

    // act
    formatter.on_equ_tag_list(equ_tag_list);
    formatter.on_tag(Tag(
        "My_ioLogik-E2243", // oops
        "ai0",
        time_from_string("2015-08-31 12:01:05.123"),
        Value(1.23),
        "V"));
    formatter.on_eof();

    // assert
    BOOST_CHECK_EQUAL(
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        "<record_list>"
        "</record_list>",
        ss.str());
}

BOOST_FIXTURE_TEST_CASE(test_XmlFormatter_on_tag_withFaultyTagRecord_ShouldNotStoreInFormatter, TestFixture)
{
    // arrange
    const string tags[] = {"ai0", "di0"};
    map<string, set<string> > equ_tag_list;
    equ_tag_list["My_ioLogik-E2242"] = set<string>(tags, tags + size(tags));

    stringstream ss;
    XmlFormatter formatter(ss, 1);

    // act
    formatter.on_equ_tag_list(equ_tag_list);
    formatter.on_tag(Tag(
        "My_ioLogik-E2242",
        "pi0", // oops
        time_from_string("2015-08-31 12:01:05.123"),
        Value(1.23),
        "V"));
    formatter.on_eof();

    // assert
    BOOST_CHECK_EQUAL(
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        "<record_list>"
        "</record_list>",
        ss.str());
}

BOOST_FIXTURE_TEST_CASE(test_XmlFormatter_on_tag_withMultiEquipment_ShouldReturnTwoCorrectRecord, TestFixture)
{
    // arrange
    map<string, set<string> > equ_tag_list;
    const string tags_1[] = {"di0", "ai0"};
    equ_tag_list["My_ioLogik-E2242"] = set<string>(tags_1, tags_1 + size(tags_1));
    const string tags_2[] = {"di1", "ai1"};
    equ_tag_list["My_other_ioLogik-E2242"] = set<string>(tags_2, tags_2 + size(tags_2));

    stringstream ss;
    XmlFormatter xml_formatter(ss, 2);
    xml_formatter.on_equ_tag_list(equ_tag_list);

    const Tag tag_1(
        "My_ioLogik-E2242",
        "di0",
        time_from_string("2015-08-31 12:01:05.123"),
        Value(static_cast<int64_t>(1))
    );
    const Tag tag_2(
        "My_other_ioLogik-E2242",
        "ai1",
        time_from_string("2015-08-31 12:01:05.123"),
        Value(1.23),
        "V"
    );

    // act
    xml_formatter.on_tag(tag_1);
    xml_formatter.on_tag(tag_2);
    xml_formatter.on_eof();

    // assert
    BOOST_CHECK_EQUAL(
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
        "<record_list>"
            "<record>"
                "<equ>My_ioLogik-E2242</equ>"
                "<at>2015-08-31T12:01:05.123Z</at>"
                "<tag_list>"
                    "<tag>"
                        "<name>di0</name>"
                        "<value type=\"int\">1</value>"
                    "</tag>"
                "</tag_list>"
            "</record>"
            "<record>"
                "<equ>My_other_ioLogik-E2242</equ>"
                "<at>2015-08-31T12:01:05.123Z</at>"
                "<tag_list>"
                    "<tag>"
                        "<name>ai1</name>"
                        "<value type=\"float\" unit=\"V\">1.23</value>"
                    "</tag>"
                "</tag_list>"
            "</record>"
        "</record_list>",
        ss.str()
    );
}

}
