#include <mxloguploader/App.hpp>

#include <set>
#include <vector>

#include <boost/filesystem.hpp>
#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>
#include <mxutil/PidFile.hpp>

#include <mxloguploader/Log.hpp>

using boost::filesystem::directory_iterator;
using boost::filesystem::path;
using boost::make_shared;
using boost::shared_ptr;

using mxutil::PidFile;

using std::set;
using std::string;
using std::vector;

namespace
{

void
    delete_dangling_symlink(const path &d);

const vector<path>
    enumerate_log_list(
        const path &log_directory,
        const path &uploaded_directory);
}

namespace mxloguploader
{

struct App::Impl
{
    Impl(
        const path &log_directory_,
        const path &working_directory_,
        shared_ptr<StatusUpdater> status_updater_,
        shared_ptr<LogConverter> log_converter_,
        shared_ptr<Uploader> uploader_,
        const ConvertExternalFilenameFunc &external_filename_,
        unsigned int file_count_)
        : log_directory(log_directory_)
        , working_directory(working_directory_)
        , status_updater(status_updater_)
        , log_converter(log_converter_)
        , uploader(uploader_)
        , external_filename(external_filename_)
        , file_count(file_count_)
    {
    }

    const path                      log_directory;
    const path                      working_directory;
    const shared_ptr<StatusUpdater> status_updater;
    const shared_ptr<LogConverter>  log_converter;
    const shared_ptr<Uploader>      uploader;
    const ConvertExternalFilenameFunc   external_filename;
    const unsigned int              file_count;

    shared_ptr<PidFile>             pidfile;
};

App::App(
    const path &log_directory,
    const path &working_directory,
    shared_ptr<StatusUpdater> status_updater,
    shared_ptr<LogConverter> log_converter,
    shared_ptr<Uploader> uploader,
    const ConvertExternalFilenameFunc &external_filename,
    unsigned int file_count)
    : pimpl(new Impl(
        log_directory,
        working_directory,
        status_updater,
        log_converter,
        uploader,
        external_filename,
        file_count))
{
    BOOST_ASSERT(pimpl->status_updater);
    BOOST_ASSERT(pimpl->log_converter);
    BOOST_ASSERT(pimpl->uploader);

    if (!is_directory(pimpl->working_directory))
        create_directories(working_directory);

    pimpl->pidfile =
        PidFile::get((pimpl->working_directory / "lock").string());
}

App::~App()
{
}

void
    App::run()
{
    const path uploaded_directory = pimpl->working_directory / "uploaded";

    // enumerate Log list
    vector<path> log_list = enumerate_log_list(
        pimpl->log_directory,
        uploaded_directory);

    // truncate Log list to fit file_count
    if (pimpl->file_count != 0 &&
        log_list.size() > pimpl->file_count)
        log_list.resize(pimpl->file_count);

    delete_dangling_symlink(uploaded_directory);

    // for each Log
    //   upload
    //   update status
    //   mark uploaded
    BOOST_FOREACH(const path &p, log_list)
    {
        Log log(
            p,
            pimpl->log_converter,
            uploaded_directory / p.filename());

        const path external_log = log.get_external_log_path();
        const string external_filename = pimpl->external_filename(
            log.internal_log_path());

        try
        {
            pimpl->uploader->upload(external_log, &external_filename);

            pimpl->status_updater->update(
                log.internal_log_path());

            log.mark_uploaded();
        }
        catch(Error &e)
        {
            pimpl->status_updater->update(
                log.internal_log_path(),
                false,
                diagnostic_information(e));

            throw;
        }
    }
}

}

namespace
{

void
    delete_dangling_symlink(const path &d)
{
    vector<path> to_remove;
    for(directory_iterator iter = directory_iterator(d);
        iter != directory_iterator();
        ++iter)
    {
        const path &p = *iter;
        if (!is_symlink(p))
            continue;

        const path target = read_symlink(p);
        if (!exists(target))
            to_remove.push_back(p);
    }

    BOOST_FOREACH(const path &p, to_remove)
    {
        remove(p);
    }
}

const set<path>
    get_uploaded_log_list(const path &uploaded_directory)
{
    set<path> rtn;

    for(directory_iterator iter = directory_iterator(uploaded_directory);
        iter != directory_iterator();
        ++iter)
    {
        const path &p = *iter;

        if (!is_symlink(p))
            continue;

        const path t = read_symlink(p);
        if (exists(t))
            rtn.insert(t);
    }

    return rtn;
}

const set<path>
    list_regular(const path &d)
{
    set<path> rtn;

    if (!exists(d))
        return rtn;

    for(directory_iterator iter = directory_iterator(d);
        iter != directory_iterator();
        ++iter)
    {
        const path &p = *iter;
        if (is_regular_file(p))
        {
            rtn.insert(p);
        }
    }

    return rtn;
}

const vector<path>
    enumerate_log_list(
        const path &log_directory,
        const path &uploaded_directory)
{
    set<path> log_list = list_regular(log_directory);
    const set<path> uploaded_list = get_uploaded_log_list(uploaded_directory);

    vector<path> rtn;

    set_difference(
        log_list.begin(),
        log_list.end(),
        uploaded_list.begin(),
        uploaded_list.end(),
        back_inserter(rtn));

    return rtn;
}

}
