#include <mxloguploader/Builder.hpp>

#include <fstream>

#include <boost/bind.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/regex.hpp>

#include <mxutil/log.hpp>

#include <mxloguploader/Configuration.hpp>
#include <mxloguploader/HttpPostUploader.hpp>
#include <mxloguploader/ProfileConfiguration.hpp>
#include <mxloguploader/Uploader.hpp>

using boost::bind;
using boost::filesystem::path;
using boost::lexical_cast;
using boost::make_shared;
using boost::regex;
using boost::smatch;
using boost::shared_ptr;

using mxloguploader::ProfileConfiguration;
using mxloguploader::HttpPostUploader;
using mxloguploader::Log;
using mxloguploader::LogConverter;
using mxloguploader::ParseError;
using mxloguploader::StatusUpdater;
using mxloguploader::Uploader;

using std::ifstream;
using std::ios_base;
using std::string;
using std::vector;

namespace
{

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_WARN("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ParseError());
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_WARN("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ParseError());
    }

    return string(buf.begin(), buf.end());
}

shared_ptr<HttpPostUploader>
    create_http_post_uploader(
        const string &url,
        bool bypass_ssl,
        shared_ptr<ProfileConfiguration::Method::HttpPost::Authentication> authentication,
        unsigned int retry_count,
        unsigned int retry_period_sec)
{
    shared_ptr<HttpPostUploader> rtn = make_shared<HttpPostUploader>(
        url,
        bypass_ssl,
        retry_count,
        retry_period_sec);

    if (authentication)
        rtn->set_authentication(
            authentication->username,
            authentication->password);

    return rtn;
}

const string
    external_filename_from_internal_path(
        const path &p,
        LogConverter::format_t format,
        LogConverter::compress_t compress)
{
    const string filename = p.filename().string();

    static regex e("^([0-9]{8}-[0-9]{4})[0-9]{2}\\.log\\.gz$");
    smatch what;
    if (!regex_match(filename, what, e))
    {
        MXLOG_WARN("unexpected path: " + p.string());
        return p.string();
    }

    string external_filename = what[1].str();

    if (LogConverter::FORMAT_CSV == format)
        external_filename += ".csv";
    else if (LogConverter::FORMAT_XML == format)
        external_filename += ".xml";
    else if (LogConverter::FORMAT_JSON == format)
        external_filename += ".json";
    else
    {
        MXLOG_EROR("unexpected format: " + lexical_cast<string>(format));
        BOOST_VERIFY(!"unexpected format");
    }

    if (LogConverter::COMPRESS_GZIP == compress)
        external_filename += ".gz";
    else if (LogConverter::COMPRESS_NONE == compress)
        ;
    else
    {
        MXLOG_EROR("unexpected compress: " + lexical_cast<string>(compress));
        BOOST_VERIFY(!"unexpected compress");
    }


    return external_filename;
}

}

namespace mxloguploader
{

struct Builder::Impl
{
    Impl()
        : file_count(0)
    {
    }

    unsigned int        file_count;

    shared_ptr<Configuration>           configuration;
    shared_ptr<ProfileConfiguration>    profile_configuration;
    shared_ptr<StatusPublisher>         status_publisher;
    shared_ptr<StatusUpdater>           status_updater;
    shared_ptr<path>                    working_directory;
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_configuration(const path &p)
{
    pimpl->configuration = make_shared<Configuration>(Configuration::parse(readfile(p)));
}

void
    Builder::set_profile(const string &profile)
{
    pimpl->profile_configuration = make_shared<ProfileConfiguration>(
        ProfileConfiguration::parse(readfile(
            "/etc/mxloguploader/conf.d/" + profile + ".conf"))
    );

    pimpl->status_publisher = make_shared<StatusPublisher>(
        profile,
        "127.0.0.1",
        1883);

    pimpl->working_directory = make_shared<path>(path(pimpl->configuration->working_directory) / profile);

    pimpl->status_updater = make_shared<StatusUpdater>(
        *pimpl->working_directory / "status",
        pimpl->status_publisher);
}

void
    Builder::set_file(const path &p)
{
    pimpl->profile_configuration = make_shared<ProfileConfiguration>(
        ProfileConfiguration::parse(readfile(p))
    );
}

void
    Builder::set_file_count(unsigned int file_count)
{
    pimpl->file_count = file_count;
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(
        pimpl->working_directory &&
        pimpl->profile_configuration &&
        pimpl->status_updater);

    const path working_directory = *pimpl->working_directory;
    const ProfileConfiguration profile_configuration = *pimpl->profile_configuration;
    const shared_ptr<StatusUpdater> status_updater = pimpl->status_updater;
    const unsigned int file_count = pimpl->file_count;

    if (!is_directory(working_directory))
        create_directories(working_directory);

    const shared_ptr<LogConverter> log_converter = make_shared<LogConverter>(
        profile_configuration.content.format,
        profile_configuration.content.compress);

    const path uploaded_directory = working_directory / "uploaded";
    if (!is_directory(uploaded_directory))
        create_directories(uploaded_directory);

    App::ConvertExternalFilenameFunc external_filename = bind(
        &external_filename_from_internal_path,
        _1,
        profile_configuration.content.format,
        profile_configuration.content.compress
    );

    return make_shared<App>(
        profile_configuration.log_directory,
        working_directory,
        status_updater,
        log_converter,
        create_http_post_uploader(
            profile_configuration.method.http_post->url,
            profile_configuration.method.http_post->bypass_ssl,
            profile_configuration.method.http_post->authentication,
            profile_configuration.retry.count,
            profile_configuration.retry.period_sec),
        external_filename,
        file_count
    );
}

shared_ptr<AppTest>
    Builder::build_test() const
{
    BOOST_ASSERT(pimpl->profile_configuration);

    const ProfileConfiguration profile_configuration = *pimpl->profile_configuration;

    return make_shared<AppTest>(
        create_http_post_uploader(
            profile_configuration.method.http_post->url,
            profile_configuration.method.http_post->bypass_ssl,
            profile_configuration.method.http_post->authentication,
            profile_configuration.retry.count,
            profile_configuration.retry.period_sec)
    );
}

shared_ptr<AppClear>
    Builder::build_clear() const
{
    BOOST_ASSERT(pimpl->working_directory && pimpl->status_publisher);

    const path working_directory = *pimpl->working_directory;
    const shared_ptr<StatusPublisher> status_publisher = pimpl->status_publisher;

    return make_shared<AppClear>(
        working_directory,
        status_publisher);
}

}
