#include <mxloguploader/Log.hpp>

#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

using boost::filesystem::path;
using boost::filesystem::temp_directory_path;
using boost::filesystem::unique_path;
using boost::make_shared;
using boost::shared_ptr;
using boost::system::error_code;

namespace mxloguploader
{

struct Log::Impl
{
    Impl(
        const path &internal_log_path_,
        shared_ptr<LogConverter> converter_,
        const path &uploaded_symlink_)
        : internal_log_path(internal_log_path_)
        , converter(converter_)
        , uploaded_symlink(uploaded_symlink_)
    {
    }

    const path                      internal_log_path;
    const shared_ptr<LogConverter>  converter;
    const path                      uploaded_symlink;

    shared_ptr<path>        external_log_path;
};

Log::Log(
    const path &internal_log_path,
    shared_ptr<LogConverter> converter,
    const path &uploaded_symlink)
    : pimpl(new Impl(
        internal_log_path,
        converter,
        uploaded_symlink))
{
    BOOST_ASSERT(pimpl->internal_log_path.is_absolute());
    BOOST_ASSERT(pimpl->converter);
    BOOST_ASSERT(pimpl->uploaded_symlink.is_absolute());
}

Log::~Log()
{
    if (pimpl->external_log_path)
    {
        if (0 != unlink(pimpl->external_log_path->string().c_str()))
        {
            MXLOG_WARN("delete " + pimpl->external_log_path->string() + " failed, ignored");
        }

        pimpl->external_log_path.reset();
    }
}

const path &
    Log::internal_log_path() const
{
    return pimpl->internal_log_path;
}

const path &
    Log::get_external_log_path() const
{
    if (!pimpl->external_log_path)
    {
        const path external_log_path =
            temp_directory_path()
            / unique_path(pimpl->internal_log_path.filename().string() + "-%%%%%%%%");

        if (exists(external_log_path))
        {
            BOOST_THROW_EXCEPTION(ConversionError());
        }

        try
        {
            pimpl->converter->convert(
                pimpl->internal_log_path,
                external_log_path);
        }
        catch(ConversionError &)
        {
            error_code ec;
            remove(external_log_path, ec); // ec ignored

            throw;
        }

        pimpl->external_log_path = make_shared<path>(external_log_path);
    }

    BOOST_ASSERT(pimpl->external_log_path);
    return *pimpl->external_log_path;
}

void
    Log::mark_uploaded()
{
    error_code ec;
    create_symlink(pimpl->internal_log_path, pimpl->uploaded_symlink, ec);

    if (ec)
    {
        BOOST_THROW_EXCEPTION(SystemError());
    }
}

}
