#include <mxloguploader/LogConverter.hpp>

#include <fstream>
#include <sstream>

#include <boost/iostreams/device/file_descriptor.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/system/error_code.hpp>

#include <mxutil/log.hpp>

using boost::filesystem::path;
using boost::lexical_cast;
using boost::iostreams::never_close_handle;
using boost::iostreams::copy;
using boost::iostreams::file_descriptor_source;
using boost::iostreams::filtering_ostream;
using boost::iostreams::gzip_compressor;
using boost::iostreams::stream_buffer;
using boost::system::error_code;

using mxloguploader::LogConverter;

using std::ios_base;
using std::istream;
using std::ofstream;
using std::string;
using std::stringstream;

namespace
{

const string
    cmd_from_format(LogConverter::format_t);

const string
    cmd_from_compress(LogConverter::compress_t);

}

namespace mxloguploader
{

struct LogConverter::Impl
{
    Impl(
        format_t format_,
        compress_t compress_)
        : format(format_)
        , compress(compress_)
    {
    }

    const format_t      format;
    const compress_t    compress;
};

LogConverter::LogConverter(
    format_t format,
    compress_t compress)
    : pimpl(new Impl(
        format,
        compress))
{
}

LogConverter::~LogConverter()
{
}

void
    LogConverter::convert(
        const path &from,
        const path &to) const
{
    stringstream cmd;
    cmd << cmd_from_format(pimpl->format)
        << " < " << from;

    FILE *f = popen(cmd.str().c_str(), "r");
    if (f == NULL)
    {
        MXLOG_WARN("popen failed: " + cmd.str());
        BOOST_THROW_EXCEPTION(ConversionError());
    }

    try
    {
        stream_buffer<file_descriptor_source> fs(fileno(f), never_close_handle);
        istream in(&fs);

        filtering_ostream out;

        switch(pimpl->compress)
        {
        case COMPRESS_NONE:
            // do nothing
            break;
        case COMPRESS_GZIP:
            out.push(gzip_compressor());
            break;
        default:
            MXLOG_EROR("unexpected pimpl->compress: " + lexical_cast<string>(pimpl->compress));
            exit(1);
        }

        ofstream outfile(to.string().c_str(), ios_base::binary);
        if (!outfile)
        {
            MXLOG_WARN("open file error: " + to.string());
            BOOST_THROW_EXCEPTION(ConversionError());
        }

        out.push(outfile);

        try
        {
            copy(in, out);
        }
        catch(boost::exception &e)
        {
            MXLOG_WARN(diagnostic_information(e));
            BOOST_THROW_EXCEPTION(ConversionError());
        }

        const int r = pclose(f);
        f = NULL;
        const int exit_status = WEXITSTATUS(r);
        if (exit_status != 0)
        {
            error_code ec;
            remove(to, ec); // ec ignored

            BOOST_THROW_EXCEPTION(ConversionError());
        }
    }
    catch(...)
    {
        if (f)
            pclose(f);

        throw;
    }
}

}

namespace
{

const string
    cmd_from_format(LogConverter::format_t f)
{
    string format;
    switch(f)
    {
    case LogConverter::FORMAT_CSV:
        format = "csv";
        break;

    case LogConverter::FORMAT_JSON:
        format = "json";
        break;

    case LogConverter::FORMAT_XML:
        format = "xml";
        break;

    default:
        BOOST_VERIFY(!"unexpected format");
    }

    return "mxlogconverter --format " + format;
}

}
