#include <mxloguploader/ProfileConfiguration.hpp>

#include <map>

#include <boost/make_shared.hpp>

#include <libconfig.h++>

#include <mxutil/log.hpp>

using boost::make_shared;

using libconfig::Config;
using libconfig::ParseException;
using libconfig::Setting;
using libconfig::SettingNotFoundException;

using mxloguploader::ProfileConfiguration;
using mxloguploader::LogConverter;
using mxloguploader::ParseError;

using std::map;
using std::string;

namespace
{

void
    validate_configuration(const ProfileConfiguration &);

LogConverter::format_t
    format_from_str(const string &);

LogConverter::compress_t
    compress_from_str(const string &);

}

namespace mxloguploader
{

const ProfileConfiguration
    ProfileConfiguration::parse(const string &content)
{
    Config config;

    ProfileConfiguration rtn;

    try
    {
        config.readString(content.c_str());

        if (config.exists("method"))
        {
            const Setting &method = config.lookup("method");

            if (method.exists("http-post"))
            {
                const Setting &http_post = method["http-post"];
                rtn.method.http_post = make_shared<ProfileConfiguration::Method::HttpPost>();
                if (!http_post.lookupValue("url", rtn.method.http_post->url))
                    BOOST_THROW_EXCEPTION(ParseError());

                if (http_post.exists("authentication"))
                {
                    const Setting &authentication = http_post["authentication"];

                    rtn.method.http_post->authentication =
                        make_shared<ProfileConfiguration::Method::HttpPost::Authentication>();

                    if (!authentication.lookupValue(
                            "username",
                            rtn.method.http_post->authentication->username) ||
                        !authentication.lookupValue(
                            "password",
                            rtn.method.http_post->authentication->password))
                        BOOST_THROW_EXCEPTION(ParseError());
                }

                if (!http_post.lookupValue("bypass-ssl", rtn.method.http_post->bypass_ssl))
                    BOOST_THROW_EXCEPTION(ParseError());
            }
        }

        if (!config.lookupValue("retry.count", rtn.retry.count))
            BOOST_THROW_EXCEPTION(ParseError());

        if (!config.lookupValue("retry.period-sec", rtn.retry.period_sec))
            BOOST_THROW_EXCEPTION(ParseError());

        if (!config.lookupValue("log-directory", rtn.log_directory))
            BOOST_THROW_EXCEPTION(ParseError());

        string format;
        if (!config.lookupValue("content.format", format))
            BOOST_THROW_EXCEPTION(ParseError());

        rtn.content.format = format_from_str(format);

        string compress;
        if (!config.lookupValue("content.compress", compress))
            BOOST_THROW_EXCEPTION(ParseError());

        rtn.content.compress = compress_from_str(compress);

        validate_configuration(rtn);

        return rtn;
    }
    catch(const ParseException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
    catch(const SettingNotFoundException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
}

}

namespace
{

bool
    has_colon(const string &s)
{
    const size_t pos = s.find(':');
    return pos != string::npos;
}

void
    validate_configuration(const ProfileConfiguration &config)
{
    if (!config.method.http_post)
        BOOST_THROW_EXCEPTION(ParseError());

    if (config.method.http_post &&
        config.method.http_post->authentication &&
        has_colon(config.method.http_post->authentication->username))
        BOOST_THROW_EXCEPTION(ParseError());
}

const map<string, LogConverter::format_t>
    get_format_map()
{
    map<string, LogConverter::format_t> rtn;
    rtn["json"] = LogConverter::FORMAT_JSON;
    rtn["csv"] = LogConverter::FORMAT_CSV;
    rtn["xml"] = LogConverter::FORMAT_XML;

    return rtn;
}

LogConverter::format_t
    format_from_str(const string &s)
{
    static const map<string, LogConverter::format_t> format_map = get_format_map();

    const map<string, LogConverter::format_t>::const_iterator iter = format_map.find(s);
    if (iter == format_map.end())
        BOOST_THROW_EXCEPTION(ParseError());

    return iter->second;
}

const map<string, LogConverter::compress_t>
    get_compress_map()
{
    map<string, LogConverter::compress_t> rtn;
    rtn["none"] = LogConverter::COMPRESS_NONE;
    rtn["gzip"] = LogConverter::COMPRESS_GZIP;

    return rtn;
}

LogConverter::compress_t
    compress_from_str(const string &s)
{
    static const map<string, LogConverter::compress_t> compress_map = get_compress_map();

    const map<string, LogConverter::compress_t>::const_iterator iter = compress_map.find(s);
    if (iter == compress_map.end())
        BOOST_THROW_EXCEPTION(ParseError());

    return iter->second;
}

}
