#include <mxloguploader/Status.hpp>

#include <rapidjson/document.h>
#include <rapidjson/stringbuffer.h>
#include <rapidjson/writer.h>

#include <mxutil/log.hpp>

#include <mxtag/tag_form.hpp>
#include <mxtag/tag_parse.hpp>

using boost::make_shared;
using boost::posix_time::ptime;

using mxloguploader::ParseError;
using mxloguploader::Status;

using mxtag::form_iso8601_time;
using mxtag::parse_iso8601_time;

using rapidjson::Document;
using rapidjson::Value;
using rapidjson::StringBuffer;
using rapidjson::Writer;

using std::string;

namespace
{

const Status::LastAttempt
    parse_last_attempt(const Value &v);

const Status::LastSuccess
    parse_last_success(const Value &v);

}

namespace mxloguploader
{
/*
Status Format: (in JSON)

{
    "lastAttempt": {
        "name": "/var/mxc/logs/20150706-153807.log.gz",
        "at": "2015-07-06T15:38:46Z",
        "status": "remote server error"
    },
    "lastSuccess": {
        "name": "/var/mxc/logs/20150706-123456.log.gz",
        "at": "2015-07-06T13:46:22Z"
    },
    "fileUploadCount": 47
}
 */

const Status
    Status::parse(const string &s)
{
    Document doc;
    doc.Parse(s.c_str());
    if (!doc.IsObject())
        BOOST_THROW_EXCEPTION(ParseError());

    Status status;

    if (doc.HasMember("lastAttempt"))
    {
        status.last_attempt = make_shared<Status::LastAttempt>(
            parse_last_attempt(doc["lastAttempt"]));
    }

    if (doc.HasMember("lastSuccess"))
    {
        status.last_success = make_shared<Status::LastSuccess>(
            parse_last_success(doc["lastSuccess"]));
    }

    if (!doc.HasMember("fileUploadCount"))
        BOOST_THROW_EXCEPTION(ParseError());

    status.file_upload_count = doc["fileUploadCount"].GetInt64();

    return status;
}

const string
    Status::str() const
{
    Document doc;
    doc.SetObject();

    if (this->last_attempt)
    {
        Value v;
        v
            .SetObject()
            .AddMember(
                Value("name", doc.GetAllocator()).Move(),
                Value(this->last_attempt->name.c_str(), doc.GetAllocator()).Move(),
                doc.GetAllocator())
            .AddMember(
                Value("at", doc.GetAllocator()).Move(),
                Value(form_iso8601_time(this->last_attempt->at).c_str(), doc.GetAllocator()).Move(),
                doc.GetAllocator())
            .AddMember(
                Value("status", doc.GetAllocator()).Move(),
                Value(this->last_attempt->status.c_str(), doc.GetAllocator()).Move(),
                doc.GetAllocator())
            ;

        doc.AddMember(
            Value("lastAttempt", doc.GetAllocator()).Move(),
            v,
            doc.GetAllocator());
    }

    if (this->last_success)
    {
        Value v;
        v
            .SetObject()
            .AddMember(
                Value("name", doc.GetAllocator()).Move(),
                Value(this->last_success->name.c_str(), doc.GetAllocator()).Move(),
                doc.GetAllocator())
            .AddMember(
                Value("at", doc.GetAllocator()).Move(),
                Value(form_iso8601_time(this->last_success->at).c_str(), doc.GetAllocator()).Move(),
                doc.GetAllocator())
            ;

        doc.AddMember(
            Value("lastSuccess", doc.GetAllocator()).Move(),
            v,
            doc.GetAllocator());
    }

    doc.AddMember(
        Value("fileUploadCount").Move(),
        Value(this->file_upload_count).Move(),
        doc.GetAllocator());

    StringBuffer buf;
    Writer<StringBuffer> writer(buf);
    doc.Accept(writer);

    return buf.GetString();
}

}

namespace
{

const Status::LastAttempt
    parse_last_attempt(const Value &v)
{
    if (!v.HasMember("name") ||
        !v["name"].IsString() ||

        !v.HasMember("at") ||
        !v["name"].IsString() ||

        !v.HasMember("status") ||
        !v["status"].IsString())
        BOOST_THROW_EXCEPTION(ParseError());

    ptime at;
    if (!parse_iso8601_time(v["at"].GetString(), at))
        BOOST_THROW_EXCEPTION(ParseError());

    return Status::LastAttempt(
        v["name"].GetString(),
        at,
        v["status"].GetString());
}

const Status::LastSuccess
    parse_last_success(const Value &v)
{
    if (!v.HasMember("name") ||
        !v["name"].IsString() ||

        !v.HasMember("at") ||
        !v["at"].IsString())
        BOOST_THROW_EXCEPTION(ParseError());

    ptime at;
    if (!parse_iso8601_time(v["at"].GetString(), at))
        BOOST_THROW_EXCEPTION(ParseError());

    return Status::LastSuccess(
        v["name"].GetString(),
        at);
}

}
