#include <mxloguploader/StatusUpdater.hpp>

#include <fstream>
#include <vector>

#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>

#include <mxloguploader/Status.hpp>

using boost::filesystem::path;
using boost::make_shared;
using boost::posix_time::ptime;
using boost::shared_ptr;

using mxloguploader::ParseError;

using std::ifstream;
using std::ios_base;
using std::ofstream;
using std::string;
using std::vector;

namespace
{

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_WARN("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ParseError());
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_WARN("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ParseError());
    }

    return string(buf.begin(), buf.end());
}

void
    writefile(
        const path &p,
        const string &content)
{
    ofstream outfile(p.string().c_str(), ios_base::binary);
    if (!outfile)
    {
        MXLOG_WARN("open file failed: " + p.string());
        return;
    }

    outfile.write(content.c_str(), content.size());
}

}

namespace mxloguploader
{

struct StatusUpdater::Impl
{
    Impl(
        const path &status_file_,
        shared_ptr<StatusPublisher> publisher_)
        : status_file(status_file_)
        , publisher(publisher_)
    {
    }

    const path                          status_file;
    const shared_ptr<StatusPublisher>   publisher;

    Status      status;
};

StatusUpdater::StatusUpdater(
    const path &status_file,
    shared_ptr<StatusPublisher> publisher)
    : pimpl(new Impl(
        status_file,
        publisher))
{
    BOOST_ASSERT(pimpl->publisher);

    try
    {
        pimpl->status = Status::parse(readfile(status_file));
    }
    catch(ParseError &)
    {
        // do nothing
    }
}

StatusUpdater::~StatusUpdater()
{
}

void
    StatusUpdater::update(
        const path &internal_log,
        bool success,
        const string &status,
        const ptime &now)
{
    // update value
    pimpl->status.last_attempt = make_shared<Status::LastAttempt>(
        internal_log.string(),
        now,
        status);

    if (success)
    {
        pimpl->status.last_success = make_shared<Status::LastSuccess>(
            internal_log.string(),
            now);

        ++pimpl->status.file_upload_count;
    }

    // store status-file
    writefile(pimpl->status_file, pimpl->status.str());

    // update mqtt
    pimpl->publisher->publish(pimpl->status);
}

}
