#include <mxmodbusgateway/Builder.hpp>

#include <fstream>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusgateway/Configuration.hpp>
#include <mxmodbusgateway/worker/ModbusWorker.hpp>

using boost::asio::io_service;
using boost::bind;
using boost::filesystem::path;
using boost::shared_ptr;

using mxmodbus::ModbusRtuClient;

using mxmodbusgateway::ArgumentError;
using mxmodbusgateway::worker::ModbusWorker;

using std::ifstream;
using std::ios_base;
using std::string;
using std::vector;

namespace
{

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_EROR("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_EROR("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    return string(buf.begin(), buf.end());
}

shared_ptr<ModbusRtuClient>
    create_modbus_rtu_client(
        const string &path,
        ModbusRtuClient::baud_t b,
        ModbusRtuClient::parity_t p,
        ModbusRtuClient::stop_bit_t s,
        int interchar_timeout_ms,
        int response_timeout_ms)
{
    return shared_ptr<ModbusRtuClient>(new ModbusRtuClient(
        path,
        b,
        p,
        s,
        interchar_timeout_ms,
        response_timeout_ms));
}

}

namespace mxmodbusgateway
{

struct Builder::Impl
{
    shared_ptr<Configuration>   configuration;
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_configuration(const path &p)
{
    try
    {
        pimpl->configuration.reset(new Configuration(
            Configuration::parse(readfile(p))
        ));
    }
    catch(ParseError &e)
    {
        MXLOG_EROR(diagnostic_information(e));
        BOOST_THROW_EXCEPTION(ArgumentError());
    }
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(pimpl->configuration);

    const shared_ptr<io_service> iosvc(new io_service());

    vector<shared_ptr<Server> > server_list;
    BOOST_FOREACH(const Configuration::Port &port, pimpl->configuration->port_list)
    {
        const Configuration::Port::Rtu &rtu = port.rtu;
        const Configuration::Port::Tcp &tcp = port.tcp;

        const shared_ptr<ModbusWorker> worker(new ModbusWorker(
            *iosvc,
            bind(
                &create_modbus_rtu_client,
                rtu.path,
                rtu.serial_config.baud,
                rtu.serial_config.parity,
                rtu.serial_config.stopbit,
                rtu.interchar_timeout_ms,
                rtu.request_timeout_ms
            ),
            rtu.interframe_delay_ms
        ));

        const shared_ptr<Server> server(new Server(
            *iosvc,
            tcp.host,
            tcp.service,
            worker));

        server_list.push_back(server);
    }

    return shared_ptr<App>(new App(
        iosvc,
        server_list
    ));
}

}
