#include <mxmodbusgateway/Configuration.hpp>

#include <map>

#include <boost/regex.hpp>

#include <libconfig.h++>

namespace mxmodbusgateway
{

using boost::regex;
using boost::regex_match;
using boost::smatch;

using libconfig::Config;
using libconfig::ParseException;
using libconfig::Setting;
using libconfig::SettingNotFoundException;

using mxmodbus::ModbusRtuClient;

using std::map;
using std::string;
using std::vector;

namespace
{

const Configuration::Port::Rtu::SerialConfig
    parse_serial_config(const string &);

}

const Configuration
    Configuration::parse(const string &content)
{
    Config config;

    try
    {
        config.readString(content.c_str());
        const Setting &port_list = config.lookup("port-list");
        vector<Configuration::Port> cfg_port_list;
        for (int i = 0; i < port_list.getLength(); i++)
        {
            const Setting &port = port_list[i];

            const Setting &tcp = port["tcp"];
            const string host = tcp["host"];
            const int service = tcp["service"];

            const Setting &rtu = port["rtu"];
            const string path = rtu["path"];
            const Configuration::Port::Rtu::SerialConfig serial_config =
                parse_serial_config(rtu["serial-config"]);
            const unsigned int interframe_delay_ms = rtu["interframe-delay-ms"];
            const unsigned int interchar_timeout_ms = rtu["interchar-timeout-ms"];
            const unsigned int request_timeout_ms = rtu["request-timeout-ms"];

            cfg_port_list.push_back(Configuration::Port(
                Configuration::Port::Tcp(
                    host,
                    service),
                Configuration::Port::Rtu(
                    path,
                    serial_config,
                    interframe_delay_ms,
                    interchar_timeout_ms,
                    request_timeout_ms)
            ));
        }

        return Configuration(cfg_port_list);
    }
    catch(const ParseException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
    catch(const SettingNotFoundException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
}

namespace
{

const Configuration::Port::Rtu::SerialConfig
    parse_serial_config(const string &str)
{
    static const regex e(
        "^((?:9600)|(?:19200)|(?:38400)|(?:57600)|(?:115200))-8-(n|e|o)-(1|2)$"
    );

    smatch what;
    if (!regex_match(
            str,
            what,
            e))
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }

    const string baud_str = what[1];
    const string parity_str = what[2];
    const string stopbit_str = what[3];

    map<string, ModbusRtuClient::baud_t> sbaud_map;
    sbaud_map["9600"] = ModbusRtuClient::BAUD_9600;
    sbaud_map["19200"] = ModbusRtuClient::BAUD_19200;
    sbaud_map["38400"] = ModbusRtuClient::BAUD_38400;
    sbaud_map["57600"] = ModbusRtuClient::BAUD_57600;
    sbaud_map["115200"] = ModbusRtuClient::BAUD_115200;

    map<string, ModbusRtuClient::parity_t> sparity_map;
    sparity_map["n"] = ModbusRtuClient::NONE;
    sparity_map["e"] = ModbusRtuClient::EVEN;
    sparity_map["o"] = ModbusRtuClient::ODD;

    map<string, ModbusRtuClient::stop_bit_t> sstopbit_map;
    sstopbit_map["1"] = ModbusRtuClient::STOP_1;
    sstopbit_map["2"] = ModbusRtuClient::STOP_2;

    return Configuration::Port::Rtu::SerialConfig(
        sbaud_map.at(baud_str),
        sparity_map.at(parity_str),
        sstopbit_map.at(stopbit_str));
}

}

}
