#include <mxmodbusgateway/ReadCoilsCommand.hpp>

#include <boost/bind.hpp>

#include <mxmodbus/tcp/modbus_tcp_protocol.hpp>

namespace mxmodbusgateway
{

using boost::asio::io_service;
using boost::bind;
using boost::shared_ptr;
using boost::system::error_code;

using mxutil::CompleteCallback;

using mxmodbus::address_t;
using mxmodbus::device_t;
using mxmodbus::exception_code;
using mxmodbus::tcp::parse_read_coils_request;
using mxmodbus::tcp::form_read_coils_exception;
using mxmodbus::tcp::form_read_coils_response;

using mxmodbusgateway::worker::ModbusWorker;

using std::vector;

shared_ptr<ReadCoilsCommand>
    ReadCoilsCommand::parse(
        vector<uint8_t> &buf,
        io_service &iosvc)
{
    uint16_t packet_id;
    device_t device;
    address_t address;
    uint16_t quantity;

    const int result = parse_read_coils_request(
        buf,
        packet_id,
        device,
        address,
        quantity);

    if (result == 0)
        return shared_ptr<ReadCoilsCommand>();

    if (result < 0)
        BOOST_THROW_EXCEPTION(ParseError());

    BOOST_ASSERT(result > 0);

    // consume bytes read
    buf.erase(buf.begin(), buf.begin() + result);

    return shared_ptr<ReadCoilsCommand>(new ReadCoilsCommand(
        iosvc,
        packet_id,
        device,
        address,
        quantity
    ));
}

struct ReadCoilsCommand::Impl
{
    Impl(
        io_service &iosvc_,
        uint16_t packet_id_,
        device_t device_,
        address_t address_,
        size_t quantity_)
        : iosvc(iosvc_)
        , packet_id(packet_id_)
        , device(device_)
        , address(address_)
        , quantity(quantity_)
    {
    }

    io_service &    iosvc;
    const uint16_t  packet_id;
    const device_t  device;
    const address_t address;
    const size_t    quantity;

    vector<bool>    read_buf;
};

ReadCoilsCommand::ReadCoilsCommand(
    io_service &iosvc,
    uint16_t packet_id,
    device_t device,
    address_t address,
    size_t quantity)
    : pimpl(new Impl(
        iosvc,
        packet_id,
        device,
        address,
        quantity))
{
}

ReadCoilsCommand::~ReadCoilsCommand()
{
}

void
    ReadCoilsCommand::async_execute(
        ModbusWorker &worker,
        vector<uint8_t> &out,
        const CompleteCallback &cc)
{
    worker.async_read_coils(
        pimpl->device,
        pimpl->address,
        pimpl->quantity,
        pimpl->read_buf,
        bind(
            &ReadCoilsCommand::on_read_complete_,
            this->shared_from_this(),
            _1,
            &out,
            cc
        )
    );
}

void
    ReadCoilsCommand::on_read_complete_(
        const error_code &ec,
        vector<uint8_t> *out,
        const CompleteCallback &cc)
{
    BOOST_ASSERT(out);

    if (ec)
    {
        *out = form_read_coils_exception(
            pimpl->packet_id,
            pimpl->device,
            errc::modbus_exception_from_error_code(ec)
        );

        pimpl->iosvc.post(bind(cc, error_code()));
    }
    else
    {
        *out = form_read_coils_response(
            pimpl->packet_id,
            pimpl->device,
            pimpl->read_buf
        );

        pimpl->iosvc.post(bind(cc, error_code()));
    }
}

}
