#include <mxmodbusgateway/Server.hpp>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusgateway/Session.hpp>

namespace mxmodbusgateway
{

using boost::asio::io_service;
using boost::asio::ip::tcp;
using boost::bind;
using boost::lexical_cast;
using boost::make_shared;
using boost::shared_ptr;
using boost::system::error_code;

using mxmodbusgateway::worker::ModbusWorker;

using mxutil::CompleteCallback;

using std::string;
using std::vector;

namespace
{

void
    dummy_callback(const error_code &ec)
{
    if (!ec)
    {
        MXLOG_WARN(ec.message());
    }
}

}

struct Server::Impl
{
    Impl(
        io_service &iosvc_,
        const string &host_,
        int port_,
        shared_ptr<ModbusWorker> worker_)
        : iosvc(iosvc_)
        , host(host_)
        , port(port_)
        , worker(worker_)
    {
    }

    io_service &                    iosvc;
    const string                    host;
    const int                       port;
    const shared_ptr<ModbusWorker>  worker;

    shared_ptr<tcp::resolver>           resolver;
    vector<shared_ptr<tcp::acceptor> >  acceptor_list;

    shared_ptr<CompleteCallback>        cc;

    void
        complete(const error_code &ec)
    {
        if (this->cc)
        {
            this->iosvc.post(bind(
                *cc,
                ec
            ));

            this->cc.reset();
        }
    }
};

Server::Server(
    io_service &iosvc,
    const string &host,
    int port,
    shared_ptr<ModbusWorker> worker)
    : pimpl(new Impl(
        iosvc,
        host,
        port,
        worker))
{
    BOOST_ASSERT(pimpl->worker);
}

Server::~Server()
{
    this->cancel();
}

void
    Server::async_run(CompleteCallback cc)
{
    BOOST_ASSERT(!pimpl->cc);

    pimpl->cc.reset(new CompleteCallback(cc));
    pimpl->resolver.reset(new tcp::resolver(pimpl->iosvc));

    const tcp::resolver::query q(
        pimpl->host,
        lexical_cast<string>(pimpl->port),
        tcp::resolver::query::passive);

    pimpl->resolver->async_resolve(
        q,
        bind(
            &Server::on_resolve_complete_,
            this->shared_from_this(),
            boost::asio::placeholders::error,
            boost::asio::placeholders::iterator
        )
    );
}

void
    Server::cancel()
{
    if (pimpl->resolver)
    {
        pimpl->resolver->cancel();
    }

    BOOST_FOREACH(shared_ptr<tcp::acceptor> acceptor, pimpl->acceptor_list)
    {
        acceptor->cancel();
    }
}

void
    Server::on_resolve_complete_(
        const error_code &ec,
        tcp::resolver::iterator endpoint_iter)
{
    if (ec)
    {
        pimpl->complete(ec);
        return;
    }

    const tcp::resolver::iterator end;
    for(; endpoint_iter != end; ++endpoint_iter)
    {
        const shared_ptr<tcp::acceptor> acceptor(new tcp::acceptor(
            pimpl->iosvc,
            *endpoint_iter
        ));

        const shared_ptr<tcp::socket> skt(new tcp::socket(pimpl->iosvc));

        acceptor->async_accept(
            *skt,
            bind(
                &Server::on_accept_,
                this->shared_from_this(),
                boost::asio::placeholders::error,
                acceptor,
                skt
            )
        );

        pimpl->acceptor_list.push_back(acceptor);
    }

    pimpl->resolver.reset();
}

void
    Server::on_accept_(
        const error_code &ec,
        shared_ptr<tcp::acceptor> acceptor,
        shared_ptr<tcp::socket> skt)
{
    if (ec)
    {
        pimpl->complete(ec);
        this->cancel();
        return;
    }

    BOOST_ASSERT(acceptor && skt);

    // create Session
    const shared_ptr<Session> session(new Session(
        pimpl->iosvc,
        skt,
        pimpl->worker
    ));

    session->async_run(&dummy_callback);

    skt.reset(new tcp::socket(pimpl->iosvc));
    acceptor->async_accept(
        *skt,
        bind(
            &Server::on_accept_,
            this->shared_from_this(),
            boost::asio::placeholders::error,
            acceptor,
            skt
        )
    );
}

}
