#include <mxmodbusgateway/Session.hpp>

#include <vector>

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusgateway/CommandFactory.hpp>
#include <mxmodbusgateway/errc.hpp>

namespace mxmodbusgateway
{

using boost::asio::buffer;
using boost::asio::deadline_timer;
using boost::asio::io_service;
using boost::asio::ip::tcp;
using boost::asio::streambuf;
using boost::asio::placeholders::error;
using boost::asio::placeholders::bytes_transferred;
using boost::bind;
using boost::lexical_cast;
using boost::posix_time::pos_infin;
using boost::posix_time::seconds;
using boost::shared_ptr;
using boost::system::error_code;

using mxmodbusgateway::worker::ModbusWorker;

using mxutil::CompleteCallback;

using std::streamsize;
using std::string;
using std::vector;

namespace
{

const int   INPUT_TIMEOUT_SEC = 60;

}

struct Session::Impl
{
    Impl(
        io_service &iosvc_,
        shared_ptr<tcp::socket> skt_,
        shared_ptr<ModbusWorker> &worker_)
        : iosvc(iosvc_)
        , skt(skt_)
        , worker(worker_)
        , input_buffer(1024)
        , input_timeout(iosvc_)
    {
    }

    io_service &                    iosvc;
    const shared_ptr<tcp::socket>   skt;
    const shared_ptr<ModbusWorker>  worker;

    shared_ptr<CompleteCallback>        cc;

    vector<uint8_t>         input_buffer;
    deadline_timer          input_timeout;

    vector<uint8_t>         output_buffer;

    CommandFactory          cmd_factory;

    void
        complete(const error_code &ec)
    {
        if (this->cc)
        {
            this->iosvc.post(bind(
                *cc,
                ec
            ));

            this->cc.reset();
        }

        this->skt->cancel();
        this->input_timeout.cancel();
    }

    bool
        stopped() const
    {
        return this->cc == NULL;
    }
};

Session::Session(
    io_service &iosvc,
    shared_ptr<tcp::socket> skt,
    shared_ptr<ModbusWorker> worker)
    : pimpl(new Impl(
        iosvc,
        skt,
        worker))
{
    BOOST_ASSERT(pimpl->skt && pimpl->worker);
}

Session::~Session()
{
    this->cancel();
}

void
    Session::async_run(CompleteCallback cc)
{
    BOOST_ASSERT(!pimpl->cc);
    pimpl->cc.reset(new CompleteCallback(cc));

    // setup timeout
    pimpl->input_timeout.async_wait(bind(
        &Session::on_input_timeout_,
        this->shared_from_this(),
        error
    ));

    pimpl->input_timeout.expires_from_now(seconds(INPUT_TIMEOUT_SEC));
    this->execute_next_command_();
}

void
    Session::cancel()
{
    pimpl->skt->cancel();
    pimpl->input_timeout.cancel();
}

void
    Session::execute_next_command_()
{
    try
    {
        const shared_ptr<Command> cmd = pimpl->cmd_factory.extract_command(pimpl->iosvc);

        if (cmd)
        {
            cmd->async_execute(
                *pimpl->worker,
                pimpl->output_buffer,
                bind(
                    &Session::on_command_complete_,
                    this->shared_from_this(),
                    error,
                    cmd
                )
            );

            // turn-off timer
            pimpl->input_timeout.expires_at(pos_infin);
        }
        else
        {
            pimpl->skt->async_read_some(
                buffer(pimpl->input_buffer),
                bind(
                    &Session::on_read_complete_,
                    this->shared_from_this(),
                    error,
                    bytes_transferred
                )
            );
        }
    }
    catch(ParseError &e)
    {
        MXLOG_WARN(diagnostic_information(e));
        pimpl->complete(error_code(errc::parse_error, errc::category()));
        return;
    }
}

void
    Session::on_read_complete_(
        const error_code &ec,
        size_t bytes_read)
{
    if (pimpl->stopped())
        return;

    if (ec)
    {
        pimpl->complete(ec);
        return;
    }

    const vector<uint8_t> buf(pimpl->input_buffer.begin(), pimpl->input_buffer.begin() + bytes_read);
    if (!buf.empty())
    {
        MXLOG_DBUG("Session::on_read_complete_(): buf.size() = " + lexical_cast<string>(buf.size()));
        // reset timer
        pimpl->input_timeout.expires_from_now(seconds(INPUT_TIMEOUT_SEC));
    }

    pimpl->cmd_factory.append_raw(buf);

    this->execute_next_command_();
}

void
    Session::on_command_complete_(
        const error_code &ec,
        shared_ptr<Command>)
{
    if (pimpl->stopped())
        return;

    if (ec)
    {
        pimpl->complete(ec);
        return;
    }

    BOOST_ASSERT(!pimpl->output_buffer.empty());

    async_write(
        *pimpl->skt,
        buffer(pimpl->output_buffer),
        bind(
            &Session::on_write_complete_,
            this->shared_from_this(),
            error
        )
    );
}

void
    Session::on_write_complete_(
        const error_code &ec)
{
    if (pimpl->stopped())
        return;

    if (ec)
    {
        pimpl->complete(ec);
        return;
    }

    pimpl->output_buffer.clear();

    // re-enable timer
    pimpl->input_timeout.expires_from_now(seconds(INPUT_TIMEOUT_SEC));
    this->execute_next_command_();
}

void
    Session::on_input_timeout_(
        const error_code &ec)
{
    if (pimpl->stopped())
        return;

    if (ec &&
        ec != boost::asio::error::operation_aborted)
    {
        pimpl->complete(ec);
        return;
    }

    if (pimpl->input_timeout.expires_at() <= deadline_timer::traits_type::now())
    {
        pimpl->complete(error_code(errc::timeout_error, errc::category()));
    }
    else
    {
        pimpl->input_timeout.async_wait(bind(
            &Session::on_input_timeout_,
            this->shared_from_this(),
            error
        ));
    }
}

}
