#include <mxmodbusgateway/WriteSingleCoilCommand.hpp>

#include <boost/bind.hpp>

#include <mxmodbus/tcp/modbus_tcp_protocol.hpp>

using boost::asio::io_service;
using boost::bind;
using boost::shared_ptr;
using boost::system::error_code;

using mxutil::CompleteCallback;

using mxmodbus::address_t;
using mxmodbus::device_t;
using mxmodbus::exception_code;
using mxmodbus::tcp::parse_write_single_coil_request;
using mxmodbus::tcp::form_write_single_coil_exception;
using mxmodbus::tcp::form_write_single_coil_response;

using mxmodbusgateway::worker::ModbusWorker;

using std::vector;

namespace mxmodbusgateway
{

shared_ptr<WriteSingleCoilCommand>
    WriteSingleCoilCommand::parse(
        vector<uint8_t> &buf,
        io_service &iosvc)
{
    uint16_t packet_id;
    device_t device;
    address_t address;
    bool value;

    const int result = parse_write_single_coil_request(
        buf,
        packet_id,
        device,
        address,
        value);

    if (result == 0)
        return shared_ptr<WriteSingleCoilCommand>();

    if (result < 0)
        BOOST_THROW_EXCEPTION(ParseError());

    BOOST_ASSERT(result > 0);

    // consume bytes read
    buf.erase(buf.begin(), buf.begin() + result);

    return shared_ptr<WriteSingleCoilCommand>(new WriteSingleCoilCommand(
        iosvc,
        packet_id,
        device,
        address,
        value
    ));
}

struct WriteSingleCoilCommand::Impl
{
    Impl(
        io_service &iosvc_,
        uint16_t packet_id_,
        device_t device_,
        address_t address_,
        bool value_)
        : iosvc(iosvc_)
        , packet_id(packet_id_)
        , device(device_)
        , address(address_)
        , value(value_)
    {
    }

    io_service &    iosvc;
    const uint16_t  packet_id;
    const device_t  device;
    const address_t address;
    const bool      value;
};

WriteSingleCoilCommand::WriteSingleCoilCommand(
    io_service &iosvc,
    uint16_t packet_id,
    device_t device,
    address_t address,
    bool value)
    : pimpl(new Impl(
        iosvc,
        packet_id,
        device,
        address,
        value))
{
}

WriteSingleCoilCommand::~WriteSingleCoilCommand()
{
}

void
    WriteSingleCoilCommand::async_execute(
        ModbusWorker &worker,
        vector<uint8_t> &out,
        const CompleteCallback &cc)
{
    worker.async_write_single_coil(
        pimpl->device,
        pimpl->address,
        pimpl->value,
        bind(
            &WriteSingleCoilCommand::on_read_complete_,
            this->shared_from_this(),
            _1,
            &out,
            cc
        )
    );
}

void
    WriteSingleCoilCommand::on_read_complete_(
        const error_code &ec,
        vector<uint8_t> *out,
        const CompleteCallback &cc)
{
    BOOST_ASSERT(out);

    if (ec)
    {
        *out = form_write_single_coil_exception(
            pimpl->packet_id,
            pimpl->device,
            errc::modbus_exception_from_error_code(ec)
        );

        pimpl->iosvc.post(bind(cc, error_code()));
    }
    else
    {
        *out = form_write_single_coil_response(
            pimpl->packet_id,
            pimpl->device,
            pimpl->address,
            pimpl->value
        );

        pimpl->iosvc.post(bind(cc, error_code()));
    }
}

}
