#include <mxmodbusgateway/worker/ModbusWorker.hpp>

#include <deque>

#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread.hpp>
#include <boost/thread/thread.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusgateway/worker/ReadCommand.hpp>
#include <mxmodbusgateway/worker/WriteSingleCommand.hpp>
#include <mxmodbusgateway/worker/WriteMultipleCommand.hpp>

using boost::asio::io_service;
using boost::bind;
using boost::condition_variable;
using boost::get_error_info;
using boost::mutex;
using boost::posix_time::microsec_clock;
using boost::posix_time::milliseconds;
using boost::posix_time::neg_infin;
using boost::posix_time::ptime;
using boost::posix_time::time_duration;
using boost::shared_ptr;
using boost::system::error_code;
using boost::this_thread::interruption_point;
using boost::thread;
using boost::thread_interrupted;

using mxmodbus::address_t;
using mxmodbus::device_t;
using mxmodbus::exception_t;
using mxmodbus::EInfoException;
using mxmodbus::ModbusClient;

using mxutil::CompleteCallback;

using std::deque;
using std::vector;

namespace mxmodbusgateway
{
namespace worker
{

typedef
    ReadCommand<vector<bool>, &ModbusClient::read_coils>
    ReadCoilsCommand;

typedef
    ReadCommand<vector<bool>, &ModbusClient::read_discrete_inputs>
    ReadDiscreteInputsCommand;

typedef
    ReadCommand<vector<uint16_t>, &ModbusClient::read_holding_registers>
    ReadHoldingRegistersCommand;

typedef
    ReadCommand<vector<uint16_t>, &ModbusClient::read_input_registers>
    ReadInputRegistersCommand;

typedef
    WriteSingleCommand<bool, &ModbusClient::write_single_coil>
    WriteSingleCoilCommand;

typedef
    WriteSingleCommand<uint16_t, &ModbusClient::write_single_register>
    WriteSingleRegisterCommand;

typedef
    WriteMultipleCommand<vector<bool>, &ModbusClient::write_multiple_coils>
    WriteMultipleCoilsCommand;

typedef
    WriteMultipleCommand<vector<uint16_t>, &ModbusClient::write_multiple_registers>
    WriteMultipleRegistersCommand;

typedef
    boost::tuple<
        shared_ptr<Command>,
        CompleteCallback
    >
    CommandCallback;

struct ModbusWorker::Impl
{
    Impl(
        io_service &iosvc_,
        const CreateModbusClientFunc &create_modbus_client_,
        unsigned int interframe_delay_ms_)
        : iosvc(iosvc_)
        , create_modbus_client(create_modbus_client_)
        , interframe_delay_ms(interframe_delay_ms_)
    {
    }

    io_service &                    iosvc;
    const CreateModbusClientFunc    create_modbus_client;
    const unsigned int              interframe_delay_ms;

    shared_ptr<thread>              thread_object;

    deque<CommandCallback>          command_queue;
    mutex                           command_queue_mtx;
    condition_variable              command_queue_cnd;

    const CommandCallback
        get_next_command()
    {
        mutex::scoped_lock lock(this->command_queue_mtx);
        while(this->command_queue.empty())
            this->command_queue_cnd.wait(lock);

        const CommandCallback rtn = this->command_queue.front();
        this->command_queue.pop_front();
        return rtn;
    }

    void
        worker_thread()
    {
        try
        {
            shared_ptr<ModbusClient> mc;
            ptime next_execution = neg_infin;

            for(;;)
            {
                interruption_point();

                const CommandCallback cnc = this->get_next_command();
                const shared_ptr<Command> cmd = cnc.get<0>();
                BOOST_ASSERT(cmd);

                const CompleteCallback cc = cnc.get<1>();

                if (!mc)
                    mc = this->create_modbus_client();

                BOOST_ASSERT(mc);

                try
                {
                    const ptime now = microsec_clock::universal_time();
                    const time_duration tdiff = next_execution - now;

                    // if interframe-delay not met, sleep for awhile
                    if (!tdiff.is_negative())
                    {
                        interruption_point();
                        usleep(tdiff.total_microseconds());
                    }

                    interruption_point();
                    cmd->execute(*mc);

                    // determine next execution time
                    next_execution =
                        microsec_clock::universal_time()
                        + milliseconds(this->interframe_delay_ms);

                    this->iosvc.post(bind(cc, error_code()));
                }
                catch(mxmodbus::Error &e)
                {
                    // determine next execution time
                    next_execution =
                        microsec_clock::universal_time()
                        + milliseconds(this->interframe_delay_ms);

                    MXLOG_WARN(diagnostic_information(e));

                    int ec = errc::modbus_error;
                    const exception_t *modbus_exception = get_error_info<EInfoException>(e);
                    if (modbus_exception)
                        ec = errc::modbus_exception_base + *modbus_exception;

                    this->iosvc.post(bind(cc, error_code(ec, errc::category())));

                    mc.reset();
                }
            }
        }
        catch(thread_interrupted &e)
        {
            MXLOG_INFO("terminating");
        }
        catch(boost::exception &e)
        {
            MXLOG_EROR(diagnostic_information(e));
            exit(1);
        }
        catch(std::exception &e)
        {
            MXLOG_EROR(e.what());
            exit(1);
        }
        catch(...)
        {
            MXLOG_EROR("unexpected exception caught");
            exit(1);
        }
    }
};

ModbusWorker::ModbusWorker(
    io_service &iosvc,
    const CreateModbusClientFunc &create_modbus_client,
    unsigned int interframe_delay_ms)
    : pimpl(new Impl(
        iosvc,
        create_modbus_client,
        interframe_delay_ms))
{
    pimpl->thread_object.reset(new thread(bind(
        &Impl::worker_thread,
        &*pimpl
    )));
}

ModbusWorker::~ModbusWorker()
{
    pimpl->thread_object->interrupt();
    pimpl->thread_object->join();
}

void
    ModbusWorker::async_read_coils(
        device_t device,
        address_t addr,
        size_t quantity,
        vector<bool> &out,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new ReadCoilsCommand(
            device,
            addr,
            quantity,
            out
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_read_discrete_inputs(
        device_t device,
        address_t addr,
        size_t quantity,
        vector<bool> &out,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new ReadDiscreteInputsCommand(
            device,
            addr,
            quantity,
            out
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_read_holding_registers(
        device_t device,
        address_t addr,
        size_t quantity,
        vector<uint16_t> &out,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new ReadHoldingRegistersCommand(
            device,
            addr,
            quantity,
            out
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_read_input_registers(
        device_t device,
        address_t addr,
        size_t quantity,
        vector<uint16_t> &out,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new ReadInputRegistersCommand(
            device,
            addr,
            quantity,
            out
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_write_single_coil(
        device_t device,
        address_t addr,
        bool value,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new WriteSingleCoilCommand(
            device,
            addr,
            value
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_write_single_register(
        device_t device,
        address_t addr,
        uint16_t value,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new WriteSingleRegisterCommand(
            device,
            addr,
            value
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_write_multiple_coils(
        device_t device,
        address_t addr,
        vector<bool> value,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new WriteMultipleCoilsCommand(
            device,
            addr,
            value
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

void
    ModbusWorker::async_write_multiple_registers(
        device_t device,
        address_t addr,
        vector<uint16_t> value,
        const CompleteCallback &cc)
{
    mutex::scoped_lock lock(pimpl->command_queue_mtx);
    pimpl->command_queue.push_back(CommandCallback(
        shared_ptr<Command>(new WriteMultipleRegistersCommand(
            device,
            addr,
            value
        )),
        cc
    ));

    pimpl->command_queue_cnd.notify_one();
}

}
}
