#include <boost/test/unit_test.hpp>

#include <mxmodbusgateway/Configuration.hpp>

namespace mxmodbusgateway
{

using mxmodbus::ModbusRtuClient;

using std::string;

namespace
{

class TestFixture
{
};

}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_ShouldReturnCorrectPortsLength, TestFixture)
{
    // arrange
    const string SUT =
        "port-list = (\n"
            "{\n"
                "tcp = {\n"
                    "host = \"127.0.0.1\"\n"
                    "service = 502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS0\"\n"
                    "serial-config = \"9600-8-n-1\"\n"
                    "interframe-delay-ms = 300\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 500\n"
                "}\n"
            "},\n"
            "{\n"
                "tcp = {\n"
                    "host = \"0.0.0.0\"\n"
                    "service = 1502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS1\"\n"
                    "serial-config = \"19200-8-n-1\"\n"
                    "interframe-delay-ms = 200\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 300\n"
                "}\n"
            "}\n"
    ")\n";

    // act
    const Configuration result = Configuration::parse(SUT);

    // assert
    BOOST_CHECK_EQUAL(2, result.port_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_ShouldReturnCorrectValue, TestFixture)
{
    // arrange
    const string SUT =
        "port-list = (\n"
            "{\n"
                "tcp = {\n"
                    "host = \"127.0.0.1\"\n"
                    "service = 502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS0\"\n"
                    "serial-config = \"9600-8-n-1\"\n"
                    "interframe-delay-ms = 300\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 500\n"
                "}\n"
            "}\n"
    ")\n";

    // act
    const Configuration result = Configuration::parse(SUT);

    // assert
    BOOST_REQUIRE_EQUAL(1, result.port_list.size());

    const Configuration::Port port = result.port_list[0];

    const Configuration::Port::Tcp tcp = port.tcp;
    BOOST_CHECK_EQUAL("127.0.0.1", tcp.host);
    BOOST_CHECK_EQUAL(502, tcp.service);

    const Configuration::Port::Rtu rtu = port.rtu;
    BOOST_CHECK_EQUAL("/dev/ttyS0", rtu.path);
    BOOST_CHECK_EQUAL(ModbusRtuClient::BAUD_9600, rtu.serial_config.baud);
    BOOST_CHECK_EQUAL(ModbusRtuClient::NONE, rtu.serial_config.parity);
    BOOST_CHECK_EQUAL(ModbusRtuClient::STOP_1, rtu.serial_config.stopbit);
    BOOST_CHECK_EQUAL(300, rtu.interframe_delay_ms);
    BOOST_CHECK_EQUAL(100, rtu.interchar_timeout_ms);
    BOOST_CHECK_EQUAL(500, rtu.request_timeout_ms);
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithFaultyConfig_ShouldReturnParseError, TestFixture)
{
    // arrange
    const string SUT =
        "port-list = (\n"
            "{\n"
                "tcp = {\n"
                    "host = \"127.0.0.1\"\n"
                    "service = 502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS0\"\n"
                    "serial-config = \"9600-8-n-1\"\n"
                    "interframe-delay-ms = 300\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 500\n"
                "}\n"
            "}\n"
    "}\n"; // oops

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithInsufficientConfig_ShouldReturnParseError, TestFixture)
{
    // arrange
    const string SUT =
        "port-list = (\n"
            "{\n"
                "tcp = {\n"
                    "#host = \"127.0.0.1\"\n" // oops
                    "service = 502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS0\"\n"
                    "serial-config = \"9600-8-n-1\"\n"
                    "interframe-delay-ms = 300\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 500\n"
                "}\n"
            "}\n"
    ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithFaultyBaud_ShouldReturnParseError, TestFixture)
{
    // arrange
    const string SUT =
        "port-list = (\n"
            "{\n"
                "tcp = {\n"
                    "host = \"127.0.0.1\"\n"
                    "service = 502\n"
                "}\n"
                "rtu = {\n"
                    "path = \"/dev/ttyS0\"\n"
                    "serial-config = \"384000-8-n-1\"\n" // oops
                    "interframe-delay-ms = 300\n"
                    "interchar-timeout-ms = 100\n"
                    "request-timeout-ms = 500\n"
                "}\n"
            "}\n"
    ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError);
}

}
