#include <exception>
#include <string>
#include <vector>

#include <boost/cstdint.hpp>
#include <boost/exception/all.hpp>

#include <mxsl/mxsl.hpp>

#include <mxutil/log.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StderrLogger.hpp>

#include <mxmodbusio/App.hpp>
#include <mxmodbusio/Builder.hpp>
#include <mxmodbusio/error.hpp>
#include <mxmodbusio/Option.hpp>

using boost::get_error_info;
using boost::logic::tribool;
using boost::shared_ptr;

using mxmodbusio::App;
using mxmodbusio::Builder;
using mxmodbusio::EInfoExitStatus;
using mxmodbusio::exit_status;
using mxmodbusio::exit_status_t;
using mxmodbusio::SystemExit;
using mxmodbusio::Option;

using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StderrLogger;

using std::string;
using std::vector;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

shared_ptr<App>
    get_app(
        const string &modbus_host,
        int modbus_port,
        int modbus_request_timeout_ms,
        uint8_t device_id,
        shared_ptr<int> read_address,
        int quantity,
        shared_ptr<int> write_address,
        vector<uint16_t> data,
        bool single);

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StderrLogger());
        logger->set_level(Logger::DBUG);
        logger->enable_line_of_file();
        Log::get().set_logger(logger);

        const shared_ptr<App> app = get_app(
            opt.host,
            opt.port,
            opt.request_timeout_ms,
            opt.device_id,
            opt.read_address,
            opt.quantity,
            opt.write_address,
            opt.data,
            opt.single);

        mxsl::validate();

        if (app)
            app->run();

        return 0;
    }
    catch(SystemExit &e)
    {
        MXLOG_EROR(diagnostic_information(e));

        const exit_status_t *ec = get_error_info<EInfoExitStatus>(e);
        if (ec)
            exit(*ec);
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

shared_ptr<App>
    get_app(
        const string &host,
        int port,
        int request_timeout_ms,
        uint8_t device_id,
        shared_ptr<int> read_address,
        int quantity,
        shared_ptr<int> write_address,
        vector<uint16_t> data,
        bool single)

{
    Builder builder;
    builder.set_modbus(
        host,
        port,
        device_id,
        request_timeout_ms);

    if (read_address && write_address)
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_ARGUMENT));

    if (read_address)
        builder.add_read_request(
            *read_address,
            quantity);

    if (write_address)
        builder.add_write_request(
            *write_address,
            data,
            single);

    return builder.build();
}

}
