#include <mxmodbusio/Builder.hpp>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>

#include <mxmodbus/ModbusClient.hpp>
#include <mxmodbus/ModbusClientProxy.hpp>
#include <mxmodbus/ModbusTcpClient.hpp>

#include <mxmodbusio/Command.hpp>
#include <mxmodbusio/CommandFactory.hpp>
#include <mxmodbusio/CommandList.hpp>
#include <mxmodbusio/CommandPrompt.hpp>

using boost::make_shared;
using boost::shared_ptr;

using mxmodbus::ModbusClient;
using mxmodbus::ModbusClientProxy;
using mxmodbus::ModbusTcpClient;

using mxmodbusio::Command;
using mxmodbusio::EInfoExitStatus;
using mxmodbusio::exit_status;
using mxmodbusio::SystemExit;

using std::cin;
using std::istream;
using std::string;
using std::vector;

namespace
{

void
    dummy_delete(istream *)
{
    // do nothing
}

}

namespace mxmodbusio
{

struct Builder::Impl
{
    Impl()
    {
    }

    uint8_t                         device_id;

    shared_ptr<ModbusClientProxy>   modbus_client_proxy;

    vector<shared_ptr<Command> >    command_list;

    shared_ptr<ModbusTcpClient>
        create_modbus_tcp_client(
            const string &modbus_host,
            int modbus_port,
            int modbus_request_timeout_ms) const
    {
        return shared_ptr<ModbusTcpClient>(new ModbusTcpClient(
            modbus_host,
            modbus_port,
            modbus_request_timeout_ms
        ));
    }
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_modbus(
        const string &host,
        int port,
        uint8_t device_id,
        int request_timeout_ms)
{
    if (port > 65535)
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_ARGUMENT));

    if (request_timeout_ms < 0)
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_ARGUMENT));

    pimpl->device_id = device_id;

    pimpl->modbus_client_proxy.reset(new ModbusClientProxy(
        bind(
            &Impl::create_modbus_tcp_client,
            &*pimpl,
            host,
            port,
            request_timeout_ms)
    ));
}

void
    Builder::add_read_request(
        int address,
        int quantity)
{
    pimpl->command_list.push_back(
        CommandFactory::instance().create_read_command(
            address,
            quantity
        )
    );
}

void
    Builder::add_write_request(
        int address,
        const vector<uint16_t> &data,
        bool single)
{
    if (single)
    {
        const vector<shared_ptr<Command> > cmd_list =
            CommandFactory::instance().create_write_single_command(
                address,
                data);

        pimpl->command_list.insert(
            pimpl->command_list.end(),
            cmd_list.begin(),
            cmd_list.end());
    }
    else
    {
        pimpl->command_list.push_back(
            CommandFactory::instance().create_write_command(
                address,
                data
            )
        );
    }
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(pimpl->modbus_client_proxy);

    shared_ptr<CommandSource> cmd_source;
    if (pimpl->command_list.empty())
    {
        cmd_source = make_shared<CommandPrompt>(
            shared_ptr<istream>(
                &cin,
                &dummy_delete
            )
        );
    }
    else
    {
        cmd_source = make_shared<CommandList>(
            pimpl->command_list);
    }

    return make_shared<App>(
        pimpl->modbus_client_proxy,
        pimpl->device_id,
        cmd_source);
}

}
