#include <mxmodbusio/CommandFactory.hpp>

#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>

#include <mxmodbusio/ReadCommand.hpp>
#include <mxmodbusio/WriteCommand.hpp>

using boost::make_shared;
using boost::shared_ptr;

using mxmodbus::ModbusClient;

using mxmodbusio::ReadCommand;
using mxmodbusio::WriteCommand;

using std::vector;

namespace
{

typedef
    ReadCommand<
        vector<bool>,
        &ModbusClient::read_coils
    >
    ReadCoilsCommand;

typedef
    ReadCommand<
        vector<bool>,
        &ModbusClient::read_discrete_inputs
    >
    ReadDiscreteInputsCommand;

typedef
    ReadCommand<
        vector<uint16_t>,
        &ModbusClient::read_input_registers
    >
    ReadInputRegistersCommand;

typedef
    ReadCommand<
        vector<uint16_t>,
        &ModbusClient::read_holding_registers
    >
    ReadHoldingRegistersCommand;

typedef
    WriteCommand<
        bool,
        &ModbusClient::write_single_coil
    >
    WriteSingleCoilCommand;

typedef
    WriteCommand<
        uint16_t,
        &ModbusClient::write_single_register
    >
    WriteSingleRegisterCommand;

typedef
    WriteCommand<
        const vector<bool> &,
        &ModbusClient::write_multiple_coils
    >
    WriteMultipleCoilsCommand;

typedef
    WriteCommand<
        const vector<uint16_t> &,
        &ModbusClient::write_multiple_registers
    >
    WriteMultipleRegistersCommand;

const vector<bool>
    bool_vector_from_uint16_vector(
        const vector<uint16_t> &buf)
{
    vector<bool> rtn;
    BOOST_FOREACH(uint16_t b, buf)
    {
        rtn.push_back(b);
    }

    return rtn;
}

}

namespace mxmodbusio
{

CommandFactory &
    CommandFactory::instance()
{
    static CommandFactory inst;

    return inst;
}

struct CommandFactory::Impl
{
};

CommandFactory::CommandFactory()
{
}

CommandFactory::~CommandFactory()
{
}

shared_ptr<Command>
    CommandFactory::create_read_command(
        int_least32_t reference,
        size_t quantity) const
{
    if (reference >= 0x00000 && reference <= 0x0FFFF)
    {
        return make_shared<ReadCoilsCommand>(
            reference,
            quantity);
    }
    else if (reference >= 0x10000 && reference <= 0x1FFFF)
    {
        return make_shared<ReadDiscreteInputsCommand>(
            reference,
            quantity);
    }
    else if (reference >= 0x30000 && reference <= 0x3FFFF)
    {
        return make_shared<ReadInputRegistersCommand>(
            reference,
            quantity);
    }
    else if (reference >= 0x40000 && reference <= 0x4FFFF)
    {
        return make_shared<ReadHoldingRegistersCommand>(
            reference,
            quantity);
    }
    else
        BOOST_THROW_EXCEPTION(
            ArgumentError()
        );
}

shared_ptr<Command>
    CommandFactory::create_write_command(
        int_least32_t reference,
        const vector<uint16_t> &data) const
{
    if (reference >= 0x0000 && reference <= 0x0FFFF)
    {
        const vector<bool> data_ = bool_vector_from_uint16_vector(data);
        return make_shared<WriteMultipleCoilsCommand>(
            reference,
            data_);
    }
    else if (reference >= 0x40000 && reference <= 0x4FFFF)
    {
        return make_shared<WriteMultipleRegistersCommand>(
            reference,
            data);
    }
    else
        BOOST_THROW_EXCEPTION(
            ArgumentError()
        );
}

const vector<shared_ptr<Command> >
    CommandFactory::create_write_single_command(
        int_least32_t reference,
        const vector<uint16_t> &data) const
{
    vector<shared_ptr<Command> > cmd_list;
    if (reference >= 0x0000 && reference <= 0x0FFFF)
    {
        const vector<bool> data_ = bool_vector_from_uint16_vector(data);
        for(size_t idx = 0; idx < data.size(); ++idx)
        {
            cmd_list.push_back(make_shared<WriteSingleCoilCommand>(
                reference + idx,
                data_[idx]));
        }
    }
    else if (reference >= 0x40000 && reference <= 0x4FFFF)
    {
        for(size_t idx = 0; idx < data.size(); ++idx)
        {
            cmd_list.push_back(make_shared<WriteSingleRegisterCommand>(
                reference + idx,
                data[idx]));
        }
    }
    else
        BOOST_THROW_EXCEPTION(
            ArgumentError()
        );

    return cmd_list;
}

}
