#include <mxmodbusio/Option.hpp>

#include <cstdlib>
#include <string>
#include <vector>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>
#include <boost/program_options.hpp>

#include <mxutil/log.hpp>

#include <config.h>

#include <mxmodbusio/error.hpp>
#include <mxmodbusio/impl/parse_command.hpp>

using boost::lexical_cast;
using boost::logic::tribool;
using boost::logic::indeterminate;
using boost::make_shared;
using boost::program_options::options_description;
using boost::program_options::value;
using boost::program_options::variables_map;
using boost::program_options::command_line_style::default_style;
using boost::program_options::command_line_style::allow_guessing;

using mxmodbusio::EInfoExitStatus;
using mxmodbusio::exit_status;
using mxmodbusio::SystemExit;
using mxmodbusio::impl::parse_data;

using std::cout;
using std::endl;
using std::string;
using std::strtol;
using std::vector;

namespace
{

int
    get_max_quantity(int ref)
{
    return
        ref >= 0x00000 && ref <= 0x0FFFF ? 2000 :
        ref >= 0x10000 && ref <= 0x1FFFF ? 2000 :
        ref >= 0x30000 && ref <= 0x3FFFF ? 125 :
        ref >= 0x40000 && ref <= 0x4FFFF ? 125 :
        -1;
}

}

namespace mxmodbusio
{

Option::Option()
    : host("127.0.0.1")
    , port(502)
    , request_timeout_ms(3000)
    , device_id(1)
    , quantity(1)
    , data(vector<uint16_t>(1, 0x0000))
    , single(false)
{
}

tribool
    Option::parse_args(
        int argc,
        const char *argv[])
{
    options_description desc("Allowed options");
    desc.add_options()
        ("read,r", value<string>(), "Do read operation.")
        ("write,w", value<string>(), "Do write operation.")
        ("data,d", value<string>(), "Use with --write option. Default 0000.")
        ("quantity,q", value<int>(),
            "Use with --read option. Default 1.\n"
            "0x00000~0x0FFFF: 1~2000\n"
            "0x10000-0x1FFFF: 1-2000\n"
            "0x30000-0x3FFFF: 1-125\n"
            "0x40000-0x4FFFF: 1-125")
        ("host,h", value<string>(), "MODBUS slave address. Default 127.0.0.1.")
        ("port,p", value<int>(), "MODBUS slave port. Default 502.")
        ("timeout-ms,t", value<int>(), "Timeout for each MODBUS command. Default 3000. (100-60000)")
        ("unit,u", value<int>(), "MODBUS slave ID. (0-255). Default 1.")
        ("single,s", "Use write_single_* command to instead of write_multiple_* command.")
        ("help", "Print this manual.")
        ("version", "Print version.");

    variables_map vm;
    try
    {
        store(
            parse_command_line(argc, argv, desc, default_style & (~allow_guessing)),
            vm
        );
    }
    catch(boost::program_options::error &e)
    {
        cout << desc << endl;
        return false;
    }

    if (vm.count("help"))
    {
        cout << desc << endl;
        return indeterminate;
    }

    if (vm.count("version"))
    {
        cout << "mxmodbusio v" << VERSION << endl;
        return indeterminate;
    }

    if (vm.count("host"))
        this->host = vm["host"].as<string>();

    if (vm.count("port"))
        this->port = vm["port"].as<int>();

    if (vm.count("timeout-ms"))
    {
        const int timeout_ms = vm["timeout-ms"].as<int>();
        if (timeout_ms < 100 || timeout_ms > 60000)
        {
            cout << desc << endl;
            return false;
        }
        this->request_timeout_ms = timeout_ms;
    }

    if (vm.count("unit"))
    {
        int device_id_ = vm["unit"].as<int>();
        if (device_id_ < 0 || device_id_ > 255)
        {
            cout << desc << endl;
            return false;
        }
        this->device_id = static_cast<uint8_t>(device_id_);
    }

    if (vm.count("read"))
    {
        const int reference = strtol(vm["read"].as<string>().c_str(), NULL, 0);
        if (reference < 0 || (reference >= 0x20000 && reference <= 0x2FFFF) ||
            reference >= 0x5FFFF)
        {
            cout << desc << endl;
            return false;
        }

        this->read_address = make_shared<int>(reference);
    }

    if (vm.count("quantity"))
        this->quantity = vm["quantity"].as<int>();

    if (vm.count("write"))
    {
        const int reference = strtol(vm["write"].as<string>().c_str(), NULL, 0);
        if (reference < 0 || (reference >= 0x10000 && reference <= 0x3FFFF) ||
            reference >= 0x5FFFF)
        {
            cout << desc << endl;
            return false;
        }

        this->write_address = make_shared<int>(reference);
    }

    if (vm.count("data"))
    {
        if (!parse_data(
                vm["data"].as<string>(),
                this->data))
        {
            cout << desc << endl;
            return false;
        }
    }

    if (vm.count("single"))
        this->single = true;

    if (this->read_address)
    {
        const int max_quantity = get_max_quantity(*this->read_address);

        if (max_quantity < 0)
        {
            MXLOG_EROR("internal error");
            return false;
        }

        if (this->quantity > max_quantity)
        {
            MXLOG_EROR("illegal --quantity: " + lexical_cast<string>(this->quantity));
            return false;
        }
    }

    if (this->write_address)
    {
        const int max_quantity = get_max_quantity(*this->write_address);

        if (max_quantity < 0)
        {
            MXLOG_EROR("internal error");
            return false;
        }

        if (this->data.size() > static_cast<size_t>(max_quantity))
        {
            MXLOG_EROR("data too long: data.size() = " + lexical_cast<string>(this->data.size()));
            return false;
        }
    }

    return true;
}

}
