#include <mxmodbusio/impl/parse_command.hpp>

#include <sstream>

#include <boost/algorithm/string.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>

using boost::is_any_of;
using boost::lexical_cast;
using boost::regex;
using boost::smatch;
using boost::split;
using boost::token_compress_on;

using std::hex;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

template <typename T>
T
    numcast(
        const string &s)
{
    stringstream ss;

    if (s.size() >= 2 &&
        s.substr(0, 2) == "0x")
    {
        ss << hex;
    }

    ss << s;

    T v;
    ss >> v;

    return v;
}

}

namespace mxmodbusio
{
namespace impl
{

bool
    parse_read_command(
        const string &line,
        int_least32_t &reference,
        size_t &quantity)
{
    static const regex e(
        "^read[ ]+((?:0x[0-9a-f]+)|(?:[0-9]+))[ ]+((?:0x[0-9a-f]+)|(?:[0-9]+))$"
    );

    smatch what;
    if (!regex_match(line, what, e))
        return false;

    reference = numcast<int_least32_t>(what[1].str());
    quantity = numcast<size_t>(what[2].str());

    return true;
}

bool
    parse_write_command(
        const string &line,
        int_least32_t &reference,
        vector<uint16_t> &data)
{
    static const regex e(
        "^write[ ]+((?:0x[0-9a-f]+)|(?:[0-9]+))[ ]+([0-9a-f:]+)$"
    );

    smatch what;
    if (!regex_match(line, what, e))
        return false;

    if (!parse_data(
            what[2].str(),
            data))
        return false;

    reference = numcast<int_least32_t>(what[1].str());

    return true;
}

bool
    parse_swrite_command(
        const string &line,
        int_least32_t &reference,
        uint16_t &data_)
{
    static const regex e(
        "^swrite[ ]+((?:0x[0-9a-f]+)|(?:[0-9]+))[ ]+([0-9a-f]+)$"
    );

    smatch what;
    if (!regex_match(line, what, e))
        return false;

    vector<uint16_t> data;
    if (!parse_data(
            what[2].str(),
            data))
        return false;

    BOOST_VERIFY(data.size() == 1);

    reference = numcast<int_least32_t>(what[1].str());
    data_ = data.front();

    return true;
}

bool
    parse_data(
        const string &s,
        vector<uint16_t> &data)
{
    vector<string> str_list;
    split(str_list, s, is_any_of(":"), token_compress_on);

    vector<uint16_t> vlist;
    BOOST_FOREACH(const string &s, str_list)
    {
        if (s.empty() ||
                s.size() > 4)
            return false;

        if (s.find_first_not_of("0123456789abcdefABCDEF") != string::npos)
            return false;

        vlist.push_back(strtol(s.c_str(), NULL, 16));
    }

    data = vlist;

    return true;
}

}
}
