#ifndef _MXMODBUSMASTER_MODBUS_REFERENCE_HPP_
#define _MXMODBUSMASTER_MODBUS_REFERENCE_HPP_

#include <stdexcept>
#include <vector>

#include <boost/cstdint.hpp>
#include <boost/scoped_ptr.hpp>

#include <mxmodbusmaster/error.hpp>

namespace mxmodbusmaster
{
namespace modbus
{

/*
 * class Reference models four sections of Modbus registers:
 *   0x00000~0x0FFFF: Coil
 *   0x10000~0x1FFFF: Discrete Input
 *   0x30000~0x3FFFF: Input Register
 *   0x40000~0x4FFFF: Holding Register
 *
 * Values read from Reference are Word oriented.
 * Coils and Discrete Input values can only be either 0x00 or 0x01.
 * Any get/set operation shall not overlap 2 or more register sections.
 */

class Reference
{
public:
    typedef
        uint32_t
        address_t;

    typedef
        uint16_t
        value_t;

    static const address_t  COIL_BASE;
    static const address_t  DISCRETE_INPUT_BASE;
    static const address_t  INPUT_REGISTER_BASE;
    static const address_t  HOLDING_REGISTER_BASE;

public:
    Reference();
    Reference(const Reference &);
    ~Reference();

    value_t
        get(address_t) const; // throws std::out_of_range, mxmodbusmaster::NotFoundError

    void
        set(address_t,
            value_t); // throws std::out_of_range

    const std::vector<value_t>
        get(address_t,
            std::size_t) const; // throws std::out_of_range, mxmodbusmaster::NotFoundError

    void
        set(address_t,
            const std::vector<value_t> &); // throws std::out_of_range

    Reference &
        operator = (const Reference &);

    bool
        operator ==(const Reference &) const;

private:
    struct Impl;
    boost::scoped_ptr<Impl> pimpl;
};

}
}

#endif
