#include <mxutil/PidFile.hpp>
#include <mxutil/Signal.hpp>
#include <mxutil/log.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StderrLogger.hpp>
#include <mxutil/log/SyslogLogger.hpp>

#include <mxsl/mxsl.hpp>
#include <mxmodbusmaster/App.hpp>
#include <mxmodbusmaster/Builder.hpp>
#include <mxmodbusmaster/Option.hpp>

using boost::logic::tribool;
using boost::shared_ptr;

using mxmodbusmaster::App;
using mxmodbusmaster::Builder;
using mxmodbusmaster::Option;

using mxutil::PidFile;
using mxutil::Signal;
using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StderrLogger;
using mxutil::log::SyslogLogger;

using std::cerr;
using std::endl;
using std::string;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);
shared_ptr<App>
    get_app(
        const string &mqtt_broker_host,
        int mqtt_broker_port);

volatile bool stop = false;
void
    on_term(int)
{
    stop = true;
}

}

volatile bool export_equipment_status = false;
void
    on_usr1(int)
{
    export_equipment_status = true;
}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StderrLogger());
        logger->set_level(Logger::DBUG);
        logger->enable_line_of_file();
        if (!opt.foreground)
        {
            const int result = daemon(0, 0);
            if (0 != result)
            {
                cerr << strerror(errno) << endl;
                return 1;
            }

            logger.reset(new SyslogLogger("mxmodbusmaster"));
            logger->set_level(Logger::WARN);
        }
        Log::get().set_logger(logger);

        const shared_ptr<App> app = get_app(
            opt.mqtt_broker_host,
            opt.mqtt_broker_port);

        mxsl::validate();

        // set up signal handler
        Signal::instance().sigterm(&on_term);
        Signal::instance().sigusr1(&on_usr1);

        shared_ptr<PidFile> pidfile;
        if (!opt.foreground)
            pidfile = PidFile::get("/var/run/mxmodbusmaster.pid");

        while(!stop)
        {
            if (export_equipment_status)
            {
                export_equipment_status = false;
                app->export_equipment_status();
            }

            app->loop();
            usleep(10 * 1000);
        }

        return 0;
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

shared_ptr<App>
    get_app(
        const string &mqtt_broker_host,
        int mqtt_broker_port)
{
    Builder builder;
    builder.set_configuration("etc/mxmodbusmaster/configuration");
    builder.add_equipment_template_directory("etc/mxmodbusmaster/template.d");
    builder.set_mqtt_broker(mqtt_broker_host, mqtt_broker_port);

    return builder.build();
}

}
