#include <mxutil/log.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/SyslogLogger.hpp>

#include <mxsl/mxsl.hpp>

#include <mxmodbusclient/App.hpp>
#include <mxmodbusclient/Builder.hpp>
#include <mxmodbusclient/Option.hpp>

using boost::filesystem::path;
using boost::get_error_info;
using boost::logic::tribool;
using boost::shared_ptr;

using mxmodbusclient::App;
using mxmodbusclient::Builder;
using mxmodbusclient::EInfoExitStatus;
using mxmodbusclient::exit_status_t;
using mxmodbusclient::SystemExit;
using mxmodbusclient::Option;

using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::SyslogLogger;

using std::string;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

shared_ptr<App>
    get_app(
        const path &equipment_template_path,
        const string &modbus_host,
        int modbus_port,
        int modbus_request_timeout_ms,
        uint8_t device_id,
        bool dry_run);

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new SyslogLogger("mxmodbusclient"));
        logger->set_level(Logger::DBUG);
        logger->enable_line_of_file();
        Log::get().set_logger(logger);

        const shared_ptr<App> app = get_app(
            opt.equipment_template_path,
            opt.modbus_host,
            opt.modbus_port,
            opt.modbus_request_timeout_ms,
            opt.device_id,
            opt.dry_run);

        mxsl::validate();

        if (!opt.dry_run)
            app->run();

        return 0;
    }
    catch(SystemExit &e)
    {
        const exit_status_t *ec = get_error_info<EInfoExitStatus>(e);
        if (ec)
            exit(*ec);
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

shared_ptr<App>
    get_app(
        const path &equipment_template_path,
        const string &modbus_host,
        int modbus_port,
        int modbus_request_timeout_ms,
        uint8_t device_id,
        bool dry_run)
{
    Builder builder;
    builder.set_equipment_template_path(equipment_template_path);
    builder.set_modbus_host_port(modbus_host, modbus_port);
    builder.set_modbus_request_timeout_ms(modbus_request_timeout_ms);
    builder.set_device_id(device_id);
    builder.set_dry_run(dry_run);

    return builder.build();
}

}
