#include <mxmodbusclient/Builder.hpp>

#include <fstream>
#include <vector>

#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusclient/ReadingCollector.hpp>

#include <mxmodbusmaster/EquipmentTemplate.hpp>
#include <mxmodbusmaster/error.hpp>
#include <mxmodbusmaster/modbus/Equipment.hpp>
#include <mxmodbusmaster/modbus/Master.hpp>
#include <mxmodbusmaster/modbus/Reader.hpp>
#include <mxmodbusmaster/modbus/trivial/TrivialReader.hpp>
#include <mxmodbusmaster/modbus/trivial/TrivialReaderBuilder.hpp>
#include <mxmodbusmaster/reading/Reading.hpp>

using boost::filesystem::path;
using boost::make_shared;
using boost::shared_ptr;

using mxmodbusclient::EInfoExitStatus;
using mxmodbusclient::exit_status;
using mxmodbusclient::SystemExit;
using mxmodbusclient::ReadingCollector;

using mxmodbusmaster::ArgumentError;
using mxmodbusmaster::EquipmentTemplate;
using mxmodbusmaster::ParseError;
using mxmodbusmaster::modbus::Equipment;
using mxmodbusmaster::modbus::EquipmentStatus;
using mxmodbusmaster::modbus::Master;
using mxmodbusmaster::modbus::Reader;
using mxmodbusmaster::modbus::trivial::TrivialReader;
using mxmodbusmaster::modbus::trivial::TrivialReaderBuilder;
using mxmodbusmaster::reading::Reading;

using std::ifstream;
using std::ios_base;
using std::string;
using std::vector;

namespace
{

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_EROR("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_EQUIPMENT_TEMPLATE));
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_EROR("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_EQUIPMENT_TEMPLATE));
    }

    return string(buf.begin(), buf.end());
}

}

namespace mxmodbusclient
{

struct Builder::Impl
{
    Impl()
        : device_id(0x00)
        , modbus_request_timeout_ms(3000)
        , modbus_host("127.0.0.1")
        , modbus_port(502)
        , dry_run(false)
    {
    }

    uint8_t                             device_id;
    int                                 modbus_request_timeout_ms;
    string                              modbus_host;
    int                                 modbus_port;
    bool                                dry_run;

    shared_ptr<EquipmentTemplate>       equipment_template;
    shared_ptr<Reader>                  reader;
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_equipment_template_path(const path &path)
{
    try
    {
        shared_ptr<EquipmentTemplate> equ_template(new EquipmentTemplate(
            EquipmentTemplate::parse(readfile(path))
        )); // throws mxmodbusmaster::ParseError

        pimpl->equipment_template = equ_template;

        TrivialReaderBuilder reader_builder;
        reader_builder.set_equipment_template(*equ_template);
        pimpl->reader = reader_builder.build(); // throws mxmodbusmaster::ArgumentError
    }
    catch(ParseError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_EQUIPMENT_TEMPLATE));
    }
    catch(ArgumentError &e)
    {
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_EQUIPMENT_TEMPLATE));
    }
}

void
    Builder::set_modbus_host_port(
        const string &host,
        int port)
{
    if (port > 65535)
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_ARGUMENT));

    pimpl->modbus_host = host;
    pimpl->modbus_port = port;
}

void
    Builder::set_modbus_request_timeout_ms(int timeout_ms)
{
    if (timeout_ms < 0)
        BOOST_THROW_EXCEPTION(SystemExit() << EInfoExitStatus(exit_status::ILLEGAL_ARGUMENT));

    pimpl->modbus_request_timeout_ms = timeout_ms;
}

void
    Builder::set_device_id(uint8_t device_id)
{
    pimpl->device_id = device_id;
}

void
    Builder::set_dry_run(bool dry_run)
{
    pimpl->dry_run = dry_run;
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(pimpl->equipment_template);

    if (pimpl->dry_run)
        return shared_ptr<App>();

    const vector<shared_ptr<Equipment> > equipment_list(
        1,
        make_shared<Equipment>(
            pimpl->equipment_template->id,
            pimpl->device_id,
            pimpl->reader)
    );

    const shared_ptr<Master> master(new Master(
        pimpl->modbus_host,
        pimpl->modbus_port,
        pimpl->modbus_request_timeout_ms,
        equipment_list
    ));

    const shared_ptr<ReadingCollector> reading_collector(new ReadingCollector(
        pimpl->equipment_template->id,
        pimpl->equipment_template->tag_list
    ));

    return shared_ptr<App>(new App(
        master,
        reading_collector
    ));
}

}
