#include <mxmodbusclient/Option.hpp>

#include <boost/program_options.hpp>

#include <config.h>

namespace mxmodbusclient
{

using boost::filesystem::path;
using boost::logic::tribool;
using boost::logic::indeterminate;
using boost::program_options::options_description;
using boost::program_options::value;
using boost::program_options::variables_map;
using boost::program_options::command_line_style::default_style;
using boost::program_options::command_line_style::allow_guessing;

using std::cout;
using std::endl;
using std::string;

Option::Option()
    : modbus_host("127.0.0.1")
    , modbus_port(502)
    , modbus_request_timeout_ms(3000)
    , device_id(0x00)
    , dry_run(false)
{
}

tribool
    Option::parse_args(
        int argc,
        const char *argv[])
{
    options_description desc("Allowed options");
    desc.add_options()
        ("equipment-template,e", value<path>(), "Path to Equipment Template configuration file.")
        ("dry-run", "Check Equipment Template validity only.")
        ("host,h", value<string>(), "Change Modbus host address")
        ("port,p", value<int>(), "Change Modbus port.")
        ("timeout-ms,t", value<int>(), "Timeout for each MODBUS command. (100-60000)")
        ("device-id,d", value<int>(), "Modbus device id. (0-255)")
        ("help", "Print this manual.")
        ("version", "Print version.");

    variables_map vm;
    try
    {
        store(
            parse_command_line(argc, argv, desc, default_style & (~allow_guessing)),
            vm
        );
    }
    catch(boost::program_options::error &e)
    {
        cout << desc << endl;
        return false;
    }

    if (vm.count("help"))
    {
        cout << desc << endl;
        return indeterminate;
    }

    if (vm.count("version"))
    {
        cout << "mxmodbusclient v" << VERSION << endl;
        return indeterminate;
    }

    if (vm.count("equipment-template"))
        this->equipment_template_path = vm["equipment-template"].as<path>();
    else
    {
        cout << desc << endl;
        return false;
    }

    if (vm.count("dry-run"))
        this->dry_run = true;

    if (vm.count("host"))
        this->modbus_host = vm["host"].as<string>();

    if (vm.count("port"))
        this->modbus_port = vm["port"].as<int>();

    if (vm.count("timeout-ms"))
    {
        const int timeout_ms = vm["timeout-ms"].as<int>();
        if (timeout_ms < 100 || timeout_ms > 60000)
        {
            cout << desc << endl;
            return false;
        }
        this->modbus_request_timeout_ms = timeout_ms;
    }

    if (vm.count("device-id"))
    {
        int device_id_ = vm["device-id"].as<int>();
        if (device_id_ < 0 || device_id_ > 255)
        {
            cout << desc << endl;
            return false;
        }
        this->device_id = static_cast<uint8_t>(device_id_);
    }

    return true;
}

}
