#include <mxmodbusmaster/Builder.hpp>

#include <fstream>

#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>
#include <mxutil/sched/DailyScheduler.hpp>

#include <mxmqtt/Worker.hpp>

#include <mxtag/mqtt/ProtobufSerializer.hpp>

#include <mxmodbusmaster/Configuration.hpp>
#include <mxmodbusmaster/modbus/CompositeReaderFactory.hpp>
#include <mxmodbusmaster/modbus/Worker.hpp>
#include <mxmodbusmaster/modbus/trivial/TrivialReaderFactory.hpp>

using boost::bind;
using boost::filesystem::path;
using boost::lexical_cast;
using boost::make_shared;
using boost::posix_time::ptime;
using boost::posix_time::milliseconds;
using boost::shared_ptr;

using mxmodbusmaster::ArgumentError;
using mxmodbusmaster::modbus::CompositeReaderFactory;
using mxmodbusmaster::modbus::Equipment;
using mxmodbusmaster::modbus::EquipmentStatus;
using mxmodbusmaster::modbus::Master;
using mxmodbusmaster::modbus::Reader;
using mxmodbusmaster::modbus::ReaderFactory;
using mxmodbusmaster::modbus::Worker;
using mxmodbusmaster::modbus::trivial::TrivialReaderFactory;
using mxmodbusmaster::reading::Reading;

using mxmqtt::Message;
typedef mxmqtt::Worker MqttWorker;

using mxtag::mqtt::ProtobufSerializer;
using mxtag::mqtt::Serializer;

using mxutil::sched::Scheduler;
using mxutil::sched::DailyScheduler;

using std::ifstream;
using std::ios_base;
using std::pair;
using std::setfill;
using std::setw;
using std::string;
using std::stringstream;
using std::vector;

namespace
{

static const int DEFAULT_REQUEST_TIMEOUT_MS = 30 * 1000;

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_EROR("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_EROR("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    return string(buf.begin(), buf.end());
}

void
    whip_worker(shared_ptr<Worker> worker)
{
    BOOST_ASSERT(worker);

    worker->whip();
}

void
    publish_reading(
        shared_ptr<MqttWorker> publisher,
        shared_ptr<Serializer> serializer,
        const string &equipment_id,
        const Reading &reading)
{
    BOOST_ASSERT(publisher && serializer);

    MXLOG_INFO(equipment_id + ":" + reading.tag_id() + ":" + reading.value().to_json_value());

    const mxtag::Tag t(
        equipment_id,
        reading.tag_id(),
        reading.timestamp(),
        reading.value().to_mxtag_value(),
        reading.unit() ? *reading.unit() : "");

    const pair<string, vector<uint8_t> > mqtt = serializer->from_tag(t);
    publisher->publish(Message(
        mqtt.first,
        mqtt.second,
        2,
        false));
}

void
    publish_status(
        shared_ptr<MqttWorker> publisher,
        const EquipmentStatus &status)
{
    BOOST_ASSERT(publisher);

    const pair<string, vector<uint8_t> > mqtt_msg = mqtt_from_equipment_status(status);
    publisher->publish(Message(
        mqtt_msg.first,
        mqtt_msg.second,
        0,
        true));
}

}

namespace mxmodbusmaster
{

struct Builder::Impl
{
    Impl()
        : mqtt_broker_host("127.0.0.1")
        , mqtt_broker_service(1883)
    {
    }

    shared_ptr<Configuration>   configuration;
    CompositeReaderFactory      reader_factory;

    string                      mqtt_broker_host;
    int                         mqtt_broker_service;
};

Builder::Builder()
    : pimpl(new Impl())
{
}

Builder::~Builder()
{
}

void
    Builder::set_configuration(const path &path)
{
    try
    {
        shared_ptr<Configuration> pconf(new Configuration(
            Configuration::parse(readfile(path))
        ));

        pimpl->configuration = pconf;
    }
    catch(ParseError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        BOOST_THROW_EXCEPTION(ArgumentError());
    }
}

void
    Builder::add_equipment_template_directory(const path &path)
{
    if (!is_directory(path))
    {
        MXLOG_WARN("not a directory: " + path.string());
    }

    pimpl->reader_factory.add_reader_factory(shared_ptr<ReaderFactory>(
        new TrivialReaderFactory(path)
    ));
}

void
    Builder::set_mqtt_broker(
        const string &host,
        int service)
{
    if (service > 65535)
        BOOST_THROW_EXCEPTION(ArgumentError());

    pimpl->mqtt_broker_host = host;
    pimpl->mqtt_broker_service = service;
}

shared_ptr<App>
    Builder::build() const
{
    BOOST_ASSERT(pimpl->configuration);

    vector<shared_ptr<Scheduler> > scheduler_list;

    shared_ptr<MqttWorker> publisher(new MqttWorker(
        pimpl->mqtt_broker_host,
        pimpl->mqtt_broker_service,
        "modbus-master"
    ));

    vector<shared_ptr<Equipment> > all_equipment_list;
    BOOST_FOREACH(const Configuration::Polling &polling, pimpl->configuration->polling_list)
    {
        vector<shared_ptr<Equipment> > equipment_list;
        BOOST_FOREACH(const Configuration::Polling::Device &device, polling.device_list)
        {
            shared_ptr<Reader> reader = pimpl->reader_factory.get(device.template_id);
            if (!reader)
            {
                MXLOG_WARN("cannot find reader: " + device.template_id);

                continue;
            }

            equipment_list.push_back(shared_ptr<Equipment>(new Equipment(
                device.equipment_name,
                device.device_id,
                reader
            )));
        }

        if (equipment_list.empty())
        {
            MXLOG_WARN("empty device-list, ignored");
            continue;
        }

        all_equipment_list.insert(
            all_equipment_list.end(),
            equipment_list.begin(),
            equipment_list.end());

        shared_ptr<Master> master(new Master(
            polling.host,
            polling.service,
            polling.request_timeout_ms,
            equipment_list
        ));

        shared_ptr<Worker> worker(new Worker(
            master,
            bind(
                publish_reading,
                publisher,
                make_shared<ProtobufSerializer>(),
                _1,
                _2
            ),
            bind(
                publish_status,
                publisher,
                _1
            )
        ));

        if (polling.polling_period_ms == 0 || polling.polling_period_ms >= 86400000)
        {
            MXLOG_WARN("illegal polling-period-ms: " + lexical_cast<string>(polling.polling_period_ms));
            continue;
        }

        scheduler_list.push_back(make_shared<DailyScheduler>(
            milliseconds(polling.polling_period_ms),
            bind(
                &whip_worker,
                worker
            )
        ));
    }

    return shared_ptr<App>(new App(
        all_equipment_list,
        pimpl->configuration->equipment_status_path,
        scheduler_list
    ));
}

}
