#include <mxmodbusmaster/Configuration.hpp>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>

#include <libconfig.h++>

#include <mxtag/validate.hpp>

using boost::lexical_cast;
using boost::regex;

using libconfig::Config;
using libconfig::ParseException;
using libconfig::Setting;
using libconfig::SettingNotFoundException;

using mxmodbusmaster::Configuration;
using mxmodbusmaster::EInfoMsg;
using mxmodbusmaster::ParseError;

using mxtag::is_valid_equipment_name;

using std::string;
using std::vector;

namespace
{

void
    validate_template_id(
        const string &template_id);
void
    validate_equipment_name(
        const string &equipment_name);
bool
    exists(
        const vector<Configuration::Polling::Device> &,
        const string &equipment_name);

}

namespace mxmodbusmaster
{

const Configuration
    Configuration::parse(const string &str)
{
    Config config;

    try
    {
        config.readString(str.c_str());

        const Setting & polling_list = config.lookup("polling-list");
        vector<Configuration::Polling> cfg_polling_list;
        for(int ax = 0; ax < polling_list.getLength(); ++ax)
        {
            const Setting &polling = polling_list[ax];

            const string host = polling["host"];
            const int service = polling["service"];
            const unsigned int polling_period_ms = polling["polling-period-ms"];
            if (polling_period_ms < 1000 || polling_period_ms >= 86400000)
                BOOST_THROW_EXCEPTION(
                    ParseError() << EInfoMsg("illegal polling-period-ms: " + lexical_cast<string>(polling_period_ms))
                );

            const unsigned int request_timeout_ms = polling["request-timeout-ms"];
            if (request_timeout_ms < 100 || request_timeout_ms > 60000)
                BOOST_THROW_EXCEPTION(
                    ParseError() << EInfoMsg("illegal request-timeout-ms: " + lexical_cast<string>(request_timeout_ms))
                );

            const Setting &device_list = polling["device-list"];

            vector<Configuration::Polling::Device> cfg_device_list;
            for(int bx = 0; bx < device_list.getLength(); ++bx)
            {
                const Setting &device = device_list[bx];

                const unsigned int device_id_ = device["device-id"];
                if (device_id_ > 255)
                    BOOST_THROW_EXCEPTION(
                        ParseError() << EInfoMsg("illegal device-id: " + device_id_)
                    );

                const uint8_t device_id = device_id_;
                const string template_id = device["template-id"];
                validate_template_id(template_id);

                const string equipment_name = device["equipment-name"];
                validate_equipment_name(equipment_name);

                if (exists(cfg_device_list, equipment_name))
                    BOOST_THROW_EXCEPTION(
                        ParseError() << EInfoMsg("duplicated equipment-name: " + equipment_name)
                    );

                cfg_device_list.push_back(Configuration::Polling::Device(
                    device_id,
                    template_id,
                    equipment_name
                ));
            }

            cfg_polling_list.push_back(Configuration::Polling(
                host,
                service,
                polling_period_ms,
                request_timeout_ms,
                cfg_device_list
            ));
        }

        string equipment_status_path = "var/mxmodbusmaster/equipment-status";
        if (config.exists("equipment-status-path"))
            if (!config.lookupValue("equipment-status-path", equipment_status_path))
                BOOST_THROW_EXCEPTION(
                    ParseError() << EInfoMsg("illegal equipment-status-path: value type error")
                );

        return Configuration(cfg_polling_list, equipment_status_path);
    }
    catch(const ParseException &e)
    {
        BOOST_THROW_EXCEPTION(
            ParseError() << EInfoMsg("format error")
        );
    }
    catch(const SettingNotFoundException &e)
    {
        BOOST_THROW_EXCEPTION(
            ParseError() << EInfoMsg("schema error")
        );
    }
}

}

namespace
{

void
    validate_template_id(const string &template_id)
{
    static const regex e(
        "^[A-Za-z0-9_-]{3,255}$"
    );

    if (!regex_match(template_id.begin(), template_id.end(), e))
        BOOST_THROW_EXCEPTION(
            ParseError() << EInfoMsg("illegal template-id: " + template_id)
        );
}

void
    validate_equipment_name(const string &equipment_name)
{
    if (!is_valid_equipment_name(equipment_name))
        BOOST_THROW_EXCEPTION(
            ParseError() << EInfoMsg("illegal equipment-name: " + equipment_name)
        );
}

bool
    exists(
        const vector<Configuration::Polling::Device> &device_list,
        const string &equipment_name)
{
    BOOST_FOREACH(const Configuration::Polling::Device &device, device_list)
    {
        if (device.equipment_name == equipment_name)
            return true;
    }

    return false;
}

}
