#include <mxmodbusmaster/EquipmentTemplate.hpp>

#include <map>

#include <boost/regex.hpp>

#include <libconfig.h++>

#include <mxtag/validate.hpp>

using boost::regex;

using libconfig::Config;
using libconfig::ParseException;
using libconfig::Setting;
using libconfig::SettingNotFoundException;

using mxmodbusmaster::EInfoMsg;
using mxmodbusmaster::ParseError;
using mxmodbusmaster::modbus::FUNC_NOTSET;
using mxmodbusmaster::modbus::FUNC_READ_COILS;
using mxmodbusmaster::modbus::FUNC_READ_DISCRETE_INPUTS;
using mxmodbusmaster::modbus::FUNC_READ_INPUT_REGISTERS;
using mxmodbusmaster::modbus::FUNC_READ_HOLDING_REGISTERS;
using mxmodbusmaster::modbus::function_t;
using mxmodbusmaster::modbus::trivial::byte_order_t;
using mxmodbusmaster::modbus::trivial::BO_BA;
using mxmodbusmaster::modbus::trivial::BO_CDAB;
using mxmodbusmaster::modbus::trivial::BO_DCBA;
using mxmodbusmaster::modbus::trivial::BO_NC;
using mxmodbusmaster::modbus::trivial::tag_type_t;
using mxmodbusmaster::modbus::trivial::TAGT_INT16;
using mxmodbusmaster::modbus::trivial::TAGT_INT32;
using mxmodbusmaster::modbus::trivial::TAGT_INT64;
using mxmodbusmaster::modbus::trivial::TAGT_UINT16;
using mxmodbusmaster::modbus::trivial::TAGT_UINT32;
using mxmodbusmaster::modbus::trivial::TAGT_UINT64;
using mxmodbusmaster::modbus::trivial::TAGT_FLOAT32;
using mxmodbusmaster::modbus::trivial::TAGT_FLOAT64;
using mxmodbusmaster::modbus::trivial::TAGT_STRING;

using mxtag::is_valid_tag_name;
using mxtag::is_valid_unit;

using std::map;
using std::string;
using std::vector;

namespace
{

bool
    lookup_numeric_value(
        const Setting &,
        const char *key,
        double &v);

byte_order_t
    byte_order_from_str(const string &);

function_t
    function_from_str(const string &);

tag_type_t
    tag_type_from_str(const string &);

void
    validate_template_id(const string &);

void
    validate_tag(const string &);

void
    validate_unit(
        const string &);

}

namespace mxmodbusmaster
{

const EquipmentTemplate
    EquipmentTemplate::parse(const string &str)
{
    Config config;

    try
    {
        config.readString(str.c_str());

        const string id = config.lookup("id");
        validate_template_id(id);

        const Setting &request_list = config.lookup("request-list");
        vector<EquipmentTemplate::Request> cfg_request_list;
        for(int idx = 0; idx < request_list.getLength(); ++idx)
        {
            const Setting &request = request_list[idx];

            const function_t function = function_from_str(request["function"]);
            if (function == FUNC_NOTSET)
                BOOST_THROW_EXCEPTION(ParseError());

            const unsigned int address = static_cast<unsigned int>(request["address"]);
            if (address > 0xFFFF)
                BOOST_THROW_EXCEPTION(ParseError());

            const unsigned int quantity = static_cast<unsigned int>(request["quantity"]);
            if (quantity > 0xFFFF)
                BOOST_THROW_EXCEPTION(ParseError());

            cfg_request_list.push_back(EquipmentTemplate::Request(
                function,
                address,
                quantity
            ));
        }

        const Setting &tag_list = config.lookup("tag-list");
        vector<EquipmentTemplate::Tag> cfg_tag_list;
        for(int idx = 0; idx < tag_list.getLength(); ++idx)
        {
            const Setting &tag = tag_list[idx];

            const string id = tag["id"];
            validate_tag(id);
            const uint32_t reference = static_cast<unsigned int>(tag["reference"]);
            const unsigned int quantity = tag["quantity"];

            const tag_type_t type = tag_type_from_str(tag["type"]);

            EquipmentTemplate::Tag tag_ = EquipmentTemplate::Tag(
                id,
                reference,
                quantity,
                type
            );

            string unit_;
            if (tag.lookupValue("unit", unit_))
            {
                validate_unit(unit_);
                tag_.unit.reset(new string(unit_));
            }

            if (tag.exists("invalid-value"))
            {
                const Setting &invalid_value_ = tag["invalid-value"];
                if (invalid_value_.getLength() != static_cast<int>(quantity))
                    BOOST_THROW_EXCEPTION(ParseError());

                tag_.invalid_value.reset(new vector<uint16_t>());
                for(int idx = 0; idx < static_cast<int>(quantity); ++idx)
                {
                    const unsigned int value = invalid_value_[idx];
                    tag_.invalid_value->push_back(value);
                }
            }

            string byte_order_;
            if (tag.lookupValue("byte-order", byte_order_))
            {
                tag_.byte_order.reset(new byte_order_t(
                    byte_order_from_str(byte_order_)
                ));
            }

            double slope;
            if (lookup_numeric_value(tag, "slope", slope))
            {
                tag_.slope.reset(new double(slope));
            }

            double offset;
            if (lookup_numeric_value(tag, "offset", offset))
            {
                tag_.offset.reset(new double(offset));
            }

            cfg_tag_list.push_back(tag_);
        }

        return EquipmentTemplate(id, cfg_request_list, cfg_tag_list);
    }
    catch(const ParseException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
    catch(const SettingNotFoundException &e)
    {
        BOOST_THROW_EXCEPTION(ParseError());
    }
}

}

namespace
{

bool
    lookup_numeric_value(
        const Setting &s,
        const char *key,
        double &v)
{
    double d;
    int i;
    long long ll;

    if (s.lookupValue(key, d))
    {
        v = d;
        return true;
    }

    if (s.lookupValue(key, i))
    {
        v = i;
        return true;
    }

    if (s.lookupValue(key, ll))
    {
        v = ll;
        return true;
    }

    return false;
}

const map<string, byte_order_t>
    get_byte_order_map()
{
    map<string, byte_order_t> rtn;
    rtn["BA"] = BO_BA;
    rtn["CDAB"] = BO_CDAB;
    rtn["DCBA"] = BO_DCBA;

    return rtn;
}

byte_order_t
    byte_order_from_str(const string &s)
{
    static const map<string, byte_order_t> byte_order_map = get_byte_order_map();

    const map<string, byte_order_t>::const_iterator iter = byte_order_map.find(s);
    if (iter == byte_order_map.end())
        BOOST_THROW_EXCEPTION(ParseError());

    return iter->second;
}

const map<string, function_t>
    get_function_map()
{
    map<string, function_t> rtn;
    rtn["read-coils"] = FUNC_READ_COILS;
    rtn["read-discrete-inputs"] = FUNC_READ_DISCRETE_INPUTS;
    rtn["read-input-registers"] = FUNC_READ_INPUT_REGISTERS;
    rtn["read-holding-registers"] = FUNC_READ_HOLDING_REGISTERS;

    return rtn;
}

function_t
    function_from_str(const string &s)
{
    static const map<string, function_t> func_map = get_function_map();

    const map<string, function_t>::const_iterator iter = func_map.find(s);
    if (iter == func_map.end())
        BOOST_THROW_EXCEPTION(ParseError());

    return iter->second;
}

const map<string, tag_type_t>
    get_tag_type_map()
{
    map<string, tag_type_t> rtn;
    rtn["int16"] = TAGT_INT16;
    rtn["int32"] = TAGT_INT32;
    rtn["int64"] = TAGT_INT64;
    rtn["uint16"] = TAGT_UINT16;
    rtn["uint32"] = TAGT_UINT32;
    rtn["uint64"] = TAGT_UINT64;
    rtn["float32"] = TAGT_FLOAT32;
    rtn["float64"] = TAGT_FLOAT64;
    rtn["string"] = TAGT_STRING;

    return rtn;
}

tag_type_t
    tag_type_from_str(const string &s)
{
    static const map<string, tag_type_t> tag_type_map = get_tag_type_map();

    const map<string, tag_type_t>::const_iterator iter = tag_type_map.find(s);
    if (iter == tag_type_map.end())
        BOOST_THROW_EXCEPTION(ParseError());

    return iter->second;
}

void
    validate_template_id(
        const string &id)
{
    static const regex e(
        "^[A-Za-z0-9_-]{3,255}$"
    );

    if (!regex_match(id.begin(), id.end(), e))
        BOOST_THROW_EXCEPTION(
            ParseError()
                << EInfoMsg("illegal template id: " + id)
        );
}

void
    validate_tag(
        const string &tag)
{
    if (!is_valid_tag_name(tag))
        BOOST_THROW_EXCEPTION(
            ParseError()
                << EInfoMsg("illegal tag: " + tag)
        );
}

void
    validate_unit(
        const string &unit)
{
    if (!is_valid_unit(unit))
        BOOST_THROW_EXCEPTION(
            ParseError()
                << EInfoMsg("illegal unit: " + unit)
        );
}

}
