#include <mxmodbusmaster/modbus/Equipment.hpp>

#include <boost/foreach.hpp>

namespace mxmodbusmaster
{
namespace modbus
{

using boost::shared_ptr;
using boost::posix_time::microsec_clock;
using boost::posix_time::ptime;

using mxmodbus::ModbusClient;

using mxmodbusmaster::reading::Reading;

using std::string;
using std::vector;

struct Equipment::Impl
{
    Impl(
        const string &id_,
        device_t device_,
        shared_ptr<Reader> reader_)
        : id(id_)
        , device(device_)
        , reader(reader_)
        , equipment_status(id_)
        , total_request_count(0)
        , total_request_success(0)
        , total_reading_count(0)
    {
    }

    const string                id;
    const device_t              device;
    const shared_ptr<Reader>    reader;

    EquipmentStatus             equipment_status;

    shared_ptr<ptime>   last_success_time;
    uint64_t            total_request_count;
    uint64_t            total_request_success;
    uint64_t            total_reading_count;
};

Equipment::Equipment(
    const string &id,
    device_t device,
    shared_ptr<Reader> reader)
    : pimpl(new Impl(
        id,
        device,
        reader))
{
    BOOST_ASSERT(pimpl->reader);
}

Equipment::~Equipment()
{
}

void
    Equipment::read_and_publish(
        ModbusClient &client,
        const PublishReadingFunction &publish_reading,
        const PublishStatusFunction &publish_status) const
{
    const ptime now = microsec_clock::universal_time();

    ReadStatus read_status;

    const vector<Reading> reading_list = pimpl->reader->read(
        pimpl->device,
        client,
        &read_status
    );

    // update accumulations
    EquipmentStatus::LastPoll last_poll(
        now,
        read_status.time_elapsed,
        read_status.request_count,
        read_status.request_success,
        reading_list.size());

    BOOST_FOREACH(const Reading &reading, reading_list)
    {
        publish_reading(pimpl->id, reading);
    }

    pimpl->equipment_status = pimpl->equipment_status.accumulate(last_poll);
    publish_status(pimpl->equipment_status);

    if (read_status.error_tag_list.size())
        BOOST_THROW_EXCEPTION(RequestError());
}

const EquipmentStatus
    Equipment::equipment_status() const
{
    return pimpl->equipment_status;
}

}
}
