#include <mxmodbusmaster/modbus/EquipmentStatus.hpp>

#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>

#include <rapidjson/document.h>
#include <rapidjson/stringbuffer.h>
#include <rapidjson/writer.h>

#include <mxtag/tag_form.hpp>

using boost::make_shared;
using boost::none;
using boost::optional;
using boost::posix_time::not_a_date_time;
using boost::posix_time::ptime;
using boost::posix_time::time_duration;
using boost::shared_ptr;

using mxtag::form_iso8601_time;

using rapidjson::Document;
using rapidjson::StringBuffer;
using rapidjson::StringRef;
using rapidjson::Value;
using rapidjson::Writer;

using std::pair;
using std::string;
using std::vector;

namespace mxmodbusmaster
{
namespace modbus
{

struct EquipmentStatus::Impl
{
    Impl(
        const string &equipment_,
        shared_ptr<LastPoll> last_poll_,
        int_least64_t total_request_count_,
        int_least64_t total_request_success_,
        int_least64_t total_reading_count_,
        const ptime &last_success_at_)
        : equipment(equipment_)
        , last_poll(last_poll_)
        , total_request_count(total_request_count_)
        , total_request_success(total_request_success_)
        , total_reading_count(total_reading_count_)
        , last_success_at(last_success_at_)
    {
    }

    const string                equipment;

    const shared_ptr<LastPoll>  last_poll;
    const int_least64_t         total_request_count;
    const int_least64_t         total_request_success;
    const int_least64_t         total_reading_count;
    const ptime                 last_success_at; // not_a_date_time if not exist
};

EquipmentStatus::EquipmentStatus(
    const string &equipment)
    : pimpl(new Impl(
        equipment,
        shared_ptr<LastPoll>(),
        0,
        0,
        0,
        not_a_date_time))
{
}

EquipmentStatus::EquipmentStatus(
    const string &equipment,
    const LastPoll &last_poll,
    int_least64_t total_request_count,
    int_least64_t total_request_success,
    int_least64_t total_reading_count,
    const ptime &last_success_at)
    : pimpl(new Impl(
        equipment,
        make_shared<LastPoll>(last_poll),
        total_request_count,
        total_request_success,
        total_reading_count,
        last_success_at))
{
}

EquipmentStatus::~EquipmentStatus()
{
}

const string &
    EquipmentStatus::equipment() const
{
    return pimpl->equipment;
}

optional<EquipmentStatus::LastPoll>
    EquipmentStatus::last_poll() const
{
    if (pimpl->last_poll)
        return *pimpl->last_poll;
    else
        return none;
}

optional<ptime>
    EquipmentStatus::last_success_at() const
{
    if (pimpl->last_success_at.is_not_a_date_time())
        return none;
    else
        return pimpl->last_success_at;
}

int_least64_t
    EquipmentStatus::total_request_count() const
{
    return pimpl->total_request_count;
}

int_least64_t
    EquipmentStatus::total_request_success() const
{
    return pimpl->total_request_success;
}

int_least64_t
    EquipmentStatus::total_reading_count() const
{
    return pimpl->total_reading_count;
}

const EquipmentStatus
    EquipmentStatus::accumulate(const LastPoll &last_poll) const
{
    return EquipmentStatus(
        this->equipment(),
        last_poll,
        pimpl->total_request_count + last_poll.request_count(),
        pimpl->total_request_success + last_poll.request_success(),
        pimpl->total_reading_count + last_poll.reading_count(),
        last_poll.request_count() == last_poll.request_success() ? last_poll.at() : pimpl->last_success_at);
}

namespace
{

void
    fill_last_poll(
        Value &v,
        const EquipmentStatus::LastPoll &last_poll,
        Document &document)
{
    v
        .SetObject()
        .AddMember(
            Value("at", document.GetAllocator()).Move(),
            Value(form_iso8601_time(last_poll.at()).c_str(), document.GetAllocator()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("elapsedMS", document.GetAllocator()).Move(),
            Value(last_poll.elapsed().total_milliseconds()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("requestCount", document.GetAllocator()).Move(),
            Value(last_poll.request_count()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("requestSuccess", document.GetAllocator()).Move(),
            Value(last_poll.request_success()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("readingCount", document.GetAllocator()).Move(),
            Value(last_poll.reading_count()).Move(),
            document.GetAllocator());
}

void
    fill_status(
        Value &v,
        const EquipmentStatus &status,
        Document &document)
{
// {
//     "lastPoll":
//     {
//         "at": "2015-06-25T17:05:24Z",
//         "elapsedMS": 203,
//         "requestCount": 4,
//         "requestSuccess": 4,
//         "readingCount": 4
//     },
//
//     "lastSuccessAt": "2015-06-25T17:05:24Z",
//     "totalRequestCount": 40024,
//     "totalRequestSuccess": 40000,
//     "totalReadingCount": 40000
// }

    v.SetObject();

    if (status.last_poll())
    {
        Value last_poll;
        fill_last_poll(last_poll, *status.last_poll(), document);

        v.AddMember(
            Value("lastPoll", document.GetAllocator()).Move(),
            last_poll,
            document.GetAllocator());
    }

    if (status.last_success_at())
    {
        v.AddMember(
            Value("lastSuccessAt", document.GetAllocator()).Move(),
            Value(form_iso8601_time(*status.last_success_at()).c_str(), document.GetAllocator()).Move(),
            document.GetAllocator());
    }
    else
        v.AddMember(
            Value("lastSuccessAt", document.GetAllocator()).Move(),
            Value().SetNull().Move(),
            document.GetAllocator());

    v
        .AddMember(
            Value("totalRequestCount", document.GetAllocator()).Move(),
            Value(status.total_request_count()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("totalRequestSuccess", document.GetAllocator()).Move(),
            Value(status.total_request_success()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("totalReadingCount", document.GetAllocator()).Move(),
            Value(status.total_reading_count()).Move(),
            document.GetAllocator());
}

void
    fill_full_status(
        Value &v,
        const EquipmentStatus &equipment_status,
        Document &document)
{
// {
//     "equipment": "My_other_ioLogik_E2242",
//     "status":
//     {
//         "lastPoll":
//         {
//             "at": "2015-06-25T17:05:22Z",
//             "elapsedMS": 1020,
//             "requestCount": 4,
//             "requestSuccess": 2,
//             "readingCount": 2
//         },
//
//         "totalRequestCount": 9764,
//         "totalRequestSuccess": 4882,
//         "totalReadingCount": 4882
//     }
// }

    Value status;
    fill_status(
        status,
        equipment_status,
        document);

    v
        .SetObject()
        .AddMember(
            Value("equipment", document.GetAllocator()).Move(),
            Value(equipment_status.equipment().c_str(), document.GetAllocator()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("status", document.GetAllocator()).Move(),
            status,
            document.GetAllocator());
}

}

const pair<string, vector<uint8_t> >
    mqtt_from_equipment_status(const EquipmentStatus &status)
{
    BOOST_ASSERT(status.last_poll());

    Document document;
    fill_status(document, status, document);

    StringBuffer payload;
    Writer<StringBuffer> writer(payload);
    document.Accept(writer);

    return make_pair(
        "/equs/" + status.equipment() + "/status",
        vector<uint8_t>(payload.GetString(), payload.GetString() + payload.GetSize()));
}

const string
    json_from_equipment_status_list(
        const vector<EquipmentStatus> &status_list,
        const ptime &at)
{
// {
//     "at": "2015-12-04T16:29:41Z",
//     "equipments": {
//         "My_ioLogik_E2242": {
//             "lastPoll":
//             {
//                 "at": "2015-06-25T17:05:24Z",
//                 "elapsedMS": 203,
//                 "requestCount": 4,
//                 "requestSuccess": 4,
//                 "readingCount": 4
//             },
//
//             "lastSuccessAt": "2015-06-25T17:05:24Z",
//             "totalRequestCount": 40024,
//             "totalRequestSuccess": 40000,
//             "totalReadingCount": 40000
//         },
//         "My_other_ioLogik_E2242": {
//             "lastPoll":
//             {
//                 "at": "2015-06-25T17:05:22Z",
//                 "elapsedMS": 1020,
//                 "requestCount": 4,
//                 "requestSuccess": 2,
//                 "readingCount": 2
//             },
//
//             "totalRequestCount": 9764,
//             "totalRequestSuccess": 4882,
//             "totalReadingCount": 4882
//         }
//     }
// }

    Document document;

    Value equipments;
    equipments.SetObject();
    BOOST_FOREACH(const EquipmentStatus &status, status_list)
    {
        Value v;
        fill_status(v, status, document);
        equipments.AddMember(
            Value(status.equipment().c_str(), document.GetAllocator()).Move(),
            v,
            document.GetAllocator());
    }

    document
        .SetObject()
        .AddMember(
            Value("at", document.GetAllocator()).Move(),
            Value(form_iso8601_time(at).c_str(), document.GetAllocator()).Move(),
            document.GetAllocator())
        .AddMember(
            Value("equipments", document.GetAllocator()).Move(),
            equipments,
            document.GetAllocator());

    StringBuffer json;
    Writer<StringBuffer> writer(json);
    document.Accept(writer);

    return json.GetString();
}

// EquipmentStatus::LastPoll //////////////////////////////////////////////////
struct EquipmentStatus::LastPoll::Impl
{
    Impl(
        const ptime &at_,
        const time_duration &elapsed_,
        int_least64_t request_count_,
        int_least64_t request_success_,
        int_least64_t reading_count_)
        : at(at_)
        , elapsed(elapsed_)
        , request_count(request_count_)
        , request_success(request_success_)
        , reading_count(reading_count_)
    {
    }

    const ptime             at;
    const time_duration     elapsed;
    const int_least64_t     request_count;
    const int_least64_t     request_success;
    const int_least64_t     reading_count;
};

EquipmentStatus::LastPoll::LastPoll(
    const ptime &at,
    const time_duration &elapsed,
    int_least64_t request_count,
    int_least64_t request_success,
    int_least64_t reading_count)
    : pimpl(new Impl(
        at,
        elapsed,
        request_count,
        request_success,
        reading_count))
{
}

EquipmentStatus::LastPoll::~LastPoll()
{
}

const ptime &
    EquipmentStatus::LastPoll::at() const
{
    return pimpl->at;
}

const time_duration &
    EquipmentStatus::LastPoll::elapsed() const
{
    return pimpl->elapsed;
}

int_least64_t
    EquipmentStatus::LastPoll::request_count() const
{
    return pimpl->request_count;
}

int_least64_t
    EquipmentStatus::LastPoll::request_success() const
{
    return pimpl->request_success;
}

int_least64_t
    EquipmentStatus::LastPoll::reading_count() const
{
    return pimpl->reading_count;
}

}
}
