#include <mxmodbusmaster/modbus/Reference.hpp>

#include <boost/foreach.hpp>
#include <boost/numeric/ublas/vector_sparse.hpp>

#include <mxmodbusmaster/modbus/SparseVector.hpp>

namespace mxmodbusmaster
{
namespace modbus
{

using std::vector;

using boost::numeric::ublas::mapped_vector;

namespace
{

struct Profile
{
    Reference::address_t                begin;
    Reference::address_t                end;
    SparseVector<Reference::value_t>    values;

    Profile(Reference::address_t begin_,
            Reference::address_t end_)
        : begin(begin_)
        , end(end_)
        , values(end_ - begin_)
    {
    }

    Profile(const Profile &rhs)
        : begin(rhs.begin)
        , end(rhs.end)
        , values(rhs.values)
    {
    }

    Profile &
        operator =(const Profile &rhs)
    {
        if (this == &rhs)
            return *this;

        this->begin = rhs.begin;
        this->end = rhs.end;
        this->values = rhs.values;
        return *this;
    }

    bool
        operator ==(const Profile &rhs) const
    {
        if (this == &rhs)
            return true;

        return
            this->begin == rhs.begin &&
            this->end == rhs.end &&
            this->values == rhs.values;
    }
};

}

const Reference::address_t Reference::COIL_BASE             = 0x00000;
const Reference::address_t Reference::DISCRETE_INPUT_BASE   = 0x10000;
const Reference::address_t Reference::INPUT_REGISTER_BASE   = 0x30000;
const Reference::address_t Reference::HOLDING_REGISTER_BASE = 0x40000;

struct Reference::Impl
{
    Impl()
    {
        this->profiles.push_back(Profile(0x00000, 0x10000));
        this->profiles.push_back(Profile(0x10000, 0x20000));
        this->profiles.push_back(Profile(0x30000, 0x40000));
        this->profiles.push_back(Profile(0x40000, 0x50000));
    }

    Impl(const Impl &rhs)
        : profiles(rhs.profiles)
    {
    }

    Profile &
        get_profile_by_addr(Reference::address_t addr)
    {
        BOOST_FOREACH(Profile &profile, this->profiles)
        {
            if (addr >= profile.begin && addr < profile.end)
                return profile;
        }

        BOOST_THROW_EXCEPTION(NotFoundError());
    }

    vector<Profile> profiles;
};

Reference::Reference()
    : pimpl(new Impl())
{
}

Reference::Reference(const Reference &rhs)
    : pimpl(new Impl(*rhs.pimpl))
{
}

Reference::~Reference()
{
}

Reference::value_t
    Reference::get(address_t addr) const
{
    const Profile &profile = pimpl->get_profile_by_addr(addr);
    return profile.values.get(addr - profile.begin);
}

void
    Reference::set( address_t addr,
                    value_t value)
{
    Profile &profile = pimpl->get_profile_by_addr(addr);
    profile.values.set(addr - profile.begin, value);
}

const vector<Reference::value_t>
    Reference::get( address_t addr,
                    size_t len) const
{
    vector<value_t> rtn;
    for(size_t ax = 0; ax < len; ++ax)
        rtn.push_back(this->get(addr + ax));

    return rtn;
}

void
    Reference::set( address_t addr,
                    const vector<value_t> &value)
{
    Profile &profile = pimpl->get_profile_by_addr(addr);

    try
    {
        // Use profile.values.has_value() to determine whether all values lies in valid address.
        for(size_t ax = 0; ax < value.size(); ++ax)
            (void)profile.values.has_value(addr - profile.begin + ax);
    }
    catch(NotFoundError &e)
    {
        throw;
    }

    for(size_t ax = 0; ax < value.size(); ++ax)
        this->set(addr + ax, value[ax]);
}

Reference &
    Reference::operator = (const Reference &rhs)
{
    if (this == &rhs)
        return *this;

    this->pimpl.reset(new Impl(*rhs.pimpl));

    return *this;
}

bool
    Reference::operator ==(const Reference &rhs) const
{
    if (this == &rhs)
        return true;

    return pimpl->profiles == rhs.pimpl->profiles;
}

}
}
