#include <mxmodbusmaster/modbus/Worker.hpp>

#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/thread/condition_variable.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>

#include <mxutil/log.hpp>

namespace mxmodbusmaster
{
namespace modbus
{

using boost::bind;
using boost::condition_variable;
using boost::function;
using boost::thread;
using boost::thread_interrupted;
using boost::mutex;
using boost::shared_ptr;

using std::cerr;
using std::endl;

struct Worker::Impl
{
    Impl(
        shared_ptr<Master> master_,
        const PublishReadingFunction &publish_reading_,
        const PublishStatusFunction &publish_status_)
        : master(master_)
        , publish_reading(publish_reading_)
        , publish_status(publish_status_)
        , work_stack(0)
    {
    }

    const shared_ptr<Master>        master;
    const PublishReadingFunction    publish_reading;
    const PublishStatusFunction     publish_status;
    unsigned int                    work_stack;
    mutex                           work_stack_mutex;
    condition_variable              work_stack_cond;

    shared_ptr<thread>              thread_object;

    void
        wait_for_work()
    {
        mutex::scoped_lock lock(this->work_stack_mutex);
        while(this->work_stack == 0)
        {
            this->work_stack_cond.wait(lock);
        }
    }

    void
        worker_thread()
    {
        try
        {
            for(;;)
            {
                this->wait_for_work();

                this->master->read_and_publish(
                    this->publish_reading,
                    this->publish_status
                );

                mutex::scoped_lock lock(this->work_stack_mutex);
                --this->work_stack;
            }
        }
        catch(thread_interrupted &e)
        {
            MXLOG_INFO("terminating");
        }
        catch(boost::exception &e)
        {
            MXLOG_EROR(diagnostic_information(e));

            exit(1);
        }
        catch(...)
        {
            MXLOG_EROR("unexpected exception caught");

            exit(1);
        }
    }
};

Worker::Worker(
    shared_ptr<Master> master,
    const PublishReadingFunction &publish_reading,
    const PublishStatusFunction &publish_status)
    : pimpl(new Impl(
        master,
        publish_reading,
        publish_status))
{
    BOOST_ASSERT(pimpl->master);

    pimpl->thread_object.reset(new thread(bind(
        &Impl::worker_thread,
        pimpl
    )));
}

Worker::~Worker()
{
    pimpl->thread_object->interrupt();
    pimpl->thread_object->join();
}

void
    Worker::whip()
{
    mutex::scoped_lock lock(pimpl->work_stack_mutex);

    if (0 == pimpl->work_stack)
    {
        ++pimpl->work_stack;
        pimpl->work_stack_cond.notify_one();
    }
}

}
}
