#include <mxmodbusmaster/modbus/trivial/Tag.hpp>

#include <boost/foreach.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusmaster/reading/Value.hpp>
#include <mxmodbusmaster/modbus/trivial/tag_convert.hpp>

namespace mxmodbusmaster
{
namespace modbus
{
namespace trivial
{

using boost::optional;
using boost::none;
using boost::posix_time::ptime;
using boost::scoped_ptr;

using mxmodbusmaster::reading::Reading;
using mxmodbusmaster::reading::Value;

using std::string;
using std::vector;

namespace
{

const vector<uint8_t>
    byte_vector_from_word_vector(const vector<uint16_t> &wv)
{
    vector<uint8_t> bv;

    BOOST_FOREACH(uint16_t w, wv)
    {
        bv.push_back((w >> 8) & 0xFF);
        bv.push_back((w >> 0) & 0xFF);
    }

    return bv;
}

}

struct Tag::Impl
{
    Impl(
        const string &id_,
        uint32_t address_,
        size_t quantity_,
        tag_type_t type_)
        : id(id_)
        , address(address_)
        , quantity(quantity_)
        , type(type_)
        , byte_order(BO_NC)
    {
    }

    const string        id;
    const uint32_t      address;
    const size_t        quantity;
    const tag_type_t    type;

    scoped_ptr<string>              unit;
    scoped_ptr<vector<uint16_t> >   invalid_value;
    byte_order_t                    byte_order;
    scoped_ptr<double>              slope;
    scoped_ptr<double>              offset;

    const Value
        value_from_binary(const vector<uint8_t> &buf) const
    {
        switch(this->type)
        {
        case TAGT_INT16:    return Value::get(static_cast<int64_t>(int16_from_binary(buf)));
        case TAGT_INT32:    return Value::get(static_cast<int64_t>(int32_from_binary(buf)));
        case TAGT_INT64:    return Value::get(int64_from_binary(buf));
        case TAGT_UINT16:   return Value::get(static_cast<uint64_t>(uint16_from_binary(buf)));
        case TAGT_UINT32:   return Value::get(static_cast<uint64_t>(uint32_from_binary(buf)));
        case TAGT_UINT64:   return Value::get(uint64_from_binary(buf));
        case TAGT_FLOAT32:  return Value::get(float32_from_binary(buf));
        case TAGT_FLOAT64:  return Value::get(float64_from_binary(buf));
        case TAGT_STRING:   return Value::get(string_from_binary(buf));
        default:
            BOOST_THROW_EXCEPTION(ArgumentError());
        }
    }
};

Tag::Tag(
    const string &  id,
    uint32_t        address,
    size_t          quantity,
    tag_type_t      type)
    : pimpl(new Impl(
        id,
        address,
        quantity,
        type
    ))
{
}

Tag::~Tag()
{
}

void
    Tag::set_unit(const string &unit)
{
    pimpl->unit.reset(new string(unit));
}

void
    Tag::set_invalid_value(const vector<uint16_t> &invalid_value)
{
    if (invalid_value.size() != pimpl->quantity)
        BOOST_THROW_EXCEPTION(ArgumentError());

    pimpl->invalid_value.reset(new vector<uint16_t>(invalid_value));
}

void
    Tag::set_byte_order(byte_order_t bo)
{
    pimpl->byte_order = bo;
}

void
    Tag::set_slope(double slope)
{
    pimpl->slope.reset(new double(slope));
}

void
    Tag::set_offset(double offset)
{
    pimpl->offset.reset(new double(offset));
}

const string &
    Tag::id() const
{
    return pimpl->id;
}

optional<Reading>
    Tag::get_reading(
        const Reference &ref,
        const ptime &now) const
{
    try
    {
        const vector<uint16_t> wbuf = ref.get(pimpl->address, pimpl->quantity); // NotFoundError

        if (pimpl->invalid_value &&
            wbuf == *pimpl->invalid_value)
        {
            MXLOG_INFO("invalid-value matched, ignore value");
            return none;
        }

        vector<uint8_t> buf = byte_vector_from_word_vector(wbuf);

        buf = reorder(pimpl->byte_order, buf);

        Value v = pimpl->value_from_binary(buf);

        if (pimpl->slope)
            v = v.multiply(*pimpl->slope);

        if (pimpl->offset)
            v = v.add(*pimpl->offset);

        if (pimpl->unit)
            return Reading(pimpl->id, now, v, *pimpl->unit);
        else
            return Reading(pimpl->id, now, v);
    }
    catch(NotFoundError &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        throw;
    }
}

}
}
}
