#include <mxmodbusmaster/modbus/trivial/TrivialReader.hpp>

#include <boost/date_time/posix_time/posix_time_types.hpp>
#include <boost/foreach.hpp>

#include <mxutil/log.hpp>

namespace mxmodbusmaster
{
namespace modbus
{
namespace trivial
{

using boost::optional;
using boost::posix_time::ptime;
using boost::posix_time::microsec_clock;
using boost::shared_ptr;

using mxmodbus::ModbusClient;

using mxmodbusmaster::modbus::Command;
using mxmodbusmaster::reading::Reading;

using std::vector;

struct TrivialReader::Impl
{
    Impl(
        const vector<shared_ptr<Command> > command_list_,
        const vector<shared_ptr<Tag> > tag_list_)
        : command_list(command_list_)
        , tag_list(tag_list_)
    {
    }

    const vector<shared_ptr<Command> >  command_list;
    const vector<shared_ptr<Tag> >      tag_list;
};

TrivialReader::TrivialReader(
    const vector<shared_ptr<Command> > &command_list,
    const vector<shared_ptr<Tag> > &tag_list)
    : pimpl(new Impl(
        command_list,
        tag_list))
{
}

TrivialReader::~TrivialReader()
{
}

const vector<Reading>
    TrivialReader::read(
        device_t device,
        ModbusClient &client,
        ReadStatus *read_status) const
{
    const ptime now = microsec_clock::universal_time();
    Reference ref;
    // run all commands
    try
    {
        BOOST_FOREACH(shared_ptr<Command> command, pimpl->command_list)
        {
            // accumulate request count
            if (read_status)
                ++read_status->request_count;

            command->run(
                device,
                client,
                ref
            ); // throws ModbusError

            // accumulate success count
            if (read_status)
                ++read_status->request_success;
        }

        // store elapse time
        if (read_status)
            read_status->time_elapsed = microsec_clock::universal_time() - now;
    }
    catch(mxmodbus::Error &e)
    {
        MXLOG_WARN(diagnostic_information(e));

        if (read_status)
            read_status->time_elapsed = microsec_clock::universal_time() - now;
    }

    // extract Reading from Reference
    vector<Reading> rtn;
    BOOST_FOREACH(shared_ptr<Tag> tag, pimpl->tag_list)
    {
        optional<Reading> reading;
        try
        {
            reading = tag->get_reading(ref, now); // throws NotFoundError
            if (!reading)
                continue;
        }
        catch(NotFoundError &e)
        {
            if (read_status)
                read_status->error_tag_list.push_back(tag->id());

            continue;
        }

        rtn.push_back(*reading);
    }

    return rtn;
}

}
}
}
