#include <mxmodbusmaster/modbus/trivial/TrivialReaderBuilder.hpp>

#include <boost/cstdint.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusmaster/modbus/ReadCoilsCommand.hpp>
#include <mxmodbusmaster/modbus/ReadDiscreteInputsCommand.hpp>
#include <mxmodbusmaster/modbus/ReadInputRegistersCommand.hpp>
#include <mxmodbusmaster/modbus/ReadHoldingRegistersCommand.hpp>
#include <mxmodbusmaster/modbus/trivial/DummyModbusClient.hpp>

using boost::lexical_cast;
using boost::make_shared;
using boost::shared_ptr;

using mxmodbusmaster::ArgumentError;
using mxmodbusmaster::modbus::Command;
using mxmodbusmaster::modbus::FUNC_READ_COILS;
using mxmodbusmaster::modbus::FUNC_READ_DISCRETE_INPUTS;
using mxmodbusmaster::modbus::FUNC_READ_INPUT_REGISTERS;
using mxmodbusmaster::modbus::FUNC_READ_HOLDING_REGISTERS;
using mxmodbusmaster::modbus::function_t;
using mxmodbusmaster::modbus::ReadCoilsCommand;
using mxmodbusmaster::modbus::ReadDiscreteInputsCommand;
using mxmodbusmaster::modbus::ReadInputRegistersCommand;
using mxmodbusmaster::modbus::ReadHoldingRegistersCommand;

using std::string;
using std::vector;

namespace
{

shared_ptr<Command>
    create_command(
        function_t function,
        uint16_t address,
        uint16_t quantity)
{
    switch(function)
    {
    case FUNC_READ_COILS:
        return shared_ptr<Command>(new ReadCoilsCommand(address, quantity));
    case FUNC_READ_DISCRETE_INPUTS:
        return shared_ptr<Command>(new ReadDiscreteInputsCommand(address, quantity));
    case FUNC_READ_INPUT_REGISTERS:
        return shared_ptr<Command>(new ReadInputRegistersCommand(address, quantity));
    case FUNC_READ_HOLDING_REGISTERS:
        return shared_ptr<Command>(new ReadHoldingRegistersCommand(address, quantity));
    default:
        MXLOG_EROR("unexpected function: " + lexical_cast<string>(function));
        BOOST_THROW_EXCEPTION(ArgumentError());
    }
}

}

namespace mxmodbusmaster
{
namespace modbus
{
namespace trivial
{

struct TrivialReaderBuilder::Impl
{
    Impl()
    {
    }

    shared_ptr<EquipmentTemplate> equipment_template;
};

TrivialReaderBuilder::TrivialReaderBuilder()
    : pimpl(new Impl())
{
}

TrivialReaderBuilder::~TrivialReaderBuilder()
{
}

void
    TrivialReaderBuilder::set_equipment_template(const EquipmentTemplate &equipment_template)
{
    pimpl->equipment_template = make_shared<EquipmentTemplate>(equipment_template);
}

shared_ptr<TrivialReader>
    TrivialReaderBuilder::build() const
{
    BOOST_ASSERT(pimpl->equipment_template);

    const EquipmentTemplate &et = *pimpl->equipment_template;

    vector<shared_ptr<Command> > command_list;
    BOOST_FOREACH(const EquipmentTemplate::Request &req, et.request_list)
    {
        command_list.push_back(create_command(
            req.function,
            req.address,
            req.quantity
        ));
    }

    vector<shared_ptr<Tag> > tag_list;
    BOOST_FOREACH(const EquipmentTemplate::Tag &tagcfg, et.tag_list)
    {
        shared_ptr<Tag> tag(new Tag(
            tagcfg.id,
            tagcfg.reference,
            tagcfg.quantity,
            tagcfg.type
        ));

        if (tagcfg.unit)
            tag->set_unit(*tagcfg.unit);

        if (tagcfg.invalid_value)   // throws ArgumentError
            tag->set_invalid_value(*tagcfg.invalid_value);

        if (tagcfg.byte_order)
            tag->set_byte_order(*tagcfg.byte_order);

        if (tagcfg.slope)
            tag->set_slope(*tagcfg.slope);

        if (tagcfg.offset)
            tag->set_offset(*tagcfg.offset);

        tag_list.push_back(tag);
    }

    const shared_ptr<TrivialReader> reader = make_shared<TrivialReader>(
        command_list,
        tag_list
    );

    DummyModbusClient dummy_modbus_client;
    ReadStatus read_status;
    reader->read(0x00, dummy_modbus_client, &read_status);
    if (read_status.error_tag_list.size())
        BOOST_THROW_EXCEPTION(ArgumentError());

    return reader;
}

}
}
}
