#include <mxmodbusmaster/modbus/trivial/TrivialReaderFactory.hpp>

#include <fstream>
#include <vector>
#include <map>

#include <boost/foreach.hpp>

#include <mxutil/log.hpp>

#include <mxmodbusmaster/EquipmentTemplate.hpp>
#include <mxmodbusmaster/modbus/trivial/TrivialReaderBuilder.hpp>

using boost::filesystem::directory_iterator;
using boost::filesystem::is_directory;
using boost::filesystem::path;
using boost::shared_ptr;

using mxmodbusmaster::ArgumentError;
using mxmodbusmaster::EquipmentTemplate;
using mxmodbusmaster::modbus::trivial::TrivialReader;
using mxmodbusmaster::modbus::trivial::TrivialReaderBuilder;

using std::ifstream;
using std::ios_base;
using std::make_pair;
using std::map;
using std::string;
using std::vector;

namespace
{

const vector<path>
    listdir(const path &dir)
{
    vector<path> v;
    copy(directory_iterator(dir), directory_iterator(), back_inserter(v));

    vector<path> rtn;
    BOOST_FOREACH(const path &p, v)
    {
        if (is_regular_file(p))
            rtn.push_back(p);
    }

    return rtn;
}

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_WARN("open file failed: " + p.string());

        return string();
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_WARN("read file failed: " + p.string());

        return string();
    }

    return string(buf.begin(), buf.end());
}

const size_t FILESIZE_LIMIT = 128 * 1024;

}

namespace mxmodbusmaster
{
namespace modbus
{
namespace trivial
{

typedef
    map<string, shared_ptr<TrivialReader> >
    IdReaderMap;

struct TrivialReaderFactory::Impl
{
    Impl()
    {
    }

    IdReaderMap   id_reader_map;
};

TrivialReaderFactory::TrivialReaderFactory(const path &config_dir)
    : pimpl(new Impl())
{
    if (!is_directory(config_dir))
        BOOST_THROW_EXCEPTION(ArgumentError());

    const vector<path> filelist = listdir(path(config_dir));
    BOOST_FOREACH(const path &f, filelist)
    {
        const size_t fs = file_size(f);
        if (fs > FILESIZE_LIMIT)
        {
            MXLOG_WARN("file too large: " + f.string() + ", ignored");

            continue;
        }

        if (f.extension().string() != ".met")
        {
            MXLOG_WARN("file ignored: " + f.string());

            continue;
        }

        try
        {
            const string cfg_str = readfile(f);

            // parse
            const EquipmentTemplate et = EquipmentTemplate::parse(cfg_str); // throws ParseError

            // check whether same id exist
            const IdReaderMap::const_iterator iter = pimpl->id_reader_map.find(et.id);
            if (iter != pimpl->id_reader_map.end())
            {
                MXLOG_WARN("duplicated template-id: file = " + f.string() + ", id = " + et.id);

                continue;
            }

            // create TrivialReader
            TrivialReaderBuilder builder;
            builder.set_equipment_template(et);
            const shared_ptr<TrivialReader> tr = builder.build();   // throws ArgumentError

            // add TrivialReader to id_reader_map
            pimpl->id_reader_map.insert(make_pair<string, shared_ptr<TrivialReader> >(
                et.id,
                tr
            ));

            MXLOG_INFO("successfully loaded: file = " + f.string());
        }
        catch(ParseError &e)
        {
            MXLOG_WARN(diagnostic_information(e));

            continue;
        }
        catch(ArgumentError &e)
        {
            MXLOG_WARN(diagnostic_information(e));

            continue;
        }
    }
}

TrivialReaderFactory::~TrivialReaderFactory()
{
}

shared_ptr<Reader>
    TrivialReaderFactory::get(const string &name) const
{
    const IdReaderMap::const_iterator iter = pimpl->id_reader_map.find(name);

    if (iter == pimpl->id_reader_map.end())
        return shared_ptr<Reader>();

    return iter->second;
}

}
}
}
