#include <boost/test/unit_test.hpp>

#include <mxmodbusmaster/modbus/Reference.hpp>

using mxmodbusmaster::modbus::Reference;

using std::vector;

namespace
{

struct TestFixture
{
    Reference   reference;
};

}

namespace mxmodbusmaster
{
namespace modbus
{

BOOST_FIXTURE_TEST_CASE(test_Reference_get_WithOutOfRangeAddress_ShouldThrowOutOfRange, TestFixture)
{
    // arrange

    // act & assert
    BOOST_CHECK_THROW(reference.get(0x20000), NotFoundError);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_get_WithNotExistAddress_ShouldThrowNotFoundError, TestFixture)
{
    // arrange

    // act & assert
    BOOST_CHECK_THROW(reference.get(0x00000), NotFoundError);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_get_ShouldPass, TestFixture)
{
    // arrange
    reference.set(0x10000, 1);

    // act
    const uint16_t value = reference.get(0x10000);

    // assert
    BOOST_CHECK_EQUAL(1, value);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_get2_ShouldPass, TestFixture)
{
    // arrange
    reference.set(0x30000, 0x1234);
    reference.set(0x30001, 0x5678);

    // act
    const vector<uint16_t> values = reference.get(0x30000, 2);

    // assert
    BOOST_REQUIRE_EQUAL(2, values.size());
    BOOST_CHECK_EQUAL(0x1234, values[0]);
    BOOST_CHECK_EQUAL(0x5678, values[1]);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_set2_WithOutOfRangeLen_ShouldThrowOutOfRange, TestFixture)
{
    // arrange

    // act & assert
    BOOST_CHECK_THROW(reference.set(0x3FFFF, vector<uint16_t>(2, 0x3456)), NotFoundError);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_set2_ShouldPass, TestFixture)
{
    // arrange

    // act
    reference.set(0x3FFFF, vector<uint16_t>(1, 0x1234));

    // assert
    BOOST_CHECK_EQUAL(0x1234, reference.get(0x3FFFF));
}

BOOST_FIXTURE_TEST_CASE(test_Reference_operator_equ_ShouldPass, TestFixture)
{
    // arrange
    reference.set(0x00000, 1);
    reference.set(0x10000, 0);
    reference.set(0x30000, 0x1234);
    reference.set(0x40000, 0x5678);

    Reference reference2;
    reference2.set(0x00000, 1);
    reference2.set(0x10000, 0);
    reference2.set(0x30000, 0x1234);
    reference2.set(0x40000, 0x5678);

    // act
    const bool result = reference == reference2;

    // assert
    BOOST_CHECK(result);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_copy_ctor_ShouldPass, TestFixture)
{
    // arrange
    reference.set(0x00000, 1);
    reference.set(0x10000, 0);
    reference.set(0x30000, 0x1234);
    reference.set(0x40000, 0x5678);

    // act
    const Reference lhs = reference;

    // assert
    BOOST_CHECK(reference == lhs);
}

BOOST_FIXTURE_TEST_CASE(test_Reference_operator_assignment_ShouldPass, TestFixture)
{
    // arrange
    reference.set(0x00000, 1);
    reference.set(0x10000, 0);
    reference.set(0x30000, 0x1234);
    reference.set(0x40000, 0x5678);

    // act
    Reference lhs;
    lhs = reference;

    // assert
    BOOST_CHECK(reference == lhs);
}

}
}
