#include <boost/test/unit_test.hpp>

#include <mxmodbusmaster/Configuration.hpp>

namespace mxmodbusmaster
{
using std::string;

namespace
{

class TestFixture
{
};
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_ShouldReturnCorrectLength, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "# same port with different polling-period\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 60000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x32\n"
                        "template-id = \"ioLogik-E2242\"\n"
                        "equipment-name = \"My_ioLogik-E2242\"\n"
                    "}\n"
                ")\n"
            "},\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 1502\n"
                "polling-period-ms = 45000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x01\n"
                        "template-id = \"ioLogik-E2242\"\n"
                        "equipment-name = \"Another_ioLogik-E2242\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act
    const Configuration result = Configuration::parse(SUT);

    // assert
    BOOST_CHECK_EQUAL(3, result.polling_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_ShouldReturnCorrectValue, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n"
        "equipment-status-path = \"/dev/shm/var/mxmodbusmaster/equipment-status\"\n";

    // act
    const Configuration result = Configuration::parse(SUT);

    // assert
    BOOST_REQUIRE_EQUAL(1, result.polling_list.size());

    const Configuration::Polling polling = result.polling_list[0];
    BOOST_CHECK_EQUAL("127.0.0.1", polling.host);
    BOOST_CHECK_EQUAL(502, polling.service);
    BOOST_CHECK_EQUAL(30000, polling.polling_period_ms);
    BOOST_CHECK_EQUAL(30000, polling.request_timeout_ms);

    BOOST_REQUIRE_EQUAL(2, polling.device_list.size());

    {
        const Configuration::Polling::Device device = polling.device_list[0];
        BOOST_CHECK_EQUAL(0x30, device.device_id);
        BOOST_CHECK_EQUAL("ioLogik-E2214", device.template_id);
        BOOST_CHECK_EQUAL("My_ioLogik-E2214", device.equipment_name);
    }

    {
        const Configuration::Polling::Device device = polling.device_list[1];
        BOOST_CHECK_EQUAL(0x31, device.device_id);
        BOOST_CHECK_EQUAL("Sunspec_Inverter", device.template_id);
        BOOST_CHECK_EQUAL("THE_building", device.equipment_name);
    }
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithFaultyConfig_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec Inverter\"\n"
                        "equipment-name = \"THE building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        "}\n" // oops
    ;

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithInsufficientConfig_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "#service = 502\n" // oops
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WhenDeviceIdTooBig_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 256\n" // oops
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithPollingPeriodTooSmall_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 300\n" // oops
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithRequestTimeoutTooSmall_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30\n" // oops
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithIllegalEquipmentName_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My ioLogik-E2214\"\n" // oops
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithIllegalTemplateId_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214 \"\n" // oops
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"THE_building\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

BOOST_FIXTURE_TEST_CASE(test_Configuration_parse_WithDuplicatedEquipmentName_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "polling-list = (\n"
            "{\n"
                "host = \"127.0.0.1\"\n"
                "service = 502\n"
                "polling-period-ms = 30000\n"
                "request-timeout-ms = 30000\n"
                "device-list = (\n"
                    "{\n"
                        "device-id = 0x30\n"
                        "template-id = \"ioLogik-E2214\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "},\n"
                    "{\n"
                        "device-id = 0x31\n"
                        "template-id = \"Sunspec_Inverter\"\n"
                        "equipment-name = \"My_ioLogik-E2214\"\n"
                    "}\n"
                ")\n"
            "}\n"
        ")\n"
        "equipment-status-path = \"/dev/shm/var/mxmodbusmaster/equipment-status\"\n";

    // act & assert
    BOOST_CHECK_THROW(Configuration::parse(SUT), ParseError)
}

}
