#include <boost/test/unit_test.hpp>

#include <mxmodbusmaster/EquipmentTemplate.hpp>

namespace mxmodbusmaster
{

using mxmodbusmaster::modbus::FUNC_READ_COILS;
using mxmodbusmaster::modbus::trivial::TAGT_UINT16;
using mxmodbusmaster::modbus::trivial::TAGT_FLOAT32;

using std::string;

namespace
{

class TestFixture
{
};
}

BOOST_FIXTURE_TEST_CASE(test_EquipmentTemplate_parse_ShouldReturnCorrectIdLength, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "},\n"
            "{\n"
                "function = \"read-input-registers\"\n"
                "address = 0x0130\n"
                "quantity = 2\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"di0\"\n"
                "reference = 0x00000\n"
                "quantity = 1\n"
                "type = \"uint16\"\n"
            "},\n"
            "{\n"
                "id = \"di1\"\n"
                "reference = 0x00001\n"
                "quantity = 1\n"
                "type = \"uint16\"\n"
            "},\n"
            "{\n"
                "id = \"di2\"\n"
                "reference = 0x00002\n"
                "quantity = 1\n"
                "type = \"uint16\"\n"
            "}\n"
        ")\n"
    ;

    // act
    const EquipmentTemplate result = EquipmentTemplate::parse(SUT);

    // assert
    BOOST_CHECK_EQUAL("ioLogik-E2242", result.id);
    BOOST_CHECK_EQUAL(2, result.request_list.size());
    BOOST_CHECK_EQUAL(3, result.tag_list.size());
}

BOOST_FIXTURE_TEST_CASE(test_EquipmentTemplate_parse_ShouldReturnCorrectValue, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"di0\"\n"
                "reference = 0x00000\n"
                "quantity = 1\n"
                "type = \"uint16\"\n"
            "},\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"float32\"\n"
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        ")\n"
    ;

    // act
    const EquipmentTemplate result = EquipmentTemplate::parse(SUT);

    // assert
    BOOST_REQUIRE_EQUAL("ioLogik-E2242", result.id);

    BOOST_REQUIRE_EQUAL(1, result.request_list.size());

    const EquipmentTemplate::Request request = result.request_list[0];
    BOOST_CHECK_EQUAL(FUNC_READ_COILS, request.function);
    BOOST_CHECK_EQUAL(0x0000, request.address);
    BOOST_CHECK_EQUAL(6, request.quantity);

    BOOST_REQUIRE_EQUAL(2, result.tag_list.size()); {
        const EquipmentTemplate::Tag tag = result.tag_list[0];
        BOOST_CHECK_EQUAL("di0", tag.id);
        BOOST_CHECK_EQUAL(0x00000, tag.reference);
        BOOST_CHECK_EQUAL(1, tag.quantity);
        BOOST_CHECK_EQUAL(TAGT_UINT16, tag.type);
    }

    {
        const EquipmentTemplate::Tag tag = result.tag_list[1];
        BOOST_CHECK_EQUAL("ai0", tag.id);
        BOOST_CHECK_EQUAL("V", *tag.unit);
        BOOST_CHECK_EQUAL(0x40130, tag.reference);
        BOOST_CHECK_EQUAL(2, tag.quantity);
        BOOST_CHECK_EQUAL(0xFFFF, (*tag.invalid_value)[0]);
        BOOST_CHECK_EQUAL(0xFFFF, (*tag.invalid_value)[1]);
        BOOST_CHECK_EQUAL(modbus::trivial::BO_CDAB, *tag.byte_order);
        BOOST_CHECK_EQUAL(TAGT_FLOAT32, tag.type);
        BOOST_CHECK_EQUAL(0.33, *tag.slope);
        BOOST_CHECK_EQUAL(3.14, *tag.offset);
    }
}

BOOST_FIXTURE_TEST_CASE(test_EquipmentTemplate_parse_WithIntSlopeAndOffset_ShouldPass, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"di0\"\n"
                "reference = 0x00000\n"
                "quantity = 1\n"
                "type = \"uint16\"\n"
            "},\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"float32\"\n"
                "slope = 3\n"
                "offset = 5\n"
            "}\n"
        ")\n"
    ;

    // act
    const EquipmentTemplate result = EquipmentTemplate::parse(SUT);

    // assert
    const EquipmentTemplate::Tag tag = result.tag_list[1];
    BOOST_REQUIRE(tag.slope);
    BOOST_CHECK_CLOSE(3, *tag.slope, 0.0001);
    BOOST_REQUIRE(tag.offset);
    BOOST_CHECK_CLOSE(5, *tag.offset, 0.0001);
}

BOOST_FIXTURE_TEST_CASE (test_EquipmentTemplate_parse_WithFaultyConfig_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"float32\"\n"
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        "}\n" // oops
    ;

    // act & assert
    BOOST_CHECK_THROW(EquipmentTemplate::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE (test_EquipmentTemplate_parse_WithInsufficientConfig_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "jd = \"ioLogik-E2242\"\n" // oops
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"float32\"\n"
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(EquipmentTemplate::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE (test_EquipmentTemplate_parse_WhenTagQuantityAndInvalidValueSizeMismatch_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 3\n" // oops
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n" // oops
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"float32\"\n"
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(EquipmentTemplate::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE (test_EquipmentTemplate_parse_WhenTagTypeInsufficient_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"CDAB\"\n"
                "type = \"int48\"\n" // oops
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(EquipmentTemplate::parse(SUT), ParseError);
}

BOOST_FIXTURE_TEST_CASE (test_EquipmentTemplate_parse_WhenTagByteOrderInsufficient_ShouldThrowParseError, TestFixture)
{
    // arrange
    const string SUT =
        "id = \"ioLogik-E2242\"\n"
        "request-list = (\n"
            "{\n"
                "function = \"read-coils\"\n"
                "address = 0x0000\n"
                "quantity = 6\n"
            "}\n"
        ")\n"
        "tag-list = (\n"
            "{\n"
                "id = \"ai0\"\n"
                "unit = \"V\"\n"
                "reference = 0x40130\n"
                "quantity = 2\n"
                "invalid-value = [\n"
                    "0xFFFF, 0xFFFF\n"
                "]\n"
                "byte-order = \"AAAA\"\n" // oops
                "type = \"float32\"\n"
                "slope = 0.33\n"
                "offset = 3.14\n"
            "}\n"
        ")\n"
    ;

    // act & assert
    BOOST_CHECK_THROW(EquipmentTemplate::parse(SUT), ParseError);
}

}
