#ifndef _MXSERIALFRAMER_FRAMER_HPP_
#define _MXSERIALFRAMER_FRAMER_HPP_

#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/noncopyable.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include <mxutil/async.hpp>
#include <mxutil/log.hpp>

namespace mxserialframer
{

template <typename ISTREAM, typename OSTREAM>
class Framer
    : private boost::noncopyable
    , public boost::enable_shared_from_this<Framer<ISTREAM, OSTREAM> >
{
public:
    Framer(
        boost::asio::io_service &iosvc,
        ISTREAM &is,
        OSTREAM &os,
        const boost::posix_time::time_duration &noinput_timeout = boost::posix_time::not_a_date_time,
        int overflow_size = -1)
        : iosvc_(iosvc)
        , is_(is)
        , os_(os)
        , noinput_timeout_(noinput_timeout)
        , overflow_size_(overflow_size)
        , noinput_timer_(iosvc)
        , is_reading_(false)
        , is_writing_(false)
    {
    }

    ~Framer()
    {
    }

    void
        async_start(mxutil::CompleteCallback cc)
    {
        using boost::make_shared;

        BOOST_ASSERT(!this->cc_);
        this->cc_ = make_shared<mxutil::CompleteCallback>(cc);

        return this->do_read();
    }

    void
        cancel()
    {
        this->is_.cancel();
        this->os_.cancel();
        this->noinput_timer_.cancel();
    }

private:
    void
        do_read()
    {
        using boost::bind;
        using boost::shared_ptr;
        using boost::make_shared;
        using std::vector;

        if (this->is_reading_)
            return;

        this->is_reading_ = true;

        const shared_ptr<vector<uint8_t> > buf = make_shared<vector<uint8_t> >(4096);
        this->is_.async_read_some(
            boost::asio::buffer(*buf),
            bind(
                &Framer::on_read_complete,
                this->shared_from_this(),
                boost::asio::placeholders::error,
                boost::asio::placeholders::bytes_transferred,
                buf
            )
        );
    }

    void
        on_read_complete(
            const boost::system::error_code &ec,
            std::size_t bytes_transferred,
            boost::shared_ptr<std::vector<uint8_t> > buf)
    {
        this->is_reading_ = false;

        if (ec)
            return this->invoke_cc(ec);

        MXLOG_DBUG(boost::lexical_cast<std::string>(this->noinput_timeout_)
            + ", " + boost::lexical_cast<std::string>(this->noinput_timeout_.is_not_a_date_time()));

        // reset timer
        if (bytes_transferred != 0 &&
            !this->noinput_timeout_.is_not_a_date_time())
        {
            MXLOG_DBUG("reset timer");

            this->noinput_timer_.cancel();
            this->noinput_timer_.expires_from_now(this->noinput_timeout_);
            this->noinput_timer_.async_wait(bind(
                &Framer::on_noinput_timeout,
                this->shared_from_this(),
                boost::asio::placeholders::error));
        }

        MXLOG_DBUG(boost::lexical_cast<std::string>(bytes_transferred) + " bytes read");

        // append to buffer
        this->buf_.insert(
            this->buf_.end(),
            buf->begin(),
            buf->begin() + bytes_transferred);

        // if buffer empty, continue reading
        if (this->buf_.empty())
            return this->do_read();

        // if buffer overflow, flush
        if (this->overflow_size_ >= 0 &&
            this->buf_.size() >= static_cast<std::size_t>(this->overflow_size_))
            return this->do_write();

        // continue reading
        return this->do_read();
    }

    void
        do_write()
    {
        using boost::shared_ptr;
        using boost::make_shared;

        using std::vector;

        if (this->is_writing_)
            return;

        this->is_writing_ = true;

        const shared_ptr<vector<uint8_t> > buf = make_shared<vector<uint8_t> >();
        swap(*buf, this->buf_);

        MXLOG_DBUG("writing " + boost::lexical_cast<std::string>(buf->size()) + " bytes");

        boost::asio::async_write(
            this->os_,
            boost::asio::buffer(*buf),
            boost::bind(
                &Framer::on_write_complete,
                this->shared_from_this(),
                boost::asio::placeholders::error,
                boost::asio::placeholders::bytes_transferred,
                buf)
        );
    }

    void
        on_write_complete(
            const boost::system::error_code &ec,
            std::size_t bytes_transferred,
            boost::shared_ptr<std::vector<uint8_t> > buf)
    {
        this->is_writing_ = false;

        if (ec)
            return this->invoke_cc(ec);

        BOOST_ASSERT(buf);

        buf->erase(buf->begin(), buf->begin() + bytes_transferred);
        if (!buf->empty())
        {
            this->is_writing_ = true;

            boost::asio::async_write(
                this->os_,
                boost::asio::buffer(*buf),
                boost::bind(
                    &Framer::on_write_complete,
                    this->shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred,
                    buf)
            );

            return;
        }

        return this->do_read();
    }

    void
        on_noinput_timeout(const boost::system::error_code &ec)
    {
        MXLOG_DBUG("noinput timeout");

        if (ec)
            return;

        return this->do_write();
    }

    void
        invoke_cc(const boost::system::error_code &ec)
    {
        if (!this->cc_)
            return;

        boost::shared_ptr<mxutil::CompleteCallback> cc;
        swap(cc, this->cc_);

        this->iosvc_.post(boost::bind(
            *cc,
            ec));
    }

private:
    boost::asio::io_service &   iosvc_;
    ISTREAM &   is_;
    OSTREAM &   os_;
    const boost::posix_time::time_duration  noinput_timeout_;
    const int                               overflow_size_;

    boost::asio::deadline_timer             noinput_timer_;

    boost::shared_ptr<mxutil::CompleteCallback> cc_;
    bool        is_reading_;
    bool        is_writing_;

    std::vector<uint8_t>    buf_;
};

}

#endif
