#include <exception>
#include <string>
#include <vector>

#include <boost/cstdint.hpp>
#include <boost/exception/all.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>

#include <mxsl/mxsl.hpp>

#include <mxutil/log.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StderrLogger.hpp>

#include <mxserialframer/Framer.hpp>
#include <mxserialframer/Option.hpp>

using boost::asio::io_service;
using boost::asio::serial_port;
using boost::asio::posix::stream_descriptor;
using boost::get_error_info;
using boost::lexical_cast;
using boost::logic::tribool;
using boost::make_shared;
using boost::posix_time::milliseconds;
using boost::posix_time::not_a_date_time;
using boost::posix_time::time_duration;
using boost::ref;
using boost::shared_ptr;

using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StderrLogger;

using mxserialframer::Framer;
using mxserialframer::Option;

using std::string;
using std::vector;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

bool go = true;

void
    error(const boost::system::error_code &ec)
{
    MXLOG_EROR("ec = " + ec.message());
    go = false;
}

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StderrLogger());
        logger->set_level(Logger::WARN);
        if (opt.verbose)
            logger->set_level(Logger::DBUG);

        logger->enable_line_of_file();
        Log::get().set_logger(logger);

        mxsl::validate();

        io_service iosvc;
        serial_port sp(iosvc, opt.device);
        sp.set_option(opt.config.baud);
        sp.set_option(opt.config.char_size);
        sp.set_option(opt.config.parity);
        sp.set_option(opt.config.stop_bits);

        stream_descriptor in(iosvc, ::dup(STDIN_FILENO));
        stream_descriptor out(iosvc, ::dup(STDOUT_FILENO));

        int default_overflow_size = -1;

        // forward unconditionally when no conditions are set
        if (!opt.noinput_timeout_ms &&
            !opt.overflow_size)
        {
            default_overflow_size = 0;
        }

        typedef Framer<serial_port, stream_descriptor> ser2stdout_framer_t;
        const shared_ptr<ser2stdout_framer_t> ser2stdout = make_shared<ser2stdout_framer_t>(
            ref(iosvc),
            ref(sp),
            ref(out),
            opt.noinput_timeout_ms ? milliseconds(*opt.noinput_timeout_ms) : time_duration(not_a_date_time),
            opt.overflow_size ? *opt.overflow_size : default_overflow_size
        );

        typedef Framer<stream_descriptor, serial_port> stdin2ser_framer_t;
        const shared_ptr<stdin2ser_framer_t> stdin2ser = make_shared<stdin2ser_framer_t>(
            ref(iosvc),
            ref(in),
            ref(sp),
            time_duration(not_a_date_time),
            0);

        ser2stdout->async_start(&error);
        stdin2ser->async_start(&error);

        while(go)
            iosvc.run_one();

        return 0;
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(e.what());
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

}
