#include <mxserialframer/Option.hpp>

#include <boost/make_shared.hpp>
#include <boost/program_options.hpp>
#include <boost/regex.hpp>

#include <config.h>

using boost::asio::serial_port_base;
using boost::logic::tribool;
using boost::logic::indeterminate;
using boost::make_shared;
using boost::program_options::options_description;
using boost::program_options::value;
using boost::program_options::variables_map;
using boost::program_options::command_line_style::default_style;
using boost::program_options::command_line_style::allow_guessing;
using boost::regex;
using boost::regex_match;
using boost::smatch;

using mxserialframer::Option;

using std::cerr;
using std::cout;
using std::endl;
using std::map;
using std::string;

namespace
{

bool
    parse_config(
        const string &str,
        Option::Config &config)
{
    static const regex e(
        "^((?:9600)|(?:19200)|(?:38400)|(?:57600)|(?:115200))-8-(n|e|o)-(1|2)$"
    );

    smatch what;
    if (!regex_match(
            str,
            what,
            e))
    {
        return false;
    }

    const string baud_str = what[1];
    const string parity_str = what[2];
    const string stopbit_str = what[3];

    map<string, serial_port_base::baud_rate> sbaud_map;
    sbaud_map["9600"] = serial_port_base::baud_rate(9600);
    sbaud_map["19200"] = serial_port_base::baud_rate(19200);
    sbaud_map["38400"] = serial_port_base::baud_rate(38400);
    sbaud_map["57600"] = serial_port_base::baud_rate(57600);
    sbaud_map["115200"] = serial_port_base::baud_rate(115200);

    map<string, serial_port_base::parity> sparity_map;
    sparity_map["n"] = serial_port_base::parity(serial_port_base::parity::none);
    sparity_map["e"] = serial_port_base::parity(serial_port_base::parity::even);
    sparity_map["o"] = serial_port_base::parity(serial_port_base::parity::odd);

    map<string, serial_port_base::stop_bits> sstopbit_map;
    sstopbit_map["1"] = serial_port_base::stop_bits(serial_port_base::stop_bits::one);
    sstopbit_map["2"] = serial_port_base::stop_bits(serial_port_base::stop_bits::two);

    config.baud = sbaud_map[baud_str];
    config.char_size = serial_port_base::character_size(8);
    config.parity = sparity_map[parity_str];
    config.stop_bits = sstopbit_map[stopbit_str];

    return true;
}

}

namespace mxserialframer
{

Option::Option()
    : verbose(false)
{
}

tribool
    Option::parse_args(
        int argc,
        const char *argv[])
{
    options_description desc("mxserialframer v" VERSION);
    desc.add_options()
        ("device", value<string>(), "Serial device path.")
        ("config", value<string>(), "Serial configuration. Like: 9600-8-n-1.")
        ("noinput-timeout-ms", value<int>(), "Flush when nothing has been read from serial port for TIMEOUT_MS.")
        ("overflow-size", value<int>(), "Flush when buffer contains SIZE or more byte of data.")
        ("verbose", "Print log messages to stderr.")
        ("help", "Print this manual.")
        ("version", "Print version.");

    variables_map vm;
    try
    {
        store(
            parse_command_line(argc, argv, desc, default_style & (~allow_guessing)),
            vm
        );
    }
    catch(boost::program_options::error &e)
    {
        cerr << desc << endl;
        return false;
    }

    if (vm.count("help"))
    {
        cout << desc << endl;
        return indeterminate;
    }

    if (vm.count("version"))
    {
        cout << "mxserialframer v" << VERSION << endl;
        return indeterminate;
    }

    if (1 != vm.count("device") ||
        1 != vm.count("config"))
    {
        cerr << desc << endl;
        return false;
    }

    this->device = vm["device"].as<string>();

    const string config = vm["config"].as<string>();
    if (!parse_config(config, this->config))
    {
        cerr << "illegal --config: " << config << endl;
        return false;
    }

    if (vm.count("noinput-timeout-ms"))
    {
        const int value = vm["noinput-timeout-ms"].as<int>();
        if (value <= 0 ||
            value >= 60 * 1000)
        {
            cerr << "illegal --noinput-timeout-ms: " << value << endl;
            return false;
        }

        this->noinput_timeout_ms = make_shared<int>(value);
    }

    if (vm.count("overflow-size"))
    {
        const int value = vm["overflow-size"].as<int>();
        if (value <= 0 ||
            value > 1024 * 1024)
        {
            cerr << "illegal --overflow-size: " << value << endl;
            return false;
        }

        this->overflow_size = make_shared<int>(value);
    }

    if (vm.count("verbose"))
    {
        this->verbose = true;
    }

    return true;
}

}
