#include <fstream>
#include <vector>

#include <boost/filesystem.hpp>
#include <boost/exception/all.hpp>
#include <boost/make_shared.hpp>

#include <mxutil/error.hpp>
#include <mxutil/log.hpp>
#include <mxutil/PidFile.hpp>
#include <mxutil/Signal.hpp>
#include <mxutil/log/Log.hpp>
#include <mxutil/log/StdoutLogger.hpp>
#include <mxutil/log/SyslogLogger.hpp>

#include <mxsl/mxsl.hpp>

#include <mxtagcollector/App.hpp>
#include <mxtagcollector/Configuration.hpp>
#include <mxtagcollector/Option.hpp>

using boost::filesystem::path;
using boost::logic::tribool;
using boost::make_shared;
using boost::shared_ptr;

using mxtagcollector::App;
using mxtagcollector::Configuration;
using mxtagcollector::Option;

using mxutil::ArgumentError;
using mxutil::PidFile;
using mxutil::Signal;
using mxutil::log::Log;
using mxutil::log::Logger;
using mxutil::log::StdoutLogger;
using mxutil::log::SyslogLogger;

using std::cerr;
using std::endl;
using std::ifstream;
using std::ios_base;
using std::string;
using std::vector;

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[]);

const Configuration
    get_configuration();

volatile bool stop = false;
void
    on_term(int)
{
    stop = true;
}

}

int
    main(
        int argc,
        const char *argv[])
{
    try
    {
        const Option opt = get_option(argc, argv);

        shared_ptr<Logger> logger(new StdoutLogger());
        logger->set_level(Logger::DBUG);
        if (!opt.foreground)
        {
            const int result = daemon(0, 0);
            if (0 != result)
            {
                cerr << strerror(errno) << endl;
                return 1;
            }

            logger.reset(new SyslogLogger("mxtagcollector"));
            logger->set_level(Logger::WARN);
        }
        Log::get().set_logger(logger);

        const Configuration config = get_configuration();

        shared_ptr<App> app = make_shared<App>(
            opt.mqtt_broker_host,
            opt.mqtt_broker_port,
            opt.dbpath,
            config.sqlite3_busy_retry_timeout_ms,
            config.sqlite3_busy_wait_ms,
            config.sqlite3_commit_period_ms
        );

        Signal::instance().sigterm(&on_term);

        shared_ptr<PidFile> pidfile;
        if (!opt.foreground)
            pidfile = PidFile::get("/var/run/mxtagcollector.pid");

        app->start();

        while(!stop)
        {
            mxsl::validate();

            app->loop();
        }

        return 0;
    }
    catch(boost::exception &e)
    {
        MXLOG_EROR(diagnostic_information(e));
    }
    catch(std::exception &e)
    {
        MXLOG_EROR(string("std::exception: ") + e.what());
    }
    catch(...)
    {
        MXLOG_EROR("unexpected exception caught");
    }

    return 1;
}

namespace
{

const Option
    get_option(
        int argc,
        const char *argv[])
{
    Option opt;
    const tribool parsed = opt.parse_args(argc, argv);
    if (parsed)
        return opt;
    else if (!parsed)
        exit(2);
    else
        exit(0);
}

const string
    readfile(const path &p)
{
    ifstream infile(p.string().c_str(), ios_base::binary);
    if (!infile)
    {
        MXLOG_EROR("open file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    vector<char> buf(file_size(p));

    infile.read(&buf.front(), buf.size());
    if (infile.gcount() != static_cast<int>(buf.size()))
    {
        MXLOG_EROR("read file failed: " + p.string());

        BOOST_THROW_EXCEPTION(ArgumentError());
    }

    return string(buf.begin(), buf.end());
}

const Configuration
    get_configuration()
{
    return Configuration::parse(readfile("etc/mxtagcollector/configuration"));
}

}
