#include <mxtagcollector/App.hpp>

#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/make_shared.hpp>

#include <mxutil/log.hpp>
#include <mxutil/sched/DailyScheduler.hpp>

#include <mxhtag/Database.hpp>

#include <mxtag/mqtt/TagSubscriber.hpp>

#include <mxtagcollector/ReadingInserter.hpp>

using boost::lexical_cast;
using boost::make_shared;
using boost::posix_time::ptime;
using boost::posix_time::milliseconds;
using boost::shared_ptr;

using mxhtag::Database;
using mxhtag::VersionError;

using mxtag::mqtt::TagSubscriber;
using mxtag::Tag;

using mxutil::sched::DailyScheduler;

using std::map;
using std::set;
using std::string;
using std::vector;

namespace mxtagcollector
{

struct App::Impl
{
    Impl(
        const string &broker_host_,
        int broker_port_,
        const string &dbpath_,
        int sqlite3_busy_retry_timeout_ms_,
        int sqlite3_busy_wait_ms_)
        : broker_host(broker_host_)
        , broker_port(broker_port_)
        , dbpath(dbpath_)
        , sqlite3_busy_retry_timeout_ms(sqlite3_busy_retry_timeout_ms_)
        , sqlite3_busy_wait_ms(sqlite3_busy_wait_ms_)
    {
    }

    const string    broker_host;
    const int       broker_port;
    const string    dbpath;
    const int       sqlite3_busy_retry_timeout_ms;
    const int       sqlite3_busy_wait_ms;

    ReadingInserter             reading_inserter;
    shared_ptr<DailyScheduler>  scheduler;

    shared_ptr<TagSubscriber>   tag_subscriber;
    shared_ptr<Database>        database;

    void
        process_message(const Tag &v)
    {
        try
        {
            this->reading_inserter.insert(v);
        }
        catch(boost::exception &e)
        {
            MXLOG_EROR(diagnostic_information(e));
            exit(1);
        }
        catch(std::exception &e)
        {
            MXLOG_EROR(string("std::exception: ") + e.what());
            exit(1);
        }
        catch(...)
        {
            MXLOG_EROR("unexpected exception caught");
            exit(1);
        }
    }

    void
        commit()
    {
        BOOST_ASSERT(this->database);

        this->reading_inserter.commit(*this->database);
    }
};

App::App(
    const string &broker_host,
    int broker_port,
    const string &dbpath,
    int sqlite3_busy_retry_timeout_ms,
    int sqlite3_busy_wait_ms,
    int sqlite3_commit_period_ms)
    : pimpl(new Impl(
        broker_host,
        broker_port,
        dbpath,
        sqlite3_busy_retry_timeout_ms,
        sqlite3_busy_wait_ms))
{
    pimpl->scheduler.reset(new DailyScheduler(
        milliseconds(sqlite3_commit_period_ms),
        bind(
            &Impl::commit,
            pimpl)
    ));
}

App::~App()
{
}

void
    App::start()
{
    map<string, set<string> > eq_tag_list;
    eq_tag_list["+"].insert("+");

    pimpl->tag_subscriber = make_shared<TagSubscriber>(
        pimpl->broker_host,
        pimpl->broker_port,
        "tag-collector",
        false,
        eq_tag_list,
        bind(
            &Impl::process_message,
            pimpl,
            _1
        )
    );

    try
    {
        pimpl->database = make_shared<Database>(
            pimpl->dbpath,
            pimpl->sqlite3_busy_retry_timeout_ms,
            pimpl->sqlite3_busy_wait_ms);
    }
    catch(mxhtag::Error &e)
    {
        if (0 != unlink(pimpl->dbpath.c_str()))
        {
            MXLOG_WARN("unlink() failed, ignored");
        }

        Database::create_tables(pimpl->dbpath);
        pimpl->database = make_shared<Database>(
            pimpl->dbpath,
            pimpl->sqlite3_busy_retry_timeout_ms,
            pimpl->sqlite3_busy_wait_ms);
    }
}

void
    App::loop()
{
    pimpl->tag_subscriber->loop();

    pimpl->scheduler->check_and_fire();
}

}
