#include <mxtagimport/Option.hpp>

#include <iostream>

#include <boost/make_shared.hpp>
#include <boost/program_options.hpp>

#include <mxtag/tag_parse.hpp>

#include <config.h>

using boost::bad_lexical_cast;
using boost::lexical_cast;
using boost::logic::indeterminate;
using boost::logic::tribool;
using boost::make_shared;
using boost::posix_time::microsec_clock;
using boost::posix_time::ptime;
using boost::program_options::options_description;
using boost::program_options::variables_map;
using boost::program_options::command_line_style::default_style;
using boost::program_options::command_line_style::allow_guessing;

namespace po = boost::program_options;

using mxidafpp::Value;

using mxtag::parse_iso8601_time;

using std::cerr;
using std::cout;
using std::endl;
using std::string;
using std::vector;

namespace
{

template <typename FROM, typename TO>
bool
    try_lexical_convert(
        const FROM &f,
        TO &t)
{
    try
    {
        t = lexical_cast<TO>(f);

        return true;
    }
    catch(bad_lexical_cast &)
    {
        return false;
    }
}

}

namespace mxtagimport
{

Option::Option()
    : mqtt_broker_host("127.0.0.1")
    , mqtt_broker_port(1883)
    , at(microsec_clock::universal_time())
{
}

tribool
    Option::parse_args(
        int argc,
        const char *argv[])
{
    options_description desc("Options");
    desc.add_options()
        ("mqtt-broker-host", po::value<string>(),
            "Broker host.\n"
            "Default \"127.0.0.1\".")
        ("mqtt-broker-port", po::value<int>(),
            "Broker port.\n"
            "Default \"1883\".")
        ("equipment", po::value<string>(),
            "Equipment name.\n"
            "Must be given.")
        ("tag", po::value<string>(),
            "Tag name.\n"
            "Must be given.")
        ("type", po::value<string>(),
            "Value type.\n"
            "If not given, mxtagimport would try to guess one.\n"
            "Allowed: int, uint, float, string")
        ("value", po::value<string>(),
            "Value.")
        ("at", po::value<string>(),
            "Timestamp.\n"
            "If not given, mxtagimport uses system time.\n")
        ("unit", po::value<string>(),
            "Unit.\n"
            "Default \"\".")
        ("help", "Print this manual.")
        ("version", "Print version.");

    variables_map vm;
    try
    {
        store(
            parse_command_line(argc, argv, desc, default_style & (~allow_guessing)),
            vm
        );
    }
    catch(boost::program_options::error &e)
    {
        cerr << desc << endl;
        return false;
    }

    if (vm.count("help"))
    {
        cout << desc << endl;
        return indeterminate;
    }

    if (vm.count("version"))
    {
        cout << "mxtagimport v" << VERSION << endl;
        return indeterminate;
    }

    if (vm.count("mqtt-broker-host"))
        this->mqtt_broker_host = vm["mqtt-broker-host"].as<string>();

    if (vm.count("mqtt-broker-port"))
        this->mqtt_broker_port = vm["mqtt-broker-port"].as<int>();

    if (!vm.count("equipment") ||
        !vm.count("tag") ||
        !vm.count("value"))
    {
        cerr << desc << endl;
        return false;
    }

    this->equipment = vm["equipment"].as<string>();
    this->tag = vm["tag"].as<string>();

    const string value_str = vm["value"].as<string>();

    if (vm.count("at"))
    {
        if (!parse_iso8601_time(
                vm["at"].as<string>(),
                this->at))
        {
            cerr << desc << endl;
            return false;
        }
    }

    if (vm.count("unit"))
        this->unit = make_shared<string>(vm["unit"].as<string>());

    if (vm.count("type"))
    {
        const string type = vm["type"].as<string>();
        try
        {
            if (type == "int")
                this->value = make_shared<Value>(lexical_cast<int64_t>(value_str));
            else if (type == "uint")
                this->value = make_shared<Value>(lexical_cast<uint64_t>(value_str));
            else if (type == "float")
                this->value = make_shared<Value>(lexical_cast<double>(value_str));
            else if (type == "string")
                this->value = make_shared<Value>(value_str);
            else
            {
                cerr << desc << endl;
                return false;
            }
        }
        catch(bad_lexical_cast &)
        {
            cerr << desc << endl;
            return false;
        }
    }
    else
    {
        // try to parse value into proper type
        uint64_t u;
        int64_t i;
        double f;
        if (try_lexical_convert(value_str, u))
            this->value = make_shared<Value>(u);
        else if (try_lexical_convert(value_str, i))
            this->value = make_shared<Value>(i);
        else if (try_lexical_convert(value_str, f))
            this->value = make_shared<Value>(f);
        else
            this->value = make_shared<Value>(value_str);
    }

    return true;
}

}
