from __future__ import division
import struct
import logging
import math
import sys
from PyCRC.CRC16DNP import CRC16DNP
from binascii import hexlify

class DNP3HeaderError(Exception):
    pass


class DNP3BlockError(Exception):
    pass


class DNP3(object):
    SYNC = int("0x0564", 0)

    def __init__(self):
        pass

    def parse_header(self, header_data):
        header = struct.unpack('!HBBHHH', header_data[:10])

        # Check 0x05 0x64
        if header[0] != self.SYNC:
            logging.debug("Not header, skip: {}".format(hexlify(bytes(header))))
            raise DNP3HeaderError("Not header, skip")

        # Check header CRC-16-DNP
        crc = header[-1:][0]
        if crc != CRC16DNP().calculate(header_data[:-2]):
            logging.debug("CRC error: {}".format(hexlify(bytes(header))))
            raise DNP3HeaderError("CRC Error")

        """
        The length field is 1 byte in length and specifies the count of user
        bytes in the frame. The CONTROL,DESTINATION and SOURCE field sizes are
        included in this count. The minimum value for this field is 5 and the
        maximum value is 255.
        """
        self.length = header[1]
        # control(1), dest(2), src(@)
        self.body_length = self.length - (1 + 2 + 2)
        blocks_count =  int(math.ceil(self.body_length / 16))

        # Total Length should be
        # SYNC(2) + CONTROL (1) + CRC (2) + CRC(blocks_count * 2)
        self.total_length = self.length + 3 + 2 + blocks_count * 2

        logging.debug("Vaild header, length: {}".format(self.length))
        logging.debug("Data payload length: {}".format(self.body_length))
        logging.debug("Blocks count: {}".format(blocks_count))
        logging.debug("Total packet size: {}".format(self.total_length))

    def parse_payload(self, packet_buffer):
        payload_data = packet_buffer[10:self.total_length]
        while len(payload_data) > 0:
            if len(payload_data) < 18:  # 16 + 2 CRC
                block_data = payload_data
                payload_data = ""
            else:
                block_data = payload_data[:18]
                payload_data = payload_data[18:]

            crc = struct.unpack("!H", block_data[-2:])[0]
            if crc != CRC16DNP().calculate(block_data[:-2]):
                logging.debug("CRC error {}".format(
                    hexlify(bytes(packet_buffer[10:self.total_length]))))
                raise DNP3BlockError("CRC Error")

    def parse_packet(self, packet_buffer):
        if len(packet_buffer) < 10:
            return None

        self.parse_header(packet_buffer[:10])
        if len(packet_buffer) < self.total_length:
            return None

        self.parse_payload(packet_buffer)
        return self


def Framer(inputfd, outfd):
    packet_buffer = ""
    dnp = None
    for indata in inputfd:
        packet_buffer += indata
        try:
            dnp = DNP3().parse_packet(packet_buffer)
            if dnp is None:
                continue
            packet_buffer = packet_buffer[dnp.total_length:]
        except (DNP3HeaderError, DNP3BlockError):
            packet_buffer = packet_buffer[1:]
        finally:
            dnp = None

if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )

    packet = "\x05\x64\x0B\xC4\x04\x00\x03\x00" +\
        "\xE4\x2B\xE5\xC0\x01\x02\x00\x06\x98\x5C" +\
        "\x00\x00\x00\x00\x00" +\
        "\x05\x64\x0A\x44\x03\x00\x04\x00" +\
        "\x7C\xAE\xE0\xC0\x81\x00\x00\xC1\xF0" +\
        "\x05\x64\x11\xc4\x78\xfb\xf4\x01" +\
        "\xbc\x41\xe6\xc9\x01\x3c\x02\x06" +\
        "\x3c\x03\x06\x3c\x04\x06\xf3\x95" +\
        "\x05\x64\x0a\x44\xf4\x01\x78\xfb" +\
        "\x35\xdb\xcc\xc9\x81\x00\x00\x0a" +\
        "\x06" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xcd\xf2\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\xb1\xb5\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xe7\xd2\x00\x5d\xfe" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xce\xf3\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x3d\x0d\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xe8\xd3\x00\xfb\xd4" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xcf\xf4\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x80\x07\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xe9\xd4\x00\x22\x4a" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xd0\xf5\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x8a\x92\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xea\xd5\x00\xa4\xcb" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xd1\xf6\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x2e\x6d\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xeb\xd6\x00\xb7\x62" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xd2\xf7\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\xa2\xd5\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xec\xd7\x00\xd1\x85" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xd3\xf8\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x54\x78\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xed\xd8\x00\x9c\x74" +\
        "\x05\x64\x3f\x44\xf4\x01\x78\xfb" +\
        "\xff\xaa\xd4\xf9\x82\x00\x00\x02" +\
        "\x01\x28\x10\x00\x0e\x00\x81\x0d" +\
        "\x00\x81\x91\xbb\x14\x00\x81\x17" +\
        "\x00\x81\x0c\x00\x81\x09\x00\x81" +\
        "\x0f\x00\x81\x10\xeb\xb5\x00\x81" +\
        "\x0e\x00\x01\x0d\x00\x01\x14\x00" +\
        "\x01\x17\x00\x01\x0c\x00\x36\xbb" +\
        "\x01\x09\x00\x01\x0f\x00\x01\x10" +\
        "\x00\x01\xf4\xcf" +\
        "\x05\x64\x08\xc4\x78\xfb\xf4\x01" +\
        "\x46\x16\xee\xd9\x00\x1a\xf5" +\
        "\x05\x64\x11\xc4\x78\xfb\xf4\x01" +\
        "\xbc\x41\xef\xca\x01\x3c\x02\x06" +\
        "\x3c\x03\x06\x3c\x04\x06\xa8\x8a" +\
        "\x05\x64\x0a\x44\xf4\x01\x78\xfb" +\
        "\x35\xdb\xd5\xca\x81\x00\x00\x96" +\
        "\xd0" +\
        "\x05\x64\x11\xc4\x78\xfb\xf4\x01" +\
        "\xbc\x41\xf0\xcb\x01\x3c\x02\x06" +\
        "\x3c\x03\x06\x3c\x04\x06\x6a\0x95"

    Framer(packet, sys.stderr)
