# Using the C API

For developers who want to integrate the ThingsPro *Configuration API* with their *C* programs, `libcurl` is a convenient library. The `libcurl` library is an easy-to-use free client-side URL transfer library. It is highly portable and can be used to build applications that can work on many different platforms.

## `libcurl`
This library provides easy interfaces to init easy-session, sets all options, and sends a request based on the option selected. After the response is received, if you do not want to reuse the session, you must close the session.

## Using `libcurl`
The `libcurl` library provides the `curl-config` tool that can find the library path and the required flags.

Including `curl.h` into your source code:
```c
#include <curl/curl.h>
```

Compiling your source code:
```
$ gcc -c example.c -o example.o `curl-config --cflags`
```

Linking your code:
```
$ gcc -o example example.o `curl-config --libs`
```

### Initialization and Cleanup

The `libcurl` library should be initialized as follows to get the easy-session handle before you can use the library:
```c
CURL *curl = curl_easy_init();
```

Cleaning up the easy-session's handle after all operations are finished:
```c
/* always cleanup */
curl_easy_cleanup(curl);
```

### Setting Basic Options

Set a resource for `CURLOPT_URL` as follows:
```c
/* Set URL */
curl_easy_setopt(curl, CURLOPT_URL, "https://localhost/api/v1/<resource>");

/* Skip the verification of the server's certificate. */
curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0L);

/* Skip to verify the certificate's name against host. */
curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0L);
```
Note: To access ThingsPro via HTTPS, you must disable `CURLOPT_SSL_VERIFYPEER` and `CURLOPT_SSL_VERIFYHOST`.

### Performing the Request

After the request is performed the status is assigned to `res`, refer to libcurl error codes (https://curl.haxx.se/libcurl/c/libcurl-errors.html) for meanings. If `CURLOPT_WRITEFUNCTION` is not set, the response will be printed to stdout by default.

```c
/* Perform the request, res will get the return code */
res = curl_easy_perform(curl);
```

## Getting Resource(s)

Assign a URL to `CURLOPT_URL` and perform the request.

## Creating a resource

To `POST` data, define the header as follows:

```c
struct curl_slist *headers = NULL;
headers = curl_slist_append(headers, "Content-Type: application/json");
```

Specify the `POST` data using the string format:
```c
curl_easy_setopt(curl, CURLOPT_POSTFIELDS, json_str);
```

## Updating a Resource

To `PUT` data, define the header as follows:
```c
struct curl_slist *headers = NULL;
headers = curl_slist_append(headers, "Content-Type: application/json");
```

Specify the `PUT` data using the string format:
```c
curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "PUT");
curl_easy_setopt(curl, CURLOPT_POSTFIELDS, json_str);
```

Specify the information of the file to include:
```c
#include <fcntl.h>
#include <sys/stat.h>

FILE *fd;
struct stat file_info;

stat(file, &file_info);
fd = fopen(file, "rb");
```

Using a file to `PUT` data:

```c
/* enable uploading */
curl_easy_setopt(curl, CURLOPT_POSTFIELDS, json_str);

/* specify which file to upload */
curl_easy_setopt(curl, CURLOPT_READDATA, fd);

/* provide the size of the upload, we specicially typecast the value
   to curl_off_t since we must be sure to use the correct data size */
curl_easy_setopt(curl, CURLOPT_INFILESIZE_LARGE,
                 (curl_off_t)file_info.st_size);
```

## Deleting a Resource

Using `CURLOPT_CUSTOMREQUEST` to specify a `DELETE` as follows:
```c
curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
```

## Examples

Download the source code in the examples from [libcurl-example.tar.gz](./libcurl-example.tar.gz). These files are intended as examples only and contain all the example sources listed here.

| Code | Description |
|:---- |:----------- |
| [https_get.c](libcurl-example/https_get.c) | Simple HTTPS GET |
| [https_get_write_cb.c](libcurl-example/https_get_write_cb.c) | Shows how the write callback function can be used after GET |
| [https_post.c](libcurl-example/https_post.c) | Simple HTTPS POST |
| [https_put.c](libcurl-example/https_put.c) | Simple HTTPS PUT |
| [https_put_by_file.c](libcurl-example/https_put_by_file.c) | HTTP PUT with easy interface from a file |
| [https_put_read_cb.c](libcurl-example/https_put_read_cb.c) | HTTP PUT with easy interface from a file and read callback |
| [https_delete.c](libcurl-example/https_delete.c) | Simple HTTPS DELETE ||
