#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os

from sanji.core import Sanji
from sanji.core import Route
from sanji.connection.mqtt import Mqtt
from iptables import IPTables
from iptables.nat.portmapping import PortMapping
from iptables.nat.ssnat import Ssnat
from iptables.firewall.wan import Wan


class Index(Sanji):
    _logger = logging.getLogger("sanji.iptables.index")

    def init(self, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(__file__))

        self.iptables = IPTables(name="iptables", path=path_root)
        self.iptables.flush()

        self.portmapping = PortMapping(
            name="nat.portmapping", path=path_root, iptables=self.iptables)
        self.ssnat = Ssnat(
            name="nat.ssnat", path=path_root, iptables=self.iptables,
            iface_key="outInterface")
        self.wan = Wan(
            name="firewall.wan", path=path_root, iptables=self.iptables,
            iface_key="inInterface")

    @Route(methods="get", resource="/iptables")
    def get_rules(self, message, response):
        return response(data=self.iptables.get_all())

    @Route(methods="get", resource="/iptables/:table")
    def get_rules_by_table(self, message, response):
        data = self.iptables.get_rules_by_table(table=message.param["table"])
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="get", resource="/iptables/:table/:id")
    def get_rule(self, message, response):
        data = self.iptables.get_rule_by_id(
            table=message.param["table"], id=int(message.param["id"]))
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="post", resource="/iptables/:table")
    def post_rule(self, message, response):
        data = self.iptables.add_rule_to_table(
            table=message.param["table"],
            newObj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="put", resource="/iptables/:table/:id")
    def put_rule(self, message, response):
        data = self.iptables.update_rule_to_table(
            table=message.param["table"],
            id=int(message.param["id"]),
            newObj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="delete", resource="/iptables/:table/:id")
    def delete_rule(self, message, response):
        self.iptables.remove_rule_from_table(
            table=message.param["table"],
            id=int(message.param["id"]))
        # TODO: response for delete?
        return response(data=None)

    # Port Forwarding (Port Mapping)
    @Route(methods="get", resource="/network/portmapping")
    def portmapping_get(self, message, response):
        return response(data=self.portmapping.getAll())

    @Route(methods="post", resource="/network/portmapping")
    def portmapping_post(self, message, response):
        data = self.portmapping.add(obj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="get", resource="/network/portmapping/:id")
    def portmapping_get_by_id(self, message, response):
        data = self.portmapping.get(id=int(message.param["id"]))
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="put", resource="/network/portmapping/:id")
    def portmapping_put(self, message, response):
        data = self.portmapping.update(
            id=int(message.param["id"]),
            newObj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="delete", resource="/network/portmapping/:id")
    def portmapping_delete(self, message, response):
        self.portmapping.remove(id=int(message.param["id"]))
        # TODO: response for delete?
        return response(data=None)

    # SSNAT
    @Route(methods="get", resource="/network/ssnat")
    def ssnat_get(self, message, response):
        return response(data=self.ssnat.getAll(hide=True))

    @Route(methods="post", resource="/network/ssnat")
    def ssnat_post(self, message, response):
        if "outInterface" not in message.data:
            message.data["outInterface"] = "wan"
        data = self.ssnat.add(obj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="get", resource="/network/ssnat/:id")
    def ssnat_get_by_id(self, message, response):
        data = self.ssnat.get(id=int(message.param["id"]), hide=True)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="put", resource="/network/ssnat/:id")
    def ssnat_put_by_id(self, message, response):
        data = self.ssnat.update(
            id=int(message.param["id"]),
            newObj=message.data)
        if data is None:
            return response(code=404)
        return response(data=data)

    @Route(methods="delete", resource="/network/ssnat/:id")
    def ssnat_delete(self, message, response):
        self.ssnat.remove(id=int(message.param["id"]))
        # TODO: response for delete?
        return response(data=None)

    # Firewall - WAN
    # TODO: search instead fixed id
    # Reference:
    # https://www.digitalocean.com/community/tutorials/how-to-set-up-a-basic-iptables-firewall-on-centos-6
    # http://www.thegeekstuff.com/2011/06/iptables-rules-examples/
    # http://www.cyberciti.biz/tips/linux-iptables-11-how-to-block-or-open-httpweb-service.html
    @Route(methods="get", resource="/network/firewall/wan/https")
    def firewall_wan_get(self, message, response):
        data = self.wan.get(id=1)
        if data is None:
            return response(code=404)
        return response(data={"enable": not data["enable"]})

    @Route(methods="put", resource="/network/firewall/wan/https")
    def firewall_wan_put(self, message, response):
        message.data["enable"] = not message.data["enable"]
        data = self.wan.update(id=1, newObj=message.data)
        if data is None:
            return response(code=404)
        return response(data={"enable": not data["enable"]})

    # WAN event
    @Route(methods="put", resource="/network/wan")
    def wan_event(self, message):
        self.ssnat.update_wan(wan=message.data["interface"])
        self.wan.update_wan(wan=message.data["interface"])


if __name__ == "__main__":
    FORMAT = '%(asctime)s - %(levelname)s - %(lineno)s - %(message)s'
    logging.basicConfig(level=0, format=FORMAT)
    logging.getLogger("sh").setLevel(logging.WARN)
    index = Index(connection=Mqtt())
    index.start()

    '''
    path_root = os.path.abspath(os.path.dirname(__file__))
    iptables = IPTables(name="iptables", path=path_root)

    print iptables.get_all()
    print iptables.get_rules_by_table(table="filter")
    print iptables.get_rules_by_table(table="nat")
    print iptables.get_rules_by_table(table="mangle")

    print iptables.remove_rule_from_table(table="filter", id=1)

    newObj = {
        "chain": "INPUT",
        "target": "REJECT",
        "protocol": "tcp",
        "match": "tcp",
        "in-interface": "eth0",
        "destination-port": "443"
    }
    print iptables.add_rule_to_table(table="filter", newObj=newObj)

    newObj = {
        "chain": "INPUT",
        "target": "ACCEPT",
        "protocol": "tcp",
        "match": "tcp",
        "in-interface": "eth0",
        "destination-port": "443"
    }
    print iptables.update_rule_to_table(id=1, table="filter", newObj=newObj)
    '''
