import logging
import sh
from threading import Event

from voluptuous import Schema, Optional, Required, In, Any, All, Length, \
    Extra, Invalid, REMOVE_EXTRA

import iptable

_logger = logging.getLogger("sanji.iptables")


def Ports(delim=":"):
    """Port range verification function.
    The format should be: <start port>[delim<end port>]

    Args:
        delim: delimiter for port range
    """
    def f(v):
        msg = "invalid port format; should be integer between 1 and 65535"
        ports = str(v).split(delim)
        if len(ports) > 2:
            raise Invalid(msg)
        for entry in ports:
            if entry.isdigit() is False:
                raise Invalid(msg)
            port = int(entry)
            if port < 1 or port > 65535:
                raise Invalid(msg)
        return str(v)
    return f

FILTER_SCHEMA = Schema({
    "id": int,
    Required("chain"): In(frozenset(["INPUT", "OUTPUT", "FORWARD"])),
    Required("target"): In(frozenset(["ACCEPT", "DROP", "REJECT"])),
    Optional("protocol"): All(Any(unicode, str), Length(1, 255)),
    Optional("match"): All(Any(unicode, str), Length(1, 255)),
    Optional("in-interface"): All(Any(unicode, str), Length(1, 255)),
    Optional("destination-port"): All(Ports()),
    Extra: object
}, extra=REMOVE_EXTRA)

NAT_SCHEMA = Schema({
    "id": int,
    Required("chain"): In(frozenset(["PREROUTING", "OUTPUT", "POSTROUTING"])),
    Required("target"): In(frozenset(["DNAT", "REDIRECT", "SNAT",
                                     "MASQUERADE"])),
    Optional("in-interface"): All(Any(unicode, str), Length(1, 255)),
    Optional("out-interface"): All(Any(unicode, str), Length(1, 255)),
    Optional("destination"): All(Any(unicode, str), Length(1, 255)),
    Optional("destination-port"): All(Ports()),
    Optional("to-destination"): All(Any(unicode, str), Length(1, 255)),
    Optional("to-ports"): All(Ports("-")),
    Optional("source"): All(Any(unicode, str), Length(1, 255)),
    Optional("to-source"): All(Any(unicode, str), Length(1, 255)),
    Extra: object
}, extra=REMOVE_EXTRA)

MANGLE_SCHEMA = Schema({
    "id": int,
    Required("chain"): In(frozenset(["PREROUTING", "OUTPUT", "INPUT",
                                     "FORWARD", "POSTROUTING"])),
    Required("target"): In(frozenset(["DSCP", "ECN", "IPMARK", "MARK",
                                      "IPV4OPTSSTRIP", "TCPMSS", "TOS",
                                      "TTL"])),
    Optional("protocol"): All(Any(unicode, str), Length(1, 255)),
    Optional("options"): All(Any(unicode, str), Length(1, 255)),
    Extra: object
}, extra=REMOVE_EXTRA)


class IPTablesBatch(object):

    def __init__(self, tables):
        self.enable = Event()
        self.tables = tables

    def __enter__(self):
        if self.enable.is_set():
            raise RuntimeError("Already in batch mode")
        self.enable.set()
        for table in self.tables:
            try:
                self.tables[table].batch().__enter__()
            except:
                pass

    def __exit__(self, type, value, traceback):
        self.enable.clear()
        for table in self.tables:
            self.tables[table].batch().__exit__(None, None, None)


class IPTablesError(Exception):
    pass


class IPTables(object):

    iptables_path = "/etc/iptables/up.rules"

    def __init__(self, *args, **kwargs):
        support_tables = [
            {"name": "filter", "schema": FILTER_SCHEMA},
            {"name": "nat", "schema": NAT_SCHEMA},
            {"name": "mangle", "schema": MANGLE_SCHEMA}
        ]

        # initialize iptables
        self.tables = {}
        for table in support_tables:
            self.tables[table["name"]] = \
                iptable.IPTable(name=table["name"], path=kwargs["path"],
                                schema=table["schema"])
        self._batch = IPTablesBatch(self.tables)

    def batch(self):
        return self._batch

    def _generate_rule_options(self, rule):
        option_mapping = [
            {"key": "in-interface", "option": "-i"},
            {"key": "protocol", "option": "-p"},
            {"key": "match", "option": "-m"},
            {"key": "source", "option": "-s"},
            {"key": "destination", "option": "--destination"},
            {"key": "destination-port", "option": "--dport"},
            {"key": "out-interface", "option": "-o"},
            {"key": "target", "option": "-j"},
            {"key": "to-destination", "option": "--to-destination"},
            {"key": "to-ports", "option": "--to-ports"},
            {"key": "options", "option": ""}
        ]
        rule_opts = []
        rule_opts.append("-A")
        rule_opts.append(rule["chain"])
        for opt in option_mapping:
            if opt["key"] in rule:
                rule_opts.append(opt["option"])
                rule_opts.append(rule[opt["key"]])
        return rule_opts

    def _generate_config_by_table(self, table, fp):
        """Generate configuration contents by table

            Args:
                table: one of the table in iptables
                fp: file pointer
        """
        fp.write("*%s\n" % table.name)
        # for rule in table.getAll():
        for rule in sorted(
                table.getAll(),
                key=lambda x: x["priority"] if "priority" in x else 1):
            fp.write(" ".join(self._generate_rule_options(rule)))
            fp.write("\n")
        fp.write("\n")
        fp.write("COMMIT\n")

    def _generate_config(self):
        """Generate configurations for `iptables` to `/etc/iptables/up.rules`
        """
        with open(self.iptables_path, "w") as fp:
            for table in self.tables:
                self._generate_config_by_table(self.tables[table], fp)
                fp.write("\n")

    def flush(self):
        """Flush all tables. """
        with self.batch():
            for table in self.tables:
                self.tables[table].flush()

    def apply(self):
        """Update iptables' rules and restore it
            $ sudo iptables-save > /etc/iptables/iptables.rules
            $ sudo iptables-restore < /etc/iptables/iptables.rules
        """
        try:
            self._generate_config()
            with open(self.iptables_path) as fp:
                sh.iptables_restore(_in=fp)
        except Exception as e:
            _logger.error(str(e))
            raise e

    def get_all(self):
        """Get all iptables' rules, format will be
            {
                "filter": [],
                "nat": [],
                ...
            }
        """
        rules = {}
        for key in self.tables:
            rules[key] = self.tables[key].getAll()
        return rules

    def get_rules_by_table(self, table=None):
        """Get by table

            Args:
                table: name of the iptable
        """
        if table is None or table not in self.tables:
            return None

        return self.tables[table].getAll()

    def get_rule_by_id(self, table=None, id=1):
        """Get by table

            Args:
                table: name of the iptable
                id: rule identifier
        """
        if table is None or table not in self.tables:
            return None

        return self.tables[table].get(id=id)

    def update_rule_to_table(
            self, table=None, id=0, newObj=None, activate=True):
        """Update an exist rule.

            Args:
                table: name of the iptable
                id: rule identifier
                newObj: rule object to be updated
           Raises:
               TODO
        """
        if table is None or table not in self.tables:
            return None

        updated = self.tables[table].set(id=id, newObj=newObj)
        if updated is None:
            return None

        try:
            if activate is True:
                self.apply()
        except Exception as e:
            self.tables[table].remove(id=id)
            raise e
        return updated

    def add_rule_to_table(self, table=None, newObj=None, activate=True):
        """Add a rule.

            Args:
                table (str): table name
                newObj: rule object to be added
            Raises:
                TODO
        """
        if table is None or table not in self.tables:
            return None

        new = self.tables[table].add(obj=newObj)
        if new is None:
            return None

        try:
            if activate is True:
                self.apply()
        except Exception as e:
            self.tables[table].remove(id=new["id"])
            raise e
        return new

    def remove_rule_from_table(self, table=None, id=1, activate=True):
        """Get by table

            Args:
                table: name of the iptable
                id: rule identifier
        """
        if table is None or table not in self.tables:
            return None

        len = self.tables[table].remove(id=id)
        if activate is True:
            self.apply()
        return len
