import logging
import sys
import os
import copy
from voluptuous import Schema, Required, Optional, Any, All, Length, Extra, \
    In, Range, REMOVE_EXTRA

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + '/../../')
    from iptables.mapping import Mapping
except ImportError as e:
    print(e)
    print("Please check the python PATH for import test module. (%s)"
          % __file__)
    exit(1)

_logger = logging.getLogger("sanji.iptables.firewall.wan")

"""
`iptableMapping` for mapping to iptables rules
"iptablesMapping": {
    "filter": [],
    "nat": [],
    "mangle": []
}
"""
# service should be unique, also a key for instead of id
WAN_SCHEMA = Schema({
    "id": int,
    Optional("service"): All(Any(unicode, str), Length(1, 255)),
    Required("enable"): bool,
    Required("behavior"): In(frozenset(["accept", "drop", "reject"])),
    Required("protocol"): In(frozenset(["tcp", "udp", "both"])),
    Required("inInterface"): All(Any(unicode, str), Length(1, 255)),
    Required("portStart"): All(Range(1, 65535)),
    Optional("portEnd"): All(Range(1, 65535)),
    Extra: object
}, extra=REMOVE_EXTRA)


class WanError(Exception):
    pass


class Wan(Mapping):
    """Firewall configuration for wan interface.
    """

    def __init__(self, *args, **kwargs):
        """Initialize mapping object

            Args:
            name (str): object name, will be the configuration prefix
            schema (Schema obj): for schema veridation
            iptables (obj): iptables object
            converter (method): convert settings to iptables' rules
        """
        kwargs["converter"] = self._convert_to_iptables_rules
        kwargs["schema"] = WAN_SCHEMA
        super(Wan, self).__init__(*args, **kwargs)

# Example commands:
# iptables -t filter -A INPUT -p tcp -m tcp -i eth0 --dport 443 -j REJECT

    def __set_rule_protocol(self, rule, protocol):
            rule["protocol"] = protocol
            rule["match"] = protocol
            return rule

    def _convert_to_iptables_rules(self, obj):
        """Convert to iptables rules
        rules = {
            "filter": [],
            "nat": [],
            "mangle": []
        }

            Args:
                obj: setting for convert to iptables' rules
            Returns:
                rules (dict): rules with iptables
        """
        filter_rule = {
            "chain": "INPUT",
            "target": "%s" % obj["behavior"].upper(),
            "destination-port": "%d:%d" % (obj["portStart"], obj["portEnd"])
        }
        filter_rule["in-interface"] = \
            self.wan if obj["inInterface"] == "wan" else obj["inInterface"]

        rules = {}
        rules["filter"] = []
        if "protocol" in obj:
            rules["filter"] += \
                self._generate_by_protocol(filter_rule, obj["protocol"])
        else:
            rules["filter"].append(filter_rule)
        return rules

    def update(self, id, newObj):
        """Update rule; use current value if optional attribes not given.
        TODO: validation
        """
        rule = copy.deepcopy(self.get(id=id))
        if rule is None:
            return None

        rule.update(newObj)
        return super(Wan, self).update(id=id, newObj=rule)
