import logging
import copy
from sanji.model import Model
from threading import Lock

_logger = logging.getLogger("sanji.iptables.mapping")

"""
`iptableMapping` for mapping to iptables rules
"iptablesMapping": {
    "filter": [],
    "nat": [],
    "mangle": []
}
"""


class MappingError(Exception):
    pass


class Mapping(Model):

    def __init__(self, *args, **kwargs):
        """Initialize mapping object

            Args:
            name (str): object name, will be the configuration prefix
            schema (Schema obj): for schema veridation
            iptables (obj): iptables object
            converter (method): convert settings to iptables' rules
            iface_key (string): interface key of the rule
        """
        self._wan_lock = Lock()

        # FIXME: default WAN
        self.wan = "eth0"

        self.iptables = kwargs.pop("iptables")
        self.converter = kwargs.pop("converter", None)
        self.iface_key = kwargs.pop("iface_key", None)

        super(Mapping, self).__init__(*args, **kwargs)
        self.name = kwargs["name"]

        self.restore()

    def _generate_port_string(self, obj, delim=":"):
        """Generate port string by delimeter.

            Args:
                obj: object with "portStart" and/or "portEnd"

            Returns:
                string: port string
        """
        if "portStart" not in obj:
            return None

        ports = []
        ports.append(int(obj["portStart"]))
        if "portEnd" in obj:
            port = int(obj["portEnd"])
            if port > ports[0]:
                ports.append(port)
        return delim.join(str(port) for port in ports)

    def _generate_by_protocol(self, rule, protocol):
        """Generate rule(s) by the given protocol. If protocol is "both",
        2 rules will be generated.

            Args:
                rule: rule object
                protocol: tcp, udp, or both

            Returns:
                array: rule array with protocol
        """
        rules = []
        if protocol == "both":
            rule["protocol"] = "tcp"
            rules.append(copy.deepcopy(rule))
            rule["protocol"] = "udp"
            rules.append(copy.deepcopy(rule))
        else:
            rule["protocol"] = protocol
            rules.append(copy.deepcopy(rule))
        return rules

    def _convert_to_iptables_rules(self, obj):
        """Convert to iptables rules
        rules = {
            "filter": [],
            "nat": [],
            "mangle": []
        }

            Args:
                obj: setting for convert to iptables' rules
            Returns:
                rules (dict): rules with iptables
        """
        if self.converter is not None:
            return self.converter(obj)
        return {}

    def _add_rules_to_iptables(self, obj, rules):
        """Add rules into iptables and save the mapping id

            Args:
                obj: setting
                rules: rules to be added
        """
        if "iptablesMapping" not in obj:
            obj["iptablesMapping"] = {}

        with self.iptables.batch():
            for table in rules:
                for rule in rules[table]:
                    added = self.iptables.add_rule_to_table(
                        table, newObj=rule, activate=False)
                    if table not in obj["iptablesMapping"]:
                        obj["iptablesMapping"][table] = []
                    obj["iptablesMapping"][table].append(added["id"])

    def _delete_rules_from_iptables(self, obj):
        """Delete rules from iptables

            Args:
                obj: settings
            Returns:
                int: deleted rule number
        """
        if "iptablesMapping" not in obj:
            return

        entry = 0
        with self.iptables.batch():
            for table in obj["iptablesMapping"]:
                for rule_id in obj["iptablesMapping"][table]:
                    self.iptables.remove_rule_from_table(
                        table, id=rule_id, activate=False)
                    entry += 1
        obj["iptablesMapping"].clear()
        return entry

    def _should_hide(self, entry):
        """Check if the entry has "hide" attribute or not.

            Args:
                entry (dict): entry to be checked

            Returns:
                boolean: True if entry should be hided, else False
        """
        return True if "hide" in entry and entry["hide"] is True else False

    def getAll(self, hide=False):
        data = super(Mapping, self).getAll()

        if hide is False:
            return data

        return [entry for entry in data if self._should_hide(entry) is False]

    def get(self, id, hide=False):
        entry = super(Mapping, self).get(id=id)

        if hide is False:
            return entry
        return entry if self._should_hide(entry) is False else None

    def add(self, obj=None):
        """Update an exist rule.

            Args:
                obj: settings
            Returns:
                obj (object): new object with id
            Raises:
                TODO
        """
        # add into iptables and save the id
        if obj["enable"] is True:
            self._add_rules_to_iptables(
                obj, self._convert_to_iptables_rules(obj))
            try:
                self.iptables.apply()
            except Exception:
                self._delete_rules_from_iptables(obj)
                return None

        return super(Mapping, self).add(obj=obj)

    def update(self, id, newObj=None):
        """Update an exist rule.

            Args:
                id: rule id to be updated
                newObj: setting object
            Returns:
                obj (object): updated object
            Raises:
                TODO
        """
        rule = self.get(id=id)
        if rule is None:
            return None

        if "enable" not in newObj:
            newObj["enable"] = rule["enable"]

        # delete original rules and add new rules into iptables
        rc = self._delete_rules_from_iptables(rule)
        if "iptablesMapping" in newObj:
            newObj["iptablesMapping"].clear()

        if newObj["enable"] is True:
            self._add_rules_to_iptables(
                newObj, self._convert_to_iptables_rules(newObj))
        if rc > 0 or newObj["enable"] is True:
            try:
                self.iptables.apply()
            except Exception:
                self._delete_rules_from_iptables(newObj)
                return None

        return super(Mapping, self).set(id=id, newObj=newObj)

    def remove(self, id):
        """Remove an exist rule.

            Args:
                id: rule id to be deleted
            Returns:
                len (int): exist rule number
        """
        rule = self.get(id=id)
        if rule is None:
            return None

        rc = self._delete_rules_from_iptables(rule)
        if rc > 0:
            self.iptables.apply()
        return super(Mapping, self).remove(id=id)

    def restore(self):
        """Restore all settings.

            Args:
            Returen:
        """
        for entry in self.getAll():
            if entry["enable"] is True:
                if "iptablesMapping" in entry:
                    entry["iptablesMapping"].clear()
                self._add_rules_to_iptables(
                    entry, self._convert_to_iptables_rules(entry))
            super(Mapping, self).set(id=entry["id"], newObj=entry)
        self.iptables.apply()

    def _update_wan(self, wan):
        """Update the interface of rules if changed

            Args:
            Returen:
        """
        self.wan = wan
        if self.iface_key is None:
            return

        for entry in self.getAll():
            if entry["enable"] is False \
                    or self.iface_key not in entry \
                    or entry[self.iface_key] != "wan":
                continue

            self._delete_rules_from_iptables(entry)
            self._add_rules_to_iptables(
                entry, self._convert_to_iptables_rules(entry))
            super(Mapping, self).set(id=entry["id"], newObj=entry)
        self.iptables.apply()

    def update_wan(self, wan):
        with self._wan_lock:
            self._update_wan(wan)
