import logging
import sys
import os
from voluptuous import Schema, Optional, Required, In, Any, All, Length, \
    Range, Extra, REMOVE_EXTRA

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + '/../../')
    from iptables.mapping import Mapping
except ImportError as e:
    print(e)
    print("Please check the python PATH for import test module. (%s)"
          % __file__)
    exit(1)

_logger = logging.getLogger("sanji.iptables.nat.portmapping")

"""
`iptableMapping` for mapping to iptables rules
"iptablesMapping": {
    "filter": [],
    "nat": [],
    "mangle": []
}
"""
PORTMAPPING_SCHEMA = Schema({
    "id": int,
    Required("enable"): bool,
    Optional("service"): All(Any(unicode, str), Length(1, 255)),
    Required("portStart"): All(Range(1, 65535)),
    Optional("portEnd"): All(Range(1, 65535)),
    Required("destIp"): All(Any(unicode, str), Length(1, 255)),
    Required("protocol"): In(frozenset(["tcp", "udp", "both"])),
    Extra: object
}, extra=REMOVE_EXTRA)


class PortMappingError(Exception):
    pass


class PortMapping(Mapping):

    def __init__(self, *args, **kwargs):
        """Initialize mapping object

            Args:
            name (str): object name, will be the configuration prefix
            schema (Schema obj): for schema veridation
            iptables (obj): iptables object
            converter (method): convert settings to iptables' rules
        """
        kwargs["converter"] = self._convert_to_iptables_rules
        kwargs["schema"] = PORTMAPPING_SCHEMA
        super(PortMapping, self).__init__(*args, **kwargs)

# Example commands:
# iptables -t nat -A PREROUTING -p tcp -i ppp0 --dport 8001 -j DNAT \
#        --to-destination 192.168.1.200:8080
# iptables -A FORWARD -p tcp -d 192.168.1.200 --dport 8080 -m state \
#        --state NEW,ESTABLISHED,RELATED -j ACCEPT

# iptables -t nat -A PREROUTING -i eth0 -p udp --dport 1234 -j DNAT \
#        --to B_SERVER_IP:1234
# iptables -t filter -A FORWARD -p udp -d B_SERVER_IP --dport 1234 -j ACCEPT

    def __set_rule_protocol(self, rule, protocol):
            rule["protocol"] = protocol
            return rule

    def _convert_to_iptables_rules(self, obj):
        """Convert to iptables rules

            Args:
                obj: portmapping setting
        """
        if "iptablesMapping" not in obj:
            obj["iptablesMapping"] = {}

        filter_rule = {
            "chain": "FORWARD",
            "target": "ACCEPT",
            "destination": "%s" % obj["destIp"]
        }
        nat_rule = {
            "chain": "PREROUTING",
            "target": "DNAT",
            "to-destination": obj["destIp"]
        }
        if "portStart" in obj:
            filter_rule["destination-port"] = self._generate_port_string(obj)
            nat_rule["destination-port"] = self._generate_port_string(obj)
            nat_rule["to-destination"] += \
                ":%s" % self._generate_port_string(obj, "-")

        rules = {}
        rules["filter"] = []
        rules["nat"] = []
        if "protocol" in obj:
            rules["nat"] += \
                self._generate_by_protocol(nat_rule, obj["protocol"])
            rules["filter"] += \
                self._generate_by_protocol(filter_rule, obj["protocol"])
        else:
            rules["nat"].append(nat_rule)
            rules["filter"].append(filter_rule)
        return rules

    def add(self, obj=None):
        """Update an exist rule.

            Args:
                obj: portmapping setting
            Returns:
                obj (object): new object with id
            Raises:
                TODO
        """
        if "portEnd" not in obj:
            obj["portEnd"] = obj["portStart"]
        return super(PortMapping, self).add(obj=obj)

    def update(self, id, newObj=None):
        """Update an exist rule.

            Args:
                id: rule id to be updated
                newObj: portmapping setting object
            Returns:
                obj (object): updated object
            Raises:
                TODO
        """
        if "portEnd" not in newObj:
            newObj["portEnd"] = newObj["portStart"]
        return super(PortMapping, self).update(id=id, newObj=newObj)
