import logging
import sys
import os
from voluptuous import Schema, Required, Optional, In, Any, All, Length, \
    Range, Extra, REMOVE_EXTRA

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + '/../../')
    from iptables.mapping import Mapping
except ImportError as e:
    print(e)
    print("Please check the python PATH for import test module. (%s)"
          % __file__)
    exit(1)

_logger = logging.getLogger("sanji.iptables.nat.ssnat")

"""
`iptableMapping` for mapping to iptables rules
"iptablesMapping": {
    "filter": [],
    "nat": [],
    "mangle": []
}
"""
SSNAT_SCHEMA = Schema({
    "id": int,
    Required("enable"): bool,
    Optional("protocol"): In(frozenset(["tcp", "udp", "both"])),
    Optional("intIp"): All(Any(unicode, str), Length(1, 255)),
    Optional("portStart"): All(Range(1, 65535)),
    Optional("portEnd"): All(Range(1, 65535)),
    Required("outInterface"): All(Any(unicode, str), Length(1, 255)),
    Extra: object
}, extra=REMOVE_EXTRA)


class SsnatError(Exception):
    pass


class Ssnat(Mapping):
    """Simple SNAP, only support MASQUERADE.
    """

    def __init__(self, *args, **kwargs):
        """Initialize mapping object

            Args:
            name (str): object name, will be the configuration prefix
            schema (Schema obj): for schema veridation
            iptables (obj): iptables object
            converter (method): convert settings to iptables' rules
        """
        kwargs["converter"] = self._convert_to_iptables_rules
        kwargs["schema"] = SSNAT_SCHEMA
        super(Ssnat, self).__init__(*args, **kwargs)

# Example commands:
# iptables -t nat
#  -A POSTROUTING -o wwan0 -j MASQUERADE
#  -A POSTROUTING -p tcp -s 192.168.93.10 -o wwan0 -j MASQUERADE \
#         --to-ports 12345
#  -A POSTROUTING -p tcp -s 192.168.93.10 --dport 12345 -o wwan0 \
#         -j MASQUERADE --to-ports 12345
#  -A POSTROUTING -o wwan0 -j SNAT --to $WWANIP
#  -A POSTROUTING -p tcp -s 192.168.93.10 --dport 12345 -o wwan0 -j SNAT \
#         --to-source 192.168.31.31:12345

    def _convert_to_iptables_rules(self, obj):
        """Convert to iptables rules
        rules = {
            "filter": [],
            "nat": [],
            "mangle": []
        }

            Args:
                obj: setting for convert to iptables' rules
            Returns:
                rules (dict): rules with iptables
        """
        nat_rule = {
            "priority": 0,
            "chain": "POSTROUTING",
            "target": "MASQUERADE"
        }
        nat_rule["out-interface"] = \
            self.wan if obj["outInterface"] == "wan" else obj["outInterface"]

        if "priority" in obj:
            nat_rule["priority"] = obj["priority"]
        if "intIp" in obj:
            nat_rule["source"] = obj["intIp"]
        if "portStart" in obj:
            nat_rule["destination-port"] = self._generate_port_string(obj)
            nat_rule["to-ports"] = self._generate_port_string(obj, "-")

        rules = {}
        rules["nat"] = []
        if "protocol" in obj:
            rules["nat"] += \
                self._generate_by_protocol(nat_rule, obj["protocol"])
        else:
            rules["nat"].append(nat_rule)
        return rules
