#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
import sys
import unittest
import glob
from mock import Mock
from mock import patch
from sanji.connection.mockup import Mockup
from sanji.message import Message

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../")
    from iptables import IPTables
    from index import Index
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    print (e)
    exit(1)


class TestIndexClass(unittest.TestCase):

    @patch.object(IPTables, "apply")
    def setUp(self, mock_apply):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../")
        self.jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        self.backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in self.jsons + self.backups:
            os.unlink(file)

        self.index = Index(connection=Mockup())
        self.index.iptables.iptables_path = \
            "%s/tests/iptables.rules" % self.root_path

    def tearDown(self):
        files = glob.glob(os.path.join(self.root_path, "data/*.json")) + \
            glob.glob(os.path.join(self.root_path, "data/*.backup")) + \
            glob.glob(os.path.join(self.index.iptables.iptables_path))
        for prevConfig in files:
            try:
                os.unlink(prevConfig)
            except:
                pass

        self.index.stop()
        self.index = None

    def test__get_rules(self):
        """test__get_rules: Get all rules"""
        result = self.index.iptables.get_all()
        resp = Mock()
        self.index.get_rules(message=None, response=resp, test=True)
        resp.assert_called_once_with(data=result)

    def test__get_rules_by_table(self):
        """test__get_rules_by_table: Get filter rules"""
        resp = Mock()
        message = Message({
            "param": {
                "table": "filter"
            }
        })
        self.index.get_rules_by_table(
            message=message, response=resp, test=True)
        resp.assert_called_once_with(
            data=self.index.iptables.get_rules_by_table(
                message.param["table"]))

    @patch("iptables.sh")
    def test__get_rule(self, mock_sh):
        """test__get_rule: Get a filter rule"""
        resp = Mock()
        message = Message({
            "param": {
                "table": "filter",
                "id": 1
            }
        })

        self.index.iptables.add_rule_to_table(
            message.param["table"],
            newObj={
                "chain": "INPUT",
                "target": "REJECT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            })
        self.index.get_rule(message=message, response=resp, test=True)
        resp.assert_called_once_with(
            data=self.index.iptables.get_rule_by_id(
                message.param["table"],
                message.param["id"]))

    @patch("iptables.sh")
    def test__post_rule(self, mock_sh):
        """test__post_rule: Add a filter rule"""
        resp = Mock()
        message = Message({
            "param": {
                "table": "filter"
            },
            "data": {
                "chain": "INPUT",
                "target": "REJECT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            }
        })

        self.index.post_rule(message=message, response=resp, test=True)
        resp.assert_called_once_with(
            data=self.index.iptables.get_rule_by_id(
                table=message.param["table"], id=1))

    @patch("iptables.sh")
    def test__put_rule(self, mock_sh):
        """test__put_rule: Update a filter rule"""
        resp = Mock()
        message = Message({
            "param": {
                "table": "filter",
                "id": 1
            },
            "data": {
                "id": 1,
                "chain": "INPUT",
                "target": "ACCEPT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            }
        })

        self.index.iptables.add_rule_to_table(
            message.param["table"],
            newObj={
                "chain": "INPUT",
                "target": "REJECT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            })
        self.index.put_rule(message=message, response=resp, test=True)
        resp.assert_called_once_with(
            data=self.index.iptables.get_rule_by_id(
                table=message.param["table"], id=1))

    @patch("iptables.sh")
    def test__delete_rule(self, mock_sh):
        """test__delete_rule: Delete a filter rule"""
        resp = Mock()
        message = Message({
            "param": {
                "table": "filter",
                "id": 1
            }
        })

        self.index.delete_rule(message=message, response=resp, test=True)
        # TODO: should be code=200?
        # resp.assert_called_once_with(code=200)
        resp.assert_called_once_with(data=None)

    # Port Mapping
    def test__portmapping_get(self):
        """test__portmapping_get: Get all portmapping rules"""
        result = self.index.portmapping.getAll()
        resp = Mock()
        self.index.portmapping_get(message=None, response=resp, test=True)
        resp.assert_called_once_with(data=result)

    @patch("iptables.sh")
    def test__portmapping_post(self, mock_sh):
        """test__portmapping_post: Add a portmapping rule"""
        resp = Mock()
        message = Message({
            "data": {
                "enable": True,
                "service": "test",
                "portStart": 80,
                "destIp": "192.168.6.127",
                "protocol": "tcp"
            }
        })

        self.index.portmapping_post(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.portmapping.get(id=1))
        # TODO: check if iptables rule added

    def test__portmapping_get_by_id(self):
        """test__portmapping_get_by_id: Get a portmapping rule by id"""
        resp = Mock()
        message = Message({
            "param": {
                "id": 1
            }
        })

        self.index.portmapping_get_by_id(
            message=message, response=resp, test=True)
        resp.assert_called_once_with(code=404)
# TODO       data=self.index.portmapping.get(id=int(message.param["id"])))

    @patch("iptables.sh")
    def test__portmapping_put(self, mock_sh):
        """test__portmapping_put: Add a portmapping rule"""
        resp = Mock()
        message = Message({
            "param": {
                "id": 1
            },
            "data": {
                "enable": True,
                "service": "test",
                "portStart": 80,
                "destIp": "192.168.6.127",
                "protocol": "tcp"
            }
        })
        self.index.portmapping.add(message.data)
        message.data["service"] = "test-put"
        message.data["portEnd"] = 100

        self.index.portmapping_put(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.portmapping.get(id=1))
        # TODO: check if iptables rule updated

    @patch("iptables.sh")
    def test__portmapping_delete(self, mock_sh):
        """test__portmapping_delete: Delete a portmapping rule"""
        data = {
            "enable": True,
            "service": "test",
            "portStart": 80,
            "destIp": "192.168.6.127",
            "protocol": "tcp"
        }
        data = self.index.portmapping.add(data)

        resp = Mock()
        message = Message({
            "param": {
                "id": data["id"]
            }
        })

        self.index.portmapping_delete(
            message=message, response=resp, test=True)
        # TODO: should be code=200?
        # resp.assert_called_once_with(code=200)
        resp.assert_called_once_with(data=None)

    # SSNAT
    @patch("iptables.sh")
    def test__ssnat_get(self, mock_sh):
        """test__ssnat_get: Get ssnat rules"""
        data = {
            "enable": True,
            "outInterface": "eth0"
        }
        data = self.index.ssnat.add(obj=data)

        resp = Mock()
        self.index.ssnat_get(message=None, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.ssnat.getAll(hide=True))

    @patch("iptables.sh")
    def test__ssnat_post(self, mock_sh):
        """test__ssnat_post: Add a ssnat rule"""
        resp = Mock()
        message = Message({
            "data": {
                "enable": True,
                "outInterface": "eth0",
                "priority": 10,
                "intIp": "192.168.7.123",
                "portStart": 12345
            }
        })

        self.index.ssnat_post(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.ssnat.get(id=2))
        # TODO: check if iptables rule added

    @patch("iptables.sh")
    def test__ssnat_get_by_id(self, mock_sh):
        """test__ssnat_get_by_id: Get ssnat rule by id"""
        data = {
            "enable": True,
            "outInterface": "eth0"
        }
        data = self.index.ssnat.add(obj=data)

        message = Message({
            "param": {
                "id": data["id"]
            }
        })

        resp = Mock()
        self.index.ssnat_get_by_id(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.ssnat.get(id=data["id"]))

    @patch("iptables.sh")
    def test__ssnat_put_by_id(self, mock_sh):
        """test__ssnat_put: update ssnat"""
        resp = Mock()
        message = Message({
            "param": {
                "id": 1
            },
            "data": {
                "id": 1,
                "enable": True,
                "outInterface": "eth0"
            }
        })
        self.index.ssnat.add(obj=message.data)

        message.data["outInterface"] = "wwan0"

        self.index.ssnat_put_by_id(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.index.ssnat.get(id=1))

    @patch("iptables.sh")
    def test__ssnat_delete(self, mock_sh):
        """test__ssnat_delete: Delete a ssnat rule"""
        data = {
            "id": 1,
            "enable": True,
            "outInterface": "eth0"
        }
        self.index.ssnat.add(obj=data)

        resp = Mock()
        message = Message({
            "param": {
                "id": data["id"]
            }
        })

        self.index.ssnat_delete(
            message=message, response=resp, test=True)
        # TODO: should be code=200?
        # resp.assert_called_once_with(code=200)
        resp.assert_called_once_with(data=None)

    # firewall for WAN
    def test__firewall_wan_get(self):
        """test__firewall_wan_get: Get firewall rule for wan"""
        resp = Mock()
        self.index.firewall_wan_get(message=None, response=resp, test=True)
        data = {"enable": not self.index.wan.get(id=1)["enable"]}
        resp.assert_called_once_with(data=data)

    @patch("iptables.sh")
    def test__firewall_wan_put(self, mock_sh):
        """test__firewall_wan_put: enable firewall for wan"""
        resp = Mock()
        message = Message({
            "data": {
                "enable": True,
            }
        })

        self.index.firewall_wan_put(message=message, response=resp, test=True)
        resp.assert_called_once_with(data={"enable": True})


if __name__ == "__main__":
    unittest.main()
