#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
import sys
import unittest
import glob
from mock import patch

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    os.environ["PATH"] = os.path.dirname(os.path.realpath(__file__)) \
        + ":" + os.environ["PATH"]
    from iptables import IPTables
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    print (e)
    exit(1)


class TestIPTablesClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        self.jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        self.backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in self.jsons + self.backups:
            os.unlink(file)

        self.iptables = IPTables(name="iptables", path=self.root_path)
        self.iptables.iptables_path = "%s/tests/iptables.rules" \
            % self.root_path

    def tearDown(self):
        files = glob.glob(os.path.join(self.root_path, "data/*.json")) + \
            glob.glob(os.path.join(self.root_path, "data/*.backup")) + \
            glob.glob(os.path.join(self.iptables.iptables_path))
        for prevConfig in files:
            try:
                os.unlink(prevConfig)
            except:
                pass

        self.iptables = None

    def test__generate_rule_options(self):
        """test__generate_rule_options: Generate rule uptions"""
        rule = {
            "chain": "POSTROUTING",
            "target": "MASQUERADE",
            "out-interface": "wwan0"
        }
        opts = ["-A", "POSTROUTING", "-o", "wwan0", "-j", "MASQUERADE"]
        rule_opts = self.iptables._generate_rule_options(rule)
        self.assertEqual(opts, rule_opts)

    def test__generate_config(self):
        """test__generate_config: Generate iptables rule file"""
        # TODO
        pass

    @patch("sh.iptables_restore")
    def test__apply(self, mock_sh):
        """test__apply: Apply the settings"""
        # TODO: find the better way
        self.iptables.apply()
        self.assertEqual(mock_sh.called, True)

    @patch("sh.iptables_restore")
    def test__apply__failed(self, mock_sh):
        """test__apply__failed: Failed to apply the settings"""
        # TODO: find the better way
        mock_sh.side_effect = IOError("error")
        with self.assertRaises(IOError):
            self.iptables.apply()
        # mock_logger.assert_called_once_with(IOError("error"))

    def test__get_all(self):
        """test__get_all: Get all rules (support 3 tables currently)"""
        result = self.iptables.get_all()
        self.assertEqual(3, len(result))

    def test__get_rules_by_table(self):
        """test__get_rules_by_table: Get filter rules"""
        rules = self.iptables.get_rules_by_table("filter")
        self.assertEqual(self.iptables.tables["filter"].getAll(), rules)

    def test__get_rule_by_id(self):
        """test__get_rule_by_id: Get a filter rule"""
        rule = self.iptables.get_rule_by_id("filter", 1)
        self.assertEqual(self.iptables.tables["filter"].get(1), rule)

    @patch("iptables.sh.iptables_restore")
    def test__add_rule_to_table(self, mock_iptables_restore):
        """test__add_rule_to_table: Add a filter rule"""
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        self.iptables.add_rule_to_table(table="filter", newObj=rule)

        rule["id"] = 1
        self.assertEqual(rule, self.iptables.tables["filter"].get(1))

    @patch("iptables.sh.iptables_restore")
    def test__add_rule_to_table__failed(self, mock_iptables_restore):
        """test__add_rule_to_table: Fail to add a filter rule"""
        mock_iptables_restore.side_effect = IOError("error")

        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }
        with self.assertRaises(IOError):
            self.iptables.add_rule_to_table(table="filter", newObj=rule)

        self.assertEqual(None, self.iptables.tables["filter"].get(1))

    @patch("iptables.sh.iptables_restore")
    def test__update_rule_to_table(self, mock_iptables_restore):
        """test__update_rule_to_table: Update a filter rule"""
        self.iptables.add_rule_to_table(
            table="filter",
            newObj={
                "chain": "INPUT",
                "target": "REJECT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            })

        rule = {
            "id": 1,
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        self.iptables.update_rule_to_table(table="filter", id=1, newObj=rule)
        self.assertEqual(rule, self.iptables.tables["filter"].get(1))

    @patch("iptables.sh.iptables_restore")
    def test__update_rule_to_table__failed(self, mock_iptables_restore):
        """test__update_rule_to_table__failed: Fail to update a filter rule"""
        self.iptables.add_rule_to_table(
            table="filter",
            newObj={
                "chain": "INPUT",
                "target": "REJECT",
                "protocol": "tcp",
                "match": "tcp",
                "in-interface": "eth1",
                "destination-port": "443"
            })

        rule = {
            "id": 1,
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        mock_iptables_restore.side_effect = IOError("error")
        with self.assertRaises(IOError):
            self.iptables.update_rule_to_table(
                table="filter", id=1, newObj=rule)
        self.assertEqual(None, self.iptables.tables["filter"].get(1))

    @patch("iptables.sh.iptables_restore")
    def test__remove_rule_from_table(self, mock_iptables_restore):
        """test__remove_rule_from_table: Delete a filter rule"""
        self.iptables.remove_rule_from_table(table="filter", id=1)
        self.assertEqual([], self.iptables.tables["filter"].getAll())


if __name__ == "__main__":
    unittest.main()
