#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
import sys
import unittest
import glob
import copy
from mock import patch

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    os.environ["PATH"] = os.path.dirname(os.path.realpath(__file__)) \
        + ":" + os.environ["PATH"]
    from iptables import IPTables
    from iptables.mapping import Mapping
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    print (e)
    exit(1)


class TestMappingClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        self.jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        self.backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in self.jsons + self.backups:
            os.unlink(file)

        self.iptables = IPTables(name="iptables", path=self.root_path)
        self.iptables.iptables_path = "%s/tests/iptables.rules" \
            % self.root_path
        self.mapping = Mapping(name="nat.portmapping", path=self.root_path,
                               iptables=self.iptables)

    def tearDown(self):
        files = glob.glob(os.path.join(self.root_path, "data/*.json")) + \
            glob.glob(os.path.join(self.root_path, "data/*.backup")) + \
            glob.glob(os.path.join(self.iptables.iptables_path))
        for prevConfig in files:
            try:
                os.unlink(prevConfig)
            except:
                pass

        self.mapping = None

    def test__generate_by_protocol__tcp(self):
        """test__generate_by_protocol: Generate rules with tcp protocol."""
        # arrange
        rule = {
            "chain": "POSTROUTING",
            "target": "MASQUERADE",
            "out-interface": "wwan0"
        }

        # act
        rules = self.mapping._generate_by_protocol(rule, "tcp")

        # assert
        rule["protocol"] = "tcp"
        self.assertEqual(rules, [rule])

    def test__generate_by_protocol__both(self):
        """test__generate_by_protocol__both: Generate rules with both tcp and
        udp."""
        # arrange
        rule = {
            "chain": "POSTROUTING",
            "target": "MASQUERADE",
            "out-interface": "wwan0"
        }

        # act
        rules = self.mapping._generate_by_protocol(rule, "both")

        # assert
        rule["protocol"] = "tcp"
        both = []
        both.append(copy.deepcopy(rule))
        rule["protocol"] = "udp"
        both.append(copy.deepcopy(rule))
        self.assertEqual(both, rules)

    @patch("iptables.sh.iptables_restore")
    def test__add_rules_to_iptables(self, mock_iptables_restore):
        """test__add_rules_to_iptables: Add a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        # act
        self.mapping._add_rules_to_iptables(rule, {"filter": [rule]})

        # assert
        rule["id"] = 1
        self.assertEqual(rule, self.iptables.tables["filter"].get(1))

    @patch("iptables.sh.iptables_restore")
    def test__delete_rules_from_iptables(self, mock_iptables_restore):
        """test__delete_rules_from_iptables: Delete a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }
        self.mapping._add_rules_to_iptables(rule, {"filter": [rule]})

        # act
        self.mapping._delete_rules_from_iptables(rule)

        # assert
        self.assertEqual([], self.iptables.tables["filter"].getAll())

    @patch("iptables.sh.iptables_restore")
    def test__add(self, mock_iptables_restore):
        """test__add: add a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        def converter(obj):
            return {"filter": [rule]}
        self.mapping.converter = converter

        # act
        obj = self.mapping.add({"enable": True})
        rule["id"] = obj["id"]

        # assert
        self.assertEqual(rule, self.iptables.tables["filter"].get(obj["id"]))

    @patch.object(IPTables, "apply")
    def test__add__failed(self, mock_apply):
        """test__add__failed: fail to add a filter rule"""
        # arrange
        mock_apply.side_effect = IOError("error")
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        def converter(obj):
            return {"filter": [rule]}
        self.mapping.converter = converter

        # act
        obj = self.mapping.add({"enable": True})

        # assert
        self.assertEqual(None, obj)

    @patch.object(IPTables, "apply")
    def test__update(self, mock_apply):
        """test__update: update a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        def converter(obj):
            return {"filter": [rule]}
        self.mapping.converter = converter
        obj = self.mapping.add({"enable": True})
        rule["id"] = obj["id"]

        # arrange - updated rule
        rule["destination-port"] = "80"

        # act
        obj = self.mapping.update(1, {"enable": True})

        # assert
        self.assertEqual(
            self.iptables.tables["filter"].get(
                obj["iptablesMapping"]["filter"][0]),
            rule)

    @patch.object(IPTables, "apply")
    def test__update__failed(self, mock_apply):
        """test__update__failed: fail to update a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        def converter(obj):
            return {"filter": [rule]}
        self.mapping.converter = converter
        obj = self.mapping.add({"enable": True})
        rule["id"] = obj["id"]

        # arrange - updated rule
        rule["destination-port"] = "80"

        # act
        mock_apply.side_effect = IOError("error")
        obj = self.mapping.update(1, {"enable": True})

        # assert
        self.assertEqual(obj, None)

    @patch.object(IPTables, "apply")
    def test__remove(self, mock_apply):
        """test__remove: remove a filter rule"""
        # arrange
        rule = {
            "chain": "INPUT",
            "target": "REJECT",
            "protocol": "tcp",
            "match": "tcp",
            "in-interface": "eth1",
            "destination-port": "443"
        }

        def converter(obj):
            return {"filter": [rule]}
        self.mapping.converter = converter
        obj = self.mapping.add({"enable": True})

        # act
        num = self.mapping.remove(obj["id"])

        # assert
        self.assertEqual(num, 1)
        self.assertEqual(self.iptables.tables["filter"].getAll(), [])


if __name__ == "__main__":
    unittest.main()
