#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import logging
import sh

from sanji.core import Sanji
from sanji.core import Route
from logprofile import LogProfiles
from logprofile import LogUploader
from logprofile import restart_mxlogarchiver
from logprofile import merge_equipment

_logger = logging.getLogger("sanji.logprofile.index")


def has_equipment(logprofile, equipmentName):
    for device in logprofile["equipmentList"]:
        if device["equipmentName"] == equipmentName:
            return True
    return False


def is_log_path_available():
    MX_DATA_LOG_PATH = os.environ.get("MX_DATA_LOG_PATH", None)
    if MX_DATA_LOG_PATH is None or len(MX_DATA_LOG_PATH) == 0:
        return False
    return True


class Index(Sanji):

    LP_CONFIG_FILE = "/etc/mxmodbusmaster/configuration"
    RESTART_SERVICE = False

    def init(self, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(__file__))
        self.logprofiles = LogProfiles(name="logprofiles", path=path_root)
        _logger.info(
            "MX_DATA_LOG_PATH: %s" % os.environ.get("MX_DATA_LOG_PATH", ""))

        if is_log_path_available() is False:
            return

        try:
            self.logprofiles.generate_configs(restart=True)
        except Exception, e:
            _logger.error(str(e))

    @Route(methods="get", resource="/mxc/logprofile")
    def get_all_logprofile(self, message, response):
        results = self.logprofiles.getAll()
        if "equipmentName" in message.query:
            equipmentName = message.query["equipmentName"]
            results = [logprofile for logprofile in results if
                       has_equipment(logprofile, equipmentName)]

        return response(data=results)

    @Route(methods="post", resource="/mxc/logprofile")
    def post_logprofile(self, message, response):
        if is_log_path_available() is False:
            return response(
                code=500, data={"message": "SD card is not available"})
        try:
            newTemplate = self.logprofiles.add(message.data)
            return response(data=newTemplate)
        except RuntimeError as e:
            return response(code=400, data={"message": str(e)})

    @Route(methods="delete", resource="/mxc/logprofile")
    def delete_logprofiles(self, message, response):
        for tid in message.data:
            self.logprofiles.remove(id=tid)
        return response(data={"message": "success"})

    @Route(methods="get", resource="/mxc/logprofile/:id")
    def get_logprofile(self, message, response):
        template = self.logprofiles.get(id=message.param["id"])
        if template is None:
            return response(code=404)

        return response(data=template)

    @Route(methods="put", resource="/mxc/logprofile/:id")
    def update_logprofile(self, message, response):
        if is_log_path_available() is False:
            return response(
                code=500, data={"message": "SD card is not available"})

        updatedLogprofile = self.logprofiles.update(
            id=int(message.param["id"]), newObj=message.data)
        if updatedLogprofile is None:
            return response(code=404)
        return response(data=updatedLogprofile)

    @Route(methods="delete", resource="/mxc/logprofile/:id")
    def delete_logprofile(self, message, response):
        self.logprofiles.remove(id=int(message.param["id"]))
        return response(data={"message": "success"})

    @Route(methods="post", resource="/mxc/test-logprofile-upload")
    def post_test_logprofile_upload(self, message, response, schema=None):
        if is_log_path_available() is False:
            return response(
                code=500, data={"message": "SD card is not available"})

        template = self.logprofiles.get(id=int(message.data["id"]))
        if template is None:
            return response(code=404)

        if not restart_mxlogarchiver(bg=False):
            return response(
                code=500, data={"message": "Restart mxlogarchiver failed."})

        try:
            sh.mxloguploader(
                "--profile", template.profilename, "--upload")
        except Exception as e:
            _logger.error(
                "Test upload logprofile:{} failed"
                .format(template.profilename))
            _logger.debug(e)
            return response(code=500)

        return response()

    @Route(methods="post", resource="/mxc/test-logprofile-connection")
    def post_test_logprofile_connection(self, message, response, schema=None):
        try:
            default_uploader = {
                "logDirectory": "/dev/null"
            }
            default_uploader.update(message.data["httpPost"])
            log_uploader = LogUploader(default_uploader)
        except Exception as e:
            _logger.debug(e)
            return response(code=400)
        _logger.info("Testing profile connection...")

        try:
            temp_conf = "/tmp/test-upload.conf"
            log_uploader.save(path=temp_conf, set_schedule=False)
            sh.mxloguploader("--file", temp_conf, "--upload-test")
        except Exception as e:
            _logger.debug(e)
            return response(code=500)

        return response()

    @Route(methods="get", resource="/mxc/equipments")
    def get_equipments(self, message, response):
        def get_equs(resource):
            try:
                return self.publish.direct.get(resource, timeout=5).data
            except Exception:
                _logger.warning("can't get: {}".format(resource))
            return []

        def add_type(eqs, eq_type):
            for eq in eqs:
                eq["type"] = eq_type

        eqs = []
        eqs_modbus = get_equs("/mxc/modbus/equipments")
        eqs_custom = get_equs("/mxc/custom/equipments")
        add_type(eqs_modbus, "modbus")
        add_type(eqs_custom, "custom")

        eqs = merge_equipment(eqs_custom, eqs_modbus)
        return response(data=eqs)

if __name__ == "__main__":
    from sanji.connection.mqtt import Mqtt
    FORMAT = "%(asctime)s - %(levelname)s - %(lineno)s - %(message)s"
    logging.basicConfig(level=0, format=FORMAT)
    _logger = logging.getLogger("sanji.modbus")
    logging.getLogger("sh").setLevel(logging.WARN)
    index = Index(connection=Mqtt())
    index.start()
