import logging
import sh
import shutil
import glob
import os
import re

from crontab import CronTab
from jsonlibconfig import encoder
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, Any, All, Length, \
    message, truth, Invalid, Range, Optional

_logger = logging.getLogger("sanji.logprofile")

NAME_PATTERN = re.compile(ur'^[A-Za-z0-9_-]+$')


@message('not a vaild name', cls=Invalid)
@truth
def IsValidName(v):
    return NAME_PATTERN.match(v)

UPLOADER_SCHEMA = Schema({
    "httpPost": Schema({
        Required("url"): Any(unicode, str),
        "enableAuth": bool,
        "authentication": dict,
        "bypassSsl": bool,
        "cronExpr": Any(unicode, str),
        "format": Any("json", "xml", "csv"),
        "compress": "gzip",
        "enableCompress": bool,
        "retryCount": int,
        "retryPeriodSec": int
    }, extra=REMOVE_EXTRA)
}, extra=REMOVE_EXTRA)

DIRECTORY_SIZE_LIMIT_DEFAULT = 100 * 1024 * 1024  # 100 MB Default
THRESHOLD_BYTE_PER_FILE_DEFAULT = 1024 * 1024  # 1 MB perfile
PERIOD_SEC_PER_FILE_DEFAULT = 3600  # 1 hour

ARCHIVER_SCHEMA = Schema({
    Optional("directory"): Any(unicode, str),
    Optional("thresholdBytePerFile", default=THRESHOLD_BYTE_PER_FILE_DEFAULT):
        Range(min=10240),
    Optional("directorySizeLimitByte", default=DIRECTORY_SIZE_LIMIT_DEFAULT):
        Range(min=1048576),
    "archivesToKeep": Range(min=0, max=1000),
    # TODO: Update apiary
    Optional("periodSecPerFile", default=PERIOD_SEC_PER_FILE_DEFAULT):
        Range(min=60, max=86400, max_included=False),
}, extra=REMOVE_EXTRA)

LP_SCHEMA = Schema({
    "id": int,
    Required("name"): All(IsValidName(), Length(2, 255)),
    Required("equipmentList"): [Schema({
        Required("equipmentName"): IsValidName(),
        Required("equipmentTags"): [Schema({
            Required("name"): Any(str, unicode),
            Required("logOnChange"): bool
        }, extra=REMOVE_EXTRA)]
    }, extra=REMOVE_EXTRA)],
    "uploader": UPLOADER_SCHEMA,
    "archiver": ARCHIVER_SCHEMA
}, extra=REMOVE_EXTRA)


def restart_mxlogarchiver(bg=True):
    """Restart mxlogarchiver via system service
        Returns:
            True: if success, vice versa.
    """
    try:
        sh.service("mxlogarchiver", "restart", _bg=bg)
    except Exception as e:
        _logger.error("service mxlogarchiver restart failed")
        _logger.debug(str(e))
        return False

    _logger.info("service mxlogarchiver restart")
    return True


def clear_mxloguploader(profile):
    command = "mxloguploader --clear --profile %s" % (profile)
    try:
        sh.mxloguploader("--profile", profile, "--clear", _bg=True)
    except Exception as e:
        _logger.error("%s failed" % (command))
        _logger.debug(e)

    _logger.info("%s success" % (command))
    return True


class Libconfig(object):
    def __init__(self, *args, **kw):
        """Create a Template instance
            Args:
                jsonconfig(dict): Input json template configuration
        """
        self._libconfig = None
        super(Libconfig, self).__init__(*args, **kw)

    def _convert(self):
        """Mapping restful logprofile to libconfig configuration"""
        raise NotImplementedError()

    def toLibconfig(self, force=True, indent=2):
        """Generate libconfig string from settings
            Args:
                force(boolean): If it's already mapped restful <> libconfig,
                    it won't be mapped again.
            Returns:
                libconfig(string): libconfig string
        """
        if self._libconfig is not None and not force:
            return self._libconfig

        self._libconfig = self._convert()
        return encoder.dumps(self._libconfig, indent=indent)

    def save(self, path):
        """Save libconfig to specified path with default name
            Args:
                path(str): output full filename
            Returns:
                dest(str): output file destination
            Raises:
                IOError
        """
        with open(path, "w") as fp:
            fp.write(self.toLibconfig())

        return path


class LogUploader(Libconfig, dict):

    UL_CONFIG_DIR = "/etc/mxloguploader/conf.d"

    def __init__(self, *args, **kwargs):
        super(LogUploader, self).__init__(*args, **kwargs)

    def destory(self):
        # remove uploader config
        uploaders = glob.glob(
            os.path.join(self.UL_CONFIG_DIR, ("%d-*" % (self["lid"],))))

        for f in uploaders:
            _logger.debug("Remove uploader config: %s", f)
            os.unlink(f)

        # remove crontab entry
        self.remove_schedule()

    @property
    def name(self):
        return "%d-MXLogUploader-%s" % (self["lid"], self["method"])

    def set_schedule(self):
        cmd = "mxloguploader --profile \"%s\" --upload" % (self.profile)
        cron = CronTab(user="root")
        cron.remove_all(comment=self.name)
        cron.write()
        item = cron.new(command=cmd, comment=self.name)
        item.setall(self["cronExpr"])
        cron.write()

    def remove_schedule(self):
        cron = CronTab(user="root")
        cron.remove_all(comment=self.name)
        cron.write()

    def save(self, path, set_schedule=True):
        """Save mxloguploader config"""
        super(LogUploader, self).save(path)
        self.profile = os.path.basename(os.path.splitext(path)[0])
        if set_schedule:
            self.set_schedule()

    def _convert(self):
        """Mapping restful interface to MX Log Uploader's configuration
            Raises:
                ValueError:
        """
        libconfig = {
            "method": {
                "http-post": {
                    "url": self.get("url"),
                    "bypass-ssl": True
                    if self.get("bypassSsl", False) is True
                    else False
                }
            },
            "retry": {
                "count": self.get("retryCount", 3),
                "period-sec": self.get("retryPeriodSec", 5)
            },
            "log-directory": self.get("logDirectory"),
            "content": {
                "format": self.get("format", "json"),
                "compress": "none" if self.get("enableCompress") is False
                            else self.get("compress", "gzip")
            }
        }

        if self.get("enableAuth", False) is True:
            libconfig["method"]["http-post"]["authentication"] = \
                self.get("authentication")

        return libconfig


class LogProfile(Libconfig, dict):

    FILE_EXT = "conf"
    LP_CONFIG_DIR = "/etc/mxlogarchiver/conf.d"
    LP_LOG_DIR = os.environ.get("MX_DATA_LOG_PATH", "/var/mxc/upload")

    def __init__(self, *args, **kwargs):
        super(LogProfile, self).__init__(*args, **kwargs)
        if self.get("name") is None:
            raise ValueError("invaild name")

        # Get rid off empty equipmentTags entry
        equipmentList = [device for device in self.get("equipmentList", [])
                      if len(device["equipmentTags"]) != 0]
        self["equipmentList"] = equipmentList

        # generate archive settings
        input_archiver = self.get("archiver", {})

        # Do not modify the original input, just make a copy
        archiver = input_archiver.copy()

        # Set upload directory by profile's name
        archiver_directory = "%d-%s" % (self.get("id"), self.get("name"))
        archiver["directory"] = os.path.join(
            self.LP_LOG_DIR, archiver_directory)
        self["archiver"] = archiver

        # generate uploader settings, we need instance's id
        httpPost = self.get("uploader").get("httpPost")
        httpPost["logDirectory"] = self["archiver"]["directory"]
        httpPost["method"] = "httpPost"
        httpPost["lid"] = self["id"]
        self.httpPostUploader = LogUploader(httpPost)

    @property
    def filename(self):
        return ("%d-%s" % (self.get("id"), self.get("name"))) + \
            ".%s" % self.FILE_EXT if self.FILE_EXT else ""

    @property
    def profilename(self):
        return ("%d-%s" % (self.get("id"), self.get("name")))

    def clean_upload_dir(self, force=False):
        """Remove upload dir if profile name has been changed
           Args:
               force(boolean): Delete directory including current directory
        """
        uploaders = glob.glob(
            os.path.join(self.LP_LOG_DIR, ("%d-*" % (self["id"],))))

        _logger.debug("Cleaning upload directory...")
        for f in uploaders:
            if f == self["archiver"]["directory"] and not force:
                _logger.debug("Current directory: %s" % f)
                continue
            _logger.debug("Remove upload directory: %s" % f)
            shutil.rmtree(f)

    def destory(self):
        # remove it's uploader
        if getattr(self, "httpPostUploader", None) is not None:
            self.httpPostUploader.destory()

        # remove config
        logprofiles = glob.glob(
            os.path.join(self.LP_CONFIG_DIR, ("%d-*" % (self["id"],))))
        for f in logprofiles:
            _logger.debug("Remove log profile: %s", f)
            os.unlink(f)

        self.clean_upload_dir()

    def save(self, path):
        self.destory()
        super(LogProfile, self).save(path)
        self.httpPostUploader.save(
            path=os.path.join(
                self.httpPostUploader.UL_CONFIG_DIR, self.filename))

    def _convert_tags(self, tags):
        """Convert tags format from json to libconfig"""
        converted_tags = []
        for tag in tags:
            converted_tags.append({
                "tag": tag["name"],
                "log-on-change": tag["logOnChange"]
            })

        return converted_tags

    def _convert(self):
        """Mapping restful interface to modbus master's configuration
            Raises:
                ValueError:
        """
        archive = self.get("archiver")
        libconfig = {
            "archive": {
                "directory": archive.get("directory"),
                "directory-size-limit-byte": archive.get(
                    "directorySizeLimitByte",
                    DIRECTORY_SIZE_LIMIT_DEFAULT),
                "threshold-byte-per-file": archive.get(
                    "thresholdBytePerFile",
                    THRESHOLD_BYTE_PER_FILE_DEFAULT),
                "period-sec":  archive.get(
                    "periodSec",
                    PERIOD_SEC_PER_FILE_DEFAULT)
                # we don't allowed user set the following properties now:
                # "archives-to-keep": archive.get("archivesToKeep")
            },
            "equipment-list": []
        }

        for device in self.get("equipmentList", []):
            libconfig["equipment-list"].append({
                "equipment": device["equipmentName"],
                "tag-list": self._convert_tags(device["equipmentTags"])
            })

        return libconfig


class LogProfiles(Model):

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = LP_SCHEMA
        kwargs["model_cls"] = LogProfile
        super(LogProfiles, self).__init__(*args, **kwargs)

    def generate_configs(self, restart=False):
        """Generate all of LogProfiles and save to LogProfiles.LP_CONFIG_DIR"""
        for logprofile in self.getAll():
            logprofile.save(
                path=os.path.join(
                    logprofile.LP_CONFIG_DIR, logprofile.filename))

        if restart:
            restart_mxlogarchiver()

    def get_by_name(self, name):
        """Get log profile by profile's name
            Args:
                name(str): query profile's name
            Returns:
                profile(log profile): or None if not found
        """
        for logprofile in self.getAll():
            if logprofile["name"] != name:
                continue
            return logprofile
        return None

    def add(self, data):
        """Add a logprofile
            Args:
                data(dict): logprofile config
            Returns:
                newLogProfile(LogProfile): new logprofile instance with id
            Raises:
                ValueError: if device's template not found
        """
        if self.get_by_name(data["name"]) is not None:
            raise ValueError(
                "LogProfile Name: %s is already exist!" % (data["name"]))

        if len(self.getAll()) + 1 > 5:
            raise RuntimeError("Up to 5 Log Profiles")

        newLogProfile = super(LogProfiles, self).add(data)
        newLogProfile.save(
            path=os.path.join(
                newLogProfile.LP_CONFIG_DIR, newLogProfile.filename))
        restart_mxlogarchiver()

        return newLogProfile

    def remove(self, id):
        """Remove an logprofile,\
           it will restart mxlogarchiver port if affected > 0
            Args:
                id(int): logprofile's id
            Returns:
                affected(int): number of affected rows
        """
        logprofile = self.get(id=id)
        if logprofile is None:
            return None

        logprofile.destory()
        logprofile.clean_upload_dir(force=True)
        clear_mxloguploader("%d-%s" % (id, logprofile["name"]))
        affected = super(LogProfiles, self).remove(id=id)
        if affected > 0:
            restart_mxlogarchiver()

        return affected

    def update(self, id, newObj):
        """Update an exist logprofile,\
           it will restart mxlogarchiver port if update successfully > 0
           Returns:
               newLogProfile(logprofile): udpated logprofile
           Raises:
               ValueError: if device list contains invaild entry
        """
        logprofile = self.get(id)
        if logprofile is None:
            return None

        if "name" in newObj and logprofile["name"] != newObj["name"]:
            clear_mxloguploader("%d-%s" % (id, logprofile["name"]))

        newLogProfile = super(LogProfiles, self).update(id=id, newObj=newObj)
        newLogProfile.save(
            path=os.path.join(
                newLogProfile.LP_CONFIG_DIR, newLogProfile.filename))
        restart_mxlogarchiver()

        return newLogProfile


def to_dict(key, data):
    result = {}
    for row in data:
        result[row[key]] = row
    return result


def merge_equipment_tags(source, target):
    tags_dict = to_dict("name", source)
    tags_index = map(lambda x: x["name"], source)

    for tag in target:
        name = tag["name"]
        if name not in tags_index:
            tags_dict[name] = tag
            continue
        tags_dict[name] = {
            "name": tags_dict[name]["name"],
            "description": tags_dict[name]["description"] +
            "; " + tag["description"]
        }

    return tags_dict


def merge_equipment(source, target):
    target_names = map(lambda x: x["equipmentName"], target)
    source_names = map(lambda x: x["equipmentName"], source)
    conflicts = list(set(target_names) & set(source_names))

    if len(conflicts) == 0:
        return target + source

    target_dict = to_dict("equipmentName", target)
    source_dict = to_dict("equipmentName", source)

    # If conflict, we merge two tag and append it's description
    for name in conflicts:
        target_dict[name]["type"] = "mixed"
        merged_tags = merge_equipment_tags(
            target_dict[name]["equipmentTags"],
            source_dict[name]["equipmentTags"]
        )
        target_dict[name]["equipmentTags"] = \
            [merged_tags[key] for key in merged_tags]

        # Sort by tag's name
        target_dict[name]["equipmentTags"] = sorted(
           target_dict[name]["equipmentTags"],
           key=lambda x: x["name"]
        )

        source_dict.pop(name)

    return [target_dict[key] for key in target_dict] + \
           [source_dict[key] for key in source_dict]
