#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import sys
import glob
import unittest

from mock import Mock
from mock import patch
from mock import ANY
from mock import call

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from logprofile import LogUploader
    from logprofile import LogProfile
    from logprofile import LogProfiles
    from logprofile import merge_equipment
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)


class TestFunctions(unittest.TestCase):
    def test_merge_equipments(self):
        equipments_a = [{
            "equipmentName": "LIGHT",
            "equipmentTags": [
                {
                    "name": "aaa",
                    "description": "aaa"
                }
            ]
        }]
        equipments_b = [{
            "equipmentName": "LIGHT",
            "equipmentTags": [
                {
                    "name": "bbb",
                    "description": "bbb"
                }
            ]
        }]

        merged = merge_equipment(equipments_a, equipments_b)
        self.assertEqual(merged[0]["equipmentName"], "LIGHT")
        self.assertEqual(len(merged[0]["equipmentTags"]), 2)

    def test_merge_equipments_tags(self):
        """Same equipment with conflicts should be merged too"""
        equipments_a = [{
            "equipmentName": "LIGHT",
            "equipmentTags": [
                {
                    "name": "aaa",
                    "description": "aaa"
                }
            ]
        }]

        equipments_b = [{
            "equipmentName": "LIGHT",
            "equipmentTags": [
                {
                    "name": "aaa",
                    "description": "bbb"
                }
            ]
        }]

        merged = merge_equipment(equipments_a, equipments_b)
        self.assertEqual(merged[0]["equipmentName"], "LIGHT")
        self.assertEqual(len(merged[0]["equipmentTags"]), 1)
        self.assertEqual(
            merged[0]["equipmentTags"][0]["description"],
            "bbb; aaa"
        )


class TestLogProfileClass(unittest.TestCase):

    def setUp(self):
        LogProfile.LP_CONFIG_DIR = "/tmp"
        LogUploader.UL_CONFIG_DIR = "/tmp"
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)

        self.logprofiles = LogProfiles(name="logprofiles", path=self.root_path)
        self.factory = {
            "name": "log-profile-1",
            "archiver": {
                "directory": "/var/mxc/upload",
                "thresholdBytePerFile": 5242800,
                "archivesToKeep": 24
            },
            "uploader": {
                "httpPost": {
                    "url": "http://host/post",
                    "enableAuth": False,
                    "authentication": {
                        "username": "uploader",
                        "password": "password"
                    },
                    "bypassSsl": True,
                    "cronExpr": "*/5 * * * *",
                    "format": "json",
                    "compress": "gzip",
                    "enableCompress": True,
                    "retryCount": 3,
                    "retryPeriodSec": 3
                }
            },
            "equipmentList": [
                {
                    "equipmentName": "MyioLogik-E2242",
                    "equipmentTags": [{
                        "name": "di0",
                        "logOnChange": False
                    }, {
                        "name": "di1",
                        "logOnChange": False
                    }, {
                        "name": "ai0",
                        "logOnChange": False
                    }]
                },
                {
                    "equipmentName": "MyotherioLogik-E2242",
                    "equipmentTags": [{
                        "name": "di0",
                        "logOnChange": False
                    }, {
                        "name": "di1",
                        "logOnChange": False
                    }, {
                        "name": "ai0",
                        "logOnChange": False
                    }]
                }
            ]
        }

    def tearDown(self):
        pass

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    def test_add_logprofile(self, restart_mxlogarchiver, CronTab):
        """Add a LogProfile, with a vaild config, \
           should return LogProfile instance"""
        m = Mock()
        CronTab.return_value = m
        data = self.factory.copy()
        data["uploader"]["httpPost"]["enableAuth"] = True
        newLogProfile = self.logprofiles.add(data)
        self.assertTrue(isinstance(newLogProfile, LogProfile))
        self.assertEqual(newLogProfile["id"], 1)
        # m.remove_all.assert_called_once_with(
        #     comment=newLogProfile.httpPostUploader.name)
        # m.new.assert_called_once_with(
        #     command=ANY,
        #     comment=newLogProfile.httpPostUploader.name)
        # m.write.assert_called_once_with()
        restart_mxlogarchiver.assert_called_once_with()

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    def test_add_logprofile_2(self, restart_mxlogarchiver, CronTab):
        """Add a LogProfile, with a duplicate name, \
           should raise ValueError"""
        m = Mock()
        CronTab.return_value = m
        data = self.factory.copy()
        self.logprofiles.add(data)
        data2 = self.factory.copy()
        restart_mxlogarchiver.reset_mock()
        with self.assertRaises(ValueError):
            self.logprofiles.add(data2)
        self.assertFalse(restart_mxlogarchiver.called)

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    def test_add_logprofile_3(self, restart_mxlogarchiver, CronTab):
        """Add 6 LogProfiles, should raise ValueError"""
        m = Mock()
        CronTab.return_value = m
        for i in xrange(0, 5):
            restart_mxlogarchiver.reset_mock()
            data = self.factory.copy()
            data["name"] = "name" + str(i)
            self.logprofiles.add(data)
            restart_mxlogarchiver.assert_called_once_with()

        data = self.factory.copy()
        data["name"] = "name%s" % 5
        with self.assertRaises(RuntimeError):
            self.logprofiles.add(data)

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    def test_add_logprofile_4(self, restart_mxlogarchiver, CronTab):
        """Add a LogProfile, with a vaild config and
           an Empty equipmentTags entry, should return LogProfile instance
           with no Empty equipmentTags entry"""
        m = Mock()
        CronTab.return_value = m
        data = self.factory.copy()
        data["uploader"]["httpPost"]["enableAuth"] = True
        data["equipmentList"][0]["equipmentTags"] = []
        newLogProfile = self.logprofiles.add(data)
        self.assertTrue(isinstance(newLogProfile, LogProfile))
        self.assertEqual(newLogProfile["id"], 1)
        self.assertEqual(len(newLogProfile["equipmentList"]), 1)
        restart_mxlogarchiver.assert_called_once_with()

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    @patch("logprofile.clear_mxloguploader")
    @patch("logprofile.LogProfile.clean_upload_dir")
    def test_remove_logprofile(
            self, clean_upload_dir, clear_mxloguploader,
            restart_mxlogarchiver, CronTab):
        """Remove a LogProfile, with a vaild config, \
           should remove mxlogarchive, mxloguploader config and crontab"""
        m = Mock()
        CronTab.return_value = m
        item_mock = Mock()
        m.new.return_value = item_mock
        data = self.factory.copy()
        newLogProfile = self.logprofiles.add(data)

        CronTab.reset_mock()
        clean_upload_dir.reset_mock()

        self.logprofiles.remove(newLogProfile["id"])
        m.remove_all.assert_called_once_with(
            comment=newLogProfile.httpPostUploader.name)
        m.write.assert_called_once_with()
        clear_mxloguploader.assert_called_once_with(
            "%d-%s" % (newLogProfile["id"], data["name"]))
        self.assertListEqual([
            call(),
            call(force=True)],
            clean_upload_dir.call_args_list)

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    def test_update_logprofile(self, restart_mxlogarchiver, CronTab):
        """Update a LogProfile, with a vaild id, \
           should return new LogProfile"""
        m = Mock()
        CronTab.return_value = m
        item_mock = Mock()
        m.new.return_value = item_mock
        data = self.factory.copy()
        newLogProfile = self.logprofiles.add(data)
        newData = self.factory.copy()
        newData["uploader"]["httpPost"]["cronExpr"] = "1 1 1 1 1"
        m.reset_mock()

        newLogProfile = self.logprofiles.update(newLogProfile["id"], newData)
        # m.remove_all.assert_called_once_with(
        #     comment=newLogProfile.httpPostUploader.name)
        # m.write.assert_called_once_with()
        m.new.assert_called_once_with(
            command=ANY,
            comment=newLogProfile.httpPostUploader.name)
        item_mock.setall.assert_called_once_with(
            newData["uploader"]["httpPost"]["cronExpr"])

    @patch("logprofile.CronTab")
    @patch("logprofile.restart_mxlogarchiver")
    @patch("logprofile.clear_mxloguploader")
    def test_update_logprofile_2(
            self, clear_mxloguploader, restart_mxlogarchiver, CronTab):
        """Update a LogProfile, with a vaild id and changed name, \
           should return new LogProfile and clean old files/folders"""
        m = Mock()
        CronTab.return_value = m
        item_mock = Mock()
        m.new.return_value = item_mock
        data = self.factory.copy()
        newLogProfile = self.logprofiles.add(data)
        newData = self.factory.copy()
        newData["name"] = "newName"
        m.reset_mock()

        newLogProfile = self.logprofiles.update(newLogProfile["id"], newData)
        m.new.assert_called_once_with(
            command=ANY,
            comment=newLogProfile.httpPostUploader.name)
        clear_mxloguploader.assert_called_once_with(
            "%d-%s" % (newLogProfile["id"], data["name"]))


if __name__ == "__main__":
    unittest.main()
