#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os

from voluptuous import All
from voluptuous import Length
from voluptuous import Required, Schema
from voluptuous import REMOVE_EXTRA

from sanji.core import Sanji
from sanji.core import Route
from sanji.model_initiator import ModelInitiator

from openvpn.openvpn import OpenVPN
from openvpn.openvpn import HTTPError
from openvpn.openvpn import OvpnError

_logger = logging.getLogger("sanji.openvpn.index")


class Index(Sanji):

    def init(self, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(__file__))
        self.model = ModelInitiator("openvpn", path_root)
        self.openvpn = OpenVPN(self.model)

    def _get_openvpn_data(self):
        ip = self.openvpn.get_ip()
        return {
            "connection": self.openvpn.is_connect(),
            "vpnLocalIp": ip[0],
            "vpnRemoteIp": ip[1]
        }

    # OpenVPN
    @Route(methods="get", resource="/network/openvpn")
    def get_openvpn_setting(self, message, response):
        return response(
            data=dict(
                self.model.db.items() + self._get_openvpn_data().items()))

    PUT_SCHEMA = Schema(
        {
            Required("enable"): bool
        },
        extra=REMOVE_EXTRA)

    @Route(methods="put", resource="/network/openvpn", schema=PUT_SCHEMA)
    def put_openvpn_setting(self, message, response):
        config = self.PUT_SCHEMA(message.data)
        if not self.openvpn.update(config):
            return response(
                code=400,
                data={"message": "Update OpenVPN setting fail."})

        return response(
            data=dict(
                self.model.db.items() + self._get_openvpn_data().items()))

    CONFIG_DOWNLOAD_SCHEMA = Schema({
        Required("url"): All(str, Length(1, 4096)),
        "headers": {}
    }, extra=REMOVE_EXTRA)

    CONFIG_UPLOAD_SCHEMA = Schema({
        Required("file"): CONFIG_DOWNLOAD_SCHEMA
    }, extra=REMOVE_EXTRA)

    # OpenVPNConfig
    @Route(
        methods="put",
        resource="/network/openvpn/config",
        schema=CONFIG_UPLOAD_SCHEMA)
    def get_openvpn_config(self, message, response):
        headers = message.data["file"].get("headers", {})
        try:
            self.openvpn.import_from_http(
                url=message.data["file"]["url"],
                headers=headers)
            return response()
        except OvpnError as e:
            _logger.error("Import from http error: {}".format(str(e)))
            return response(code=500, data={"message": str(e)})
        except Exception as e:
            _logger.error("Import from http error")
            _logger.error(str(e))
            raise e

    @Route(
        methods="post",
        resource="/network/openvpn/config",
        schema=CONFIG_DOWNLOAD_SCHEMA)
    def post_openvpn_config(self, message, response):
        headers = message.data.get("headers", {})
        try:
            url = self.openvpn.export_to_http(
                url=message.data["url"],
                headers=headers)
            return response(data={"url": url})
        except HTTPError as e:
            _logger.error("Export to http error: {}".format(str(e)))
            return response(code=500, data={"message": str(e)})
        except OvpnError as e:
            _logger.error("Export to http error: {}".format(str(e)))
            return response(code=404, data={"message": str(e)})
        except Exception as e:
            _logger.error("Export to http error")
            _logger.error(str(e))
            raise e

    @Route(methods="delete", resource="/network/openvpn/config")
    def delete_openvpn_config(self, message, response):
        try:
            self.openvpn.delete_ovpn()
            return response(data={})
        except OvpnError as e:
            _logger.error("Delete ovpn error: {}".format(str(e)))
            return response(code=404, data={"message": str(e)})
        except Exception as e:
            _logger.error("Delete ovpn error")
            _logger.error(str(e))
            raise e

    @Route(
        methods="post",
        resource="network/openvpn/sample",
        schema=CONFIG_DOWNLOAD_SCHEMA)
    def post_openvpn_sample(self, message, response):
        headers = message.data.get("headers", {})
        try:
            url = self.openvpn.export_to_http(
                url=message.data["url"],
                headers=headers,
                sample=True)
            return response(data={"url": url})
        except HTTPError as e:
            _logger.error("Export sample to http error: {}".format(str(e)))
            return response(code=500, data={"message": str(e)})
        except Exception as e:
            _logger.error("Export sample to http error")
            _logger.error(str(e))
            raise e


if __name__ == "__main__":
    from sanji.connection.mqtt import Mqtt
    FORMAT = "%(asctime)s - %(levelname)s - %(lineno)s - %(message)s"
    logging.basicConfig(level=0, format=FORMAT)
    _logger = logging.getLogger("sanji.openvpn")
    logging.getLogger("sh").setLevel(logging.WARN)
    index = Index(connection=Mqtt())
    index.start()
