#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os
import requests
import shutil

import sh
from retrying import RetryError
from retrying import retry

from management import Management

_logger = logging.getLogger("sanji.openvpn.openvpn")
thisFilePath = os.path.abspath(os.path.dirname(__file__))


class HTTPError(Exception):
    pass


class OvpnError(Exception):
    pass


class OpenVPN(object):

    MGMT_HOST = 'localhost'
    MGMT_PORT = 7505
    OVPN_PATH = os.path.join(thisFilePath, '..', 'data', 'client.ovpn')
    SAMPLE_OVPN_PATH = os.path.join(thisFilePath, '..', 'data', 'sample.ovpn')

    def __init__(self, model):
        self.model = model
        self.mgmt = Management(self.MGMT_HOST, self.MGMT_PORT)

        if self.model.db["enable"]:
            self.restart()
        else:
            self.stop()

    def start(self):
        ovpnfile = self.OVPN_PATH if self.model.db["ovpnFileExist"] \
            else self.SAMPLE_OVPN_PATH
        ret = None
        try:
            ret = sh.openvpn(
                '--daemon',
                '--management', self.MGMT_HOST, self.MGMT_PORT,
                '--config', ovpnfile,
                '--dev', 'tun0',
                '--up', os.path.join(
                    thisFilePath, '..', 'scripts', 'ovpn-up.sh'),
                '--down', os.path.join(
                    thisFilePath, '..', 'scripts', 'ovpn-down.sh'),
                '--script-security', 2,
                '--up-restart')
        except sh.ErrorReturnCode as e:
            ret = e
        _logger.info(
            'OpenVPN daemon is started. CODE: {}'.format(ret.exit_code))

        return ret.exit_code

    def stop(self):
        if not self._is_up():
            return 0

        ret = sh.killall('openvpn')
        _logger.info('Killing OpenVPN. CODE: {}'.format(ret.exit_code))

        self.force_stop()

        return ret.exit_code

    def force_stop(self):
        try:
            self._confirm_openvpn_down()
        except RetryError:
            sh.killall('-SIGKILL', 'openvpn')

    def restart(self):
        self.stop()
        retcode = self.start()
        _logger.info('OpenVPN daemon is restarted. CODE: {}'.format(retcode))

        return retcode

    def update(self, config):
        db = dict(
            self.model.db.items() + config.items())

        if db["enable"]:
            if self.restart() == 0:
                self.model.db = db
                self.model.save_db()
                return True
            else:
                return False
        else:
            self.stop()
            self.model.db = db
            self.model.save_db()
            return True

    def is_connect(self):
        if not self._is_up():
            return False

        try:
            rtn = self.mgmt.command('state\n')
            if rtn is False:
                return False
            if 'CONNECTED,SUCCESS' in rtn:
                _logger.debug('OpenVPN is connected.')
                return True
        except Exception:
            _logger.error(
                'OpenVPN can not get connection status. {}'.format(repr(rtn)))
            return False

    def get_ip(self):
        """
        Retuen tuple of (local_ip, remote_ip)
        """
        fail_ret = ('n/a', 'n/a')
        if not self._is_up():
            return fail_ret

        try:
            rtn = self.mgmt.command('state\n')
            if rtn is False:
                return fail_ret
            arr = rtn.replace('\r\n', '').split(',')
            if not arr[3] or not arr[4]:
                raise ValueError
        except Exception:
            _logger.error('OpenVPN can not get ip. {}'.format(repr(rtn)))
            return fail_ret

        return arr[3], arr[4]

    def export_to_http(self, url, headers={}, sample=False):
        if sample:
            ovpnfile = self.SAMPLE_OVPN_PATH
        else:
            if not self.model.db["ovpnFileExist"]:
                raise OvpnError("ovpn file not exist, please upload it first.")
            ovpnfile = self.OVPN_PATH

        r = requests.post(
            url=url,
            files={"output": open(ovpnfile, "rb")},
            headers=headers,
            verify=False
        )

        if r.status_code != requests.codes.ok:
            _logger.debug("Response code: {}".format(str(r.status_code)))
            raise HTTPError("Can't upload files.")

        resp = r.json()
        if "url" not in resp:
            _logger.debug("Resp: {}".format(str(resp)))
            raise HTTPError("Can't get file link.")

        return resp["url"]

    def import_from_http(
            self, url, headers={}, tmpfile="/tmp/openvpn.download"):
        r = requests.get(
            url=url,
            stream=True,
            headers=headers,
            verify=False
        )
        chunk_size = 1024
        if r.status_code != requests.codes.ok:
            raise HTTPError("Can't download file.")

        with open(tmpfile, 'wb') as fd:
            for chunk in r.iter_content(chunk_size):
                fd.write(chunk)
        shutil.copy2(tmpfile, self.OVPN_PATH)
        os.unlink(tmpfile)

        self.model.db["ovpnFileExist"] = True
        self.model.save_db()

        if self.model.db["enable"] and self.restart() != 0:
            self.model.db["enable"] = False
            self.model.save_db()
            raise OvpnError("Can't apply with setting.")

    def delete_ovpn(self):
        if not self.model.db["ovpnFileExist"]:
            raise OvpnError("Ovpn file not exist.")

        os.remove(self.OVPN_PATH)

        # Update ovpnFileExist value
        self.model.db["ovpnFileExist"] = False
        self.model.save_db()

        if self.model.db["enable"] and self.restart() != 0:
            self.model.db["enable"] = False
            self.model.save_db()
            _logger.info('Disable OpenVPN service after delete ovpn file.')

    def _is_up(self):
        try:
            sh.grep(
                sh.grep(
                    sh.ps('wwaux'),
                    'openvpn --daemon'),
                '-v',
                'grep')
            _logger.debug('OpenVPN is up.')
            return True
        except Exception:
            _logger.debug('OpenVPN is not up.')
            return False

    @retry(
        retry_on_result=lambda x: x is True,
        stop_max_delay=3000)
    def _confirm_openvpn_down(self):
        return self._is_up()
