import os
import sys
import unittest

from mock import ANY
from mock import Mock
from mock import patch
import requests

from sanji.model_initiator import ModelInitiator

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from openvpn.openvpn import OpenVPN
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)


class ClientOvpn(object):

    CONFIG = (
        "dev tun\n"
        "client\n"
        "remote 192.168.1.1 1194\n"
        "protp tcp\n"
    )

    def __init__(self):
        path_root = os.path.dirname(os.path.realpath(__file__)) + "/../../"
        with open(path_root + "data/client.ovpn", "w+") as f:
            f.write(self.CONFIG)


class TestOpenVPNClass(unittest.TestCase):

    @patch('openvpn.openvpn.Management')
    def setUp(self, mgmt_mock):
        path_root = os.path.dirname(os.path.realpath(__file__)) + "/../../"
        try:
            os.unlink(path_root + "data/openvpn.json")
            os.unlink(path_root + "data/openvpn.json.backup")
            os.unlink(path_root + "data/client.ovpn")
        except:
            pass
        self.model = ModelInitiator("openvpn", path_root)
        self.openvpn = OpenVPN(self.model)

        self.mgmt = mgmt_mock.return_value

    def tearDown(self):
        self.openvpn = None

    @patch("openvpn.openvpn.OpenVPN._is_up")
    def test_is_connect(self, _is_up_mock):
        # arrange
        _is_up_mock.return_value = True

        self.mgmt.command.return_value = (
            '1455518584,CONNECTED,SUCCESS,10.9.8.6,192.168.31.114\r\n')

        # act
        rtn = self.openvpn.is_connect()

        # assert
        self.assertTrue(rtn)

    @patch("openvpn.openvpn.OpenVPN._is_up")
    def test_get_ip(self, _is_up_mock):
        # arrange
        _is_up_mock.return_value = True

        self.mgmt.command.return_value = (
            '1455518584,CONNECTED,SUCCESS,10.9.8.6,192.168.31.114\r\n')

        # act
        rtn = self.openvpn.get_ip()

        # assert
        self.assertEqual(rtn[0], "10.9.8.6")
        self.assertEqual(rtn[1], "192.168.31.114")

    @patch("openvpn.openvpn.sh")
    def test_start(self, sh_mock):
        # arrange
        sh_mock.openvpn().exit_code = 0

        # act
        rtn = self.openvpn.start()

        # assert
        self.assertEqual(rtn, 0)
        self.assertTrue(sh_mock.openvpn.called)

    @patch("openvpn.openvpn.sh")
    @patch("openvpn.openvpn.OpenVPN._is_up")
    def test_stop(self, _is_up_mock, sh_mock):
        # arrange
        _is_up_mock.return_value = True
        sh_mock.killall().exit_code = 0

        # act
        rtn = self.openvpn.stop()

        # assert
        self.assertEqual(rtn, 0)
        self.assertTrue(sh_mock.killall.called)

    @patch("openvpn.openvpn.OpenVPN.restart")
    def test_update(self, restart_mock):
        # arrange
        restart_mock.return_value = 0

        self.model.db["enable"] = False
        config = {"enable": True}

        # act
        rtn = self.openvpn.update(config)

        # assert
        self.assertTrue(rtn)
        self.assertTrue(self.model.db["enable"])
        restart_mock.assert_called_once_with()

    @patch("requests.get")
    def test_import_from_http(self, get_mock):
        # arrange
        resp_get_mock = Mock()
        resp_get_mock.status_code = requests.codes.ok
        resp_get_mock.iter_content = Mock(return_value=[])
        get_mock.return_value = resp_get_mock

        # act
        self.openvpn.import_from_http(
            url="http://localhost/download",
            headers={"accessToken": "access-token"})

        # assert
        self.assertTrue(get_mock.called)
        self.assertTrue(self.model.db["ovpnFileExist"])

    @patch("requests.post")
    def test_export_to_http_withSampleOvpn(self, post_mock):
        # arrange
        resp_post_mock = Mock()
        resp_post_mock.status_code = requests.codes.ok
        resp_post_mock.json.return_value = {"url": "http://localhost/download"}
        post_mock.return_value = resp_post_mock

        # act
        rtn = self.openvpn.export_to_http(
            url="http://localhost/upload",
            headers={"accessToken": "access-token"},
            sample=True)

        # assert
        self.assertEqual(rtn, "http://localhost/download")

    @patch("requests.post")
    def test_export_to_http_withClientOvpn(self, post_mock):
        # arrange
        ClientOvpn()
        self.model.db["ovpnFileExist"] = True
        resp_post_mock = Mock()
        resp_post_mock.status_code = requests.codes.ok
        resp_post_mock.json.return_value = {"url": "http://localhost/download"}
        post_mock.return_value = resp_post_mock

        # act
        rtn = self.openvpn.export_to_http(
            url="http://localhost/upload",
            headers={"accessToken": "access-token"},
            sample=False)

        # assert
        self.assertEqual(rtn, "http://localhost/download")

    @patch("os.remove")
    def test_delete_ovpn(self, remove_mock):
        # arrange
        self.model.db["ovpnFileExist"] = True
        self.model.save_db()

        # act
        self.openvpn.delete_ovpn()

        # assert
        remove_mock.assert_called_once_with(ANY)
        self.assertFalse(self.model.db["ovpnFileExist"])


if __name__ == '__main__':
    unittest.main()
