import logging
import re
import os
import sh
from sh import ErrorReturnCode

from string import Template
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, Any, All, Length, \
    message, truth, Invalid, Range


NAME_PATTERN = re.compile(ur'^[A-Za-z0-9_-]+$')


@message('Not a vaild name', cls=Invalid)
@truth
def IsValidName(v):
    return NAME_PATTERN.match(v)

RELAY_SCHEMA = Schema({
    Required("enable"): bool,
    Required("name"): All(IsValidName(), Length(2, 255)),
    Required("totalInactivityTimeout"): Range(1, 3600),
    Required("nodes"): [Schema({
        Required("name"): Any("Upstream", "Downstream"),
        Required("type"): Any("client", "server", "serial"),

        "clientOpts": Schema({
            Required("host"): Any(str, unicode),
            Required("port"): Range(1, 65535),
            Required("protocol"): Any("udp", "tcp"),
            Required("enableSourcePort"): bool,
            Required("sourceport"): Range(1, 65535)
        }, extra=REMOVE_EXTRA),

        "serverOpts": Schema({
            Required("port"): Range(1, 65535),
            Required("protocol"): Any("udp", "tcp")
        }, extra=REMOVE_EXTRA),

        "serialOpts": Schema({
            Required("dev"): Any(str, unicode),
            Required("baudrate"): Any(115200, 57600, 38400, 19200, 9600),
            Required("parity"): Any("none", "even", "odd"),
            Required("data"): Any(8, 7),
            Required("stop"): Any(1, 2),
            Required("enableFramer"): bool,
            Required("framerOpts"): Schema({
                Required("framerType"): Any("general", "dnp3"),
                Required("noinputTimeoutMS"): Range(0, 65535),
                Required("overflowSize"): Range(0, 65535)
            }, extra=REMOVE_EXTRA)
        }, extra=REMOVE_EXTRA)
    }, extra=REMOVE_EXTRA)]
}, extra=REMOVE_EXTRA)


def read_init_script():
    path = os.path.abspath(
        os.path.dirname(__file__) + "/../data/init-script.tmpl")
    out = ""
    with open(path) as f:
        out = f.read()

    return out


class RelayError(Exception):
    pass


class Relay(dict):

    INITD_DIR = "/etc/init.d/"
    DEFAULT_DIR = "/etc/default/"
    RELAY_BIN_DIR = "/usr/bin"
    INIT_SCRIPT_TMPL = Template(read_init_script())
    RELAY_BIN_TMPL = """#!/bin/bash
while true; do
    socat -d -d -d -lf /var/log/relay-{INSTANCE_ID}.log {OPTIONS} 1>> /var/log/relay-{INSTANCE_ID}-stdout.log 2>> /var/log/relay-{INSTANCE_ID}-stderr.log &
    SO_PID=$!
    trap "pkill -TERM -P $SO_PID; pkill -TERM -P $$; exit" SIGTERM SIGINT
    wait
    sleep 5
done
"""  # noqa

    _logger = logging.getLogger("sanji.relay")

    def __init__(self, *args, **kwargs):
        super(Relay, self).__init__(*args, **kwargs)
        # nodes = self.get("nodes")
        # if len(nodes) > 2:
        #     self._logger.warning(
        #         "%s has more than two nodes." % (self.service_name))

    def get_nodes_opts(self):
        """Get nodes opts for socat"""
        opts = []
        upstream = self._get_node_opts(self["nodes"][0])
        downstream = self._get_node_opts(self["nodes"][1])

        # Append global options
        if self["totalInactivityTimeout"] != 0:
            opts.append("-T %d" % (self["totalInactivityTimeout"]))

        # Push nodes opts
        opts.append(",".join(upstream))
        opts.append(",".join(downstream))

        # Socat always open first address,
        # so udp/tcp client always be the second address.
        if opts[-1].find("-connect") == -1:
            opts[-1], opts[-2] = opts[-2], opts[-1]

        out_opts = " ".join(opts)
        self._logger.debug("%s: %s" % (self.service_name, out_opts))

        return out_opts

    def _get_node_opts(self, node):
        """Get a single node opts for socat
            Returns:
                opts(list): List of opts
        """
        opts = []
        typestring = node.get("type")
        if typestring == "serial":
            opts = self._get_serial_opts(node)
        else:
            opts = self._get_socket_opts(node)

        return opts

    def _get_serial_opts(self, node):
        """Get serial type opts
            Returns:
                opts(list): List of opts
        """
        opts = []
        serialOpts = node["serialOpts"]

        # Enable Framer
        if serialOpts["enableFramer"]:
            return self._get_serial_framer_opts(node)

        opts.append("file:%s" % (serialOpts["dev"]))
        waitlock_name = serialOpts["dev"].replace("/", "_")
        opts.append("waitlock=/var/run/%s.lock" % (waitlock_name))
        opts.append("b%d" % (serialOpts["baudrate"]))
        opts.append("cs%d" % (serialOpts["data"]))
        opts.append("raw")
        opts.append("echo=0")

        if serialOpts["parity"] == "none":
            opts.append("parenb=0")
        elif serialOpts["parity"] == "odd":
            opts.append("parenb=1,parodd=1")
        elif serialOpts["parity"] == "even":
            opts.append("parenb=1,parodd=0")

        if serialOpts["stop"] == 1:
            opts.append("cstopb=0")
        elif serialOpts["stop"] == 2:
            opts.append("cstopb=1")

        return opts

    def _get_serial_framer_opts(self, node):
        opts = []
        serialOpts = node["serialOpts"]
        framer_opts = {}

        if serialOpts["parity"] == "none":
            framer_opts["parity"] = "n"
        elif serialOpts["parity"] == "odd":
            framer_opts["parity"] = "o"
        elif serialOpts["parity"] == "even":
            framer_opts["parity"] = "e"

        framer_opts["stop"] = serialOpts["stop"]
        framer_opts["device"] = serialOpts["dev"]
        framer_opts["baudrate"] = serialOpts["baudrate"]
        framer_opts["data"] = serialOpts["data"]
        framer_opts["noinputTimeoutMS"] = \
            serialOpts["framerOpts"]["noinputTimeoutMS"]
        framer_opts["overflowSize"] = serialOpts["framerOpts"]["overflowSize"]

        if serialOpts["framerOpts"]["framerType"] == "general":
            framer_opts_string = "--device={device} " + \
                "--config={baudrate}-{data}-{parity}-{stop} " + \
                "--noinput-timeout-ms={noinputTimeoutMS} " + \
                "--overflow-size={overflowSize}"
            opts.append("EXEC:\"mxserialframer {}\""
                        .format(framer_opts_string.format(**framer_opts)))
        else:
            framer_opts_string = \
                "--config={baudrate}-{data}-{parity}-{stop} " + \
                "--noinput-timeout-ms={noinputTimeoutMS} " + \
                "--overflow-size={overflowSize} {device}"
            opts.append("EXEC:\"mxserialdnp3framer {}\""
                        .format(framer_opts_string.format(**framer_opts)))

        return opts

    def _get_socket_opts(self, node):
        """Get udp/tcp server/client type opts
            Returns:
                opts(list): List of opts
        """
        opts = []
        relay_type = node["type"]
        if relay_type == "server":
            serverOpts = node["serverOpts"]
            protocol = node[relay_type + "Opts"]["protocol"]
            opts.append("%s-listen:%d" % (protocol, serverOpts["port"]))
            opts.append("reuseaddr")
            if protocol == "tcp":
                opts.append("fork")
        elif relay_type == "client":
            clientOpts = node["clientOpts"]
            protocol = node[relay_type + "Opts"]["protocol"]
            opts.append("%s-connect:%s:%d" %
                        (protocol, clientOpts["host"], clientOpts["port"]))
            if protocol == "tcp":
                opts.append("forever")
                opts.append("interval=10")

            if clientOpts["enableSourcePort"]:
                opts.append("sourceport={}".format(clientOpts["sourceport"]))

        else:
            raise RelayError("Unknow Type: %s" % (relay_type))

        return opts

    def install(self, start=True):
        self._install_bin()
        self._install_service()
        if start:
            self.start()

    def _install_bin(self):
        self._logger.info(
            "Installing '{}' executable".format(self.service_name))
        output_str = self.RELAY_BIN_TMPL.format(
            INSTANCE_ID=self["id"],
            OPTIONS=self.get_nodes_opts()
        )
        filepath = os.path.join(self.RELAY_BIN_DIR, self.service_name)
        with open(filepath, "w") as f:
            f.write(output_str)
            sh.chmod("+x", filepath)

    def _install_service(self):
        self._logger.info("Installing '%s' service" % (self.service_name))
        output_str = self.INIT_SCRIPT_TMPL.safe_substitute(
            INSTANCE_ID=self["id"])
        filepath = os.path.join(self.INITD_DIR, self.service_name)
        with open(filepath, "w") as f:
            f.write(output_str)
            sh.chmod("+x", filepath)

    @property
    def service_name(self):
        return "relay-{}".format(self.get("id"))

    def uninstall(self):
        """Uninstall relay service script from /etc/init.d and /etc/defaults"""
        self.stop()
        try:
            sh.rm("-rf", os.path.join(self.INITD_DIR, self.service_name))
            sh.rm("-rf", os.path.join(self.DEFAULT_DIR, self.service_name))
        except ErrorReturnCode:
            self._logger.warning("Uninstall %s failed" % (self.service_name))

        return 0

    def start(self):
        """Start relay service
           Returns:
               exit_code(int): Return exit code
        """
        try:
            output = sh.service(self.service_name, "start")
            self._logger.info(
                "Relay service '%s' started" % (self.service_name))
            return output.exit_code
        except ErrorReturnCode as e:
            self._logger.info(
                "Relay service '%s' started failed" % (self.service_name))
            self._logger.debug(str(e))
            return e.exit_code

    def stop(self):
        """Stop relay service
           Returns:
               exit_code(int): Return exit code
        """
        try:
            output = sh.service(self.service_name, "stop")
            self._logger.info(
                "Relay service '%s' stopped" % (self.service_name))
            return output.exit_code
        except ErrorReturnCode as e:
            self._logger.info(
                "Relay service '%s' stopped failed" % (self.service_name))
            self._logger.debug(str(e))
            return e.exit_code

    def restart(self):
        """Restart relay service
           Returns:
               exit_code(int): Return exit code
        """
        try:
            output = sh.service(self.service_name, "restart")
        except ErrorReturnCode as e:
            return e.exit_code
        self._logger.info("Relay service '%s' restarted" % (self.service_name))
        return output.exit_code


class Relays(Model):

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = RELAY_SCHEMA
        kwargs["model_cls"] = Relay
        super(Relays, self).__init__(*args, **kwargs)

    def install_all(self):
        for relay in self.getAll():
            relay.install(start=relay["enable"])

    def add(self, data):
        raise RuntimeError("Not support Add method")

    def remove(self, id):
        raise RuntimeError("Not support Remove method")

    def update(self, id, newObj):
        """Update an exist ralay,\
           it will restart relay port if update successfully > 0
           Returns:
               newLogProfile(logprofile): udpated logprofile
           Raises:
               ValueError: if device list contains invaild entry
        """
        relay = self.get(id)
        if relay is None:
            return None

        newRelay = super(Relays, self).update(id=id, newObj=newObj)
        relay.uninstall()
        newRelay.install(start=newRelay["enable"])

        return newRelay
