#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import sys
import glob
import unittest

from mock import Mock
from mock import patch
from mock import mock_open
from mock import ANY
from mock import call

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from relay import Relay
    from relay import Relays
    from relay import RelayError
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)


class TestRelayClass(unittest.TestCase):

    def setUp(self):
        self.relay = Relay({
            "id": 1,
            "enable": False,
            "totalInactivityTimeout": 360,
            "nodes": [
              {
                "type": "server",
                "serverOpts": {
                  "protocol": "udp",
                  "port": 55688
                }
              },
              {
                "type": "client",
                "clientOpts": {
                  "protocol": "tcp",
                  "port": 55688,
                  "host": "192.168.0.100",
                  "enableSourcePort": False,
                  "sourceport": 55688
                }
              }
            ]
        })

    def tearDown(self):
        pass

    @patch("relay.sh")
    def test__install_service(self, sh):
        """Install system service, should write to /etc/init.d/relay-$id"""
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.relay._install_service()
        m.assert_called_once_with(
            "/etc/init.d/relay-%d" % (self.relay["id"]), "w")
        h = m()
        h.write.assert_called_once_with(ANY)

    @patch("relay.sh")
    def test__install_bin(self, sh):
        """Install executable, should write to /bin/usr/relay-$id"""
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.relay._install_bin()
        filepath = "/usr/bin/relay-{}".format(self.relay["id"])
        m.assert_called_once_with(filepath, "w")
        h = m()
        h.write.assert_called_once_with(ANY)
        sh.chmod.assert_called_once_with("+x", filepath)

    def test_get_nodes_opts(self):
        """Get nodes options, should return a options string for socat"""
        pass

    def test__get_node_opts_1(self):
        """Given a serial type node, should call serial opts"""
        node = {
            "type": "serial"
        }
        self.relay._get_serial_opts = Mock()
        self.relay._get_node_opts(node)
        self.relay._get_serial_opts.assert_called_once_with(node)

    def test__get_node_opts_2(self):
        """Given a non-serial type node, should call socket opts"""
        node = {
            "type": "serial"
        }
        self.relay._get_socket_opts = Mock()
        self.relay._get_socket_opts(node)
        self.relay._get_socket_opts.assert_called_once_with(node)

    def test__get_serial_opts(self):
        node = {
            "type": "server",
            "serialOpts": {
                "dev": "/dev/ttyM0",
                "baudrate": 115200,
                "stop": 1,
                "data": 8,
                "parity": "none",
                "enableFramer": False
            }
        }
        result = self.relay._get_serial_opts(node)
        self.assertIn("echo=0", result)
        self.assertIn("file:%s" % (node["serialOpts"]["dev"]), result)
        self.assertIn("cs%d" % (node["serialOpts"]["data"]), result)
        self.assertIn("cstopb=0", result)
        self.assertIn(
            "waitlock=/var/run/%s.lock" %
            (node["serialOpts"]["dev"].replace("/", "_")), result)

    def test__get_serial_opts_general_framer(self):
        node = {
            "type": "server",
            "serialOpts": {
                "dev": "/dev/ttyM0",
                "baudrate": 115200,
                "stop": 1,
                "data": 8,
                "parity": "none",
                "enableFramer": True,
                "framerOpts": {
                    "framerType": "general",
                    "noinputTimeoutMS": 20,
                    "overflowSize": 100
                }
            }
        }
        result = self.relay._get_serial_opts(node)
        self.assertIn(
            "EXEC:\"mxserialframer " +
            "--device=/dev/ttyM0 --config=115200-8-n-1 " +
            "--noinput-timeout-ms=20 --overflow-size=100\"",
            result)

    def test__get_serial_opts_dnp3_framer(self):
        node = {
            "type": "server",
            "serialOpts": {
                "dev": "/dev/ttyM0",
                "baudrate": 115200,
                "stop": 1,
                "data": 8,
                "parity": "none",
                "enableFramer": True,
                "framerOpts": {
                    "framerType": "dnp3",
                    "noinputTimeoutMS": 20,
                    "overflowSize": 100
                }
            }
        }
        result = self.relay._get_serial_opts(node)
        self.assertIn(
            "EXEC:\"mxserialdnp3framer " +
            "--config=115200-8-n-1 " +
            "--noinput-timeout-ms=20 --overflow-size=100 /dev/ttyM0\"",
            result)

    def test__get_socket_opts_1(self):
        """Given a server node, should get a protocol-listen opts string"""
        node = {
            "type": "server",
            "serverOpts": {
                "protocol": "udp",
                "port": 1234
            }
        }
        result = self.relay._get_socket_opts(node)
        self.assertIn(
            "%s-listen:%d" % (node["serverOpts"]["protocol"],
                              node["serverOpts"]["port"]),
            result)
        self.assertIn("reuseaddr", result)
        self.assertNotIn("so-bindtodevice=wwan0", result)
        self.assertNotIn("fork", result)

    def test__get_socket_opts_2(self):
        """Given a client node, should get a protocol-connect opts string"""
        node = {
            "type": "client",
            "clientOpts": {
                "protocol": "tcp",
                "port": 55688,
                "host": "192.168.1.1",
                "enableSourcePort": False,
                "sourceport": 55688
            }
        }
        result = self.relay._get_socket_opts(node)
        self.assertIn(
            "%s-connect:%s:%d" % (node["clientOpts"]["protocol"],
                                  node["clientOpts"]["host"],
                                  node["clientOpts"]["port"]),
            result)

    def test__get_socket_opts_3(self):
        """Given a invaild node, should raise RelayError"""
        node = {
            "type": "hahaha",
        }

        with self.assertRaises(RelayError):
            self.relay._get_socket_opts(node)

    def test_service_name(self):
        self.assertEqual(self.relay.service_name, "relay-1")

    @patch("relay.sh")
    @patch("relay.Relay.stop")
    def test_uninstall(self, stop, sh):
        self.relay.uninstall()
        self.assertEqual(
            sh.rm.call_args_list,
            [
                call("-rf", "/etc/init.d/relay-1"),
                call("-rf", "/etc/default/relay-1")
            ]
        )
        stop.assert_called_once_with()

    @patch("relay.sh")
    def test_start(self, sh):
        self.relay.start()
        sh.service.assert_called_once_with(self.relay.service_name, "start")

    @patch("relay.sh")
    def test_stop(self, sh):
        self.relay.stop()
        sh.service.assert_called_once_with(self.relay.service_name, "stop")

    @patch("relay.sh")
    def test_restart(self, sh):
        self.relay.restart()
        sh.service.assert_called_once_with(self.relay.service_name, "restart")

    @patch("relay.Relay._install_service")
    @patch("relay.Relay._install_bin")
    @patch("relay.Relay.start")
    def test_install(self, start, install_bin, install_service):
        """Install a relay, should call install service and default"""
        self.relay.install()
        install_bin.assert_called_once_with()
        install_service.assert_called_once_with()
        start.assert_called_once_with()


class TestRelaysClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)
        self.relays = Relays(name="relays", path=self.root_path)

    def tearDown(self):
        pass

    def test_add(self):
        """Not support add"""
        with self.assertRaises(RuntimeError):
            self.relays.add({})

    def test_remove(self):
        """Not support remove"""
        with self.assertRaises(RuntimeError):
            self.relays.remove(1)

    @patch("relay.Relay.install")
    @patch("relay.Relay.uninstall")
    def test_update(self, uninstall, install):
        """Update a relay"""
        relay = {
            "name": "123",
            "enable": False,
            "totalInactivityTimeout": 360,
            "nodes": [
              {
                "name": "Upstream",
                "type": "server",
                "serverOpts": {
                  "protocol": "udp",
                  "port": 55688
                }
              },
              {
                "name": "Downstream",
                "type": "client",
                "clientOpts": {
                  "protocol": "tcp",
                  "port": 55688,
                  "host": "192.168.0.100",
                  "enableSourcePort": False,
                  "sourceport": 55688
                }
              }
            ]
        }

        new_relay = self.relays.update(1, relay)
        self.assertIsInstance(new_relay, Relay)
        install.assert_called_once_with(start=False)
        uninstall.assert_called_once_with()

    @patch("relay.Relay.install")
    def test_install_all(self, install):
        self.relays.install_all()
        self.assertListEqual([
            call(start=False),
            call(start=False)],
            install.call_args_list
        )
        # install.assert_called_once_with(start=False)

if __name__ == "__main__":
    unittest.main()
