import logging
import sh

from sh import ErrorReturnCode
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, Any, All, Length
_logger = logging.getLogger("sanji.service")

SERVICE_SCHEMA = Schema({
    "id": int,
    "name": All(Any(unicode, str), Length(1, 255)),
    Required("enable"): bool
}, extra=REMOVE_EXTRA)


class Service(object):
    _logger = logging.getLogger("sanji.service")

    def __init__(self, service_name):
        self.service_name = service_name

    def __getattr__(self, command):
        def do_command(bg=False):
            try:
                output = sh.service(self.service_name, command, _bg=bg)
                return output.exit_code
            except ErrorReturnCode as e:
                if command == "status":
                    return 1
                self._logger.info(
                    "Service '%s' %s failed" %
                    (self.service_name, command))
                self._logger.debug(str(e))
                return e.exit_code

        def register(bg=False):
            try:
                if command == "register":
                    output = sh.insserv(self.service_name)
                elif command == "deregister":
                    output = sh.insserv("-r", self.service_name)
                return output.exit_code
            except ErrorReturnCode as e:
                self._logger.info(
                    "Service '%s' %s failed" %
                    (self.service_name, command))
                self._logger.debug(str(e))
                return e.exit_code

        if command in ["start", "restart", "stop", "status"]:
            return do_command
        elif command in ["register", "deregister"]:
            return register

        return super(Service, self).__getattr__(command)


class ServiceError(Exception):
    pass


class SService(Model):

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = SERVICE_SCHEMA
        super(SService, self).__init__(*args, **kwargs)

        self.services = {}
        for service in self.getAll():
            self.services[service["name"]] = Service(service["name"])
            self.apply(service["name"], service["enable"])

    def apply(self, name, enable=True):
        """Update register information and start or stop it."""
        # TODO: bg=True required?
        if enable is True:
            self.services[name].register()
            self.services[name].start()
        else:
            self.services[name].stop()
            self.services[name].deregister()

    def add(self, data):
        raise RuntimeError("Not support Add method")

    def remove(self, id):
        raise RuntimeError("Not support Remove method")

    def _get_status(self, name):
        # status may raise exception if service not inserted
        try:
            rc = self.services[name].status()
        except:
            rc = 1
        return True if rc == 0 else False

    def get_all(self):
        services = self.getAll()
        for entry in services:
            entry["status"] = self._get_status(entry["name"])
        return services

    def get(self, id=0, name=None):
        """Get by id or name
        """
        service = {}
        if id == 0 and name is None:
            return None
        elif id != 0:
            service = super(SService, self).get(id=id)
        else:
            for entry in self.getAll():
                if entry["name"] == name:
                    service = entry
                    break
            else:
                return None

        service["status"] = self._get_status(service["name"])
        return service

    def update(self, id=0, name=None, newObj=None):
        """Update an exist service status.

           Raises:
               TODO
        """
        service = self.get(id=id, name=name)
        if service is None:
            return None

        if "name" in newObj and service["name"] != newObj["name"]:
            raise ServiceError("Can't change service's name")

        if service["enable"] != newObj["enable"]:
            service["enable"] = newObj["enable"]
            updated = super(SService, self).update(
                id=service["id"], newObj=service)
            if updated is None:
                return None
            self.apply(updated["name"], updated["enable"])
            updated["status"] = self._get_status(updated["name"])
            return updated

        service["status"] = self._get_status(service["name"])
        return service
