#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import logging
import time
import sh
from sanji.core import Sanji
from sanji.core import Route
from sanji.connection.mqtt import Mqtt
from sanji.model_initiator import ModelInitiator
from voluptuous import Schema, Required, Optional, All, Length, REMOVE_EXTRA


class Firmware(Sanji):
    """
    A model to handle firmware upgrade.

    Attributes:
        model: database with json format.
    """
    _logger = logging.getLogger("sanji.upgrade")

    UPLOAD_SCHEMA = Schema({
        Required("url"): All(str, Length(1, 4096)),
        Required("physicalPath"): All(str, Length(1, 4096)),
        "headers": {}
    }, extra=REMOVE_EXTRA)

    UPGRADE_SCHEMA = Schema({
        Optional("delay"): int,
        Required("file"): UPLOAD_SCHEMA
    }, extra=REMOVE_EXTRA)

    def init(self, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(__file__))
        try:
            self.load(path_root)
        except:
            self.stop()
            raise IOError("Cannot load any configuration.")

    def run(self):
        if "upgrading" in self.model.db:
            if self.model.db["upgrading"] == -1:
                self.publish.event.put(
                    "/system/firmware",
                    data={"code": "FW_UPGRADE_FAIL", "type": "event"})
            else:
                self.publish.event.put(
                    "/system/firmware",
                    data={"code": "FW_UPGRADE_SUCCESS", "type": "event"})
            self.model.db.pop("upgrading")
            self.save()

    def load(self, path):
        """
        Load the configuration. If configuration is not installed yet,
        initialise them with default value.

        Args:
            path: Path for the bundle, the configuration should be located
                under "data" directory.
        """
        self.model = ModelInitiator("firmware", path, backup_interval=-1)
        if self.model.db is None:
            raise IOError("Cannot load any configuration.")
        self.save()

    def save(self):
        """
        Save and backup the configuration.
        """
        self.model.save_db()
        self.model.backup_db()

    def upgrade(self, file):
        # set flags to show the upgrading status
        """
        self.publish.event.put(
            "/system/firmware",
            data={"code": "FW_UPGRADING", "type": "event"})
        """
        self.model.db["upgrading"] = 1
        self.save()

        try:
            self._logger.info("Upgrading...")
            sh.sys_upgrade(file)
            self._logger.info("Upgrading success, reboot now.")
            self.model.db["upgrading"] = 0
        except:
            self._logger.error("Upgrading failed, please check if the file is"
                               " correct.")
            self._logger.error("Reboot now to recover the system.")
            self.model.db["upgrading"] = -1
        self.save()

        sh.reboot()

    def setdef(self):
        # TODO: stop the services that may have side effect when setdef
        self.model.db["defaulting"] = 1
        self.save()

        time.sleep(1)
        try:
            sh.setdef()
            self._logger.info(
                "Resetting to factory default success, reboot now.")
            self.model.db["defaulting"] = 0
            self.save()
            sh.reboot()
        except:
            self._logger.error("Resetting failed.")
            self.model.db["defaulting"] = -1
            self.save()

    @Route(methods="post", resource="/system/firmware", schema=UPGRADE_SCHEMA)
    def post(self, message, response):
        """
        Only save the configuration if server updated.
        {
            "delay": 3,
            "file": {
                "physicalPath": "/xxx/xxx/xxx",
                "url": "http://192.168.31.81:5000/api/v1/download/1234",
                "headers": {
                    "accessToken": "MOXAMOXA-TEST-SYSTEM"
                }
        }
        """
        response()

        firmware_file = message.data["file"]["physicalPath"]
        if "delay" in message.data and message.data["delay"] > 0:
            time.sleep(message.data["delay"])

        try:
            self.upgrade(firmware_file)
        except Exception, e:
            self._logger.error(e, exc_info=True)
            sh.reboot()
        return


if __name__ == "__main__":  # pragma: no cover
    FORMAT = "%(asctime)s - %(levelname)s - %(lineno)s - %(message)s"
    logging.basicConfig(level=0, format=FORMAT)
    _logger = logging.getLogger("sanji.firmware")

    firmware = Firmware(connection=Mqtt())
    firmware.start()
