#!/usr/bin/env python
# -*- coding: UTF-8 -*-


import os
import sys
import logging
import unittest
import glob

from mock import patch
from mock import MagicMock
from sanji.connection.mockup import Mockup

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../")
    os.environ["PATH"] = os.path.dirname(os.path.realpath(__file__)) \
        + ":" + os.environ["PATH"]
    from firmware import Firmware
except ImportError as e:
    print os.path.dirname(os.path.realpath(__file__)) + "/../"
    print sys.path
    print e
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)

dirpath = os.path.dirname(os.path.realpath(__file__))
logger = logging.getLogger()


class TestFirmwareClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../")
        self.jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        self.backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in self.jsons + self.backups:
            os.unlink(file)

        self.name = "firmware"
        self.bundle = Firmware(connection=Mockup())
        self.bundle.publish = MagicMock()

    def tearDown(self):
        self.bundle.stop()
        self.bundle = None

        files = glob.glob(os.path.join(self.root_path, "data/*.json")) + \
            glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for prevConfig in files:
            try:
                os.unlink(prevConfig)
            except:
                pass

    def test__init__no_conf(self):
        """
        init: no configuration file
        """
        with self.assertRaises(IOError):
            with patch("firmware.ModelInitiator") as mock_modelinit:
                mock_modelinit.side_effect = IOError
                self.bundle.init()

    def test__run(self):
        """
        run: normal
        """
        self.bundle.run()

    def test__run__upgrading_success(self):
        """
        run: upgrading success
        """
        self.bundle.model.db["upgrading"] = 1
        self.bundle.run()

    def test__run__upgrading_failed(self):
        """
        run: upgrading failed
        """
        self.bundle.model.db["upgrading"] = -1
        self.bundle.run()

    def test__load__current_conf(self):
        """
        load: load current configuration
        """
        self.bundle.load(dirpath + "/../")
        self.assertEqual(self.bundle.model.db, {})

    def test__save(self):
        """
        save: tested in init()
        """
        # Already tested in init()
        pass

    @patch("firmware.time.sleep")
    @patch("firmware.sh.sys_upgrade")
    @patch("firmware.sh.reboot")
    def test__upgrade(self, mock_reboot, mock_upgrade, mock_sleep):
        """
        upgrade: success
        """
        mock_upgrade.return_value = 0

        self.bundle.upgrade("firmware.frm")
        self.assertEqual(0, self.bundle.model.db["upgrading"])

    @patch("firmware.time.sleep")
    @patch("firmware.sh.sys_upgrade")
    @patch("firmware.sh.reboot")
    def test__upgrade__failed(self, mock_reboot, mock_upgrade, mock_sleep):
        """
        upgrade: failed
        """
        mock_upgrade.return_value = 1
        mock_upgrade.side_effect = Exception("error")

        self.bundle.upgrade("firmware.frm")
        self.assertEqual(-1, self.bundle.model.db["upgrading"])

    @patch("firmware.time.sleep")
    @patch("firmware.sh.setdef")
    @patch("firmware.sh.reboot")
    def test__setdef(self, mock_reboot, mock_setdef, mock_sleep):
        """
        setdef: success
        """
        self.bundle.setdef()
        self.assertEqual(0, self.bundle.model.db["defaulting"])

    @patch("firmware.time.sleep")
    @patch("firmware.sh.setdef")
    def test__setdef__failed(self, mock_setdef, mock_sleep):
        """
        setdef: failed
        """
        mock_setdef.side_effect = Exception("error")

        self.bundle.setdef()
        self.assertEqual(-1, self.bundle.model.db["defaulting"])


if __name__ == "__main__":
    FORMAT = "%(asctime)s - %(levelname)s - %(lineno)s - %(message)s"
    logging.basicConfig(level=20, format=FORMAT)
    logger = logging.getLogger("Firmware Test")
    unittest.main()
