#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import logging
import itertools
from sanji.core import Sanji
from sanji.core import Route
from modbus.equipment import get_equipment_name_list
from modbus.equipment import get_logprofile_deps
from modbus.equipment import Equipments
from modbus.equipment import EquipmentError
from modbus.equipment import HTTPError
from modbus.interface import Interfaces
from modbus.gateway import Gateways
from modbus.client import read_from_rtu
from modbus.client import read_from_tcp
from modbus import restart_mxmodbusmaster

from voluptuous import Schema, Required, REMOVE_EXTRA, All, Length, Range, Any

_logger = logging.getLogger("sanji.modbus.index")


class Index(Sanji):

    IF_CONFIG_FILE = "/etc/mxmodbusmaster/configuration"

    EXPORT_SCHEMA = Schema({
        Required("url"): All(str, Length(1, 4096)),
        "headers": dict
    }, extra=REMOVE_EXTRA)

    IMPORT_SCHEMA = Schema({
        Required("file"): EXPORT_SCHEMA
    }, extra=REMOVE_EXTRA)

    TEST_SCHEMA = Schema({
        Required("type"): Any("modbus-tcp", "modbus-rtu"),
        Required("equipmentTemplateId"): int,
        "interfaceId": int,
        Required("deviceId"): Range(0, 255),
        "config": Schema({
            Required("host"): All(Any(unicode, str), Length(1, 255)),
            Required("port"): All(int, Range(1, 65535))
        }, extra=REMOVE_EXTRA)
    }, extra=REMOVE_EXTRA)

    def init(self, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(__file__))
        # Init equipment templates, laod and export config
        self.equipments = Equipments(name="equipments", path=path_root)

        # Init interface config, laod and export config
        self.gateways = Gateways(name="gateways", path=path_root)
        self.interfaces = Interfaces(
            name="interfaces", path=path_root, gateways=self.gateways)

        self.interfaces.set_equipments(self.equipments)
        self.equipments.set_interfaces(self.interfaces)

        try:
            self.equipments.clean_templates()
            self.equipments.generate_configs()
        except Exception, e:
            _logger.error(str(e))

        try:
            self.interfaces.generate_configs()
        except Exception, e:
            _logger.error(str(e))

        restart_mxmodbusmaster()

    # Equipments
    @Route(methods="get", resource="/mxc/modbus/equipment-templates")
    def get_all_equipments(self, message, response):
        return response(data=self.equipments.getAll())

    @Route(methods="post", resource="/mxc/modbus/equipment-templates")
    def post_equipment(self, message, response):
        try:
            newTemplate = self.equipments.add(message.data)
            return response(data=newTemplate)
        except (EquipmentError, RuntimeError) as e:
            return response(code=400, data={"message": str(e)})

    @Route(methods="delete", resource="/mxc/modbus/equipment-templates")
    def delete_equipments(self, message, response):
        try:
            for tid in message.data:
                self.equipments.remove(id=tid)
            return response(data={"message": "success"})
        except EquipmentError as e:
            return response(code=400, data={"message": str(e)})

    @Route(methods="get", resource="/mxc/modbus/equipment-templates/:id")
    def get_equipment(self, message, response):
        template = self.equipments.get(id=message.param["id"])
        if template is None:
            return response(code=404)

        return response(data=template)

    def check_tags_deps(self, equipment_tmpl):
        interfaces = self.interfaces.get_by_eqname(equipment_tmpl["name"])
        equipment_name_list = [get_equipment_name_list(interface["deviceList"])
                               for interface in interfaces]
        # fastest flatten list
        # ref: http://stackoverflow.com/a/953097/1301231
        flatten_list = list(itertools.chain.from_iterable(equipment_name_list))
        return self.check_logprofile_deps(flatten_list)

    @Route(methods="put", resource="/mxc/modbus/equipment-templates/:id")
    def update_equipment(self, message, response):
        # Only check tagList update for logprfile deps
        # delete equipment template will check interface deps first
        if "tagList" in message.data:
            equipment_tmpl = self.equipments.get(id=int(message.param["id"]))
            if equipment_tmpl is None:
                return response(code=404)
            deps_list = self.check_tags_deps(equipment_tmpl)
            if len(deps_list) > 0:
                err_msg = "Equipment Name: %s in use. " % \
                           (" ,".join(deps_list)) + \
                          "Please uncheck all of it's tags from " + \
                          "Log Profile before saving it."
                response(
                    code=400,
                    data={"message": err_msg})
                return

        try:
            updatedTemplate = self.equipments.update(
                id=int(message.param["id"]), newObj=message.data)
            if updatedTemplate is None:
                return response(code=404)
            return response(data=updatedTemplate)
        except EquipmentError as e:
            return response(code=400, data={"message": str(e)})

    @Route(methods="delete", resource="/mxc/modbus/equipment-templates/:id")
    def delete_equipment(self, message, response):
        try:
            self.equipments.remove(id=int(message.param["id"]))
            return response(data={"message": "success"})
        except EquipmentError as e:
            return response(code=400, data={"message": str(e)})

    @Route(methods="post", resource="/mxc/modbus/export-equipment-templates")
    def export_equipment_templates(
            self, message, response, schema=EXPORT_SCHEMA):
        headers = message.data.get("headers", {})
        ids = message.query.get("ids", None)

        if ids is not None and ids is not True and len(ids) > 0:
            ids = [int(n) for n in ids.split(",")]
            _logger.info("Export templates: %s" % ids)
        else:
            ids = None

        try:
            url = self.equipments.export_to_http(
                url=message.data["url"],
                headers=headers,
                ids=ids
            )
            return response(data={"url": url})
        except HTTPError as he:
            _logger.error("Export to http error: %s" % str(he))
            return response(code=500, data={"message": str(he)})
        except Exception as e:
            _logger.error("Export to http error")
            _logger.error(str(e))
            raise e

    @Route(methods="put", resource="/mxc/modbus/import-equipment-templates")
    def import_equipment_templates(self, message, response):
        headers = message.data["file"].get("headers", {})
        override = message.data.get("override", False)
        try:
            self.equipments.import_from_http(
                url=message.data["file"]["url"],
                headers=headers,
                override=override)
            _logger.info("Equipment Templates imported.")
            response()
        except HTTPError as he:
            _logger.error("Import from http error: %s" % str(he))
            return response(code=500, data={"message": str(he)})
        except Exception as e:
            _logger.error("Import from http error")
            _logger.error(str(e))
            return response(code=500, data={"message": "Import failed."})

    # Interfaces
    @Route(methods="get", resource="/mxc/modbus/interfaces")
    def get_all_interfaces(self, message, response):
        return response(data=self.interfaces.getAll(embed_status=True))

    @Route(methods="post", resource="/mxc/modbus/interfaces")
    def post_interface(self, message, response):
        newInterface = self.interfaces.add(message.data)
        return response(data=newInterface)

    @Route(methods="delete", resource="/mxc/modbus/interfaces")
    def delete_interfaces(self, message, response):
        for iid in message.data:
            interface = self.interfaces.get(id=int(iid))
            if interface is None:
                continue
            equip_name_list = get_equipment_name_list(
                interface["deviceList"])
            deps_list = self.check_logprofile_deps(equip_name_list)
            if len(deps_list) > 0:
                err_msg = "Equipment Name: %s in use" % (" ,".join(deps_list))
                response(
                    code=400,
                    data={"message": err_msg})
                return
            self.interfaces.remove(id=iid)
        return response(data={"message": "success"})

    @Route(methods="get", resource="/mxc/modbus/interfaces/:id")
    def get_interface(self, message, response):
        interface = self.interfaces.get(id=int(message.param["id"]))
        if interface is None:
            return response(code=404)
        return response(data=interface)

    def check_logprofile_deps(self, equip_name_list):
        # Get logprofile
        resp = self.publish.direct.get("/mxc/logprofile")
        logprofile = resp.data

        # Get logprofile deps
        deps_list = get_logprofile_deps(logprofile, equip_name_list)
        return deps_list

    @Route(methods="put", resource="/mxc/modbus/interfaces/:id")
    def update_interface(self, message, response):
        if "deviceList" in message.data:
            # Get current deviceList diff with new deviceList
            new_equip_name_list = get_equipment_name_list(
                message.data["deviceList"])
            interface = self.interfaces.get(id=int(message.param["id"]))
            equip_name_list = get_equipment_name_list(
                interface["deviceList"])
            diff_equip_name_list = frozenset(equip_name_list).difference(
                frozenset(new_equip_name_list))

            deps_list = self.check_logprofile_deps(diff_equip_name_list)
            if len(deps_list) > 0:
                err_msg = "Equipment Name: %s in use" % (" ,".join(deps_list))
                response(
                    code=400,
                    data={"message": err_msg}
                )
                return

        newInterface = self.interfaces.update(
            id=int(message.param["id"]), newObj=message.data)
        if newInterface is None:
            return response(code=404)
        return response(data=newInterface)

    @Route(methods="delete", resource="/mxc/modbus/interfaces/:id")
    def delete_interface(self, message, response):
        interface = self.interfaces.get(id=int(message.param["id"]))
        equip_name_list = get_equipment_name_list(
            interface["deviceList"])
        deps_list = self.check_logprofile_deps(equip_name_list)
        if len(deps_list) > 0:
            err_msg = "Equipment: %s in use" % \
                (" ,".join(deps_list))
            response(
                code=400,
                data={"message": err_msg}
            )
            return

        self.interfaces.remove(id=int(message.param["id"]))
        return response(data={"message": "success"})

    @Route(methods="post", resource="/mxc/modbus/test")
    def post_test(self, message, response, schema=TEST_SCHEMA):
        eqtmpl = self.equipments.get(
            id=int(message.data["equipmentTemplateId"]))
        if eqtmpl is None:
            return response(
                code=404, data={"message": "Equipment Template not found."})
        eqtmpl_path = os.path.join(Equipments.EQ_CONFIG_DIR, eqtmpl.filename)
        report = ""

        if message.data["type"] == "modbus-tcp":
            report = read_from_tcp(
                eqtmpl_path,
                message.data["config"]["host"],
                message.data["config"]["port"],
                message.data["deviceId"]
            )
        else:
            interface = self.interfaces.get(id=message.data["interfaceId"])
            if interface is None:
                return response(
                    code=404, data={"message": "Interface not found."})
            report = read_from_rtu(
                eqtmpl_path, interface,
                self.gateways,
                message.data["deviceId"]
            )

        response(data=report)

    @Route(methods="get", resource="/mxc/modbus/equipments")
    def get_equipments(self, message, response):
        eqs = []
        for modbus_interface in self.interfaces.getAll():
            for device in modbus_interface["deviceList"]:
                template = self.equipments.get_by_name(
                    device["equipmentTemplate"])
                if template is None:
                    _logger.warning(
                        "{} template can't be found."
                        .format(device["equipmentTemplate"])
                    )
                    continue

                tags = [
                    {
                        "name": tag["name"],
                        "description": tag.get("description", "")
                    }
                    for tag in template["tagList"]
                ]

                eqs.append({
                    "equipmentName": device["equipmentName"],
                    "equipmentTags": tags
                })

        return response(data=eqs)

if __name__ == "__main__":
    from sanji.connection.mqtt import Mqtt
    FORMAT = "%(asctime)s - %(levelname)s - %(lineno)s - %(message)s"
    logging.basicConfig(level=0, format=FORMAT)
    _logger = logging.getLogger("sanji.modbus")
    logging.getLogger("sh").setLevel(logging.WARN)
    index = Index(connection=Mqtt())
    index.start()
