import logging
import sh
import os
import time
import simplejson as json
from jsonlibconfig import encoder
from threading import Lock
_logger = logging.getLogger("sanji.modbus")

__fetch_status_lock = Lock()


def fetch_status(status_file="/var/mxmodbusmaster/equipment-status"):
    with __fetch_status_lock:
        try:
            sh.rm("-rf", status_file, _timeout=5)
            master_pid = int(sh.pidof("mxmodbusmaster").strip())
            sh.kill("-USR1", master_pid)
        except Exception as e:
            _logger.error("fetch mxmodbusmaster status failed")
            _logger.debug(str(e))
            return None

        # wait until timeout or file has been generated
        timeout = 5
        while not os.path.exists(status_file) and timeout > 0:
            time.sleep(1)
            timeout = timeout - 1

        if not os.path.exists(status_file):
            _logger.warning("mxmodbusmaster status timeout")
            return None

        with open(status_file, "r") as f:
            return json.load(f)


def restart_mxmodbusmaster():
    """Restart mxmodbusmaster via system service
        Returns:
            True: if success, vice versa.
    """
    try:
        sh.service("mxmodbusmaster", "restart", _bg=True)
    except Exception as e:
        _logger.error("service mxmodbusmaster restart failed")
        _logger.debug(str(e))
        return False

    _logger.info("service mxmodbusmaster restart")
    return True


def restart_mxmodbusgateway():
    """Restart mxmodbusgateway via system service
        Returns:
            True: if success, vice versa.
    """
    try:
        sh.service("mxmodbusgateway", "restart", _bg=True)
    except Exception as e:
        _logger.error("service mxmodbusgateway restart failed")
        _logger.debug(str(e))
        return False

    _logger.info("service mxmodbusgateway restart")
    return True


class Libconfig(object):
    def __init__(self, *args, **kw):
        """Create a Template instance
            Args:
                jsonconfig(dict): Input json template configuration
        """
        self._libconfig = None
        super(Libconfig, self).__init__(*args, **kw)

    def _convert(self):
        """Mapping restful interface to libconfig configuration"""
        raise NotImplementedError()

    def toLibconfig(self, force=True, indent=2):
        """Generate libconfig string from settings
            Args:
                force(boolean): If it's already mapped restful <> libconfig,
                    it won't be mapped again.
            Returns:
                libconfig(string): libconfig string
        """
        if self._libconfig is not None and not force:
            return self._libconfig

        self._libconfig = self._convert()
        return encoder.dumps(self._libconfig, indent=indent)

    def save(self, path):
        """Save libconfig to specified path with default name
            Args:
                path(str): output full filename
            Returns:
                dest(str): output file destination
            Raises:
                IOError
        """
        with open(path, "w") as fp:
            fp.write(self.toLibconfig())

        return path
