import logging
import glob
import os
import re
import sh
import requests
import datetime
import json

from modbus import Libconfig
from modbus import restart_mxmodbusmaster
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, Any, All, Length, \
    Range, message, truth, Invalid

_logger = logging.getLogger("sanji.modbus.equipments")
NAME_PATTERN = re.compile(ur'^[A-Za-z0-9_-]+$')


@message('not a vaild name', cls=Invalid)
@truth
def IsValidName(v):
    return NAME_PATTERN.match(v)


EQ_TAG_SCHEMA = Schema({
    # tag-list.id
    Required("name"): All(
        Any(unicode, str), IsValidName(), Length(min=2, max=255)),
    Required("function"): Any("read-coils",
                              "read-input-registers",
                              "read-discrete-inputs",
                              "read-holding-registers"),
    Required("address"): All(int, Range(0, 65535)),
    Required("quantity"): All(int, Range(1, 65535)),
    Required("type"): Any("uint16", "uint32", "uint64", "int16", "int32",
                          "int64", "float32", "float64", "string", "boolean"),
    "description": All(Any(unicode, str), Length(0, 65535)),
    "unit": All(Any(unicode, str), Length(0, 255)),
    "enableByteOrder": bool,
    "byteOrder": Any("BA", "CDAB", "DCBA"),
    "enableInvalidValue": bool,
    "invalidValue": [int],
    "enableAutoScaling": bool,
    "autoScalingMode": Any("intercept", "point"),
    "autoScaling": Any(
        Schema({
            Required("slope"): Any(float, int),
            Required("offset"): Any(float, int),
            Required("sourceMin"): Any(float, int),
            Required("sourceMax"): Any(float, int),
            Required("targetMin"): Any(float, int),
            Required("targetMax"): Any(float, int)
        }, extra=REMOVE_EXTRA),
        Schema({
            Required("slope"): Any(float, int),
            Required("offset"): Any(float, int)
        }, extra=REMOVE_EXTRA),
        Schema({
            Required("sourceMin"): Any(float, int),
            Required("sourceMax"): Any(float, int),
            Required("targetMin"): Any(float, int),
            Required("targetMax"): Any(float, int)
        }, extra=REMOVE_EXTRA)
    )
}, extra=REMOVE_EXTRA)

EQ_SCHEMA = Schema({
    "id": int,
    Required("name"): All(
        Any(unicode, str), IsValidName(), Length(min=3, max=255)),
    "tagList": [EQ_TAG_SCHEMA]
}, extra=REMOVE_EXTRA)


def get_equipment_name_list(deviceList):
    """Get all equipmentNames from deviceList"""
    return [device["equipmentName"] for device in deviceList]


def get_logprofile_deps(logprofile, diff_equip_name_list):
    """Give logprofile dict list and diff equipmentName list returns deps"""
    logprofile_equip_name_list = set()
    for row in logprofile:
        logprofile_equip_name_list = logprofile_equip_name_list.union(
            get_equipment_name_list(row["equipmentList"]))
    deps_list = frozenset(diff_equip_name_list).intersection(
        logprofile_equip_name_list)

    return deps_list


class EquipmentError(Exception):
    pass


class Equipment(Libconfig, dict):

    FILE_EXT = "met"

    def __init__(self, *args, **kwargs):
        super(Equipment, self).__init__(*args, **kwargs)
        self._libconfig = self._convert()

    @property
    def filename(self):
        return ("%d-%s" % (self.get("id"), self.get("name"))) + \
            ".%s" % self.FILE_EXT if self.FILE_EXT else ""


# * `id`: Required, string, criteria `[A-Za-z0-9_-]{3,255}`.
# * `request-list`: Required, object-list.
#     * `function`: Required, string, can be `read-coils`, `read-discrete-inputs`,
#       `read-input-registers`, `read-holding-registers`.
#     * `address`: Required, integer, criteria `0x0000 <= value <= 0xFFFF`.
#     * `quantity`: Required, integer, criteria `value > 0`.
# * `tag-list`: Required, object-list.
#     * `id`: Required, string, criteria `[A-Za-z0-9_-]{2,255}`.
#     * `reference`: Required, integer, criteria
#       `0 <= value < 0x20000 || 0x30000 <= value < 0x50000`.
#     * `quantity`: Required, integer, criteria `value > 0`.
#     * `invalid-value`: Optional, integer-array, criteria `0 <= value <= 0xFFFF`.
#     * `byte-order`: Optional, string, can be `BA`, `CDAB`, `DCBA`.
#     * `type`: Required, string, can be `uint16`, `uint32`, `uint64`, `int16`,
#       `int32`, `int64`, `float32`, `float64`, `string`.
#     * `slope`: Optional, float.
#     * `offset`: Optional, float.

    def convert_request_entry(seld, tag):
        """Input tag json data, returns mxmodbusmaster config"""
        request = {
            "function": tag["function"],
            "address": str.format('0x{:05X}', tag["address"]),
            "quantity": tag["quantity"]
        }

        # Boolean is just an uint16 with quantity 1
        if tag["type"] == "boolean":
            request["quantity"] = 1

        return request

    def convert_tag_entry(self, tag):
        """Input tag json data, returns mxmodbusmaster config"""
        libconfig_tag = {
            "id": tag["name"],
            "quantity": tag["quantity"],
            "type": tag["type"]
        }

        # Boolean is just an uint16 with quantity 1
        if tag["type"] == "boolean":
            libconfig_tag["type"] = "uint16"
            libconfig_tag["quantity"] = 1

        # For address mapping to reference
        # more detail: http://www.csimn.com/CSI_pages/Modbus101.html
        # 0x = Coil = 00000-09999
        # 1x = Discrete Input = 10000-19999
        # 3x = Input Register = 30000-39999
        # 4x = Holding Register = 40000-49999
        # `read-coils`, `read-discrete-inputs`, `read-input-registers`,
        # `read-holding-registers`
        if tag["function"] == "read-coils":
            libconfig_tag["reference"] = str.format("0x{:05X}", tag["address"])

        elif tag["function"] == "read-discrete-inputs":
            libconfig_tag["reference"] = str.format(
                '0x{:05X}', tag["address"] + int("0x10000", 0))

        elif tag["function"] == "read-input-registers":
            libconfig_tag["reference"] = str.format(
                '0x{:05X}', tag["address"] + int("0x30000", 0))

        elif tag["function"] == "read-holding-registers":
            libconfig_tag["reference"] = str.format(
                '0x{:05X}', tag["address"] + int("0x40000", 0))

        else:
            _logger.warning("%s's tag list contains invalid function" %
                            self.get("name"))
            raise EquipmentError("%s's tag list contains invalid function" %
                                 self.get("name"))

        if tag.get("enableInvalidValue", False) is True \
           and "invalidValue" in tag:
                libconfig_tag["invalid-value"] = tag["invalidValue"]

        if tag.get("enableByteOrder", False) is True and "byteOrder" in tag:
            libconfig_tag["byte-order"] = tag["byteOrder"]

        if "unit" in tag:
            libconfig_tag["unit"] = tag["unit"]

        if tag.get("enableAutoScaling", False) is True \
           and "autoScalingMode" in tag:
            if tag["autoScalingMode"] == "intercept":
                libconfig_tag["slope"] = tag["autoScaling"]["slope"]
                libconfig_tag["offset"] = tag["autoScaling"]["offset"]
            elif tag["autoScalingMode"] == "point":
                scaling = tag["autoScaling"]
                libconfig_tag["slope"] = \
                    float(scaling["targetMax"] - scaling["targetMin"]) / \
                    float(scaling["sourceMax"] - scaling["sourceMin"])
                libconfig_tag["offset"] = \
                    (float((-scaling["sourceMin"]) * scaling["targetMax"]) +
                     float(scaling["sourceMin"] * scaling["targetMin"])) / \
                    float(scaling["sourceMax"] - scaling["sourceMin"]) + \
                    float(scaling["targetMin"])
            else:
                raise EquipmentError(
                    "autoScalingMode should be intercept or point")

        return libconfig_tag

    def _convert(self):
        """Mapping restful interface to modbus master's configuration
            Raises:
                EquipmentError
        """
        libconfig = {
            "id": self.get("name"),
            "request-list": [],
            "tag-list": []
        }

        for tag in self.get("tagList"):
            # TODO: Optimize request list (ex: merge)
            request = self.convert_request_entry(tag)
            libconfig["request-list"].append(request)
            libconfig_tag = self.convert_tag_entry(tag)
            libconfig["tag-list"].append(libconfig_tag)

        return libconfig


class HTTPError(Exception):
    pass


class Equipments(Libconfig, Model):

    EQ_CONFIG_DIR = "/etc/mxmodbusmaster/template.d/"

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = EQ_SCHEMA
        kwargs["model_cls"] = Equipment
        super(Equipments, self).__init__(*args, **kwargs)

    def set_interfaces(self, interfaces):
        self.interfaces = interfaces

    def clean_templates(self):
        """Clean up all templates with pattern *-*.met in config directory"""
        glob_str = "%s*-*.met" % (self.EQ_CONFIG_DIR)
        _logger.debug("clean templates: %s" % glob_str)
        sh.rm("-rf", sh.glob(glob_str))

    def generate_configs(self):
        """Generate all of equipment templates and \
           save to Equipments.EQ_CONFIG_DIR"""
        for equipment in self.getAll():
            equipment.save(
                path=os.path.join(self.EQ_CONFIG_DIR, equipment.filename))

    def _remove_eq_tmpl(self, id):
        """Remove equipment template by id
            Args:
                id(int): equipment template's id
        """
        files = glob.glob(
            os.path.join(self.EQ_CONFIG_DIR, ("%d-*" % (id,))))
        for f in files:
            _logger.debug("Remove %s", f)
            os.unlink(f)

    def get_by_name(self, name):
        """Get equipment template by template's name
            Args:
                name(str): query template's name
            Returns:
                equipment(Euipment): or None if not found
        """
        for template in self.getAll():
            if template["name"] != name:
                continue
            return template
        return None

    def add(self, data, restart=True):
        """Add a new equipment template
            Args:
                data(dict): equipment template dict
            Returns:
                newTemplate(Equipment): the new equipment instance
        """
        if self.get_by_name(data["name"]) is not None:
            raise EquipmentError(
                "Template Name: %s is already exist!" % (data["name"]))

        newTemplate = super(Equipments, self).add(data)
        self._remove_eq_tmpl(newTemplate["id"])
        newTemplate.save(
            path=os.path.join(self.EQ_CONFIG_DIR, newTemplate.filename))

        if restart:
            restart_mxmodbusmaster()

        return newTemplate

    def _check_deps_interfaces(self, equipment):
        """Check dependency of interfaces and equipments"""
        deps_interfaces = self.interfaces.get_by_eqname(equipment["name"])
        if len(deps_interfaces) > 0:
            names = ", ".join(map(lambda i: i["name"], deps_interfaces))
            raise EquipmentError(
                "Interface: %s depends on this template." % (names))

    def remove(self, id, check_deps=True):
        """Remove a equipment's template, it will also check interface's
            deviceList.

            Args:
                id(int): id of template you want to delete
            Raises:
                EquipmentError: If there is a dependency issue
        """
        eq = self.get(id=id)
        if eq is None:
            return None

        if check_deps:
            self._check_deps_interfaces(eq)
        self._remove_eq_tmpl(id)
        super(Equipments, self).remove(id)

    def remove_all(self, force=False):
        """Remove all equipments"""
        for eq in self.getAll():
            self.remove(eq["id"], check_deps=(not force))

    def update(self, id, newObj):
        #  TODO: check there is interace depends on this template
        #        should restart mxmodbusservice
        old_eq = self.get(id=id)
        if old_eq is None:
            return None

        if "name" in newObj and old_eq["name"] != newObj["name"]:
            raise EquipmentError("Can't change template's name")

        updated_template = super(Equipments, self).update(id=id, newObj=newObj)
        if updated_template is None:
            return None

        self._remove_eq_tmpl(updated_template["id"])
        updated_template.save(
            path=os.path.join(self.EQ_CONFIG_DIR, updated_template.filename))

        # If equipment template has been assigned to any interfaces,
        # restart mxmodbusmaster to apply settings on the fly
        deps_interfaces = self.interfaces.get_by_eqname(
            updated_template["name"])
        if len(deps_interfaces) > 0:
            restart_mxmodbusmaster()

        return updated_template

    def export_data(
        self,
        output="/run/shm/export-equipment-template-%s.json" %
        (datetime.datetime.now().strftime("%Y%m%d%H%M")),
        ids=None
    ):
        """Export equipment templates
            Args:
                outpue(string): Output file.
                                Defaults to "/run/shm/export-equipment-template-201512111054.json"
                ids(array): ID of seleted templates' id.
                            If ids=None, all templates would be exported.
        """  # noqa
        export_eqs = []
        if ids is None:
            export_eqs = self.getAll()
        else:
            for eqtemplate_id in ids:
                eq = self.get(id=eqtemplate_id)
                if eq is None:
                    _logger.warning(
                        "Export equipment template id: %d not found."
                        % eqtemplate_id)
                    continue
                export_eqs.append(self.get(id=eqtemplate_id))

        with open(output, "w") as f:
            json.dump(export_eqs, f)

        return output

    def import_data(self, input_file, override=False):
        with open(input_file, "r") as f:
            new_eqtemplates = json.load(f)

        conflict_templates = [self.get_by_name(eqtemplate["name"])
                              for eqtemplate in new_eqtemplates
                              if self.get_by_name(eqtemplate["name"])
                              is not None]

        if len(conflict_templates) > 0:
            if override is False:
                raise EquipmentError("Import templates have conflicts.")

            # Remove all conflict templates if override sets to True
            with self.batch():
                for template in conflict_templates:
                    self.remove(id=template["id"])

        with self.batch():
            for eq in new_eqtemplates:
                self.add(eq, restart=False)
        restart_mxmodbusmaster()

    def export_to_http(self, url, headers={}, ids=None):
        """Export selected templates to web server"""
        output = self.export_data(ids=ids)
        r = requests.post(
            url=url,
            files={"output": open(output, "rb")},
            headers=headers,
            verify=False
        )

        if r.status_code != requests.codes.ok:
            _logger.debug("Response code: %s" % str(r.status_code))
            raise HTTPError("Can't upload files.")

        resp = r.json()
        if "url" not in resp:
            _logger.debug("Resp: %s" % str(resp))
            raise HTTPError("Can't get file link.")

        return resp["url"]

    def import_from_http(
            self, url,
            headers={}, override=False, tmpfile="/tmp/eqimport.tmp"):
        """Import templates from web server, if found name duplicated and
        override set to false it will raise errors"""
        r = requests.get(
            url=url,
            stream=True,
            headers=headers,
            verify=False
        )
        chunk_size = 1024
        if r.status_code != requests.codes.ok:
            raise HTTPError("Can't download file.")

        with open(tmpfile, 'wb') as fd:
            for chunk in r.iter_content(chunk_size):
                fd.write(chunk)

        return self.import_data(input_file=tmpfile, override=override)
