import socket
import logging
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, All, Length, Range, Any
from modbus import restart_mxmodbusgateway
from modbus import Libconfig
from modbus.interface import RTU_SCHEMA

_logger = logging.getLogger("sanji.modbus.interfaces")


def check_port_avaliable(host, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(0.1)
    try:
        sock.connect((host, port))
        sock.close()
        return False
    except Exception:
        return True


GATEWAY_SCHEMA = Schema({
    Required("tcp"): Schema({
        Required("host"): All(Any(unicode, str), Length(1, 255)),
        Required("port"): All(int, Range(1, 65535))
    }, extra=REMOVE_EXTRA),
    Required("rtu"): RTU_SCHEMA
}, extra=REMOVE_EXTRA)


class Gateway(Libconfig, dict):

    def __init__(self, *args, **kwargs):
        super(Gateway, self).__init__(*args, **kwargs)
        self._libconfig = self._convert()

    @staticmethod
    def get_serial_config_str(rtu):
        """baudrate-charsize-parity-stopbits"""
        return "%d-%d-%s-%d" % (
            rtu.get("baudRate"),
            8,
            rtu.get("parity")[0],
            rtu.get("stopBits")
        )

    def _convert(self):
        tcp = self.get("tcp")
        rtu = self.get("rtu")
        libconfig = {
            "tcp": {
                "host": tcp.get("host"),
                "service": tcp.get("port")
            },
            "rtu": {
                "path": rtu.get("dev"),
                "serial-config": Gateway.get_serial_config_str(rtu),
                "interframe-delay-ms": rtu.get("interframeDelayMs"),
                "interchar-timeout-ms": rtu.get("intercharTimeoutMs"),
                "request-timeout-ms": rtu.get("rtuRequestTimeoutMS")
            }
        }

        return libconfig


class Gateways(Libconfig, Model):

    GW_CONFIG_FILE = "/etc/mxmodbusgateway/configuration"

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = GATEWAY_SCHEMA
        kwargs["model_cls"] = Gateway
        self.maxPort = 500
        super(Gateways, self).__init__(*args, **kwargs)

    def _convert(self):
        libconfig = {
            "port-list": []
        }

        for gateway in self.getAll():
            libconfig["port-list"].append(gateway._convert())

        return libconfig

    def generate_configs(self, restart=False):
        self.save(path=self.GW_CONFIG_FILE)
        if restart:
            restart_mxmodbusgateway()

    def is_dev_unique(self, dev):
        for gateway in self.getAll():
            if gateway["rtu"]["dev"] == dev:
                return False

        return True

    def get_by_dev(self, dev):
        gateway = [gateway for gateway in self.getAll()
                   if gateway["rtu"]["dev"] == dev]
        if len(gateway) == 0:
            return None
        return gateway[0]

    def get_next_avaliable_tcp_port(self):
        while True:
            self.maxPort = self.maxPort + 1
            if check_port_avaliable("127.0.0.1", self.maxPort):
                port = self.maxPort
                break
            _logger.debug("PORT %d is not available." % (self.maxPort))
        _logger.debug("PORT %d is available." % (port))
        return port

    def add(self, obj, restart=True):
        dev = obj["rtu"]["dev"]
        if not self.is_dev_unique(dev):
            raise ValueError("RTU dev: %s is already exist!" % (dev))

        if obj.get("tcp") is None:
            obj["tcp"] = {
                "host": "127.0.0.1",
                "port": self.get_next_avaliable_tcp_port()
            }

        newGateway = super(Gateways, self).add(obj=obj)
        self.generate_configs(restart=restart)
        return newGateway

    def remove(self, id, restart=True):
        affected = super(Gateways, self).remove(id)
        if affected > 0:
            self.generate_configs(restart=restart)
        return affected

    def update(self, id, newObj):
        raise NotImplementedError("Can't modify gateway")
