import logging
import copy
from modbus import Libconfig
from modbus import fetch_status
from modbus import restart_mxmodbusmaster
from modbus import restart_mxmodbusgateway
from modbus.equipment import IsValidName
from sanji.model import Model
from voluptuous import Schema, REMOVE_EXTRA, Required, Any, All, Length, \
    Range, Optional

_logger = logging.getLogger("sanji.modbus.interfaces")

TCP_SCHEMA = Schema({
    Required("host"): All(Any(unicode, str), Length(1, 255)),
    Required("port"): All(int, Range(1, 65535))
}, extra=REMOVE_EXTRA)

RTU_SCHEMA = Schema({
    Required("dev"):  All(Any(unicode, str), Length(1, 255)),
    Required("baudRate"): All(int, Range(1, 655350)),
    Required("parity"): Any("even", "odd", "none"),
    Required("stopBits"): Any(0, 1, 2),
    Required("intercharTimeoutMs"): All(int, Range(min=0)),
    Optional("interframeDelayMs", default=0): All(int, Range(min=0)),
    Optional("devDisplayName"): Any(unicode, str),
    Optional("rtuRequestTimeoutMS"): int,
}, extra=REMOVE_EXTRA, required=True)

IF_SCHEMA = Schema({
    "id": int,
    Required("name"): All(Any(unicode, str), Length(1, 255)),
    Required("type"): Any("modbus-tcp", "modbus-rtu"),
    Required("config"): Any(RTU_SCHEMA, TCP_SCHEMA),
    Required("requestTimeoutMs"): All(Range(min=100, max=60000)),
    Required("pollingPeriodMs"): All(
        int, Range(min=1000, max=86400000, max_included=False)),
    Required("deviceList"): [Schema({
        Required("deviceId"): All(int, Range(0, 255)),
        Required("equipmentTemplate"): Any(unicode, str),
        Required("equipmentName"): All(
            Any(unicode, str), IsValidName(), Length(min=3, max=255)),
    }, extra=REMOVE_EXTRA)]
}, extra=REMOVE_EXTRA)


class InterfaceError(Exception):
    pass


class Interface(Libconfig, dict):
    """Interface
        name
        type
        pollingPeriodMs
        requestTimeoutMs
        deviceList
        config
    """
    def __init__(self, *args, **kwargs):
        if len(args) == 0 or not isinstance(args[0], dict):
            raise InterfaceError("Must init with a dict")

        if args[0].get("type") == "modbus-tcp":
            self.__class__ = TCPInterface
        elif args[0].get("type") == "modbus-rtu":
            self.__class__ = RTUInterface
        else:
            raise InterfaceError(
                "Interface only accepts type: modbus-tcp, modbus-rtu")

        super(Interface, self).__init__(*args, **kwargs)

    def set_device_status(self, device_status):
        equipments_status = device_status.get("equipments", {})
        for device in self.get("deviceList", []):
            status = equipments_status.get(device["equipmentName"], {
                "lastPoll": {
                    "at": "2000-12-31T00:00:00.000Z",
                    "elaspsedMS": 0,
                    "requestCount": 0,
                    "requestSuccess": 0,
                    "readingCount": 0
                },
                "lastSuccessAt": "2000-12-31T00:00:00.000Z",
                "totalRequestCount": 0,
                "totalRequestSuccess": 0,
                "totalReadingCount": 0
            })
            device["status"] = status

        return self


class TCPInterface(Interface):
    """TCPInterface
       config:
                host(str): server host/ip
                port(int): server port
    """
    def __init__(self, *args, **kwargs):
        if len(args) == 0 or not isinstance(args[0], dict):
            raise InterfaceError("Must init with a dict")

        if self.__class__ is TCPInterface and \
           args[0].get("type") != "modbus-tcp":
            raise InterfaceError(
                "TCPInterface only accepts: {type=modbus-tcp}")
        super(TCPInterface, self).__init__(*args, **kwargs)

    def _convert(self, *args, **kwargs):
        config = self.get("config") if kwargs.get("config") \
            is None else kwargs.get("config")

        request_timeout_ms = self["requestTimeoutMs"]
        if self["type"] == "modbus-rtu":
            request_timeout_ms = self["requestTimeoutMs"] + 100

        libconfig = {
            "host": config["host"],
            "service": config["port"],
            "polling-period-ms": self["pollingPeriodMs"],
            "request-timeout-ms": request_timeout_ms,
            "device-list": []
        }

        for device in self.get("deviceList"):
            libconfig["device-list"].append({
                "device-id": str.format('0x{:02X}', device["deviceId"]),
                "template-id": device["equipmentTemplate"],
                "equipment-name": device["equipmentName"]
            })

        return libconfig

    def save(self):
        raise NotImplementedError("Can't save a single %s" % self.__class__)


class RTUInterface(TCPInterface):
    """RTUInterface
       config:
                dev(str):
                baudRate(int):
                parity(str):
                stopBits(int):
                intercharTimeoutMs(int):
                interframeDelayMs(int):
    """
    def _convert(self, *args, **kwargs):
        # TODO: Lookup which local modbus gateway we will use
        gateways = kwargs.get("gateways")
        if gateways is None:
            raise InterfaceError("missing kwargs[\"gateways\"]")

        restart = kwargs.pop("restart", True)
        rtu_config = copy.copy(self.get("config"))
        rtu_config["rtuRequestTimeoutMS"] = self.get("requestTimeoutMs")
        gateway = gateways.add({"rtu": rtu_config}, restart=restart)
        return super(RTUInterface, self)._convert(config=gateway["tcp"])


class Interfaces(Libconfig, Model):

    IF_CONFIG_FILE = "/etc/mxmodbusmaster/configuration"
    STATUS_FILE = "/run/shm/var/mxmodbusmaster/equipments-status"

    def __init__(self, *args, **kwargs):
        kwargs["schema"] = IF_SCHEMA
        kwargs["model_cls"] = Interface
        self.gateways = kwargs.pop("gateways", None)
        if self.gateways is None:
            raise InterfaceError("missing kwargs[\"gateways\"]")
        super(Interfaces, self).__init__(*args, **kwargs)

    def getAll(self, embed_status=False):
        interfaces = super(Interfaces, self).getAll()
        if not embed_status:
            return interfaces

        try:
            device_status = fetch_status(status_file=self.STATUS_FILE)
        except Exception:
            pass

        if device_status is None:
            device_status = {}

        for interface in interfaces:
            interface.set_device_status(device_status=device_status)

        return interfaces

    def _convert(self):
        libconfig = {
            "polling-list": [],
            "equipment-status-path": self.STATUS_FILE
        }

        self.gateways.removeAll()
        for interface in self.getAll():
            gateways = None
            if isinstance(interface, RTUInterface):
                gateways = self.gateways
            try:
                libconfig["polling-list"].append(
                    interface._convert(gateways=gateways, restart=False))
            except Exception as e:
                _logger.error(
                    "Gateway for RTU interface: %s error" % interface["name"])
                _logger.debug(e)

        restart_mxmodbusgateway()
        return libconfig

    def generate_configs(self, restart=False):
        self.save(path=self.IF_CONFIG_FILE)
        if restart:
            restart_mxmodbusmaster()

    def get_by_eqname(self, name):
        """Get interface ids by query equipmentTemplate in deviceList
            Args:
                name(str): equipment template name
            Returns:
                interfaces(list): matched interface's instances, it would
                    return empyy list if there is no interface matched.
        """
        interfaces = []
        for interface in self.getAll():
            for eq in interface["deviceList"]:
                if eq["equipmentTemplate"] != name:
                    continue
                interfaces.append(interface)

        return interfaces

    def set_equipments(self, equipments):
        self.equipments = equipments

    def is_equipment_template_exist(self, device):
        """Check equipment template is exist
            Arguments:
                device: device object
            Returns:
                True: If equipment template is exist, vice versa
        """
        if self.equipments.get_by_name(device["equipmentTemplate"]) is None:
            return False
        return True

    def is_equipment_name_unique(self, device, id=None):
        """Check equipment name is unique
            Arguments:
                device: device object
            Returns:
                True: If equipment name is exist, vice versa
        """
        # If we have tons of instances, we should build index for this
        for instance in self.getAll():
            # Update
            if instance["id"] == id:
                continue

            for idevice in instance["deviceList"]:
                if idevice["equipmentName"] == device["equipmentName"]:
                    return False

        return True

    def check_device_list(self, device_list, id=None):
        """Check device list is vaild, including eq-template(exist), eq-name(unique)
            Arguments:
                device_list: list of device objects
                id: ID of instance for update operation
            Returns:
                True: If equipment name is exist, vice versa
            Raises:
                InterfaceError: If anyone of rules is not fulfilled.
        """
        for device in device_list:
            # Check the equipment is exist
            if self.is_equipment_template_exist(device) is not True:
                raise InterfaceError(
                    "Equipment Template: %s is not found" %
                    device["equipmentTemplate"])

            # Check duplicated interface? same host with same port?
            if self.is_equipment_name_unique(device, id=id) is not True:
                raise InterfaceError(
                    "Equipment Name: %s is not unique" %
                    device["equipmentName"])

        return device_list

    def add(self, data):
        """Add a interface
            Args:
                data(dict): interface config
            Returns:
                newInterface(Interface): new interface instance with id
            Raises:
                ValueError: if device's template not found
        """

        self.check_device_list(data["deviceList"])
        newInterface = super(Interfaces, self).add(data)
        self.generate_configs(restart=True)
        return newInterface

    def remove(self, id):
        """Remove an interface,\
           it will restart mxmodbusmaster port if affected > 0
            Args:
                id(int): interface's id
            Returns:
                affected(int): number of affected rows
        """
        affected = super(Interfaces, self).remove(id=id)
        if affected > 0:
            self.generate_configs(restart=True)

        return affected

    def update(self, id, newObj):
        """Update an exist interface,\
           it will restart mxmodbusmaster port if update successfully > 0
           Returns:
               newInterface(interface): udpated interface
           Raises:
               ValueError: if device list contains invaild entry
        """
        old_interface = self.get(id)
        if old_interface is None:
            return None

        if "deviceList" in newObj:
            self.check_device_list(id=id, device_list=newObj["deviceList"])

        newInterface = super(Interfaces, self).update(id=id, newObj=newObj)
        if newInterface is not None:
            self.generate_configs(restart=True)

        return newInterface
