#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
import sys
import unittest
import glob
import simplejson as json
from mock import Mock
from mock import ANY
from mock import patch
from sanji.connection.mockup import Mockup
from sanji.message import Message

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../")
    from index import Index
    from modbus.equipment import Equipments
    from modbus.interface import Interfaces
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    print (e)
    exit(1)


class TestIndexClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)
        files = glob.glob(os.path.join(self.root_path, "data/*.json")) + \
            glob.glob(os.path.join(self.root_path, "data/*.backup"))

        for prevConfig in files:
            try:
                os.unlink(prevConfig)
            except:
                pass

        Index.EQ_CONFIG_DIR = "/tmp/"
        Equipments.EQ_CONFIG_DIR = "/tmp/"
        Interfaces.IF_CONFIG_FILE = "/tmp/configuration"

        with open(os.path.join(
             self.root_path, "data", "interfaces.json.factory")) as fp:
            self.interfaces = json.load(fp)
        with open(os.path.join(
             self.root_path, "data", "equipments.json.factory")) as fp:
            self.equipments = json.load(fp)

        self.index = Index(connection=Mockup())
        self.index.interfaces.gateways.add = Mock(
            return_value={
                "tcp": {
                   "host": "host",
                   "port": 1234
                }
            })

    def tearDown(self):
        self.index.stop()
        self.index = None

    def test_get_equipment(self):
        """Get one of templates by id"""
        resp = Mock()
        message = Message({
            "param": {
                "id": 1
            }
        })

        self.index.get_equipment(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.equipments[0])

    def test_get_all_equipments(self):
        """Get all of templates"""
        result = self.equipments
        resp = Mock()
        self.index.get_all_equipments(message=None, response=resp, test=True)
        resp.assert_called_once_with(data=result)

    @patch("modbus.sh")
    def test_post_equipment(self, sh):
        """Create an equipment template, \
            should respond created template with id"""
        message = Message({
            "data": {
                "name": "ioLogik-test",
                "tagList": [
                    {
                        "name": "abc",
                        "function": "read-coils",
                        "address": 0,
                        "quantity": 1,
                        "invalidValue": [0],
                        "type": "uint16",
                        "description": "alarm input"
                    }
                ]
            }
        })

        def resp(data):
            self.assertIn("id", data)
            self.assertIn("invalidValue", data["tagList"][0])

        m = Mock(side_effect=resp)
        self.index.post_equipment(message=message, response=m, test=True)
        m.assert_called_once_with(data=ANY)

    @patch("modbus.sh")
    @patch("index.Index.check_logprofile_deps")
    @patch("index.Index.check_tags_deps")
    def test_update_equipment(
            self, check_tags_deps, check_logprofile_deps, sh):
        """Update an equipment template, should respond updated content"""
        message1 = Message({
            "param": {
                "id": 2
            },
            "data": {
                "id": 55688,
                "name": "ioLogik-E1211",
                "tagList": [
                    {
                        "name": "updated",
                        "function": "read-coils",
                        "address": 0,
                        "quantity": 2,
                        "type": "uint16",
                        "description": "alarm input",
                        "invalidValue": [123, 456]
                    }
                ]
            }
        })

        m = Mock()
        self.index.update_equipment(
            message=message1, response=m, test=True)
        m.assert_called_once_with(data={
            "id": 2,
            "name": "ioLogik-E1211",
            "tagList": [
                {
                    "name": "updated",
                    "function": "read-coils",
                    "address": 0,
                    "quantity": 2,
                    "type": "uint16",
                    "description": "alarm input",
                    "invalidValue": [123, 456]
                }
            ]
        })

    def test_get_interface(self):
        """Get one of interface by id"""

        result = self.interfaces[0]
        resp = Mock()
        message = Message({
            "param": {
                "id": 1
            }
        })

        self.index.get_interface(message=message, response=resp, test=True)
        resp.assert_called_once_with(data=result)

    def test_get_all_interfaces(self):
        """Get all of interfaces"""
        resp = Mock()
        message = Message({})

        self.index.get_all_interfaces(
            message=message, response=resp, test=True)
        resp.assert_called_once_with(data=self.interfaces)

    def test_post_interface(self):
        """Create an interface , \
            should respond created interface with id"""
        message = Message({
            "data": {
                "name": "Modbus Gateway TEST",
                "type": "modbus-rtu",
                "config": {
                  "dev": "/dev/ttyS0",
                  "baudRate": 9600,
                  "parity": "even",
                  "stopBits": 1,
                  "intercharTimeoutMs": 300,
                  "requestTimeoutMs": 5000,
                  "interframeDelayMs": 300
                },
                "pollingPeriodMs": 1500,
                "requestTimeoutMs": 100,
                "deviceList": [
                  {
                    "deviceId": 31,
                    "equipmentTemplate": "ioLogik-E1210",
                    "equipmentName": "My-ioLogik-E1210"
                  }
                ]
            }
        })

        def resp(data):
            self.assertIn("id", data)

        m = Mock(side_effect=resp)
        self.index.post_interface(message=message, response=m, test=True)
        m.assert_called_once_with(data=ANY)

    @patch("index.Index.check_logprofile_deps")
    def test_delete_interfaces(self, check_logprofile_deps):
        """Delete interface configs, should respond success message"""
        message1 = Message({
            "data": [1, 2, 3, 4, 5]
        })

        m = Mock()
        self.index.delete_interfaces(
            message=message1, response=m, test=True)
        m.assert_called_once_with(data={"message": "success"})

        message2 = Message({
            "param": {
                "id": 1
            }
        })
        m = Mock()
        self.index.get_interface(
            message=message2, response=m, test=True)
        m.assert_called_once_with(code=404)

    @patch("index.Index.check_logprofile_deps")
    def test_delete_interface(self, check_logprofile_deps):
        """Delete an interface config, should respond success message"""
        message1 = Message({
            "param": {
                "id": 1
            }
        })

        m = Mock()
        self.index.delete_interface(
            message=message1, response=m, test=True)
        m.assert_called_once_with(data={"message": "success"})

        message2 = Message({
            "param": {
                "id": 1
            }
        })
        m = Mock()
        self.index.get_interface(
            message=message2, response=m, test=True)
        m.assert_called_once_with(code=404)

    @patch("index.Index.check_logprofile_deps")
    def test_update_interface(self, check_logprofile_deps):
        """Update an interface config, should respond updated content"""
        message1 = Message({
            "param": {
                "id": 2
            },
            "data": {
                "id": 1,
                "name": "Modbus Gateway TEST",
                "type": "modbus-rtu",
                "config": {
                  "dev": "/dev/ttyS0",
                  "baudRate": 9600,
                  "parity": "even",
                  "stopBits": 1,
                  "intercharTimeoutMs": 300,
                  "interframeDelayMs": 300
                },
                "requestTimeoutMs": 123,
                "pollingPeriodMs": 65530,
                "deviceList": [
                  {
                    "deviceId": 31,
                    "equipmentTemplate": "ioLogik-E1210",
                    "equipmentName": "MyioLogik-E1210"
                  },
                  {
                    "deviceId": 32,
                    "equipmentTemplate": "ioLogik-E1210",
                    "equipmentName": "MyioLogik-E1210"
                  }
                ]
              }
        })

        m = Mock()
        self.index.update_interface(
            message=message1, response=m, test=True)
        m.assert_called_once_with(data=ANY)

if __name__ == "__main__":
    unittest.main()
