#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import sys
import unittest
import glob
import requests
import simplejson as json

from mock import mock_open
from mock import patch
from mock import Mock
from mock import ANY

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from modbus.equipment import Equipment
    from modbus.equipment import Equipments
    from modbus.equipment import EquipmentError
    from modbus.equipment import get_logprofile_deps
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)


class TestEquipmentsClass(unittest.TestCase):
    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)
        self.factory = None

        with open(os.path.join(
             self.root_path, "data", "equipments.json.factory")) as fp:
            self.factory = json.load(fp)
        self.interfaces = Mock()
        Equipments.EQ_CONFIG_DIR = "/tmp/"
        self.equipments = Equipments(name="equipments", path=self.root_path)
        self.equipments.set_interfaces(self.interfaces)

    def tearDown(self):
        pass

    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    @patch("modbus.equipment.Equipment.save")
    def test_add_equipment(self, save, _remove_eq_tmpl):
        """Add a equipment template, should return equipment instance"""
        newEq = self.factory[0].copy()
        newEq["name"] = "IAmNewEq-tmpl"
        newEqWithId = self.equipments.add(newEq)
        save.assert_called_once_with(path=ANY)
        _remove_eq_tmpl.assert_called_once_with(newEqWithId["id"])
        self.assertEqual(newEqWithId["id"], len(self.factory) + 1)

    @patch("modbus.equipment.restart_mxmodbusmaster")
    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    @patch("modbus.equipment.Equipment.save")
    def test_add_equipment_with_no_description(
            self, save, _remove_eq_tmpl, restart_mxmodbusmaster):
        """Add a equipment template with no description,\
           should not raise EquipmentError"""
        data = self.factory[0].copy()
        data["name"] = "helloEQ"
        data["tagList"][0].pop("description")
        newEqWithId = self.equipments.add(data)
        save.assert_called_once_with(path=ANY)
        _remove_eq_tmpl.assert_called_once_with(newEqWithId["id"])
        self.assertEqual(newEqWithId["id"], len(self.factory) + 1)
        restart_mxmodbusmaster.assert_called_once_with()

    @patch("modbus.equipment.restart_mxmodbusmaster")
    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    @patch("modbus.equipment.Equipment.save")
    def test_add_equipment_with_zero_length_description(
            self, save, _remove_eq_tmpl, restart_mxmodbusmaster):
        """Add a equipment template with no description,\
           should not raise EquipmentError"""
        data = self.factory[0].copy()
        data["name"] = "helloEQ"
        data["tagList"][0]["description"] = ""
        newEqWithId = self.equipments.add(data)
        save.assert_called_once_with(path=ANY)
        _remove_eq_tmpl.assert_called_once_with(newEqWithId["id"])
        self.assertEqual(newEqWithId["id"], len(self.factory) + 1)
        restart_mxmodbusmaster.assert_called_once_with()

    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    @patch("modbus.equipment.Equipment.save")
    def test_add_equipment_duplicate_name(self, save, _remove_eq_tmpl):
        """Add a equipment template with duplicate name,\
           should raise EquipmentError"""
        with self.assertRaises(EquipmentError):
            self.equipments.add(self.factory[0])
        self.assertFalse(save.called)
        self.assertFalse(_remove_eq_tmpl.called)

    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    def test_remove_equipment(self, _remove_eq_tmpl):
        """Remove a equipment template with id, should remove config file"""
        self.interfaces.get_by_eqname.return_value = []
        self.equipments.remove(id=1)
        _remove_eq_tmpl.assert_called_once_with(1)
        self.assertEqual(None, self.equipments.get(id=1))

    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    def test_remove_equipment_with_deps(self, _remove_eq_tmpl):
        """Remove a equipment template with id, should remove config file"""
        self.interfaces.get_by_eqname.return_value = [{"name": "dev xxx"}]
        with self.assertRaises(EquipmentError):
            self.equipments.remove(id=1)
        self.assertFalse(_remove_eq_tmpl.called)
        self.assertNotEqual(None, self.equipments.get(id=1))

    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    def test_remove_equipment_with_non_exist_id(self, _remove_eq_tmpl):
        """Remove a non-exist equipment template, should return None"""
        self.interfaces.get_by_eqname.return_value = [{"name": "dev xxx"}]
        newEq = self.equipments.remove(id=len(self.factory) + 10)
        self.assertEqual(None, newEq)
        self.assertFalse(_remove_eq_tmpl.called)

    @patch("modbus.equipment.restart_mxmodbusmaster")
    @patch("modbus.equipment.Equipments.save")
    @patch("modbus.equipment.Equipments._remove_eq_tmpl")
    def test_update_equipment(self, _remove_eq_tmpl, save,
                              restart_mxmodbusmaster):
        """Update a equipment, should return new equipment template"""
        self.interfaces.get_by_eqname.return_value = [{}]
        eq = self.factory[0].copy()
        eq["tagList"] = []
        new_eq = self.equipments.update(id=1, newObj=eq)
        self.assertEqual(new_eq["tagList"], [])
        restart_mxmodbusmaster.assert_called_once_with()

    def test_update_equipment_with_non_exist_id(self):
        """Update a equipment with non exist id, should return None"""
        new_eq = self.equipments.update(id=55688, newObj=self.factory[0])
        self.assertEqual(new_eq, None)

    def test_update_equipment_change_name(self):
        """Update a equipment with name changed,\
           should return new equipment template"""
        eq = self.factory[0].copy()
        eq["name"] = "changed name"
        with self.assertRaises(EquipmentError):
            self.equipments.update(id=1, newObj=eq)

    def test_get_by_name(self):
        """Get Equipment instance by given name, should return equipment"""
        eq = self.equipments.get_by_name(name="ioLogik-E1210")
        self.assertTrue(isinstance(eq, Equipment))

    def test_get_by_name_non_exist(self):
        """Get Equipment non-exist instance by given name,\
           should return None"""
        eq = self.equipments.get_by_name(name="ioLogik-E2242222")
        self.assertEqual(None, eq)

    def test_export_selected_ids(self):
        """Export seleted equipment templates with given ids"""
        ids = [1, 2, 3]
        output = self.equipments.export_data(ids=ids)
        with open(output, "r") as f:
            eqtemplates = json.load(f)

        self.assertEquals(len(ids), len(eqtemplates))

    def test_export_all(self):
        """Export all equipment templates"""
        output = self.equipments.export_data()
        with open(output, "r") as f:
            eqtemplates = json.load(f)

        self.assertEquals(len(self.equipments.getAll()), len(eqtemplates))

    def test_remove_all(self):
        """Equipment Template should be able to removed"""
        self.interfaces.get_by_eqname = Mock(return_value=[])
        self.equipments.remove_all()
        self.assertEqual(self.equipments.getAll(), [])

    def test_remove_all_error(self):
        """Equipment Template should be able to removed"""
        self.interfaces.get_by_eqname = Mock(
            return_value=[{"name": "test"}, {"name": "test2"}])
        self.equipments._remove_eq_tmpl = Mock()
        with self.assertRaises(EquipmentError):
            self.equipments.remove_all()
        self.assertFalse(self.equipments._remove_eq_tmpl.called)

    @patch("modbus.equipment.json.load")
    def test_import_data_override(self, jsonload):
        neweq = [{
            "id": 5,
            "name": "ioLogik-test",
            "tagList": [
                {
                    "name": "di0",
                    "function": "read-discrete-inputs",
                    "address": 0,
                    "quantity": 1,
                    "type": "boolean",
                    "description": ""
                }
            ]
        }]
        self.interfaces.get_by_eqname = Mock(
            return_value=[{"name": "test"}, {"name": "test2"}])
        jsonload.return_value = neweq
        m = mock_open()
        self.equipments._check_deps_interfaces = Mock(return_value=False)
        with patch("__builtin__.open", m, create=True):
            self.equipments.import_data("/tmp/input.json", override=True)
        neweq[0]["id"] = 1
        self.assertEqual(
            self.equipments.get_by_name(neweq[0]["name"])["tagList"],
            neweq[0]["tagList"]
        )

    @patch("requests.post")
    @patch("modbus.equipment.Equipments.export_data")
    def test_export_to_http(self, export_data, mock_post):
        """Equipment Template should be able to export and
        upload to web server"""
        mock_resp_post = Mock()
        mock_resp_post.status_code = requests.codes.ok
        mock_resp_post.json.return_value = {"url": "testurl"}
        mock_post.return_value = mock_resp_post

        export_data.return_value = []
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.equipments.export_to_http(url=None, headers={})
            m.assert_called_with(ANY, "rb")

    @patch("requests.get")
    @patch("modbus.equipment.Equipments.import_data")
    def test_import_from_http(self, import_data, mock_get):
        """Equipment Template should be able to import and
        download from web server"""
        mock_resp_get = Mock()
        mock_resp_get.status_code = requests.codes.ok
        mock_resp_get.iter_content = Mock(return_value=iter([]))
        mock_get.return_value = mock_resp_get

        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.equipments.import_from_http(url=None)
            m.assert_called_with(ANY, "wb")


class TestEquipmentClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        self.factory = None

        with open(os.path.join(
             self.root_path, "data", "equipments.json.factory")) as fp:
            self.factory = json.load(fp)

        self.equipment = Equipment(self.factory[0])

    def tearDown(self):
        pass

    def test_filename(self):
        """Get filename of equipment instance, should return str"""
        filename = self.equipment.filename
        self.assertTrue(isinstance(filename, str))

    def test__convert(self):
        """Convert a Equipment to libconfig dict, \
           should return a dict fulfilled spec"""
        libconfig = self.equipment._convert()
        self.assertIn("request-list", libconfig)
        self.assertIn("function", libconfig["request-list"][0])
        self.assertIn("address", libconfig["request-list"][0])
        self.assertIn("quantity", libconfig["request-list"][0])

        self.assertIn("tag-list", libconfig)
        self.assertIn("id", libconfig["tag-list"][0])
        self.assertIn("quantity", libconfig["tag-list"][0])
        self.assertIn("type", libconfig["tag-list"][0])

    def test_toLibconfig(self):
        """Translate Equipment to libconfig string,\
           should return a vaild libconfig string"""
        libconfigStr = self.equipment.toLibconfig()
        self.assertTrue(isinstance(libconfigStr, str))
        # TODO: vaild string is vaild for libconfig

    def test_save(self):
        """Equipment Template should be able to be saved to a file"""
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.equipment.save("/tmp/1234")
        m.assert_called_once_with("/tmp/1234", "w")
        h = m()
        h.write.assert_called_once_with(self.equipment.toLibconfig())


class TestFunctionClass(unittest.TestCase):

    def test_get_logprofile_deps(self):
        """diff_equip_list used in logprofile should return as deps"""
        logprofile = [{
            "equipmentList": [
                {
                    "equipmentName": "aaa",
                    "equipmentTags": ["di0", "di1", "di2"]
                },
                {
                    "equipmentName": "bbb",
                    "equipmentTags": ["di0", "di1", "di2"]
                },
                {
                    "equipmentName": "ccc",
                    "equipmentTags": ["di0", "di1", "di2"]
                }
            ]
        }]

        diff_equip_list = ["aaa"]

        self.assertIn("aaa", get_logprofile_deps(logprofile, diff_equip_list))


if __name__ == "__main__":
    unittest.main()
