#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import sys
import simplejson as json
import unittest
import glob

from mock import mock_open
from mock import patch
from mock import call

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from modbus.gateway import Gateways
    from modbus.gateway import Gateway
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    print (e)
    exit(1)


class TestGatewayClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)
        self.factory = None

        with open(os.path.join(
             self.root_path, "data", "interfaces.json.factory")) as fp:
            self.factory = json.load(fp)

        Gateways.GW_CONFIG_FILE = "/tmp/configuration"
        self.gateways = Gateways(name="gateways", path=self.root_path)

    def tearDown(self):
        pass

    def test_add_gateway(self):
        """Add a gateway, should return Gateway instance"""
        rtu = self.factory[0]["config"]
        newGateway = self.gateways.add({
            "rtu": rtu
        })
        self.assertTrue(isinstance(newGateway, Gateway))
        self.assertEqual(newGateway["id"], 1)
        self.assertIn("tcp", newGateway)

    def test_remove_gateway(self):
        """Remove a gateway, should restart mxmodbusgateway"""
        rtu = self.factory[0]["config"]
        self.gateways.add({"rtu": rtu})

        with patch("modbus.gateway.restart_mxmodbusgateway") as rm:
            self.gateways.remove(id=1)
            rm.assert_called_once_with()

    @patch("modbus.gateway.restart_mxmodbusgateway")
    def test_remove_gateway_not_restart(self, restart_mxmodbusgateway):
        """Remove a gateway, should NOT restart mxmodbusgateway"""
        self.gateways.remove(id=5)
        self.assertFalse(restart_mxmodbusgateway.called)

    def test__convert(self):
        """Convert json data to libconfig, should respond a dict with
           vaild spec"""
        libconfig = self.gateways._convert()
        self.assertIn("port-list", libconfig)

    def test_save(self):
        """Save all gateways config into a single configuration file"""
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.gateways.save(path="/tmp/test")
        m.assert_called_once_with("/tmp/test", "w")
        h = m()
        h.write.assert_called_once_with(self.gateways.toLibconfig())

    @patch("modbus.gateway.check_port_avaliable")
    def test_get_next_avaliable_tcp_port(self, check_port_avaliable):
        """Should find next avaliable port"""
        return_list = [False, False, False, True]
        check_port_avaliable.side_effect = return_list
        port = self.gateways.maxPort
        self.gateways.get_next_avaliable_tcp_port()
        self.assertEqual(check_port_avaliable.call_count,
                         len(return_list))
        self.assertListEqual([
            call('127.0.0.1', port + 1),
            call('127.0.0.1', port + 2),
            call('127.0.0.1', port + 3),
            call('127.0.0.1', port + 4)],
            check_port_avaliable.call_args_list)


if __name__ == "__main__":
    unittest.main()
