#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import sys
import glob
import simplejson as json
import unittest

from mock import Mock
from mock import mock_open
from mock import patch

try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/../../")
    from modbus.interface import TCPInterface
    from modbus.interface import RTUInterface
    from modbus.interface import Interfaces
    from modbus.interface import InterfaceError
except ImportError as e:
    print "Please check the python PATH for import test module. (%s)" \
        % __file__
    exit(1)


class TestInterfacesClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        jsons = glob.glob(os.path.join(self.root_path, "data/*.json"))
        backups = glob.glob(os.path.join(self.root_path, "data/*.backup"))
        for file in jsons + backups:
            os.unlink(file)
        self.factory = None

        with open(os.path.join(
             self.root_path, "data", "interfaces.json.factory")) as fp:
            self.factory = json.load(fp)

        self.factory.append({
          "id": 3,
          "name": "MOXA HQ",
          "type": "modbus-tcp",
          "config": {
            "host": "192.168.31.123",
            "port": 502
          },
          "pollingPeriodMs": 5000,
          "requestTimeoutMs": 111,
          "deviceList": [
            {
              "deviceId": 1,
              "equipmentTemplate": "ioLogik-E2242",
              "equipmentName": "SYS-FRONT-DOOR"
            },
            {
              "deviceId": 2,
              "equipmentTemplate": "ioLogik-E2242",
              "equipmentName": "SYS-FRONT-DOOR-2"
            }
          ]
        })

        self.factory[0]["deviceList"].append({
            "deviceId": 1,
            "equipmentTemplate": "ioLogik-R2140",
            "equipmentName": "My-ioLogik-R2140-2"
        })

        self.rtuinterface = RTUInterface(self.factory[0])
        self.tcpinterface = TCPInterface(self.factory[2])
        self.equipments = Mock()
        self.gateways = Mock()
        self.gateways.add.return_value = {
            "tcp": {
               "host": "host",
               "port": 1234
            }
        }
        Interfaces.IF_CONFIG_FILE = "/tmp/configuration"
        self.interfaces = Interfaces(
            name="interfaces", path=self.root_path, gateways=self.gateways)
        self.interfaces.set_equipments(self.equipments)

    def tearDown(self):
        pass

    @patch("modbus.sh")
    def test_add_tcpinterface(self, sh):
        """Add a tcpinterface, with a vaild template, \
           should return TCPInterface instance"""
        data = self.factory[2].copy()
        data["deviceList"][0]["equipmentName"] = "test-name"
        newInterface = self.interfaces.add(self.factory[2])
        self.assertTrue(isinstance(newInterface, TCPInterface))
        self.assertEqual(newInterface["id"], 3)

    # @patch("modbus.sh")
    # def test_add_tcpinterface_failed(self, sh):
    #     """Add a tcpinterface, with a invaild template, \
    #        should raise InterfaceError"""
    #     data = self.factory[2].copy()
    #     data["deviceList"][0]["equipmentName"] = "test-name"
    #     newInterface = self.interfaces.add(self.factory[2])
    #     self.assertTrue(isinstance(newInterface, TCPInterface))
    #     self.assertEqual(newInterface["id"], 3)

    @patch("modbus.sh")
    def test_add_tcpinterface_with_invalid_eqtmpl(self, sh):
        """Add a TCPInterface, with a non-exist template, \
           should raise InterfaceError"""
        self.equipments.get_by_name.return_value = None
        interface = self.factory[2].copy()
        interface["deviceList"][0]["equipmentTemplate"] = "None"
        with self.assertRaises(InterfaceError):
            self.interfaces.add(interface)
        self.assertEqual(2, self.interfaces.maxId)
        self.assertEqual(None, self.interfaces.get(id=4))

    def test_add_rtuinterface(self):
        """Add a RTUInterface, with a vaild template, \
           should return RTUInterface instance"""
        data = self.factory[0].copy()
        data["deviceList"][0]["equipmentName"] = "55677"
        newInterface = self.interfaces.add(data)
        self.assertTrue(isinstance(newInterface, RTUInterface))
        self.assertEqual(newInterface["id"], 3)

    @patch("modbus.interface.restart_mxmodbusmaster")
    def test_remove_tcpinterface(self, restart_mxmodbusmaster):
        """Remove an interface, should restart mxmodbusmaster"""
        self.interfaces.remove(id=1)
        restart_mxmodbusmaster.assert_called_once_with()

    @patch("modbus.interface.restart_mxmodbusmaster")
    def test_remove_tcpinterface_not_restart(self, restart_mxmodbusmaster):
        """Remove an interface, should NOT restart mxmodbusmaster"""
        self.interfaces.remove(id=5)
        self.assertFalse(restart_mxmodbusmaster.called)

    @patch("modbus.interface.restart_mxmodbusmaster")
    def test_update_tcpinterface(self, restart_mxmodbusmaster):
        """Update an interface, should restart service"""
        data = self.factory[0].copy()
        data["name"] = "Updated Name"
        newInterface = self.interfaces.update(id=1, newObj=data)
        restart_mxmodbusmaster.assert_called_once_with()
        self.assertEqual(newInterface["name"], data["name"])

    @patch("modbus.interface.restart_mxmodbusmaster")
    def test_update_tcpinterface_not_restart(self, restart_mxmodbusmaster):
        """Update a non-exist interface, should not restart service"""
        data = self.factory[0].copy()
        newInterface = self.interfaces.update(id=5, newObj=data)
        self.assertFalse(restart_mxmodbusmaster.called)
        self.assertEqual(newInterface, None)

    @patch("modbus.interface.restart_mxmodbusmaster")
    def test_update_tcpinterface_invaild_eqtmpl(self, restart_mxmodbusmaster):
        """Update a interface with invaild eq tmpl,\
           should not restart service and raise InterfaceError"""
        data = self.factory[0].copy()
        self.equipments.get_by_name.return_value = None
        with self.assertRaises(InterfaceError):
            self.interfaces.update(id=2, newObj=data)
        self.assertFalse(restart_mxmodbusmaster.called)

    @patch("modbus.interface.restart_mxmodbusgateway")
    def test__convert(self, restart_mxmodbusgateway):
        """Convert json data to libconfig, should respond a dict with
           vaild spec and restart mxmodbusgateway service"""
        # add a tcp interface
        data = self.factory[2].copy()
        data["deviceList"][0]["equipmentName"] = "test-name"
        self.interfaces.add(self.factory[2])
        restart_mxmodbusgateway.reset_mock()
        libconfig = self.interfaces._convert()
        self.assertIn("polling-list", libconfig)
        self.assertEqual(len(libconfig["polling-list"]), 3)
        restart_mxmodbusgateway.assert_called_once_with()

    def test_save(self):
        """Save all interfaces config into a single configuration file"""
        m = mock_open()
        with patch("__builtin__.open", m, create=True):
            self.interfaces.save(path="/tmp/test")
        h = m()
        h.write.assert_called_once_with(self.interfaces.toLibconfig())

    @patch("modbus.interface.fetch_status")
    def test_getAll(self, fetch_status):
        # Add some devices
        data = self.factory[0].copy()
        self.interfaces.update(id=1, newObj=data)

        # Prepare
        fetch_status.return_value = {
          "at": "2015-12-30T10:20:14.865Z",
          "equipments": {
            "My-ioLogik-R2140-2": {
              "lastPoll": {
                "at": "2015-12-30T10:20:10.006Z",
                "elaspsedMS": 168,
                "requestCount": 4,
                "requestSuccess": 4,
                "readingCount": 4
              },
              "lastSuccessAt": "2015-12-30T10:20:10.006Z",
              "totalRequestCount": 20,
              "totalRequestSuccess": 20,
              "totalReadingCount": 20
            }
          }
        }

        interfaces = self.interfaces.getAll(embed_status=True)
        self.assertEqual(
            interfaces[0]["deviceList"][0]["status"],
            fetch_status.return_value["equipments"]["My-ioLogik-R2140-2"])


class TestTCPInterfaceClass(unittest.TestCase):

    def setUp(self):
        self.root_path = os.path.abspath(os.path.dirname(__file__) + "/../../")
        self.factory = None

        with open(os.path.join(
             self.root_path, "data", "interfaces.json.factory")) as fp:
            self.factory = json.load(fp)

        self.factory.append({
          "id": 3,
          "name": "MOXA HQ",
          "type": "modbus-tcp",
          "config": {
            "host": "192.168.31.123",
            "port": 502
          },
          "requestTimeoutMs": 111,
          "pollingPeriodMs": 5000,
          "deviceList": [
            {
              "deviceId": 0,
              "equipmentTemplate": "ioLogik-E2242",
              "equipmentName": "SYS-FRONT-DOOR"
            }
          ]
        })
        self.tcpinterface = TCPInterface(self.factory[2])

    def tearDown(self):
        pass

    def test__convert(self):
        """Convert a TCPInterface to libconfig dict, \
           should return a dict fulfilled spec"""
        libconfig = self.tcpinterface._convert()
        self.assertIn("host", libconfig)
        self.assertIn("service", libconfig)
        self.assertIn("polling-period-ms", libconfig)
        self.assertIn("request-timeout-ms", libconfig)
        self.assertIn("device-list", libconfig)
        self.assertIn("device-id", libconfig["device-list"][0])
        self.assertIn("template-id", libconfig["device-list"][0])
        self.assertIn("equipment-name", libconfig["device-list"][0])

    def test_toLibconfig(self):
        """Translate TCPInterface to libconfig string,\
           should return a vaild libconfig string"""
        libconfigStr = self.tcpinterface.toLibconfig()
        self.assertTrue(isinstance(libconfigStr, str))
        # TODO: vaild string is vaild for libconfig

    def test_save(self):
        """Try to save a single TCPInterface,\
           should raise a NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self.tcpinterface.save()

if __name__ == "__main__":
    unittest.main()
