(function() {
  'use strict';

  angular.module('webapp')
  .directive('mxImageProcessor', mxImageProcessor);

  function redraw(scope, context, canvasElem, image, obj) {
    context.save();
    context.setTransform(1, 0, 0, 1, 0, 0);
    context.clearRect(0, 0, canvasElem.width, canvasElem.height);
    context.restore();
    context.drawImage(image, obj.x, obj.y, obj.width, obj.height);

    scope.imageFile = canvasElem.toDataURL();

    // this is neccessary because angular doesn't know image.onload occurs
    scope.$apply();
  }

  function mxImageProcessor($document, notificationService, $log) {

    var cursor = {
      isDragging: false,
      x: null,
      y: null
    };
    var obj = {};
    var delay = 200;
    var unit = 1;
    var timeoutId;
    var delta = 0.996;
    var zoomDelta = 0.05;

    return {
      templateUrl: 'mx-image-processor.html',
      restrict: 'E',
      scope: {
        imageFile: '='
      },
      replace: true,
      link: function(scope, element) {

        scope.showImageBox = false;

        var canvas = element.find('canvas');
        var btnZoomIn = element.find('#btn-zoom-in');
        var btnZoomOut = element.find('#btn-zoom-out');
        var canvasElem = canvas[0];
        var context = canvasElem.getContext('2d');
        var image = new Image();

        image.onload = function() {

          obj.x = 0;
          obj.y = 0;
          obj.width = image.width;
          obj.height = image.height;

          if ((obj.width > canvasElem.width) && (obj.height > canvasElem.height)) {

            var scale;

            if (obj.height < obj.width) {
              scale = obj.height / canvasElem.height;
              obj.width = obj.width / scale;
              obj.height = canvasElem.height;
            } else {
              scale = obj.width / canvasElem.width;
              obj.height = obj.height / scale;
              obj.width = canvasElem.width;
            }
          }

          context.translate(0, 0);
          context.setTransform(1, 0, 0, 1, 0, 0);
          redraw(scope, context, canvasElem, image, obj);
        };

        image.onerror = function() {
          $log.error('Error loading profile image');
        };

        if (angular.isString(scope.imageFile)) {
          scope.showImageBox = true;
          image.src = scope.imageFile;
        }

        // mouse / touch down
        function down(e) {
          cursor.isDragging = true;
          cursor.x = e.offsetX;
          cursor.y = e.offsetY;
        }

        // mouse / touch up
        function up() {
          cursor.isDragging = false;

          if (scope.imageFile) {
            scope.imageFile = canvasElem.toDataURL();
          }
        }

        function move(e) {

          if (! cursor.isDragging) {
            cursor.x = e.offsetX;
            cursor.y = e.offsetY;
            return;
          }
          var deltaX = e.offsetX - cursor.x;
          var deltaY = e.offsetY - cursor.y;

          obj.x += deltaX;
          obj.y += deltaY;

          redraw(scope, context, canvasElem, image, obj);

          cursor.x = e.offsetX;
          cursor.y = e.offsetY;
        }

        function createLoopFunc(func) {

          return function(e) {
            func(e);
            (function loop() {

              if (delay > 30) {
                unit += 10;
                delay -= unit;
              }

              if (delay < 30) {
                delay = 30;
              }

              timeoutId = setTimeout(function() {
                func(e);
                loop();
              }, delay);
            })();
          };
        }

        function zoomIn(options) {

          options = options || {};

          var scale = 1 + zoomDelta;

          if (options.scrolling) {
            scale = 1 + zoomDelta + 0.05;
          }

          var translateX = options.scrolling ? cursor.x : canvas.width / 2;
          var translateY = options.scrolling ? cursor.y : canvas.height / 2;

          context.translate(translateX, translateY);
          context.scale(scale, scale);
          context.translate(-translateX, -translateY);
          redraw(scope, context, canvasElem, image, obj);
        }

        function zoomOut(options) {

          options = options || {};
          var scale = 1 - zoomDelta;

          if (options.scrolling) {
            scale = 1 - zoomDelta - 0.05;
          }

          var translateX = options.scrolling ? cursor.x : canvas.width / 2;
          var translateY = options.scrolling ? cursor.y : canvas.height / 2;

          context.translate(translateX, translateY);
          context.scale(scale, scale);
          context.translate(-translateX, -translateY);
          redraw(scope, context, canvasElem, image, obj);
        }

        function handleScroll(e) {

          e.preventDefault();

          var delta = e.wheelDelta ? e.wheelDelta / 40 : e.detail ? -e.detail : 0;

          if (delta > 0) {
            zoomIn({scrolling: true});
          } else {
            zoomOut({scrolling: true});
          }
        }

        var zoomInDownCb = createLoopFunc(zoomIn);
        var zoomOutDownCb = createLoopFunc(zoomOut);

        function stopLooping() {
          delay = 200;
          unit = 1;
          window.clearTimeout(timeoutId);
          scope.imageFile = canvasElem.toDataURL();
        }

        if ('ontouchstart' in window) {
          canvas.on('touchstart', down);
          $document.on('touchend', up);
          canvas.on('touchmove', move);

          btnZoomIn.on('touchstart', zoomInDownCb);
          btnZoomIn.on('touchleave touchend', stopLooping);

          btnZoomOut.on('touchstart', zoomOutDownCb);
          btnZoomOut.on('touchleave touchend', stopLooping);
        } else {
          canvas.on('mousedown', down);
          $document.on('mouseup', up);
          canvas.on('mousemove', move);

          btnZoomIn.on('mousedown', zoomInDownCb);
          btnZoomIn.on('mouseleave mouseup', stopLooping);

          btnZoomOut.on('mousedown', zoomOutDownCb);
          btnZoomOut.on('mouseleave mouseup', stopLooping);

          canvasElem.addEventListener('DOMMouseScroll', handleScroll, false);
          canvasElem.addEventListener('mousewheel', handleScroll, false);
        }

        scope.onFileSelect = function($files) {

          if (0 === $files.length) {
            return false;
          }

          var file = $files[0];

          if (! file.type.match(/image\/.+/)) {
            notificationService.notice('Unsurported file type.', 'bottom_right');
            return false;
          }

          scope.showImageBox = true;

          var fileReader = new FileReader();
          fileReader.readAsDataURL(file);
          fileReader.onload = function(e) {
            image.src = e.target.result;
          };

        };
      }
    };
  }
})();
