'use strict';

var gulp = require('gulp');
var del = require('del');
var runSequence = require('run-sequence');

var $ = require('gulp-load-plugins')({
  pattern: ['gulp-*', 'main-bower-files', 'uglify-save-license', 'del', 'browser-sync']
});

gulp.task('styles', function () {

  return gulp.src('client/assets/styles/main.scss')
    .pipe($.plumber())
    .pipe($.sass({
      includePaths: require('node-bourbon').with('client/bower_components')
    }))
    .pipe($.autoprefixer('last 1 version'))
    .pipe(gulp.dest('.tmp/styles'))
    .pipe($.browserSync.reload({stream: true}))
    .pipe($.size());
});

gulp.task('scripts', function () {
  return gulp.src('client/app/**/*.js')
    .pipe($.jshint())
    .pipe($.jshint.reporter('jshint-stylish'))
    .pipe($.size());
});

gulp.task('partials', function () {
  return gulp.src(['!client/bower_components/**', 'client/app/**/*.html'])
    .pipe($.minifyHtml({
      empty: true,
      spare: true,
      quotes: true
    }))
    .pipe($.ngHtml2js({
      moduleName: 'webapp',
      prefix: '/app/'
    }))
    .pipe(gulp.dest('.tmp/partials'))
    .pipe($.size());
});

gulp.task('html', ['styles', 'scripts', 'partials'], function (cb) {
  var jsFilter = $.filter('**/*.js');
  var cssFilter = $.filter('**/*.css');
  var assets = $.useref.assets();

  return gulp.src(['client/index.html', '!client/bower_components/**'])
    .pipe($.inject(gulp.src('.tmp/partials/**/*.js', {read: false}), {
      starttag: '<!-- inject:partials -->',
      // addRootSlash: true,
      // addPrefix: '../',
      relative: false,
      transform: function(filepath, file, index, length, targetFile) {
        filepath = filepath.slice(5);
        return $.inject.transform.call($.inject.transform, filepath, file, index, length, targetFile);
      }
    }))
    .pipe(assets)
    .pipe($.rev())
    .pipe(jsFilter)
    .pipe($.ngAnnotate({
      add: true
    }))
    .pipe($.uglify({preserveComments: $.uglifySaveLicense}))
    .pipe(jsFilter.restore())
    .pipe(cssFilter)
    .pipe($.replace('bower_components/bootstrap-sass-official/vendor/assets/fonts/bootstrap','fonts'))
    .pipe($.csso())
    .pipe(cssFilter.restore())
    .pipe(assets.restore())
    .pipe($.useref())
    .pipe($.revReplace())
    .pipe($.rename(function(path) {

      if (path.dirname.match(/^(styles|scripts)$/)) {
        path.dirname = '../' + path.dirname;
      }
    }))
    .pipe(gulp.dest('dist/public/views'))
    .pipe($.size());
});

gulp.task('images', function () {
  return gulp.src('client/assets/images/**/*')
    // .pipe($.cache($.imagemin({
      // progressive: true,
      // interlaced: true,
    // })))
    .pipe($.newer('dist/public/assets/images'))
    .pipe($.imagemin({
      progressive: true,
      interlaced: true
    }))
    .pipe(gulp.dest('dist/public/assets/images'))
    .pipe($.size());
});

gulp.task('fonts', function () {
 // return gulp.src($.mainBowerFiles())
  return gulp.src('client/assets/fonts/**/*')
    .pipe($.filter('**/*.{eot,svg,ttf,woff,otf,woff2}'))
    .pipe($.flatten())
    .pipe(gulp.dest('dist/public/assets/fonts'))
    .pipe($.size());
});

gulp.task('shell', function () {
  return gulp.src('scripts/docker/*')
    .pipe(gulp.dest('dist/scripts/docker'))
    .pipe($.size());
});

gulp.task('https', function () {
  return gulp.src('ssl/*')
    .pipe(gulp.dest('dist/ssl'))
    .pipe($.size());
});

gulp.task('clean', function (cb) {
  del(['.tmp', 'dist'], cb);
});

gulp.task('server-files', function() {
  return gulp.src(['server/**/*.js'])
    .pipe(gulp.dest('dist/server'));
});

gulp.task('languages', function() {
  return gulp.src(['client/assets/languages/**/*'])
    .pipe(gulp.dest('dist/public/assets/languages'));
});

gulp.task('toDist', function() {
  return gulp.src(['server.js', 'package.json'], {base: '.'})
    .pipe(gulp.dest('dist'))
    .pipe($.size());
});

gulp.task('deb', function() {
  return gulp.src(['debian/**/*', 'Makefile'], {base: '.'})
    .pipe(gulp.dest('dist'))
    .pipe($.size());
});

gulp.task('ci', function() {
  return gulp.src([
      'scripts/jenkins/arm-build.sh',
      'scripts/jenkins/crosscompile.sh'
    ])
    .pipe(gulp.dest('dist'))
    .pipe($.size());
});

gulp.task('favicon', function() {
  return gulp.src('client/favicon.ico')
    .pipe(gulp.dest('dist/public'));
});

gulp.task('vendor', function() {
  return gulp.src('client/vendor/**/*')
    .pipe(gulp.dest('dist/public/vendor'));
});

gulp.task('build:stage', function(cb) {
  runSequence('clean', [
    'html',
    'partials',
    'images',
    'fonts',
    'shell',
    'https',
    'server-files',
    'languages',
    'favicon',
    'vendor',
    'toDist'
  ]);
});

gulp.task('build:jenkins', function(cb) {
  runSequence('clean', [
    'html',
    'partials',
    'images',
    'fonts',
    'https',
    'server-files',
    'languages',
    'favicon',
    'vendor',
    'deb',
    'ci',
    'toDist'
  ]);
});

gulp.task('build', function(cb) {
  runSequence('clean', [
    'html',
    'partials',
    'images',
    'fonts',
    'https',
    'server-files',
    'languages',
    'favicon',
    'vendor',
    'deb',
    'toDist'
  ]);
});
