'use strict';

var db = require(__dirname + '/../../models'),
    _ = require('lodash'),
    Bluebird = require('bluebird'),
    User = db.User,
    UserEav = db.UserEav,
    jwt = require('jwt-simple'),
    jwtConst = require(__dirname + '/../../constants/jwt'),
    appService = require(__dirname + '/../../services/appService'),
    helperService = require(__dirname + '/../../services/helperService'),
    socketService = require(__dirname + '/../../services/socketService'),
    log = require('sanji-logger')('webapp');

exports.checkUserIp = function(req, res, next) {
  var remoteIp = req.ip;

  if ('127.0.0.1' === remoteIp) {
    res.locals.isLocalReq = true;
    return next();
  }
  return next();
};

exports.checkUpdateUser = function(req, res, next) {

  var locals = res.locals,
      tokenUser = locals.tokenUser,
      user = locals.user;

  // can only update themselves
  if (user.get('id') === tokenUser.get('id')) {
    return next();
  }
};

exports.findTokenData = function(req, res, next) {

  var reqAccessToken = req.params.accessToken,
      accessToken = reqAccessToken ? reqAccessToken : req.get('X-Mx-AccessToken') + '',
      tokenData, now;

  if (res.locals.isLocalReq) {
    return next();
  }

  try {
    tokenData = jwt.decode(accessToken, jwtConst.TOKEN_SECRET);
  } catch (e) {
    return res.status(401)
      .send({message: 'Mailformed access token.'});
  }

  now = +new Date();

  if (now - tokenData.createdAt > (jwtConst.EXPIRRATION_TIME * 1000)) {
    return res.status(419)
      .send({message: 'Access token expired.'});
  } else {
    res.locals.accessToken = accessToken;
    res.locals.tokenData = tokenData;
    return next();
  }
};

exports.getProfile = function(req, res) {

  var tokenUser, locals;

  locals = res.locals;
  tokenUser = locals.tokenUser;
  tokenUser.getProfile(locals.accessToken)
    .then(function(userProfile) {
      return res.status(200).send(userProfile);
    });

};

exports.findTokenUser = function(req, res, next) {

  if (res.locals.isLocalReq) {
    return next();
  }

  User.findOne({
    where: {id: res.locals.tokenData.userId},
    include: [UserEav]
  })
  .then(function(tokenUser) {

    if (! tokenUser) {
      return Bluebird.reject({
        code: 'TOKEN_USER_NOT_FOUND',
        message: 'Token user not found'
      });
    }

    res.locals.tokenUser = tokenUser;
    return next();
  })
  .catch(function(ret) {

    if ('TOKEN_USER_NOT_FOUND' === ret.code) {
      return res.status(404)
        .send({message: ret.message});
    }

    return res.sendStatus(401);
  });
};

exports.findUser = function(req, res, next) {

  var userId = req.params.userId;

  User.find({
    where: {id: userId},
    include: [UserEav],
  })
  .then(function(user) {

    if (! user) {
      return Bluebird.reject();
    }

    res.locals.user = user;
    next();
  })
  .catch(function() {
    res.sendStatus(400);
  });
};

/**
 * login
 */
exports.login = function(req, res) {

  var post = req.body;

  User.find({
    where: {email: post.email},
    include: [UserEav]
  })
  .then(function(user) {
    if (user && appService.verifyPassword(post.password, user.get('password'))) {
      return user;
    }
    if (user) {
      user.log(req, 'Cannot complete login due to an incorrect password');
    }
    return Bluebird.reject();
  })
  .then(function(user) {

    var accessToken;

    accessToken = jwt.encode({
      userId: user.get('id'),
      createdAt: +new Date(),
    }, jwtConst.TOKEN_SECRET);

    user.log(req, 'Login successfully');

    return user.getProfile(accessToken);
  })
  .then(function(userProfile) {
    return res.json(userProfile);
  })
  .catch(function(err) {
    log.error(err);
    return res.status(401).send({message: 'Please check your account and password.'});
  });
};

exports.listUsers = function(req, res) {

  User.findAll({attributes: ['id', 'email']})
  .then(function(users) {
    res.status(200).send(users);
  })
  .catch(function(err) {
    log.error(err);
    res.sendStatus(500);
  });
};

exports.getUser = function(req, res) {

  var user = res.locals.user;
  user.dataValues.meta = user.getMetaData();

  res.status(200).send(user);
};

exports.getUserLogs = function(req, res) {

  var tokenUser;

  tokenUser = res.locals.tokenUser;

  tokenUser.getUserLogs()
    .then(function(userLogs) {
      return res.status(200).send(userLogs);
    })
    .catch(function(err) {
      log.errors(err);
      return res.status(500).send(500);
    });
};

exports.updateUser = function(req, res) {

  var post = req.body,
      locals = res.locals,
      tokenUser = locals.tokenUser,
      user = locals.user,
      email = post.email,
      password = post.password,
      profileImage = post.profileImage,
      attrs = {};

  if (email) {
    attrs.email = email;
  }

  if (password) {
    attrs.password = appService.hashPassword(password);
  }

  user.updateAttributes(attrs)
    .then(function() {
      if (profileImage) {
        return user.updateProfileImage(db, profileImage);
      } else {
        return Bluebird.resolve(user);
      }
    })
    .then(function(row) {

      if ('UserEav' === row.Model.name) {
        user.dataValues.meta = JSON.parse(row.get('value'));
      }
      return Bluebird.resolve();
    })
    .then(function() {

      tokenUser.log(req, 'Update user data of ' + user.get('email'));

      res.status(200).send({
        message: 'User has been updated successfully',
        user: user
      });
    })
    .catch(function(err) {
      log.error(err);
      res.sendStatus(500);
    });
};

exports.updateMetaData = function(req, res) {
  var post = req.body,
      locals = res.locals,
      user = locals.user;

  user.getUserEav()
    .then(function(userEav) {
      var meta = JSON.parse(userEav.get('value'));
      _.extend(meta, post.meta);

      userEav.set('value', JSON.stringify(meta));
      return userEav.save();
    })
    .then(function(meta) {
      res.status(200).send({
        message: 'User meta has been updated successfully'
      });
    })
    .catch(function(err) {
      log.error(err);
      res.sendStatus(500);
    });

};
