'use strict';

var expressValidator = require('express-validator'),
    config = require('./config/config'),
    log = require('sanji-logger')('webapp'),
    path = require('path'),
    sanjiExpress = require('sanji-express'),
    filebin = require('express-filebin'),
    user = require(__dirname + '/api/user/user.controller'),
    SocketService = require(__dirname + '/services/socketService');

/**
 * Application routes
 */
module.exports = function(app) {

  app.use(function(req, res, next) {

    if (! req.get('Origin')) {
      return next();
    }
    res.set('Access-Control-Allow-Origin', '*');
    res.set('Access-Control-Allow-Methods', 'POST, GET, OPTIONS, DELETE, PUT');
    res.set('Access-Control-Allow-Headers', 'X-Requested-With, Content-Type, X-Mx-AccessToken');

    if ('OPTIONS' === req.method) {
      return res.sendStatus(200);
    }
    return next();
  });


  app.use(expressValidator({
    customValidators: {
      isTimestamp: function(timestamp) {
        return (new Date(+timestamp)).getTime() > 0;
      },
      isIntArray: function(arr) {

        if (! Array.isArray(arr)) {
          return false;
        }
        var filteredArr = arr.filter(function(elem) {
          return isNaN(parseInt(elem, 10));
        });
        return 0 === filteredArr.length;
      }
    }
  }));

  app.use(config.apiBase + '/users', require('./api/user'));
  app.use(config.apiBase + '/login', require('./api/login'));
  app.use(config.apiBase + '/notifications', require('./api/notification'));
  app.use(config.apiBase + '/profile', require('./api/profile'));
  app.use(config.apiBase + '/userLogs', require('./api/userlog'));
  // app.use(config.apiBase + '/site', require('./api/site'));
  app.use(config.apiBase + '/files', [
    user.checkUserIp,
    user.findTokenData,
    user.findTokenUser,
    filebin({
      baseUrl: config.apiBase + '/files',
      uploadOptions: {
        limits: {
          fileSize: 1024 * 1024 * 500
        }
      }
    })
  ]);

  if (!config.isMock) {
    app.use(config.apiBase, [
      user.checkUserIp,
      user.findTokenData,
      user.findTokenUser,
      sanjiExpress({
        bundlesHome: config.bundlesHome,
        brokerHost: config.sanjiBrokerIp,
        brokerPort: config.sanjiBrokerPort,
        io: SocketService.getIoInstance()
      })
    ]);
  }

  // All undefined api routes should return a 404
  app.get(config.apiBase + '/*', function(req, res) {
    res.status(404).send({message: 'Resource not found.'});
  });

  // All other routes to use Angular routing in app/scripts/app.js
  app.get('/*', function(req, res, next) {

    if (path.extname(req.originalUrl)) {
      log.warn("Error rendering partial '" + req.originalUrl + "'\n");
      res.sendStatus(404);
    }

    try {
      var data = JSON.stringify({
        env: config.env,
        mqttBridge: config.isBridgeMode,
        isMock: config.isMock
      });
      res.render('index', {config: data});
    } catch(err) {
      next(err);
    }
  });

  app.use(require('./handlers').onError);

};
